/*-----------------------------------------------------------------------

File  : clb_ptrees.c

Author: Stephan Schulz

Contents
 
  Functions for pointer storing AVL trees.

Changes

<1> Wed Dec 17 21:17:34 MET 1997
    New

-----------------------------------------------------------------------*/

#include "clb_ptrees.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: rotate_left()
//
//   Rotate the unbalanced tree at *update left, adjusting weights
//   accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left(PTree_p *update)
{
   PTree_p store;
   int b1,b2;

   VERBOUT10("Rotate tree left\n");

   store = (*update)->rson;
   (*update)->rson = (*update)->rson->lson;
   store->lson = *update;
   *update = store;
   
   b1 = (*update)->balance;
   b2 = ((*update)->lson)->balance;
   
   switch(b1)
   {
   case -1:
	 assert(b2 == 1);
	 (*update)->balance = -2;
	 ((*update)->lson)->balance = 0;
	 break;
   case 0:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 0;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 1;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 1:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = -1;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = 0;
	    ((*update)->lson)->balance = 0;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 2:
	 assert(b2 == 2);
	 (*update)->balance = 0;
	 ((*update)->lson)->balance = -1;
	 break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right()
//
//   Symmetrical case to rotate_left()
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right(PTree_p *update)
{
   PTree_p store;
   int b1, b2;

   VERBOUT10("Rotate tree right\n");

   store = (*update)->lson;
   (*update)->lson = (*update)->lson->rson;
   store->rson = *update;
   *update = store;

   b1 = (*update)->balance;
   b2 = ((*update)->rson)->balance;

   switch(b1)
   {
   case 1:
         assert(b2 == -1);
         (*update)->balance = 2;
         ((*update)->rson)->balance = 0;
         break;
   case 0:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 0;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = -1;
         }
         else
         {
            assert(false);
         }
         break;
   case -1:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 1;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 0;
            ((*update)->rson)->balance = 0;
         }
         else
         {
            assert(false);
         }
         break;
   case -2:
         assert(b2 == -2);
         (*update)->balance = 0;
         ((*update)->rson)->balance = 1;
         break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right_left()
//
//   Perform the right-left rotation to balance a tree. More exactly:
//   Rotate the subtree at (*update)->rson right and the tree at
//   *update left, and adjust the balances accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right_left(PTree_p *update)
{
   VERBOUT10("Rotate tree right left\n");

   rotate_right(&((*update)->rson));
   rotate_left(update);
}


/*-----------------------------------------------------------------------
//
// Function: rotate_left_right()
//
//   Perform the left-right rotation to balance a tree. See
//   rotate_right_left() 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left_right(PTree_p *update)
{
   VERBOUT10("Rotate tree left right\n");

   rotate_left(&((*update)->lson));
   rotate_right(update);
}

/*-----------------------------------------------------------------------
//
// Function: rebalance_tree()
//
//   Rebalance the tree at the passed position. Assumes that the
//   weights in the tree are correct. Returns true if the tree depth
//   has changed.
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool rebalance_tree(PTree_p *update)
{
   int b2 = 0;
   bool shrunk;

   if((*update)->balance == 2)
   {
      b2 = (*update)->rson->balance;
      assert((*update)->rson);
      if((*update)->rson->balance == -1)
      {
	 rotate_right_left(update);
      }
      else
      {
	 rotate_left(update);
      }
   }
   else if((*update)->balance == -2)
   {
      b2 = (*update)->lson->balance;
      assert((*update)->lson);
      if((*update)->lson->balance == 1)
      {
	 rotate_left_right(update);
      }
      else
      {
	 rotate_right(update);
      }
   }

   shrunk = (b2 != 0);
   
   return shrunk;
}

/*-----------------------------------------------------------------------
//
// Function: find_min_cell()
//
//   Find the position of the pointer to the minimal cell in the tree
//   anchored at root. See find_max_cell().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static PTree_p *find_min_cell(PTree_p *root, bool remove_mode)
{
   if(!*root)
   {
      return NULL;
   }

   while((*root)->lson)
   {
      (*root)->passed = -1;
      root = &((*root)->lson);
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: find_cell_pos()
//
//   Starting at root, search the tree for an entry with 'key', setting
//   'passed' values along the search papth and keeping *update
//   pointing to a potential update position. If update_on_zero is
//   true, set update to the lowest cell with balance 0, otherwise to
///   the lowest cell with balance != 0. Return the pointer to
//   where the entry in anchored, or the position where it would be
//   inserted if no such entry exists.
//
// Global Variables: -
//
// Side Effects    : Sets 'passed' values and *update
//
/----------------------------------------------------------------------*/

static PTree_p *find_cell_pos(PTree_p *root, void* key, 
 			      PTree_p **update, bool remove_mode)
{
   int cmpres;

   while(*root)
   {
      cmpres = PCmp(key, (*root)->key);
      if(cmpres == 0)
      {
	 return root;
      }
      
      if(!remove_mode && (*root)->balance)
      {
	 *update = root;
      }
      if(cmpres > 0)
      {
	 (*root)->passed = 1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == -1))
	 {
	    *update = root;
	 }
	 root = &((*root)->rson);
      }
      else
      {
	 (*root)->passed = -1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == 1))
	 {
	    *update = root;
	 }
	 root = &((*root)->lson);
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: isolate_and_mark_cell()
//
//   Find the cell with the key and move it to a position in the tree
//   where it can be deleted (i.e. a position where it has only one
//   successor). Set the passed-values leading to the new position,
//   and set the passed-value of the relevant cell to 0 so that it can
//   be found without (expensive) key comparisons.
//
// Global Variables: -
//
// Side Effects    : Changes tree and some cell entries.
//
/----------------------------------------------------------------------*/

static PTree_p isolate_and_mark_cell(PTree_p *root, void* key)
{
   PTree_p *elpos = root;
   PTree_p *replace = NULL;
   PTree_p *dummy;
   PTree_p element, repl, t_lson, t_rson;
   long      help;
   
   elpos = find_cell_pos(root, key, &dummy, true);
   if(!(*elpos))
   {
      return NULL;
   }
   element = *elpos;
   
   if(element->rson&&element->lson)
   {
      replace = find_min_cell(&(element->rson), true);
      repl = *replace;
      
      t_lson = repl->lson;
      t_rson = repl->rson;
      repl->lson = element->lson;
      
      if(repl == element->rson)
      {
	 repl->rson =  element;
      }
      else
      {
	 repl->rson = element->rson;
	 *replace = element;
      }
      
      *elpos = repl;
      element->lson = t_lson;
      element->rson = t_rson;
      
      repl->passed = 1;
      element->passed = 0;
      help = repl->balance;
      repl->balance = element->balance;
      element->balance = help; 
   }
   else 
   {
      element->passed = 0;
   }
   
   return element;
}


/*-----------------------------------------------------------------------
//
// Function: extract_marked_cell()
//
//   Descend the tree guided by the passed-entries until the marked
//   cell (passed == 0) has been found, extract it, store it in *cell,
//   and ascend the tree, rebalancing as necessary. Return true if the
//   tree height has changes. The tree has to be preprocessed by
//   isolate_and_mark_cell().
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool extract_marked_cell(PTree_p *root, PTree_p *cell)
{
   bool shrunk = false;

   assert(*root);

   if(!(*root)->passed)
   {
      *cell = *root;
      if((*cell)->lson)
      {
	 *root = (*cell)->lson;
      }
      else
      {
	 *root = (*cell)->rson;
      }
      shrunk = true;
   }
   else
   {
      if((*root)->passed == 1)
      {
	 if(extract_marked_cell(&((*root)->rson), cell))
	 {
	    shrunk = true;
	    (*root)->balance--;
	 }
      }
      else if((*root)->passed == -1)
      {
	 if(extract_marked_cell(&((*root)->lson), cell))
	 {
	    shrunk = true;
	    (*root)->balance++;
	 }
      }

      switch((*root)->balance)
      {
      case 2:
      case -2:
	    shrunk = rebalance_tree(root);
	    break;
      case 1:
      case -1:
	    shrunk = false;
	    break;
      case 0:
	    break;
      default:
	    assert(false);
	    break;
      }
   }
   return shrunk;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: PCmpFun()
//
//   Compare two pointers, return 1 if the first one is bigger, 0 if
//   both are equal, and -1 if the second one is bigger. Might be
//   machine dependend and of limited portability (comparing two
//   arbitrary pointers is not ANSI kosher, but the compiler has no
//   way to detect this, as pointers to the same array can be compared
//   under ANSI), but should be easy to hack on any
//   machine. Subtracting pointers and casting the result to int is
//   probably more efficient, but even more dangerous.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int PCmpFun(void* p1, void*p2)
{
   if(p1 > p2)
   {
      return 1;
   }
   else if(p1 < p2)
   {
      return -1;
   }
   assert(p1 == p2);
   return 0;
}
      


/*-----------------------------------------------------------------------
//
// Function: PTreeCellAllocEmpty()
//
//   Allocate a empty, initialized PTreeCell. Pointers to children
//   are NULL, int values are 0 (and pointer values in ANSI-World
//   undefined, in practice NULL on 32 bit machines)(This comment is
//   superfluous!). The balance field is (correctly) set to 0.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

PTree_p PTreeCellAllocEmpty()
{
   PTree_p handle = PTreeCellAlloc();
   
   handle->lson = handle->rson = NULL;
   handle->balance = 0;

   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: PTreeFree()
//
//   Free a Ptree (including the keys, but not potential objects
//   pointed to in the val fields
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PTreeFree(PTree_p junk)
{
   if(!junk)
   {
      return;
   }
   PTreeFree(junk->lson);
   PTreeFree(junk->rson);
   PTreeCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PTreeInsert()
//
//   If an entry with key *new->key exists in the tree return a
//   pointer to it. Otherwise insert *new in the tree and return
//   NULL. 
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

PTree_p PTreeInsert(PTree_p *root, PTree_p new)
{
   PTree_p *insert = root;
   PTree_p *update = root;
   PTree_p help;

   insert = find_cell_pos(root, new->key, &update, false);
   if(*insert)
   {
      return *insert;
   }
   
   *insert = new;
   new->lson = new->rson = NULL;
   new->balance = 0;
   
   help = *update;
   while(help != new)
   {
      help->balance += help->passed;
      help = help->passed == 1 ? help->rson : help->lson;
   }
   rebalance_tree(update);
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: PTreeStore()
//
//   Insert a cell associating key with val1 and val2 into the
//   tree. Return false if an entry for this key exists, true
//   otherwise. 
//
// Global Variables: -
//
// Side Effects    : Changes tree
//
/----------------------------------------------------------------------*/

bool PTreeStore(PTree_p *root, void* key)
{
   PTree_p handle, new;

   handle = PTreeCellAlloc();
   handle->key = key;
 
   new = PTreeInsert(root, handle);

   if(new)
   {
      PTreeCellFree(handle);
      return false;
   }
   return true;
}

/*-----------------------------------------------------------------------
//
// Function: PTreeFind()
//
//   Find the entry with key key in the tree and return it. Return
//   NULL if no such key exists.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

PTree_p PTreeFind(PTree_p root, void* key)
{
   int cmpres;
   
   while(root)
   {
      cmpres = PCmp(key,root->key);

      if(cmpres == 0)
      {
	 break;
      }
      if(cmpres > 0)
      {
	 root = root->rson;
      }
      else
      {
	 root = root->lson;
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: PTreeExtractEntry()
//
//   Find the entry with key key, remove it from the tree, rebalance
//   the tree, and return the pointer to the removed element. Return
//   NULL if no matching element exists.
//
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/


PTree_p PTreeExtractEntry(PTree_p *root, void* key)
{
   PTree_p cell;
   
   if(!isolate_and_mark_cell(root, key))
   {
      return NULL;
   }
   extract_marked_cell(root, &cell);

   cell->lson = cell->rson = NULL;
   return cell;
}


/*-----------------------------------------------------------------------
//
// Function: PTreeExtractKey()
//
//   Extract the entry with key key, delete the PTree-Node and return
//   the key.
//
// Global Variables: -
//
// Side Effects    : Memory operations, by PTreeExtractEntry()
//
/----------------------------------------------------------------------*/

void* PTreeExtractKey(PTree_p *root, void* key)
{
   PTree_p handle;
   void*   res = NULL;

   handle = PTreeExtractEntry(root, key);
   if(handle)
   {
      res = handle->key;
      PTreeCellFree(handle);
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: PTreeExtractRootKey()
//
//   Extract the root node of the tree, delete it and return the
//   key. Return NULL if the tree is empty.
//
// Global Variables: -
//
// Side Effects    : Changes tree
//
/----------------------------------------------------------------------*/

void* PTreeExtractRootKey(PTree_p *root)
{
   if(*root)
   {
      return PTreeExtractKey(root, (*root)->key);
   }
   return NULL;
}

/*-----------------------------------------------------------------------
//
// Function: PTreeDeleteEntry()
//
//   Delete the entry with key key from the tree. Return true, if the
//   key existed, false otherwise.
//
// Global Variables: -
//
// Side Effects    : By PTreeExtract(), memory operations
//
/----------------------------------------------------------------------*/

bool PTreeDeleteEntry(PTree_p *root, void* key)
{
   PTree_p cell;
   
   cell = PTreeExtractEntry(root, key);
   if(cell)
   {
      PTreeFree(cell);
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: PTreeMerge()
//
//   Merge the two trees, i.e. destroy the second one and add it's
//   element to the first one.
//
// Global Variables: -
//
// Side Effects    : Changes both trees.
//
/----------------------------------------------------------------------*/

void PTreeMerge(PTree_p *root, PTree_p add)
{
   PStack_p stack = PStackAlloc();
   PTree_p  res;

   PStackPushP(stack, add);

   while(!PStackEmpty(stack))
   {
      add = PStackPopP(stack);
      if(add)
      {
	 PStackPushP(stack, add->lson);
	 PStackPushP(stack, add->rson);
	 res = PTreeInsert(root, add);
	 assert(!res); /* Pointers should never be in two trees at
			  once for my intended application */
      }
   } 
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: PTreeNodes()
//
//   Return the number of nodes in the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long PTreeNodes(PTree_p root)
{
   PStack_p stack = PStackAlloc();
   long     res   = 0;

   PStackPushP(stack, root);
   
   while(!PStackEmpty(stack))
   {
      root = PStackPopP(stack);
      if(root)
      {
	 PStackPushP(stack, root->lson);
	 PStackPushP(stack, root->rson);
	 res++;
      }
   }
   PStackFree(stack);

   return res;   
}


AVL_TRAVERSE_DEFINITION(PTree, PTree_p)

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


