/*-----------------------------------------------------------------------

File  : clb_stringtrees.c

Author: Stephan Schulz

Contents
 
  Functions for string-indexed AVL trees.

Changes

<1> Thu Sep 25 02:36:58 MET DST 1997
    New

-----------------------------------------------------------------------*/

#include "clb_stringtrees.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


#ifdef CLB_DEBUG_FUNCTIONS


/*-----------------------------------------------------------------------
//
// Function: tree_size()
//
//   Return the number of nodes in the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static int tree_size(StrTree_p  tree)
{
   if(!tree)
   {
      return 0;
   }
   return 1+tree_size(tree->lson)+tree_size(tree->rson);
}


/*-----------------------------------------------------------------------
//
// Function: tree_depth()
//
//   Return the depth of the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static int tree_depth(StrTree_p  tree)
{
   if(!tree)
   {
      return 0;
   }
   return 1+max(tree_depth(tree->lson), tree_depth(tree->rson));
}


/*-----------------------------------------------------------------------
//
// Function: calc_balance()
//
//   Recompute the balance at the top node of the tree.
//
// Global Variables: -
//
// Side Effects    : Changes balance.
//
/----------------------------------------------------------------------*/

static void calc_balance(StrTree_p tree)
{
   assert(tree);
   tree->balance = tree_depth(tree->rson)-tree_depth(tree->lson);
   if(abs(tree->balance)>1)
   {
      printf("#\n");
   }
   else if(abs(tree->balance)>2)
   {
      printf("!\n");
   }      
}


/*-----------------------------------------------------------------------
//
// Function: verify_all_balances()
//
//   As the name implies. Returns tree depth.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static long verify_all_balances(StrTree_p tree)
{
   long ldepth, rdepth;

   if(!tree)
   {
      return 0;
   }
   ldepth = verify_all_balances(tree->lson);
   rdepth = verify_all_balances(tree->rson);
   
   if(tree->balance != (rdepth-ldepth))
   {
      printf("Bad balance %d (shoud be %ld) at %s\n", tree->balance,
	     rdepth-ldepth, tree->key);
      StrTreeDebugPrint(stdout, tree, true);
      assert(tree->balance == (rdepth-ldepth));
   }
   if(abs(tree->balance)>2)
   {
      printf("Illegal balance %d at cell %s\n", tree->balance,
	     tree->key);
      assert(abs(tree->balance)<=2);
   }
   
   return max(rdepth, ldepth)+1;
}
   

/*-----------------------------------------------------------------------
//
// Function: recompute_all_balances()
//
//   As the name implies. Returns tree depth.
//
// Global Variables: -
//
// Side Effects    : Changes balances
//
/----------------------------------------------------------------------*/

static long recompute_all_balances(StrTree_p tree)
{
   long ldepth, rdepth;

   if(!tree)
   {
      return 0;
   }
   ldepth = recompute_all_balances(tree->lson);
   rdepth = recompute_all_balances(tree->rson);
   
   if(tree->balance != (rdepth-ldepth))
   {
      printf("Bad balance %d (shoud be %ld) at %s\n", tree->balance,
	     rdepth-ldepth, tree->key);
      tree->balance = rdepth-ldepth;
      StrTreeDebugPrint(stdout, tree, true);
   }
   if(abs(tree->balance)>2)
   {
      printf("Illegal balance %d at cell %s\n", tree->balance,
	     tree->key);
   }
   
   return max(rdepth, ldepth)+1;
}
   
#endif


/*-----------------------------------------------------------------------
//
// Function: rotate_left()
//
//   Rotate the unbalanced tree at *update left, adjusting weights
//   accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left(StrTree_p *update)
{
   StrTree_p store;
   int b1,b2;

   VERBOUT10("Rotate tree left\n");

   store = (*update)->rson;
   (*update)->rson = (*update)->rson->lson;
   store->lson = *update;
   *update = store;
   
   b1 = (*update)->balance;
   b2 = ((*update)->lson)->balance;
   
   switch(b1)
   {
   case -1:
	 assert(b2 == 1);
	 (*update)->balance = -2;
	 ((*update)->lson)->balance = 0;
	 break;
   case 0:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 0;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 1;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 1:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = -1;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = 0;
	    ((*update)->lson)->balance = 0;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 2:
	 assert(b2 == 2);
	 (*update)->balance = 0;
	 ((*update)->lson)->balance = -1;
	 break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right()
//
//   Symmetrical case to rotate_left()
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right(StrTree_p *update)
{
   StrTree_p store;
   int b1, b2;

   VERBOUT10("Rotate tree right\n");

   store = (*update)->lson;
   (*update)->lson = (*update)->lson->rson;
   store->rson = *update;
   *update = store;

   b1 = (*update)->balance;
   b2 = ((*update)->rson)->balance;

   switch(b1)
   {
   case 1:
         assert(b2 == -1);
         (*update)->balance = 2;
         ((*update)->rson)->balance = 0;
         break;
   case 0:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 0;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = -1;
         }
         else
         {
            assert(false);
         }
         break;
   case -1:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 1;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 0;
            ((*update)->rson)->balance = 0;
         }
         else
         {
            assert(false);
         }
         break;
   case -2:
         assert(b2 == -2);
         (*update)->balance = 0;
         ((*update)->rson)->balance = 1;
         break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right_left()
//
//   Perform the right-left rotation to balance a tree. More exactly:
//   Rotate the subtree at (*update)->rson right and the tree at
//   *update left, and adjust the balances accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right_left(StrTree_p *update)
{
   VERBOUT10("Rotate tree right left\n");

   rotate_right(&((*update)->rson));
   rotate_left(update);
}


/*-----------------------------------------------------------------------
//
// Function: rotate_left_right()
//
//   Perform the left-right rotation to balance a tree. See
//   rotate_right_left() 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left_right(StrTree_p *update)
{
   VERBOUT10("Rotate tree left right\n");

   rotate_left(&((*update)->lson));
   rotate_right(update);
}

/*-----------------------------------------------------------------------
//
// Function: rebalance_tree()
//
//   Rebalance the tree at the passed position. Assumes that the
//   weights in the tree are correct. Returns true if the tree depth
//   has changed.
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool rebalance_tree(StrTree_p *update)
{
   int b2 = 0;
   bool shrunk;

   if((*update)->balance == 2)
   {
      b2 = (*update)->rson->balance;
      assert((*update)->rson);
      if((*update)->rson->balance == -1)
      {
	 rotate_right_left(update);
      }
      else
      {
	 rotate_left(update);
      }
   }
   else if((*update)->balance == -2)
   {
      b2 = (*update)->lson->balance;
      assert((*update)->lson);
      if((*update)->lson->balance == 1)
      {
	 rotate_left_right(update);
      }
      else
      {
	 rotate_right(update);
      }
   }

   shrunk = (b2 != 0);
   
   return shrunk;
}

#ifdef CLB_NEVER_DEFINED /* Funktion is there for symetry only */

/*-----------------------------------------------------------------------
//
// Function: find_max_cell()
//
//   Find the position of the pointer to the maximal cell in the tree
//   anchored at root and return it. Return NULL if the tree is
//   empty. Update the "passed" values of passed cells.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static StrTree_p *find_max_cell(StrTree_p *root, bool remove_mode)
{
   if(!*root)
   {
      return NULL;
   }
   
   while((*root)->rson)
   {
      (*root)->passed = 1;
      root = &((*root)->rson);
   }
   return root;
}

#endif


/*-----------------------------------------------------------------------
//
// Function: find_min_cell()
//
//   Find the position of the pointer to the minimal cell in the tree
//   anchored at root. See find_max_cell().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static StrTree_p *find_min_cell(StrTree_p *root, bool remove_mode)
{
   if(!*root)
   {
      return NULL;
   }

   while((*root)->lson)
   {
      (*root)->passed = -1;
      root = &((*root)->lson);
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: find_cell_pos()
//
//   Starting at root, search the tree for an entry with 'key', setting
//   'passed' values along the search papth and keeping *update
//   pointing to a potential update position. If update_on_zero is
//   true, set update to the lowest cell with balance 0, otherwise to
///   the lowest cell with balance != 0. Return the pointer to
//   where the entry in anchored, or the position where it would be
//   inserted if no such entry exists.
//
// Global Variables: -
//
// Side Effects    : Sets 'passed' values and *update
//
/----------------------------------------------------------------------*/

static StrTree_p *find_cell_pos(StrTree_p *root, char* key, 
				StrTree_p **update,
				bool remove_mode)
{
   int cmpres;

   while(*root)
   {
      cmpres = strcmp(key, (*root)->key);
      if(cmpres == 0)
      {
	 return root;
      }
      
      if(!remove_mode && (*root)->balance)
      {
	 *update = root;
      }
      if(cmpres > 0)
      {
	 (*root)->passed = 1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == -1))
	 {
	    *update = root;
	 }
	 root = &((*root)->rson);
      }
      else
      {
	 (*root)->passed = -1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == 1))
	 {
	    *update = root;
	 }
	 root = &((*root)->lson);
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: isolate_and_mark_cell()
//
//   Find the cell with the key and move it to a position in the tree
//   where it can be deleted (i.e. a position where it has only one
//   successor). Set the passed-values leading to the new position,
//   and set the passed-value of the relevant cell to 0 so that it can
//   be found without (expensive) key comparisons.
//
// Global Variables: -
//
// Side Effects    : Changes tree and some cell entries.
//
/----------------------------------------------------------------------*/

static StrTree_p isolate_and_mark_cell(StrTree_p *root, char* key)
{
   StrTree_p *elpos = root;
   StrTree_p *replace = NULL;
   StrTree_p *dummy;
   StrTree_p element, repl, t_lson, t_rson;
   long      help;
   
   elpos = find_cell_pos(root, key, &dummy, true);
   if(!(*elpos))
   {
      return NULL;
   }
   element = *elpos;
   
   if(element->rson&&element->lson)
   {
      replace = find_min_cell(&(element->rson), true);
      repl = *replace;
      
      t_lson = repl->lson;
      t_rson = repl->rson;
      repl->lson = element->lson;
      
      if(repl == element->rson)
      {
	 repl->rson =  element;
      }
      else
      {
	 repl->rson = element->rson;
	 *replace = element;
      }
      
      *elpos = repl;
      element->lson = t_lson;
      element->rson = t_rson;
      
      repl->passed = 1;
      element->passed = 0;
      help = repl->balance;
      repl->balance = element->balance;
      element->balance = help; 
   }
   else 
   {
      element->passed = 0;
   }
   
   return element;
}


/*-----------------------------------------------------------------------
//
// Function: extract_marked_cell()
//
//   Descend the tree guided by the passed-entries until the marked
//   cell (passed == 0) has been found, extract it, store it in *cell,
//   and ascend the tree, rebalancing as necessary. Return true if the
//   tree height has changes. The tree has to be preprocessed by
//   isolate_and_mark_cell().
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool extract_marked_cell(StrTree_p *root, StrTree_p *cell)
{
   bool shrunk = false;

   assert(*root);

   if(!(*root)->passed)
   {
      *cell = *root;
      if((*cell)->lson)
      {
	 *root = (*cell)->lson;
      }
      else
      {
	 *root = (*cell)->rson;
      }
      shrunk = true;
   }
   else
   {
      if((*root)->passed == 1)
      {
	 if(extract_marked_cell(&((*root)->rson), cell))
	 {
	    shrunk = true;
	    (*root)->balance--;
	 }
      }
      else if((*root)->passed == -1)
      {
	 if(extract_marked_cell(&((*root)->lson), cell))
	 {
	    shrunk = true;
	    (*root)->balance++;
	 }
      }

      switch((*root)->balance)
      {
      case 2:
      case -2:
	    shrunk = rebalance_tree(root);
	    break;
      case 1:
      case -1:
	    shrunk = false;
	    break;
      case 0:
	    break;
      default:
	    assert(false);
	    break;
      }
   }
   return shrunk;
}

/*-----------------------------------------------------------------------
//
// Function: strtree_print()
//
//   Print the tree with the appropriate indent level and return the
//   number of nodes.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static long strtree_print(FILE* out, StrTree_p tree, bool keys_only,
			  int indent)
{
   DStr_p indstr;
   int i, size;

   indstr = DStrAlloc();
   for(i=0;i<indent;i++)
   {
      DStrAppendStr(indstr, "  ");
   }

   if(!tree)
   {
      fprintf(out, "%s[]\n", DStrView(indstr));
      size = 0;
   }
   else 
   {
      if(keys_only)
      {
	 fprintf(out, "%s%s %d\n", DStrView(indstr), tree->key, tree->balance);
      }
      else
      {
	 fprintf(out, "%s%s\n", DStrView(indstr), tree->key);
	 fprintf(out, "%s balance: %d  passed: %d\n", DStrView(indstr),
		 tree->balance, tree->passed);
	 fprintf(out, "%s Val1: %ld  Val2: %ld\n", DStrView(indstr),
		 tree->val1.i_val, tree->val2.i_val);
	 fprintf(out, "%s lson: %d  rson: %d\n", DStrView(indstr),
		 (int)tree->lson, (int)tree->rson);
      }
      size = 1;
      if(tree->lson||tree->rson)
      {
	 size += strtree_print(out, tree->lson, keys_only, indent+2);
	 size += strtree_print(out, tree->rson, keys_only, indent+2);
      }
   }
   DStrFree(indstr);
   return size;   
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: StrTreeCellAllocEmpty()
//
//   Allocate a empty, initialized StrTreeCell. Pointers to children
//   are NULL, int values are 0 (and pointer values in ANSI-World
//   undefined, in practice NULL on 32 bit machines)(This comment is
//   superfluous!). The balance field is (correctly) set to 0.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

StrTree_p StrTreeCellAllocEmpty()
{
   StrTree_p handle = StrTreeCellAlloc();
   
   handle->val1.i_val = handle->val2.i_val = 0;
   handle->lson       = handle->rson       = NULL;
   handle->balance = 0;

   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: StrTreeFree()
//
//   Free a stringtree (including the keys, but not potential objects
//   pointed to in the val fields
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void StrTreeFree(StrTree_p junk)
{
   if(!junk)
   {
      return;
   }
   StrTreeFree(junk->lson);
   StrTreeFree(junk->rson);
   FREE(junk->key);
   StrTreeCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: StrTreeInsert()
//
//   If an entry with key *new->key exists in the tree return a
//   pointer to it. Otherwise insert *new in the tree and return
//   NULL. 
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

StrTree_p StrTreeInsert(StrTree_p *root, StrTree_p new)
{
   StrTree_p *insert = root;
   StrTree_p *update = root;
   StrTree_p help;

   insert = find_cell_pos(root, new->key, &update, false);
   if(*insert)
   {
      return *insert;
   }
   
   *insert = new;
   new->lson = new->rson = NULL;
   new->balance = 0;
   
   help = *update;
   while(help != new)
   {
      help->balance += help->passed;
      help = help->passed == 1 ? help->rson : help->lson;
   }
   rebalance_tree(update);
   return NULL;
}

/*-----------------------------------------------------------------------
//
// Function: StrTreeDebugPrint()
//
//   Print the tree in an unattractive but debug-friendly way.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

long StrTreeDebugPrint(FILE* out, StrTree_p tree, bool keys_only)
{
   long size;
   size = strtree_print(out, tree, keys_only, 0);
   fprintf(out, "Tree size: %ld\n", size);
   return size;
}


/*-----------------------------------------------------------------------
//
// Function: StrTreeFind()
//
//   Find the entry with key key in the tree and return it. Return
//   NULL if no such key exists.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

StrTree_p StrTreeFind(StrTree_p root, char* key)
{
   int cmpres;
   
   while(root)
   {
      cmpres = strcmp(key, root->key);

      if(cmpres == 0)
      {
	 break;
      }
      if(cmpres > 0)
      {
	 root = root->rson;
      }
      else
      {
	 root = root->lson;
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: StrTreeExtractEntry()
//
//   Find the entry with key key, remove it from the tree, rebalance
//   the tree, and return the pointer to the removed element. Return
//   NULL if no matching element exists.
//
//   Basic Algorithm (obsolete, replaced by recursive Version):
//
//   - Search for element el to remove, keeping potential update position
//   - Search for the one-child element mv to take its position, again
//     keeping potential update postition:
//        + if balance == -1, search for largest element in el->lson
//        + Otherwise search for the smallest element in el->rson
//   - for update to mv update balance
//   - replace el by mv and update mv's parent.
//   - Rebalance tree at *update and hope that it works ;-)
//   - Oh yes, return el...
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/


StrTree_p StrTreeExtractEntry(StrTree_p *root, char* key)
{
   StrTree_p cell;
   
   if(!isolate_and_mark_cell(root, key))
   {
      return NULL;
   }
   extract_marked_cell(root, &cell);

   cell->lson = cell->rson = NULL;
   return cell;
}


/*-----------------------------------------------------------------------
//
// Function: StrTreeDeleteEntry()
//
//   Delete the entry with key key from the tree. 
//
// Global Variables: -
//
// Side Effects    : By StrTreeExtract(), memory operations
//
/----------------------------------------------------------------------*/

bool StrTreeDeleteEntry(StrTree_p *root, char* key)
{
   StrTree_p cell;
   
   cell = StrTreeExtractEntry(root, key);
   if(cell)
   {
      StrTreeFree(cell);
      return true;
   }
   return false;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


