/*-----------------------------------------------------------------------

File  : ccl_eqn.c

Author: Stephan Schulz

Contents
 
  The termpair datatype: Rules, Equations, positive and negative
  literals. 

Changes

<1> Fri Mar 13 17:23:31 MET 1998

-----------------------------------------------------------------------*/

#include "ccl_eqn.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool EqnUseInfix = true;
bool EqnFullEquationalRep = false;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: compare_oriented_eqns()
//
//  Compare two oriented equations.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_oriented_eqns(OCB_p ocb, 
					   Eqn_p eq1, Eqn_p eq2)
{
   CompareResult res;

   assert(EqnIsOriented(eq1));
   assert(EqnIsOriented(eq2));
   
   res = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
   
   switch(res)
   {
   case to_uncomparable:
	 break;
   case to_equal:
	 res = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS,
			 DEREF_ALWAYS); 
	 break;
   case to_greater:
   case to_lesser:
	 break;
   default: 
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: compare_oriented_unoriented()
//
//   Compare an oriented equation and an unoriented one.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_oriented_unoriented(OCB_p ocb, Eqn_p eq1,
						 Eqn_p eq2)
{
   CompareResult res, res2;

   assert(EqnIsOriented(eq1));
   assert(!EqnIsOriented(eq2));
   
   res = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
   
   switch(res)
   {
   case to_uncomparable:
	 /* It is still possible that eq2->rterm >= eq1->lterm! */
	 res2 = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS,
			  DEREF_ALWAYS); 
	 switch(res2)
	 {
	 case to_uncomparable:
	       /* Stays at uncomparable, of course */
	       break;
	 case to_equal:
	       /* eq1->lterm, eq2->rterm cancel each other out */
	       res = TOCompare(ocb, eq1->rterm, eq2->lterm,
			       DEREF_ALWAYS, DEREF_ALWAYS);
	       break;
	 case to_greater:
	       /* But {eq1->lterm, eq1->rterm} !> {eq2->lterm}! */
	       res = to_uncomparable;
	       break;
	 case to_lesser:
	       /* Then also eq2->rterm > eq1->lterm, hence eq2 > eq1 */
	       res = to_lesser;
	       break;
	 default:
	       assert(false);
	       break;
	 }	 
	 break;
   case to_equal:
	 res = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS,
			 DEREF_ALWAYS); 
	 break;
   case to_greater:
	 /* But is eq1>eq2->rterm? */
	 res2 = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
	 switch(res2)
	 {
	 case to_uncomparable:
	       /* ergo {eq1->lterm, eq1->rterm} !> {eq2->rterm}! */
	       res = to_uncomparable;
	       break;
	 case to_equal:
	       /* eq1->lterm, eq2->rterm cancel each other out */
	       res = TOCompare(ocb, eq1->rterm, eq2->lterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
	       break;
	 case to_greater:
	       /* Everything is fine, eq1->lterm > eq2 */
	       break;
	 case to_lesser:
	       /* Then also eq2->rterm > eq1->rterm, hence eq2 > eq1 */
	       res = to_lesser;
	       break;
	 default:
	       assert(false);
	       break;
	 }	 	 
	 break;
   case to_lesser:
	 /* Then eq2->rterm also > eq1->rterm, i.e. eq2 > eq1! */
	 break;
   default: 
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: combine_indep_results()
//
//   Combine the results of two independed term comparisons.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult combine_indep_results(CompareResult res1,
					   CompareResult res2)
{
   CompareResult res = to_uncomparable;

   switch(res1)
   {
   case to_uncomparable:
	 break;
   case to_equal:
	 res = res2;
   case to_greater:
   case to_lesser:
	 if(res2 == to_equal || res2 == res1)
	 {
	    res = res1;
	 }
	 else
	 {
	    res = to_uncomparable;
	 }
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: compare_unoriented_eqns1()
//
//   Compare two unoriented equations (by StS).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_unoriented_eqns1(OCB_p ocb,
					     Eqn_p eq1, Eqn_p eq2)
{
   CompareResult res, res_ll, res_rr, res_lr, res_rl;

   assert(!EqnIsOriented(eq1));
   assert(!EqnIsOriented(eq2));

   res_ll = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
   res_rr = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);

   res = combine_indep_results(res_ll, res_rr);

   if(res != to_uncomparable)
   {
      return res;
   }

   res_lr = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);

   if(res_ll == to_greater && res_lr == to_greater)
   {
      return to_greater;
   }
   if(res_rr == to_lesser && res_lr == to_lesser)
   {
      return to_lesser;
   }

   res_rl = TOCompare(ocb, eq1->rterm, eq2->lterm, DEREF_ALWAYS,
		   DEREF_ALWAYS);
   
   res = combine_indep_results(res_lr, res_rl);

   if(res != to_uncomparable)
   {
      return res;
   }

   if(res_ll == to_lesser && res_rl == to_lesser)
   {
      return to_lesser;
   }
   if(res_rr == to_greater && res_rl == to_greater)
   {
      return to_greater;
   }

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: compare_unoriented_eqns()
//
//   Compare two unoriented equations (by JS):
//
//     {l1,r1} == {l2,r2}  <==>  (l1=l2 & r1=r2) v (l1=r2 & r1=l2)
//
//     Assume that {l1,r1} =/= {l2,r2}. Then,
//
//     {l1,r1} >> {l2,r2}  <==>  (l1>=l2 v r1>=l2) & (l1>=r2 v r1>=r2)
//     {l1,r1} << {l2,r2}  <==>  (l1<=l2 v l1<=r2) & (r1<=l2 v r1<=r2)
//
//     Otherwise, {l1,r1} and {l2,r2} are incomparable.
//
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_unoriented_eqns(OCB_p ocb,
					     Eqn_p eq1, Eqn_p eq2)
{
   CompareResult ll, rr, lr, rl;

   assert(!EqnIsOriented(eq1));
   assert(!EqnIsOriented(eq2));

   ll = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   lr = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);
   rl = TOCompare(ocb, eq1->rterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   rr = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);

   if (((ll == to_equal) && (rr == to_equal)) ||
       ((lr == to_equal) && (rl == to_equal))) {
      return to_equal;
   }

   if (((ll == to_greater || ll == to_equal) ||
	(rl == to_greater || rl == to_equal))   &&
       ((lr == to_greater || lr == to_equal) ||
	(rr == to_greater || rr == to_equal))) {
      return to_greater;
   }

   if (((ll == to_lesser || ll == to_equal) ||
	(lr == to_lesser || lr == to_equal))   &&
       ((rl == to_lesser || rl == to_lesser) ||
	(rr == to_lesser || rr == to_equal))) {
      return to_lesser;
   }

   return to_uncomparable;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: EqnAlloc()
//
//   Allocate a literal with the given (shared terms). References for
//   the terms will be added to the bank.
//
// Global Variables: -
//
// Side Effects    : Adds references
//
/----------------------------------------------------------------------*/

Eqn_p EqnAlloc(Term_p lterm, Term_p rterm, TB_p bank,  bool positive)
{
   Eqn_p handle = EqnCellAlloc();

   handle->type =  EqUnoriented;
   handle->positive = positive;
   handle->maximal = false;
   handle->equ_lit = !TBTermEqual(rterm, bank->true_term);
   handle->bank = bank;

   handle->next = NULL;

   TermGetRef(&(handle->lterm), lterm);
   TermGetRef(&(handle->rterm), rterm);

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnFree()
//
//   Free the space taken by an equation, release the term refs and
//   try to delete the terms.
//
// Global Variables: -
//
// Side Effects    : May change term bank, removes references
//
/----------------------------------------------------------------------*/

void EqnFree(Eqn_p junk)
{
   TermReleaseRef(&(junk->lterm));
   TBDelete(junk->bank, junk->lterm);
   TermReleaseRef(&(junk->rterm));
   TBDelete(junk->bank, junk->rterm);
   EqnCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function:  EqnParse()
//
//   Parse an equation and return a pointer to the resulting cell. If
//   EqnUseInfix is true, = and != (and =>, !=>) are used as infix
//   operators, otherwise eq() and ~eq() are used. Non-equational
//   literals are transformed into equational form. Note that
//   EqnStrictEquationalRep is not enforced (it is only used for
//   output). 
//
// Global Variables: EqnUseInfix
//
// Side Effects    : Input, memory management.
//
/----------------------------------------------------------------------*/

Eqn_p EqnParse(Scanner_p in, TB_p bank)
{
   Term_p  lterm;
   Term_p  rterm;
   bool    positive = true;
   EqnType type = EqUnoriented;
   Eqn_p   handle;
   

   if(TestInpTok(in, TildeSign))
   {
      NextToken(in);
      positive = false;
   }
   if(EqnUseInfix && !TestInpId(in, EQUAL_PREDICATE))
   {
      lterm = TBTermParse(in, bank);
      if(TestInpTok(in, Exclamation | EqualSign))
      {
	 if(!positive)
	 {
	    Warning("Negation sign ('~') used in infix equational literal");
	 }
	 if(TestInpTok(in, Exclamation))
	 {
	    NextToken(in);
	    positive = !positive;
	 }
	 AcceptInpTok(in, EqualSign);	 
	 if(TestInpTok(in, GreaterSign))
	 {
	    AcceptInpTokNoSkip(in, GreaterSign);
	    type = EqOriented;
	 }
	 rterm = TBTermParse(in, bank);
      }
      else
      {
	 rterm = bank->true_term; /* Non-Equational literal */
      }
   }
   else
   {
      if(TestInpId(in, EQUAL_PREDICATE))
      {	 
	 NextToken(in);
	 AcceptInpTok(in, OpenBracket);
	 lterm = TBTermParse(in, bank);
	 AcceptInpTok(in, Comma);
	 rterm = TBTermParse(in, bank);
	 AcceptInpTok(in, CloseBracket);
      }
      else
      {
	 lterm = TBTermParse(in, bank);
	 rterm = bank->true_term; /* Non-Equational literal */
      }
   }
   handle = EqnAlloc(lterm, rterm, bank, positive);
   handle->type = type;
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnPrint()
//
//   Print an equation. If negated is true, print the negated
//   equation.
//
// Global Variables: EqnUseInfix, EqnFullEquationalRep
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EqnPrint(FILE* out, Eqn_p eqn, bool negated,  bool fullterms)
{
   if(EqnUseInfix && 
      (EqnFullEquationalRep || 
       eqn->rterm!=eqn->bank->true_term ||
       eqn->lterm==eqn->bank->true_term))
   {
      TBPrintTerm(out, eqn->bank, eqn->lterm, fullterms);
      if(XOR(!eqn->positive, negated))
      {
	 fputc('!', out);
      }
      if(eqn->type == EqOriented)
      {
	 fprintf(out, "=>");
      }
      else
      {
	 fprintf(out, "=");
      }
      TBPrintTerm(out, eqn->bank, eqn->rterm, fullterms);
   }
   else
   {
      if(XOR(!eqn->positive, negated))
      {
	 fputc('~', out);
      }
      if((eqn->rterm!=eqn->bank->true_term) ||
	 EqnFullEquationalRep ||
	 eqn->lterm==eqn->bank->true_term)
      {
	 fprintf(out, EQUAL_PREDICATE"(");
	 TBPrintTerm(out, eqn->bank, eqn->lterm, fullterms);
	 fprintf(out, ", ");
	 TBPrintTerm(out, eqn->bank, eqn->rterm, fullterms);
	 fputc(')', out);
      }
      else
      {
	 TBPrintTerm(out, eqn->bank, eqn->lterm, fullterms);
      }
   }
#ifdef NEVER_DEFINED
   fprintf(out, "/* %s, %s */", eqn->maximal? "Max" : "Dom",
	   eqn->equ_lit? " = " : " ~ ");
#endif
}


/*-----------------------------------------------------------------------
//
// Function: EqnSwapSidesSimple()
//
//   Exchange the two sides of the equation. This will lead to
//   inconsistent states if not used carefully (i.e. only temporary or
//   via EqnSwapSides() (which takes care of the attached strings).
//
// Global Variables: -
//
// Side Effects    : Exchanges sides, !!! Causes inconsistent state!!!
//
/----------------------------------------------------------------------*/

void EqnSwapSidesSimple(Eqn_p eq)
{
   Term_p term;

   term = eq->lterm;
   eq->lterm = eq->rterm;
   eq->rterm = term;
}


/*-----------------------------------------------------------------------
//
// Function: EqnSwapSides()
//
//   Exchange the two sides of an equation. Will update type and
//   references.  
//
// Global Variables: -
//
// Side Effects    : Exchanges the sides, changes type.
//
/----------------------------------------------------------------------*/

void EqnSwapSides(Eqn_p eq)
{
   if(eq->type == EqOriented)
   {
      eq->type = EqUnoriented;
   }
   switch(eq->demod_sides)
   {
   case LeftSide:
	 eq->demod_sides = RightSide;
	 break;
   case RightSide:
	 eq->demod_sides = LeftSide;
	 break;
   default:
	 break;
   }

   TermReleaseRef(&(eq->lterm));
   TermReleaseRef(&(eq->rterm));

   EqnSwapSidesSimple(eq);

   TermGetRef(&(eq->lterm), eq->lterm);
   TermGetRef(&(eq->rterm), eq->rterm);
}


/*-----------------------------------------------------------------------
//
// Function: EqnRecordTermDates()
//
//   Set eqn->[lr]date to eqn->[lr]term->tb_date.
//
// Global Variables: -
//
// Side Effects    : As above
//
/----------------------------------------------------------------------*/

void EqnRecordTermDates(Eqn_p eqn)
{
   assert(eqn);
   assert(eqn->lterm);
   assert(eqn->rterm);

   eqn->ldate = eqn->lterm->tb_date;
   eqn->rdate = eqn->rterm->tb_date;
}


/*-----------------------------------------------------------------------
//
// Function:  EqnHasChanged()
//
//   Return MaxSide if a potential maximal side of eqn has been
//   rewritten, MinSide if a non-maximal side has been rewritten,
//   and NoSide if equation is unchanged.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

EqnSide EqnHasChanged(Eqn_p eqn)
{
   if(SysDateCompare(eqn->ldate, eqn->lterm->tb_date)!=DateEqual)
   {
      return MaxSide;
   }
   if(SysDateCompare(eqn->rdate, eqn->rterm->tb_date)!=DateEqual)
   {
      if(EqnIsOriented(eqn))
      {
	 return MinSide;
      }
      return MaxSide;
   }
   return NoSide;
}


/*-----------------------------------------------------------------------
//
// Function: EqnUpdateOrientation()
//
//   If maximal side of an equation has been rewritten, change type to
//   EqUnoriented.
//
// Global Variables: -
//
// Side Effects    : Changes eqn->type
//
/----------------------------------------------------------------------*/

bool EqnUpdateOrientation(Eqn_p eqn)
{
   if(EqnHasChanged(eqn)==LeftSide)
   {
      eqn->type = EqUnoriented;
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EqnCopy()
//
//   Create a copy of eqn with terms from bank. Does not copy the
//   next pointer. If keeporient is true, the copy will have the same
//   type as the old equation, otherwise, if the original equation was
//   oriented, the copy is so as well, otherwise it is
//   unoriented. This is only correct if the original equation is at
//   most modified by instantiation! 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eqn_p EqnCopy(Eqn_p eqn, TB_p bank, bool keeporient)
{
   Eqn_p  handle;
   Term_p lterm, rterm;

   lterm = TBInsert(bank, eqn->lterm, DEREF_ALWAYS);
   rterm = TBInsert(bank, eqn->rterm, DEREF_ALWAYS);

   handle = EqnAlloc(lterm, rterm, bank, eqn->positive);

   handle->maximal = eqn->maximal;
   if(keeporient)
   {
      handle->type = eqn->type;
#ifdef DEMOD_VAR_RESTRICTION
      handle->demod_sides = eqn->demod_sides;
#endif
   }
   else
   {
      handle->type = eqn->type == EqOriented?EqOriented:EqUnoriented;
#ifdef DEMOD_VAR_RESTRICTION
      handle->demod_sides = EqOriented?eqn->type:BothSides; /* potentially */
#endif
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnEqualDirected()
//
//   Test wether two equations are identical (modulo EqualTest). Does
//   not check for positive/negative, evaluates only terms
//   (i.e. treats equations as unsigned term pairs).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnEqualDirected(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest)
{
   bool res = EqualTest(eq1->lterm, eq2->lterm);
   
   if(!res)
   {
      return res;
   }
   res = EqualTest(eq1->rterm, eq2->rterm);
   
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnEqual()
//
//   Test wether two equations are equivalen (modulo EqualTest and
//   commutativity). Treats equations as unsigned term sets.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnEqual(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest)
{
   bool res;

   res = EqnEqualDirected(eq1, eq2, EqualTest);
   
   if(res || ((eq1->type == EqOriented) && (eq2->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(eq2);
   res = EqnEqualDirected(eq1, eq2, EqualTest);
   EqnSwapSidesSimple(eq2);
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnSubsumeDirected()
//
//   Test wether an equation subsumes another one. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodified. Don't deal with commutativity of
//   equality. 
//
// Global Variables: -
//
// Side Effects    : -
//
//----------------------------------------------------------------------*/

bool EqnSubsumeDirected(Eqn_p subsumer, Eqn_p subsumed, Subst_p
			 subst, TermEqualTestFun EqualTest)
{
   PStackPointer backtrack = PStackGetSP(subst);
   bool res;

   res = SubstComputeMatch(subsumer->lterm, subsumed->lterm, subst,
			   EqualTest);
   if(res)
   {
      res = SubstComputeMatch(subsumer->rterm,
			      subsumed->rterm, subst,
			      EqualTest);
   }
   if(!res)
   {
      SubstBacktrackToPos(subst, backtrack);
   }
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: EqnSubsume()
//
//   Test wether an equation subsumes another one. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodifies. Equations are treated as 2-sets of
//   terms unless both are oriented.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnSubsume(Eqn_p subsumer, Eqn_p subsumed, Subst_p subst,
		TermEqualTestFun EqualTest)
{
   bool   res = EqnSubsumeDirected(subsumer, subsumed, subst, EqualTest);
   
   if(res || 
      ((subsumer->type == EqOriented) && 
       (subsumed->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(subsumer);
   res = EqnSubsumeDirected(subsumer, subsumed, subst, EqualTest);
   EqnSwapSidesSimple(subsumer);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnSubsumeP()
//
//   Test wether subsumer subsumes subsumed, undo all side effects.
//
// Global Variables: -
//
// Side Effects    : - (I hope)
//
/----------------------------------------------------------------------*/

bool EqnSubsumeP(Eqn_p subsumer, Eqn_p subsumed, TermEqualTestFun
		 EqualTest) 
{
   Subst_p subst = SubstAlloc();
   bool    res = EqnSubsume(subsumer, subsumed, subst, EqualTest);

   SubstDelete(subst);
   
   return res;   
}


/*-----------------------------------------------------------------------
//
// Function: LiteralSubsumeP()
//
//   Return true if subsumer subsumes subsumed, false
///  otherwise. Checks signs!
//
// Global Variables: -
//
// Side Effects    : Only temporary
//
/----------------------------------------------------------------------*/

bool LiteralSubsumeP(Eqn_p subsumer, Eqn_p subsumed, TermEqualTestFun
		     EqualTest) 
{
   if(subsumer->positive == subsumed->positive)
   {
      return EqnSubsumeP(subsumer, subsumed, EqualTest);
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EqnUnifyDirected()
//
//   Test wether two equations can be unified. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodified. Don't deal with commutativity of
//   equality. 
//
// Global Variables: -
//
// Side Effects    : -
//
//----------------------------------------------------------------------*/

bool EqnUnifyDirected(Eqn_p eq1, Eqn_p eq2, Subst_p subst)
{
   PStackPointer backtrack = PStackGetSP(subst);
   bool res;

   res = SubstComputeMgu(eq1->lterm, eq2->lterm, subst);
   if(res)
   {
      res = SubstComputeMgu(eq1->rterm,
			      eq2->rterm, subst);
   }
   if(!res)
   {
      SubstBacktrackToPos(subst, backtrack);
   }
   return res;
}




/*-----------------------------------------------------------------------
//
// Function: EqnUnify()
//
//   Test wether two equations are unifyable. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodifies. Equations are treated as 2-sets of
//   terms unless both are oriented.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnUnify(Eqn_p eq1, Eqn_p eq2, Subst_p subst)
{
   bool   res = EqnUnifyDirected(eq1, eq2, subst);
   
   if(res || 
      ((eq1->type == EqOriented) && 
       (eq2->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(eq1);
   res = EqnUnifyDirected(eq1, eq2, subst);
   EqnSwapSidesSimple(eq1);
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: EqnUnifyP()
//
//   Test wether two equations are unifiable, undo all side effects.
//
// Global Variables: -
//
// Side Effects    : - (I hope)
//
/----------------------------------------------------------------------*/

bool  EqnUnifyP(Eqn_p eq1, Eqn_p eq2)
{
   Subst_p subst = SubstAlloc();
   bool    res = EqnUnify(eq1, eq2, subst);

   SubstDelete(subst);
   
   return res;   
}


/*-----------------------------------------------------------------------
//
// Function: EqnOrient()
//
//   Orient an equation. Return true, if sides are exchanged, false
//   otherwise. If unorientable, also find potential demod sides.
//
// Global Variables: -
//
// Side Effects    : May change term references
//
/----------------------------------------------------------------------*/

bool EqnOrient(OCB_p ocb, Eqn_p eqn)
{
   CompareResult relation = to_uncomparable;
   bool res = false;

   assert(eqn->type != EqOriented);

   if(TBTermEqual(eqn->lterm, eqn->rterm))
   {
      relation = to_equal;
   }
   else if(TBTermEqual(eqn->lterm, eqn->bank->true_term))
   {
      relation = to_lesser;
   }
   else if(TBTermEqual(eqn->rterm, eqn->bank->true_term))
   {
      relation = to_greater;
   }
   else
   {
      /* printf("EqnOrient: ");
      TermPrint(stdout, eqn->lterm, eqn->bank->sig, DEREF_ALWAYS);
      printf(" # ");
      TermPrint(stdout, eqn->rterm, eqn->bank->sig, DEREF_ALWAYS);      
      printf("\n"); */
      relation = TOCompare(ocb, eqn->lterm, eqn->rterm, DEREF_ALWAYS,
			   DEREF_ALWAYS);
   }
   
   switch(relation)
   {
   case to_uncomparable:
   case to_equal:
	 eqn->type = EqUnorientable;
#ifdef DEMOD_VAR_RESTRICTION
	 eqn->demod_sides = TermPairFindDemodSides(eqn->lterm,
						   eqn->rterm);
	 /* EqnPrintOriginal(stdout, eqn); printf(" -> %d\n",
	    eqn->demod_sides); */
#endif
	 break;
   case to_greater:
	 eqn->type = EqOriented;
	 break;
   case to_lesser:
	 EqnSwapSides(eqn);
	 res = true;
	 eqn->type = EqOriented;
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnCompare()
//
//   Compare two equations (as multisets of terms) and return the
//   result.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult EqnCompare(OCB_p ocb, Eqn_p eqn1, Eqn_p eqn2)
{
   CompareResult res, res1;

   if(EqnIsOriented(eqn1))
   {
      if(EqnIsOriented(eqn2))
      {
	 res = compare_oriented_eqns(ocb, eqn1, eqn2);
      }
      else
      {
	 res = compare_oriented_unoriented(ocb, eqn1, eqn2);
      }
   }
   else
   {
      if(EqnIsOriented(eqn2))
      {
	 res = OCBPrecedenceInverse(
	    compare_oriented_unoriented(ocb, eqn2, eqn1)
	    );
      }
      else
      {
	 res =  compare_unoriented_eqns(ocb, eqn1, eqn2);
#ifndef NDEBUG
	 res1 = compare_unoriented_eqns1(ocb, eqn1, eqn2);
	 assert(res==res1);
#endif
      }
   }
   
/*   printf("EqnCompare: ");
   EqnPrintOriginal(stdout, eqn1);
   printf(" <---> ");
   EqnPrintOriginal(stdout, eqn2);
   printf(" = %s\n", TOCompareSymbol[res]);
   */ 
  return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnGreater()
//
//   Return true if eqn1 is greater than eqn2, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnGreater(OCB_p ocb, Eqn_p eqn1, Eqn_p eqn2)
{
   bool res;

   res = EqnCompare(ocb, eqn1, eqn2) == to_greater;

   return res;
}



/*-----------------------------------------------------------------------
//
// Function: LiteralCompare()
//
//   Compare two signed literals (if the terms are equivalent, the
//   negative one is bigger, otherwise as for equations).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult LiteralCompare(OCB_p ocb, Eqn_p eqn1, Eqn_p eqn2)
{
   CompareResult res;

   res = EqnCompare(ocb, eqn1, eqn2);
   if(res == to_equal)
   {
      if(eqn1->positive && !eqn2->positive)
      {
	 res = to_lesser;
      }
      else if(!eqn1->positive && eqn2->positive)
      {
	 res = to_greater;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: LiteralGreater()
//
//   Return true if eqn1 is greater than eqn2, takes as (signed)
//   literals, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool LiteralGreater(OCB_p ocb, Eqn_p eqn1, Eqn_p eqn2)
{
   bool res;
   
   res = LiteralCompare(ocb, eqn1, eqn2) == to_greater;
   
   return res;   
}

/*-----------------------------------------------------------------------
//
// Function: SubstNormEqn()
//
//   Instantiate all variables in eqn with normed variables.
//
// Global Variables: -
//
// Side Effects    : By SubstNormTerm()
//
/----------------------------------------------------------------------*/

FunCode SubstNormEqn(Eqn_p eqn, Subst_p subst, VarBank_p vars)
{
  FunCode res;

   SubstNormTerm(eqn->lterm, subst, vars);
   res = SubstNormTerm(eqn->rterm, subst, vars);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnWeight()
//
//   Compute the weight of an equation. Weights of potentially maximal
//   sides are multiplied by max_multiplier.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

double EqnWeight(Eqn_p eqn, double max_multiplier, long vweight, long
		  fweight)
{
   double res;

   if(EqnIsOriented(eqn))
   {
      res = (double)TermWeight(eqn->rterm, vweight, fweight);
   }
   else
   {
      res = (double)TermWeight(eqn->rterm, vweight, fweight) *
	 max_multiplier;
   }
   res += ((double)TermWeight(eqn->lterm, vweight, fweight) * max_multiplier);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: LiteralWeight()
//
//   Return weight of a literal. max_multipler is applied to maximal
//   sides _and_ maximal literals, pos_multiplier is applied to
//   positive literals
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double LiteralWeight(Eqn_p eqn, double max_multiplier, double
		      pos_multiplier, long vweight, long fweight)
{
   double res;

   res = EqnWeight(eqn, max_multiplier, vweight, fweight);
      
   if(eqn->maximal)
   {
      res = res*max_multiplier;
   }
   if(eqn->positive)
   {
      res = res*pos_multiplier;
   }

   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


