/*-----------------------------------------------------------------------

File  : ccl_pdtrees.c

Author: Stephan Schulz

Contents
 
  Perfect discrimination trees for optimized rewriting.

Changes

<1> Wed Jun 24 00:55:29 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_pdtrees.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool PDTreePreferGeneral = false;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: pdtree_select_alternative()
//
//   Given a pdtree and a term node, return the alternative in the
//   tree or NULL if no alternative exists up to now.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static PDTree_p pdtree_select_alternative(PDTree_p root, Term_p term)
{
   NumTree_p cell;

   if(TermIsVar(term))
   {
      cell = NumTreeFind(root->v_alternatives, term->f_code);
      return cell?cell->val1.p_val:NULL;
   }
   assert(term->f_code < root->sig_size);
   return root->f_alternatives[term->f_code];   
}


/*-----------------------------------------------------------------------
//
// Function: remove_demod_sides()
//
//   Remove all clause positions indexed by term from the tree at
//   *root. If restrict != NULL only remove entries containing the
//   given clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

static long remove_demod_sides(PTree_p *root, Term_p term, Clause_p
			       restrict)
{
   PStack_p    stack, key_stack = PStackAlloc();
   PTree_p     cell;
   long        res = 0;
   ClausePos_p pos;
   bool        dont_delete = false;
   bool        test;

   stack = PTreeTraverseInit(*root);
   while((cell = PTreeTraverseNext(stack)))
   {
      pos = cell->key;
      if(TBTermEqual(ClausePosGetSide(pos), term))
      {	 
	 if(!restrict || (restrict == pos->clause))
	 {
	    PStackPushP(key_stack, pos);
	    /* printf("Deleting: ");
	    ClausePrint(stdout, pos->clause, true);
	    printf(" (%d, C= %d)\n", (int)pos, (int)pos->clause); */
	    ClausePosFree(pos);
	    res++;
	 }
	 else
	 {
	    /* printf("Don't deleted due to clause restr: ");
	    ClausePrint(stdout, pos->clause, true);
	    printf(" (%d, C= %d)\n", (int)pos, (int)pos->clause); */
	    dont_delete = true; /* There still is a clause left which
				   has this term in the index */				
	 }
      }
      else
      {
	 /* printf("Don't deleted due to Term inequality: ");
	 ClausePrint(stdout, pos->clause, true);
	 printf(" (%d, C= %d)\n", (int)pos, (int)pos->clause); */
      }
   }

   PTreeTraverseExit(stack);
   
   while(!PStackEmpty(key_stack))
   {
      pos = PStackPopP(key_stack);
      test = PTreeDeleteEntry(root, pos);
/*      printf("PTreeDeleteEntry() -> %d\n", (int)pos); */
      assert(test);
   }
   PStackFree(key_stack);
   
   if(!dont_delete)
   {
      term->is_indexed = false;
   }
   /* printf("res = %ld\n", res); */
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: pd_tree_delete_alternative()
//
//   Given a PDTreeCell and an alternative, delete the corresponding
//   sub-tree-cell. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

static void pd_tree_delete_alternative(PDTree_p root, FunCode f_code)
{
   NumTree_p cell;

   if(f_code<0)
   {
      cell = NumTreeExtractEntry(&(root->v_alternatives),
				 f_code);
      PDTreeDefaultCellFree(cell->val1.p_val);
      NumTreeCellFree(cell);
   }
   else
   {
      assert(f_code < root->sig_size);
      PDTreeDefaultCellFree(root->f_alternatives[f_code]);
      root->f_alternatives[f_code] = NULL;
   }
}


/*-----------------------------------------------------------------------
//
// Function: push_all_alternatives()
//
//   Push all alternatives in alt_pdt compatible with term onto the
//   stack in triples binding, variable, PDTCell.
//
// Global Variables: -
//
// Side Effects    : Changes stack
//
/----------------------------------------------------------------------*/

static void push_all_alternatives(PStack_p stack, PDTree_p alt_pdt, Term_p term)
{
   PStack_p  local_stack;
   NumTree_p cell;
   Term_p    var;
   PDTree_p  new_alt;

   if(!term)
   {
            return;      
   }

   if(PDTreePreferGeneral)
   {
      if((term->f_code > 0) && alt_pdt->f_alternatives[term->f_code])
      {
	 PStackPushP(stack, NULL);  /* No variable */
	 PStackPushP(stack, NULL);  /* and not bound... */
	 PStackPushP(stack, alt_pdt->f_alternatives[term->f_code]);
      } 
   }

   local_stack = NumTreeTraverseInit(alt_pdt->v_alternatives);

   while((cell = NumTreeTraverseNext(local_stack)))
   {
      new_alt = cell->val1.p_val;
      var     = cell->val2.p_val;
      if(!var->binding)
      {
	 PStackPushP(stack, NULL);
	 PStackPushP(stack, var);
	 PStackPushP(stack, new_alt);
      }
      else if(TBTermEqual(var->binding, term))
      {
	 PStackPushP(stack, var->binding);
	 PStackPushP(stack, var);
	 PStackPushP(stack, new_alt);
      }      
   }
   NumTreeTraverseExit(local_stack);
   
   if(!PDTreePreferGeneral)
   {
      if((term->f_code > 0) && alt_pdt->f_alternatives[term->f_code])
      {
	 PStackPushP(stack, NULL);  /* No variable */
	 PStackPushP(stack, NULL);  /* and not bound... */
	 PStackPushP(stack, alt_pdt->f_alternatives[term->f_code]);
      }
   }
}

/*-----------------------------------------------------------------------
//
// Function: pdtree_backtrack()
//
//   Backtrack over a PDTreeChoicepoint. Return false if backtracked
//   over the inital choice point, true otherwise.
//
// Global Variables: -
//
// Side Effects    : Memory operations, may change subst.
//
/----------------------------------------------------------------------*/

static bool pdtree_backtrack(PDSearch_p state, Subst_p subst)
{
   PDChP_p chp;
   Term_p  term;

   assert(!PStackEmpty(state->chp_stack));

   chp = PStackPopP(state->chp_stack);

   if(PStackEmpty(state->chp_stack))
   {
      assert(PStackEmpty(state->term_proc));
      PDChPFree(chp);
      return false;
   }
   
   term = PStackPopP(state->term_proc);
   if(chp->variable)
   {
      PStackPushP(state->term_state, term);
      if(!chp->binding)
      {
	 SubstBacktrackSingle(subst);
      }
   }
   else
   {
      TermLRTraversePrev(state->term_state, term);
   }
   PDChPFree(chp);
   return true;
}


/*-----------------------------------------------------------------------
//
// Function: pdtree_forward()
//
//   
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static PDTree_p pdtree_forward(PDSearch_p state, Subst_p subst)
{
   PDTree_p alternative = NULL;
   Term_p   var, binding, term;
   PDChP_p  choicepoint, new;
   
   choicepoint = PStackTopP(state->chp_stack);
   alternative = PStackPopP(choicepoint->choices);
   var         = PStackPopP(choicepoint->choices);
   binding     = PStackPopP(choicepoint->choices);
   new         = PDChPAlloc(var, binding);

   if(var)
   {
      term = PStackPopP(state->term_state);
      /* printf("Var %ld...\n", var->f_code); */
      if(!binding)
      {
	 /* printf("!Binding...\n"); */
	 SubstAddBinding(subst, var, term);
      }
   }
   else
   {
      /* printf("No var\n"); */
      term = TermLRTraverseNext(state->term_state);
   }
   PStackPushP(state->term_proc, term);
   
   term = PStackEmpty(state->term_state) ?
      NULL:PStackTopP(state->term_state);
   push_all_alternatives(new->choices, alternative,
			 term);
   PStackPushP(state->chp_stack, new);

   return alternative;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: PDTreeAlloc()
//
//   Allocate an empty, initialized PDTreeCell for indexing terms with
//   up to max_symbol function symbols.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

PDTree_p PDTreeAlloc(FunCode sig_size)
{
   PDTree_p handle = PDTreeCellAlloc();
   FunCode  i;
   
   handle->sig_size       = sig_size;
   handle->ref_count      = 0;
   handle->f_alternatives =
      SizeMalloc(sizeof(PDTree_p)*handle->sig_size);
   for(i=0; i<handle->sig_size; i++)
   {
      handle->f_alternatives[i] = NULL;
   }
   handle->v_alternatives = NULL;
   handle->demod_terms    = NULL;
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeDefaultCellFree()
//
//   Free a initilized but empty PDTreeCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PDTreeDefaultCellFree(PDTree_p junk)
{
   assert(junk);
   assert(junk->f_alternatives);
   assert(!junk->v_alternatives);
   SizeFree(junk->f_alternatives, sizeof(PDTree_p)*junk->sig_size);
   PDTreeCellFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function: PDTreeFree()
//
//   Return the memory taken by a PDTree.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PDTreeFree(PDTree_p junk)
{
   FunCode      i;
   ClausePos_p  tmp;
   PStack_p     stack;
   NumTree_p    cell;
   
   for(i=0; i<junk->sig_size; i++)
   {
      if(junk->f_alternatives[i])
      {
	 PDTreeFree(junk->f_alternatives[i]);
      }
   }  
   stack = NumTreeTraverseInit(junk->v_alternatives);
   while((cell = NumTreeTraverseNext(stack)))
   {
      PDTreeFree(cell->val1.p_val);
   }
   NumTreeTraverseExit(stack);
   NumTreeFree(junk->v_alternatives);
   junk->v_alternatives = NULL;
   
   while(junk->demod_terms)
   {
      tmp = PTreeExtractRootKey(&junk->demod_terms);
      ClausePosFree(tmp);
   }   
   PDTreeDefaultCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PDChPAlloc()
//
//   Allocate an initialized PDChoicepoint cell.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

PDChP_p PDChPAlloc(Term_p variable, Term_p binding)
{
   PDChP_p handle = PDChPCellAlloc();

   handle->variable = variable;
   handle->binding  = binding;
   handle->choices  = PStackAlloc();

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDChPFree()
//
//   Return a PDChoicepointcell.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

void PDChPFree(PDChP_p junk)
{
   PStackFree(junk->choices);
   
   PDChPCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PDSearchAlloc()
//
//   Allocate an initialized PDSearchState.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

PDSearch_p PDSearchAlloc(PDTree_p index, Term_p term)
{
   PDSearch_p handle = PDSearchCellAlloc();
   PDChP_p    init_state;
 
   assert(index);
   assert(term); 

   handle->chp_stack  = PStackAlloc();
   init_state         = PDChPAlloc(NULL, NULL);
   push_all_alternatives(init_state->choices, index, term);
   PStackPushP(handle->chp_stack, init_state);

   handle->term_state = TermLRTraverseInit(term);
   handle->term_proc  = PStackAlloc();
   handle->demod_store = PStackAlloc();

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDSearchFree()
//
//   Free a PDSearchstate.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void PDSearchFree(PDSearch_p junk)
{
   while(!PStackEmpty(junk->chp_stack))
   {
      PDChPFree(PStackPopP(junk->chp_stack));
   }
   PStackFree(junk->chp_stack);
   TermLRTraverseExit(junk->term_state);
   PStackFree(junk->term_proc);
   PStackFree(junk->demod_store);
   
   PDSearchCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseInit()
//
//   Initialize a stack for left-right term traversal (i.e. traversal
//   in wrinting order) and return it.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

PStack_p TermLRTraverseInit(Term_p term)
{
   PStack_p stack = PStackAlloc();
   
   PStackPushP(stack, term);
   
   return stack;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseNext()
//
//   Return the next term node in LR-ordering and update the
//   stack. Return NULL if term traveral is complete.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

Term_p TermLRTraverseNext(PStack_p stack)
{
   int    i;
   Term_p handle;
   
   if(PStackEmpty(stack))
   {
      return NULL;
   }
   handle = PStackPopP(stack);
   for(i=handle->arity-1; i>=0; i--)
   {
      PStackPushP(stack, handle->args[i]);
   }
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraversePrev()
//
//   Undo a TermLRTraverseNext() operation by replacing terms args on
//   the stack with term.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

Term_p TermLRTraversePrev(PStack_p stack, Term_p term)
{
   Term_p tmp;
   int    i;

   for(i=0; i<term->arity; i++)
   {
      tmp = PStackPopP(stack);
      assert(tmp == term->args[i]);
   }
   PStackPushP(stack, term);

   return term;
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeInsert()
//
//   Insert a clause position into the index.
//
// Global Variables: -
//
// Side Effects    : Changes PDTree
//
/----------------------------------------------------------------------*/

void PDTreeInsert(PDTree_p root, ClausePos_p demod_side)
{
   Term_p   term = ClausePosGetSide(demod_side);
   PStack_p stack;
   PDTree_p new;
   IntOrP   new_cell, var;
   
   assert(root);

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeInsert()...\n");

   term->is_indexed = true;
   stack = TermLRTraverseInit(term);
   while((term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(root, term);
      if(!new)
      {
	 new = PDTreeAlloc(root->sig_size);

	 if(TermIsVar(term))
	 {
	    new_cell.p_val = new;
	    var.p_val      = term;
	    NumTreeStore(&(root->v_alternatives), term->f_code,
			 new_cell, var);
	 }
	 else
	 {
	    assert(term->f_code < root->sig_size);
	    root->f_alternatives[term->f_code] = new;
	 }
      }
      new->ref_count++;
      root = new;
   }
   TermLRTraverseExit(stack);
   /* printf("In index: %d\n", (int)demod_side); */
   
   /* printf("Storing %d, %c\n", (int)demod_side, demod_side->side ==
      LeftSide?'l':'r'); */
   PTreeStore(&(root->demod_terms), demod_side);

   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeInsert()\n");
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeDelete()
//
//   Remove a term from the index.
//
// Global Variables: -
//
// Side Effects    : Modifies term.
//
/----------------------------------------------------------------------*/

void PDTreeDelete(PDTree_p root, Term_p term, Clause_p restrict)
{
   PStack_p stack, delete_stack;
   long     ref_diff;
   PDTree_p tmp, new;
   FunCode  f_code;
   Term_p   trav_term;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeDelete()...\n");

   assert(root);

   if(!term->is_indexed)
   {
      return;
   }
   
   delete_stack = PStackAlloc();
   PStackPushP(delete_stack, root);
   tmp = root;
   
   trav_term = term;
   stack = TermLRTraverseInit(trav_term);
   while((trav_term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(tmp, trav_term);
      assert(new);
      PStackPushInt(delete_stack, trav_term->f_code);
      PStackPushP(delete_stack, new);
      tmp = new;
   }
   TermLRTraverseExit(stack);
   
   ref_diff = remove_demod_sides(&(tmp->demod_terms), term, restrict);
   
   /* assert(ref_diff > 0); */
   
   while(PStackTopP(delete_stack)!=root)
   {
      tmp = PStackPopP(delete_stack);
      f_code = PStackPopInt(delete_stack);
      tmp->ref_count-=ref_diff;
      if(!tmp->ref_count)
      {
	 pd_tree_delete_alternative(PStackTopP(delete_stack), f_code);	 
      }
   }
   PStackFree(delete_stack);

   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeDelete()\n");
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeFindNextIndexed()
//
//    Given a search state and a substitution, find the next search
//    state describing a term matching the term encoded in
//    state->term_state and state->term_proc.
//
//    Starting conditions:
//
//    state->term_state describes the part of the term still to
//    process, with PStackTopP(state->term_state) being the next
//    subterm to process.
//
//    state->term_proc contains subterms already worked off.
//
//    state->chp_stack contains the sequence of choicepoints already
//    traversed, with PStackTopP(state->chp_state) being the state
//    whose alternatives correspond to
//    PStackTopP(state->term_state). 
//
// Global Variables: -
//
// Side Effects    : Memory operations, binds variables
//
/----------------------------------------------------------------------*/
   
PDTree_p PDTreeFindNextIndexedTerm(PDSearch_p state, Subst_p subst)
{
   PDTree_p alternative = NULL;
   PDChP_p  choicepoint;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeFindNextIndexed()...\n");

   assert(state);
   assert(subst);
   
   if(PStackEmpty(state->term_state))
   {
      pdtree_backtrack(state, subst);
   }

   while(!PStackEmpty(state->term_state))
   {
      choicepoint = PStackTopP(state->chp_stack);
      if(PStackEmpty(choicepoint->choices))
      {
	 if(!pdtree_backtrack(state, subst))
	 {
	    alternative = NULL;
	    break;
	 }
      }
      else
      {
	 alternative = pdtree_forward(state, subst);
	 assert(alternative);
      }
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeFindNextIndexed()\n");
   return alternative;
}


/*-----------------------------------------------------------------------
//
// Function:  PDTreeFindNextDemodulator()
//
//   Return the next potential demodulator frpm the index.
//
// Global Variables: -
//
// Side Effects    : Changes state and subst
//
/----------------------------------------------------------------------*/

ClausePos_p PDTreeFindNextDemodulator(PDSearch_p state, Subst_p
				      subst)
{
   PDTree_p    alternative;
   PStack_p    stack;
   PTree_p     node;
   ClausePos_p res;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeFindNextDemodulator()...\n");

   assert(state);
   assert(subst);
   
   if(PStackEmpty(state->demod_store))
   {
      alternative = PDTreeFindNextIndexedTerm(state, subst);
      if(alternative)
      {
	 stack = PTreeTraverseInit(alternative->demod_terms);
	 while((node = PTreeTraverseNext(stack)))
	 {
	    PStackPushP(state->demod_store, node->key);
	 }
	 PTreeTraverseExit(stack);
      }
      else
      {
	 DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeFindNextDemodulator()\n");   
	 return NULL;
      }
   }
   assert(!PStackEmpty(state->demod_store));

   res = PStackPopP(state->demod_store);   

   
   DEBUGMARK(PDT_INTERFACE_WATCH,
	     "...PDTreeFindNextDemodulator()\nDemod: ");
   DEBUG(PDT_INTERFACE_WATCH, 
	 ClausePrint(stdout, res->clause, true);
	 printf(" (%c)\n---> ", res->side == LeftSide? 'l':'r');
	 TermPrint(stdout, ClausePosGetSide(res),
		   res->clause->literals->bank->sig, DEREF_ONCE);
	 printf(" =?= ");
         TermPrint(stdout, PStackElementP(state->term_proc, 0),
		   res->clause->literals->bank->sig, DEREF_NEVER);
	 printf("\n"););
   assert(VerifyMatch(ClausePosGetSide(res),
		      PStackElementP(state->term_proc, 0)));

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: 
//
//   
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

PDTree_p PDTreeFindTerm(PDTree_p index, Term_p term)
{
   PStack_p stack;
   PDTree_p tmp, new;
   Term_p   trav_term;
   
   assert(index);
   
   tmp = index;   
   trav_term = term;
   stack = TermLRTraverseInit(trav_term);
   while((trav_term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(tmp, trav_term);
      assert(new);
      tmp = new;
   }
   TermLRTraverseExit(stack);
   
   return tmp;
}

/*-----------------------------------------------------------------------
//
// Function: 
//
//   
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void PDTreeStoreClauses(PTree_p *tree, PDTree_p index, Term_p term)
{
   PDTree_p    node;
   PStack_p    stack;
   ClausePos_p pos;
   PTree_p     entry;

   node = PDTreeFindTerm(index, term);
   
   stack = PTreeTraverseInit(node->demod_terms);
   while((entry = PTreeTraverseNext(stack)))
   {
      pos = entry->key;
      /* printf("PDTreeStoreClauses: %d C=%d\n", (int)pos,
	 (int)pos->clause); */
      if(TBTermEqual(ClausePosGetSide(pos), term))
      {
	 PTreeStore(tree, pos->clause);
      }
   }
   PTreeTraverseExit(stack);
}

/*-----------------------------------------------------------------------
//
// Function: ClauseHasIndexedTerm()
//
//   Return true if the clause has an indexed term.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseHasIndexedTerm(Clause_p clause)
{   
   Eqn_p handle = clause->literals;

   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 return true;
      }
      if(handle->rterm->is_indexed)
      {
	 return true;
      }
      handle = handle->next;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveUnchangedIndexedTerms()
//
//   Remove all indexed terms which are not modified of clause from
//   index. Modified terms should have been removed during
//   rewriting... 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseRemoveUnchangedIndexedTerms(Clause_p clause, PDTree_p index)
{   
   Eqn_p handle = clause->literals;
   long  i = 0;

   DEBUGMARK(PDT_INTERFACE_WATCH, "ClauseRemoveUnchangedIndexedTerms()...\n");
   assert(index);
   
   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 if(!EqnSideHasChanged(handle, LeftSide))
	 {
	    /* printf("Removed: \n--->");
	    ClausePrint(stdout, clause, true);
	    printf("\n"); */
	    PDTreeDelete(index, handle->lterm, clause);
	    i++;
	 }
      }
      if(handle->rterm->is_indexed)
      {
	 if(!EqnSideHasChanged(handle, RightSide))
	 {	    
	    PDTreeDelete(index, handle->rterm, clause);
	    /* printf("Removed: \n--->");
	    ClausePrint(stdout, clause, true);
	    printf("\n"); */
	    i++;
	 }
      }
      handle = handle->next;
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...ClauseRemoveUnchangedIndexedTerms()\n");
   return i;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveFromIndexed()
//
//   Remove all indexed terms of clause from index.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseRemoveFromIndex(Clause_p clause, PDTree_p index)
{   
   Eqn_p handle = clause->literals;
   long  i = 0;

   DEBUGMARK(PDT_INTERFACE_WATCH, "ClauseFromIndexed()...\n");
   assert(index);
   
   /* printf("Removing clause %d\n", (int) clause); */

   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 /* printf("Removed: \n--->");
	 ClausePrint(stdout, clause, true);
	 printf("\n"); */
	 PDTreeDelete(index, handle->lterm, clause);
	 i++;
      }
      if(handle->rterm->is_indexed)
      {
	 /* printf("Removed: \n--->");
	 ClausePrint(stdout, clause, true);	 
	 printf("\n"); */
	 PDTreeDelete(index, handle->rterm, clause);		 
	 i++;
      }
      handle = handle->next;
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...ClauseRemoveFromIndex()\n");
   return i;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





