/*-----------------------------------------------------------------------

File  : ccl_clauses.c

Author: Stephan Schulz

Contents
 
  Clauses - Infrastructure functions

Changes

<1> Thu Apr 16 19:38:16 MET DST 1998
    New

-----------------------------------------------------------------------*/


#include "ccl_clauses.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

static long global_clause_counter = 0; /* Counts calls to
					  ClauseAlloc()  and
					  ClauseCopy */

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: EmptyClauseAlloc()
//
//   Return a pointer to an empty clause initialized with rational
//   values.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

Clause_p EmptyClauseAlloc()
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = NULL;
   handle->tot_lit_no  = 0;
   handle->neg_lit_no  = 0;
   handle->pos_lit_no  = 0;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->fanned      = false;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAlloc()
//
//   Create a new clause with the literal list list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Clause_p ClauseAlloc(Eqn_p literals)
{
   Clause_p handle = EmptyClauseAlloc();

   handle->literals = literals;
   handle->ident = ++global_clause_counter;

   while(literals)
   {
      handle->tot_lit_no++;
      if(literals->positive)
      {
	 handle->pos_lit_no++;
      }
      else
      {
	 handle->neg_lit_no++;
      }
      literals = literals->next;
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseFree()
//
//   Free a clause. Does not take care of parents, children, etc., but
//   releases the memory directly taken by the clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseFree(Clause_p junk)
{
   assert(!junk->set);
   EvalListFree(junk->evaluations);   
   EqnListFree(junk->literals);
   PTreeFree(junk->children);
   ClauseCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseCopy()
//
//   Create a semantically equivalent clause and return a pointer to
//   it. Does not copy parents, children, etc. Terms come from the new
//   term bank and are instantiated. Evaluations are not copied as
//   well. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseCopy(Clause_p clause, TB_p bank)
{
   Clause_p handle = ClauseCellAlloc();

   handle->literals    = EqnListCopy(clause->literals, bank, true);
   handle->ident       = ++global_clause_counter;
   handle->tot_lit_no  = clause->tot_lit_no;
   handle->neg_lit_no  = clause->neg_lit_no;
   handle->pos_lit_no  = clause->pos_lit_no;
   handle->evaluations = NULL;
   handle->set         = NULL;
   handle->fanned      = clause->fanned;
   handle->procedural  = clause->procedural;
   handle->children    = NULL;
   handle->parent1     = NULL;
   handle->parent2     = NULL;
   handle->pred        = NULL;
   handle->succ        = NULL;

   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: ClausePrintList()
//
//   Print a clause as a declarative list of literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintList(FILE* out, Clause_p clause, bool fullterms)
{
   EqnListPrint(out, clause->literals, "; ", normal, fullterms);
   fprintf(out, " <-.");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintAxiom()
//
//   Print a clause as a declarative axiom in normal form, i.e. print
//   positive literals first, then <-, then negative literals.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintAxiom(FILE* out, Clause_p clause, bool fullterms)
{
   Eqn_p handle;
   int   i;

   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(handle->positive)
      {
	 EqnPrint(out, handle, normal, fullterms);
	 i++;
	 if(i<clause->pos_lit_no)
	 {
	    fprintf(out, "; ");
	 }
      }
   }
   fprintf(out, " <- ");
   
   i = 0;
   for(handle = clause->literals; handle; handle = handle->next)
   {
      if(!handle->positive)
      {
	 EqnPrint(out, handle, special, fullterms);
	 i++;
	 if(i<clause->neg_lit_no)
	 {
	    fprintf(out, ", ");
	 }
      }
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintRule()
//
//   Print a clause as a rule, with the head literal as the conclusion
//   and the remaining literals as preconditions. Select <- or :-
//   depending on wether the clause is procedural or not. If a
//   procedural clause has a single literal only, print it as a fact.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintRule(FILE* out, Clause_p clause, bool fullterms)
{
   if(clause->literals)
   {
      EqnPrint(out, clause->literals, normal, fullterms);
      if(clause->literals->next|| !clause->procedural)
      {
	 fprintf(out, clause->procedural? " :- " : " <- ");
	 EqnListPrint(out, clause->literals->next, ", ", special,
		      fullterms);
      }
   }
   else
   {
      fprintf(out, " <- ");
   }
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintGoal()
//
//   Print a clause as a goal, i.e. put all literals behind <-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintGoal(FILE* out, Clause_p clause, bool fullterms)
{
   fprintf(out, "<- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrintQuery()
//
//   Print a clause as a procedural query, i.e. put all literals
//   behind ?-.
//
// Global Variables: Only indirectly (Equation and term formats)
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrintQuery(FILE* out, Clause_p clause, bool fullterms)
{
   fprintf(out, "?- ");
   EqnListPrint(out, clause->literals, ", ", special, fullterms);
   fprintf(out, ".");
}


/*-----------------------------------------------------------------------
//
// Function: ClausePrint()
//
//   Print a clause in the most canonical representation.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClausePrint(FILE* out, Clause_p clause, bool fullterms)
{
   DEBUG(1023, fprintf(out, "/*(%3ld)*/ ", clause->ident););
   if(clause->procedural)
   {
      if(clause->pos_lit_no)
      {
	 ClausePrintRule(out, clause, fullterms);
      }
      else
      {
	 ClausePrintQuery(out, clause, fullterms);
      }
   }
   else
   {
      if(clause->fanned)
      {
	 ClausePrintRule(out, clause, fullterms);
      }
      else
      {
	 ClausePrintAxiom(out, clause, fullterms);
      }
   }
   DEBUG(64, EvalListPrintComment(out, clause->evaluations););
   DEBUG(64, printf("Address: %d", (int)clause););
}


/*-----------------------------------------------------------------------
//
// Function: ClauseStartsMaybe()
//
//   Return true if a clause possibly starts on the current position
//   in the input, i.e. if TermStartToken, TildeSign, ?-, or <- are
//   present on the input stream.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseStartsMaybe(Scanner_p in)
{
   if(TestInpTok(in, TermStartToken|TildeSign))
   {
      return true;
   }
   if(TestInpTok(in, LesserSign|QuestionMark))
   {
      if(!LookToken(in,1)->skipped && TestTok(LookToken(in, 1), Hyphen))
      {
	 return true;
      }
   }
   return false;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseParse()
//
//   Parse a clause.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations
//
/----------------------------------------------------------------------*/

Clause_p ClauseParse(Scanner_p in, TB_p bank)
{
   Eqn_p    concl, precond;
   bool     procedural = false;
   Clause_p handle;

   VarBankClearExtNames(bank->vars);
   
   concl = EqnListParse(in, bank, Semicolon);
   if(TestInpTok(in, Colon))
   {
      if(EqnListLength(concl)>1)
      {
	 AktTokenError(in, 
		       "Procedural rule cannot have more than one "
		       "head literal",
		       false);
      }
      procedural = true;
   }
   else if(TestInpTok(in, QuestionMark))
   {
      if(EqnListLength(concl)>0)
      {
	 AktTokenError(in, 
		       "Query should consist only of tail literals",
		       false);
      }
      procedural = true;
   }
   if(TestInpTok(in, Fullstop))
   {
      if(EqnListLength(concl)>1)
      {
	 AktTokenError(in, 
		       "Procedural fact cannot have more than one "
		       "literal",
		       false);
      }
      procedural = true;
   }
   else
   {
      AcceptInpTok(in, LesserSign | Colon | QuestionMark);
      AcceptInpTokNoSkip(in, Hyphen);
      
      precond = EqnListParse(in, bank, Comma);
   
      if(procedural && EqnListLength(precond)==0)
      {
	  AktTokenError(in, 
			"Procedural rule or query needs at least one "
			"tail literal (Hey! I did not make this "
			" syntax! -StS)", 
			false);
      }
      EqnListNegateEqns(precond);
      EqnListAppend(&concl, precond);
   }
   handle = ClauseAlloc(concl);
   handle->procedural = procedural;
   AcceptInpTok(in, Fullstop);

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkMaximalTerms()
//
//   Orient literals, mark maximal literals.
//
// Global Variables: -
//
// Side Effects    : Changes clause structure, orients equations
//
/----------------------------------------------------------------------*/

void ClauseMarkMaximalTerms(OCB_p ocb, Clause_p clause)
{
   EqnListOrient(ocb, clause->literals);
   EqnListMaximalLiterals(ocb, clause->literals);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseMarkRestrictedTerms()
//
//   Mark terms on which rewriting is restricted.
//
// Global Variables: -
//
// Side Effects    : Changes term bank...
//
/----------------------------------------------------------------------*/

void ClauseMarkRestrictedTerms(Clause_p clause)
{
   EqnListMarkRestrictedTerms(clause->literals);
}



/*-----------------------------------------------------------------------
//
// Function: ClauseRecordTermDates()
//
//   Record the term dates for the literals of the clause.
//
// Global Variables: -
//
// Side Effects    : Changes literals [lr]date
//
/----------------------------------------------------------------------*/

void ClauseRecordTermDates(Clause_p clause)
{
   Eqn_p handle;

   DEBUGMARK(4, "ClauseRecordTermDates()...\n");
   DEBUG(4, ClausePrint(stdout, clause, true);printf("\n"););
   handle = clause->literals;

   while(handle)
   {
      EqnRecordTermDates(handle);
      handle = handle->next;
   }
   DEBUGMARK(4, "ClauseRecordTermDates()...\n");
}


/*-----------------------------------------------------------------------
//
// Function: ClauseDetachParents()
//
//   Discouple a clause from it's parents, i.e. make the parents
//   forget the clause and vice versa.
//
// Global Variables: -
//
// Side Effects    : Changes parens children-tree.
//
/----------------------------------------------------------------------*/

void ClauseDetachParents(Clause_p clause)
{
   bool result;

   if(clause->parent1)
   {
      result = PTreeDeleteEntry(&(clause->parent1->children), clause);
      assert(result);
      clause->parent1 = NULL;
   }
   if(clause->parent2)
   {
      result = PTreeDeleteEntry(&(clause->parent2->children), clause);
      assert(result);
      clause->parent2 = NULL;
   }   
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRegisterChild()
//
//   Make child a child of clause. Clause already should be a parent
//   to child!
//
// Global Variables: -
//
// Side Effects    : Changes clause->children
//
/----------------------------------------------------------------------*/

void ClauseRegisterChild(Clause_p clause, Clause_p child)
{
   assert(child->parent1==clause || child->parent2==clause);
   
   PTreeStore(&(clause->children), child);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveSuperfluousLiterals()
//
//   Remove duplicate and trivial negative literals from the clause.
//
// Global Variables: -
//
// Side Effects    : Changes clause, termbanks
//
/----------------------------------------------------------------------*/

bool ClauseRemoveSuperfluousLiterals(Clause_p clause)
{
   Eqn_p handle;

   EqnListRemoveResolved(&(clause->literals), TBTermEqual);
   EqnListRemoveDuplicates(clause->literals, TBTermEqual);
   
   clause->tot_lit_no = 0;
   clause->neg_lit_no = 0;
   clause->pos_lit_no = 0;   
   handle = clause->literals;

   while(handle)
   {
      clause->tot_lit_no++;
      if(handle->positive)
      {
	 clause->pos_lit_no++;
      }
      else
      {
	 clause->neg_lit_no++;
      }
      handle = handle->next;
   }
   return (clause->tot_lit_no == 0);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAddEvalCell()
//
//   Add an evaluation cell (as the first evaluation) to a clause. 
//
// Global Variables: -
//
// Side Effects    : Changes eval list, makes evaluation->object point
//                   to clause.
//
/----------------------------------------------------------------------*/

void ClauseAddEvalCell(Clause_p clause, Eval_p evaluation)
{
   evaluation->object  = clause;
   evaluation->next    = clause->evaluations;
   clause->evaluations = evaluation;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveEvaluations()
//
//   Remove the evaluations from a clause and free the EvalCells.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseRemoveEvaluations(Clause_p clause)
{
   EvalListFree(clause->evaluations);
   clause->evaluations = NULL;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseWeight()
//
//   Compute the weight of a clause by counting function symbols and
//   variables and applying various modifiers.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double ClauseWeight(Clause_p clause, double max_multiplier, double
		      pos_multiplier, long vweight, long fweight)
{
   Eqn_p  handle;
   double res = 0;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      res += LiteralWeight(handle, max_multiplier, pos_multiplier,
			   vweight, fweight);
   }
   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


