/*-----------------------------------------------------------------------

File  : ccl_clausesets.c

Author: Stephan Schulz

Contents
 
  Implementation of clause sets based on AVL trees.

Changes

<1> Sun May 10 03:03:20 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_clausesets.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ClauseSetAlloc()
//
//   Allocate an empty clause set that uses setclock for time-keeping.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

ClauseSet_p ClauseSetAlloc()
{
   ClauseSet_p handle;
   int         i;

   handle = ClauseSetCellAlloc();
   
   handle->members = 0;
   handle->anchor = ClauseCellAlloc();
   handle->anchor->literals = NULL;
   handle->anchor->pred = handle->anchor->succ = handle->anchor;
   handle->date = SysDateCreationTime();
   handle->demod_index = NULL;
   for(i=0; i<CLAUSE_MAX_EVALUATIONS; i++)
   {
      handle->eval_indices[i] = NULL;
   }
   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseSetFree()
//
//   Delete a clauseset.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetFree(ClauseSet_p junk)
{
   Clause_p handle;
   
   assert(junk);

   handle = junk->anchor->succ;
   while(handle!=junk->anchor)
   {
      handle = handle->succ;
      handle->pred->set = NULL;
      ClauseFree(handle->pred);
   }
   if(junk->demod_index)
   {
      PDTreeFree(junk->demod_index);
   }
   ClauseCellFree(junk->anchor);
   ClauseSetCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetInsert() 
//
//   Insert a clause into the clauseset.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetInsert(ClauseSet_p set, Clause_p new)
{
   int    i;
   Eval_p handle, test;

   assert(!new->set);

   new->succ = set->anchor->succ;
   new->pred = set->anchor;
   set->anchor->succ->pred = new;
   set->anchor->succ = new;
   new->set = set;
   set->members++;

   i=0;
   for(handle = new->evaluations; handle; handle = handle->next)
   {
      assert(i<CLAUSE_MAX_EVALUATIONS);
      test = EvalTreeInsert(&(set->eval_indices[i]), handle);
      assert(!test);
      i++;
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedInsert()
//
//   Insert a demodulator into the set and the sets index.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedInsert(ClauseSet_p set, Clause_p new)
{
   ClausePos_p pos;

   assert(set->demod_index);
   assert(ClauseIsDemodulator(new));
   
   ClauseSetInsert(set, new);
   pos = ClausePosAlloc();
   pos->clause = new;
   pos->literal = new->literals;
   pos->side = LeftSide;
   PDTreeInsert(set->demod_index, pos);
   if(!EqnIsOriented(new->literals))
   {
      pos = ClausePosAlloc();
      pos->clause = new;
      pos->literal = new->literals;
      pos->side = RightSide;
      PDTreeInsert(set->demod_index, pos);
   }
}

/*-----------------------------------------------------------------------
//
// Function: ClauseSetExtractEntry()
//
//   Remove a clause from a clause set.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetExtractEntry(Clause_p clause)
{
   int     i;
   Eval_p  handle, test;

   assert(clause->set);
   assert(clause->set->members);
   assert(ClauseSetFind(clause->set, clause));
   
   /* printf("ClauseSetExtractEntry: %d\n", (int)clause); */

   i=0;
   for(handle = clause->evaluations; handle; handle = handle->next)
   {
      assert(i<CLAUSE_MAX_EVALUATIONS);
      test = EvalTreeExtractEntry(&(clause->set->eval_indices[i]),
				  handle); 
      assert(test);
      assert(test->object == clause);
      i++;
   }
   clause->pred->succ = clause->succ;
   clause->succ->pred = clause->pred;
   clause->set->members--;
   clause->set = NULL;
   clause->succ = NULL;
   clause->pred = NULL;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedExtractEntry()
//
//   Remove a clause from an indexed clause set.
//
// Global Variables: -
//
// Side Effects    : Changes set
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedExtractEntry(Clause_p clause)
{
   if(clause->set->demod_index)
   {
      assert(ClauseIsDemodulator(clause));
      /* printf("ClauseSetIndexedExtractEntry() left\n"); */
      PDTreeDelete(clause->set->demod_index, clause->literals->lterm,
		   clause); 
      if(!EqnIsOriented(clause->literals))
      {
	 /* printf("ClauseSetIndexedExtractEntry() right\n"); */
	 PDTreeDelete(clause->set->demod_index,
		      clause->literals->rterm, clause);
      }
   }
   ClauseSetExtractEntry(clause);
}  


/*-----------------------------------------------------------------------
//
// Function: ClauseSetExtractFirst()
//
//   Extract the first element of the set and return it. Return NULL
//   if set is empty.
//
// Global Variables: -
//
// Side Effects    : As above
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetExtractFirst(ClauseSet_p set)
{
   Clause_p handle;

   if(ClauseSetEmpty(set))
   {
      return NULL;
   }
   handle = set->anchor->succ;
   assert(handle->set == set);
   ClauseSetExtractEntry(handle);
   
   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetDeleteEntry()
//
//   Delete a clause from the clause set. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetDeleteEntry(Clause_p clause)
{
   ClauseSetExtractEntry(clause);
   ClauseFree(clause);
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetIndexedDeleteEntry()
//
//   Delete a clause from the clause set. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

void ClauseSetIndexedDeleteEntry(Clause_p clause)
{
   ClauseSetIndexedExtractEntry(clause);
   ClauseFree(clause);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFindBest()
//
//   Find the best clause (i.e. the clause with the smallest
//   evaluation).
//
// Global Variables: -
//
// Side Effects    : Changes clause set
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFindBest(ClauseSet_p set, int index)
{
   Clause_p clause;
   Eval_p   evaluation;

   assert(index<CLAUSE_MAX_EVALUATIONS);

   evaluation = EvalTreeFindSmallest(set->eval_indices[index]);
   if(!evaluation)
   {
      assert(set->anchor->succ == set->anchor);
      return NULL;
   }
   assert(evaluation->object);
   clause = evaluation->object;
   assert(clause->set == set);
   return clause;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetPrint()
//
//   Print the clause set to the given stream.
//
// Global Variables: Only for term output
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ClauseSetPrint(FILE* out, ClauseSet_p set, bool fullterms)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClausePrint(out, handle, fullterms);
      fputc('\n', out);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetParseList()
//
//   Parse a list of clauses into the set. Clauses are not
//   evaluated. Returns number of clauses parsed.
//
// Global Variables: Only for parsing of terms.
//
// Side Effects    : Input, memory operations, changes set.
//
/----------------------------------------------------------------------*/

long ClauseSetParseList(Scanner_p in, ClauseSet_p set, TB_p bank)
{
   long     count = 0;
   Clause_p clause;
   
   while(ClauseStartsMaybe(in))
   {
      clause = ClauseParse(in, bank);
      count++;
      ClauseSetInsert(set, clause);
   }
   return count;
}



/*-----------------------------------------------------------------------
//
// Function: ClauseSetMarkMaximalTerms()
//
//   Orient all literals and mark all maximal terms and literals in
//   the set.
//
// Global Variables: -
//
// Side Effects    : Orients and marks literals
//
/----------------------------------------------------------------------*/

void ClauseSetMarkMaximalTerms(OCB_p ocb, ClauseSet_p set)
{
   Clause_p handle;

   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseMarkMaximalTerms(ocb, handle);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseKillChildren()
//
//   Remove and delete children of a clause.
//
// Global Variables: -
//
// Side Effects    : Changes term sets
//
/----------------------------------------------------------------------*/

void ClauseKillChildren(Clause_p clause)
{
   Clause_p current;

   while(clause->children)
   {
      current = clause->children->key;
      assert(clause == current->parent1 || clause == current->parent2);
      /* printf("ClauseKillChildren(%d) -> %d\n",
	 (int)clause,(int)current); */
      ClauseDetachParents(current);
      if(current->set->demod_index)
      {
	 ClauseSetIndexedDeleteEntry(current);
      }
      else
      {
	 ClauseSetDeleteEntry(current);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetRecordTermDates()
//
//   Record the term dates for all clauses in the set (so that
//   rewritten terms can be recognized).
//
// Global Variables: -
//
// Side Effects    : Changes [lr]date in literals
//
/----------------------------------------------------------------------*/

void ClauseSetRecordTermDates(ClauseSet_p set)
{
   Clause_p handle;
   
   DEBUGMARK(2,"ClauseSetRecordTermDates()\n");
   DEBUG(2,ClauseSetPrint(stdout, set,
			  true);)
   for(handle = set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      ClauseRecordTermDates(handle);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFind()
//
//   Given a clause and a clause set, try to find the clause in the
//   set. This is only useful for debugging, as usually clause should
//   know about the set it is in!
//
// Global Variables: -
//
// Side Effects    : - 
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFind(ClauseSet_p set, Clause_p clause)
{
   Clause_p handle, res = NULL;
   
   assert(set);
   assert(clause);

   for(handle = set->anchor->succ; handle!=set->anchor;
       handle=handle->succ)
   {
      if(handle == clause)
      {
	 res = handle;
	 assert(clause->set == set);
	 break;
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetVerifyDemod()
//
//   Return true if pos->clause is in clause set, is a demodulator,
//   and if pos describes a potential maximal side in clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseSetVerifyDemod(ClauseSet_p demods, ClausePos_p pos)
{
   assert(demods);
   assert(pos);
   assert(TermPosIsTopPos(pos->pos));

   if(!ClauseSetFind(demods, pos->clause))
   {
      return false;
   }
   if(!ClauseIsDemodulator(pos->clause))
   {
      return false;
   }
   assert(pos->clause->literals->type != EqUnoriented);
   if((pos->side == RightSide) && EqnIsOriented(pos->clause->literals))
   {
      return false;
   }
   return true;      
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeVerifyIndex()
//
//   Check if all clauses in index are in demod as well.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool PDTreeVerifyIndex(PDTree_p root, ClauseSet_p demods)
{
   PStack_p stack = PStackAlloc(), trav;
   int              i;
   ClausePos_p      pos;
   PTree_p          entry;
   bool             res = true;
   
   if(root)
   {
      PStackPushP(stack, root);
      
      while(!PStackEmpty(stack))
      {
	 root = PStackPopP(stack);
	 
	 if(!root->demod_terms)
	 {
	    for(i=0; i<root->sig_size; i++)
	    {
	       if(root->f_alternatives[i])
	       {
		  PStackPushP(stack, root->f_alternatives[i]);
	       }
	    }
	    for(i=1; i<=root->max_var; i++)
	    {
	       if(PDArrayElementP(root->v_alternatives, i))
	       {
		  PStackPushP(stack, PDArrayElementP(root->v_alternatives, i));
	       }
	    }
	 }
	 else
	 {
	    trav = PTreeTraverseInit(root->demod_terms);
	    while((entry = PTreeTraverseNext(trav)))
	    {
	       pos = entry->key;
	       if(!ClauseSetFind(demods, pos->clause))
	       {
		  res = false;
		  /* printf("Fehler: %d\n", (int)pos->clause);
		  ClausePrint(stdout, pos->clause, true);
		  printf("\n"); */
	       }
	    }
	    PTreeTraverseExit(trav);
	    for(i=0; i<root->sig_size; i++)
	    {
	       assert(!root->f_alternatives[i]);
	    }
	 }
      }
   }
   PStackFree(stack);

   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


