/*-----------------------------------------------------------------------

File  : ccl_eqn.c

Author: Stephan Schulz

Contents
 
  The termpair datatype: Rules, Equations, positive and negative
  literals. 

Changes

<1> Fri Mar 13 17:23:31 MET 1998

-----------------------------------------------------------------------*/

#include "ccl_eqn.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool EqnUseInfix = true;
bool EqnFullEquationalRep = false;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: compare_pos_eqns()
//
//   Compare two positive equations l1=r1 and l2=r2:
//
//     (1) {l1,r1} == {l2,r2}  <==>  (l1=l2 & r1=r2) v (l1=r2 & r1=l2)
//
//     Assume that {l1,r1} =/= {l2,r2}. Then,
//
//     (2) {l1,r1} >> {l2,r2}  <==>  (l1>l2 & l1>r2) v
//                                   (l1>=l2 & r1>=r2) v
//                                   (r1>=l2 & l1>=r2) v
//                                   (r1>l2 & r1>r2)
//
//     (3) {l1,r1} << {l2,r2}  <==>  (l1<l2 & r1<l2) v
//                                   (l1<=l2 & r1<=r2) v
//                                   (r1<=l2 & l1<=r2) v
//                                   (l1<r2 & r1<r2)
//
//     (4) Otherwise, {l1,r1} and {l2,r2} are incomparable.
//
//   Assume that l1>r1 holds. Then (2) and (3) of the above comparison
//   can be simplified in the following way:
//
//     (5) {l1,r1} >> {l2,r2}  <==>  (l1>l2 & l1>r2) v
//                                   (l1>=l2 & r1>=r2)
//
//     (6) {l1,r1} << {l2,r2}  <==>  l1<l2 v
//                                   (l1<=l2 & r1<=r2) v
//                                   (r1<=l2 & l1<=r2) v
//                                   l1<r2
//
//   Assume that l2>r2 holds. Then (2) and (3) of the above comparison
//   can be simplified in the following way:
//
//     (7) {l1,r1} >> {l2,r2}  <==>  l1>l2 v
//                                   (l1>=l2 & r1>=r2) v
//                                   (r1>=l2 & l1>=r2) v
//                                   r1>l2
//
//     (8) {l1,r1} << {l2,r2}  <==>  (l1<l2 & r1<l2) v
//                                   (r1<=l2 & l1<=r2)
//
//   Assume that l1>r1 and l2>r2 hold. Then (1), (2) and (3) of the
//   above comparison can be simplified in the following way:
//
//     (9)  {l1,r1} == {l2,r2}  <==>  l1=l2 & r1=r2
//
//     Assume that {l1,r1} =/= {l2,r2}. Then,
//
//     (10) {l1,r1} >> {l2,r2}  <==>  l1>l2 v (l1=l2 & r1>r2)
//     (11) {l1,r1} << {l2,r2}  <==>  l1<l2 v (l1=l2 & r1<r2)
//
//   The simplified versions (5)-(11) are not yet implemented!
//
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_pos_eqns(OCB_p ocb, Eqn_p eq1, Eqn_p eq2)
{
   CompareResult l1l2, l1r2, r1l2, r1r2;   

   assert(eq1->positive == eq2->positive);

   l1l2 = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   l1r2 = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);
   r1l2 = TOCompare(ocb, eq1->rterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   r1r2 = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);

   if (((l1l2 == to_equal) && (r1r2 == to_equal)) ||
       ((l1r2 == to_equal) && (r1l2 == to_equal))) {
      return to_equal;                                  /* Case (1) */
   }

   if (((l1l2 == to_greater) && (l1r2 == to_greater)) ||
       (((l1l2 == to_greater) || (l1l2 == to_equal)) &&
	((r1r2 == to_greater) || (r1r2 == to_equal))) ||
       (((r1l2 == to_greater) || (r1l2 == to_equal)) &&
	((l1r2 == to_greater) || (l1r2 == to_equal))) ||
       ((r1l2 == to_greater) && (r1r2 == to_greater))) {
      return to_greater;                                /* Case (2) */
   }

   if (((l1l2 == to_lesser) && (r1l2 == to_lesser)) ||
       (((l1l2 == to_lesser) || (l1l2 == to_equal)) &&
	((r1r2 == to_lesser) || (r1r2 == to_equal))) ||
       (((r1l2 == to_lesser) || (r1l2 == to_equal)) &&
	((l1r2 == to_lesser) || (l1r2 == to_equal))) ||
       ((l1r2 == to_lesser) && (r1r2 == to_lesser))) {
      return to_lesser;                                 /* Case (3) */
   }

   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: compare_poseqn_negeqn()
//
//   Compare a positive equations l1=r1 and a negative equation
//   l2=/=r2:
//
//   (1)  {{l1},{r1}} == {{l2,r2}}:   This case is impossible!
//
//   (2)  {{l1},{r1}} >> {{l2,r2}}  <==>  (l1>l2 & l1>r2) v
//                                                     (r1>l2 & r1>r2)
//   (3)  {{l1},{r1}} << {{l2,r2}}  <==>  (l1<=l2 v l1<=r2) &
//                                                   (r1<=l2 v r1<=r2)
//
//   (4)  Otherwise, {{l1},{r1}} and {{l2,r2}} are incomparable.
//
//   Assume that l1>r1 holds. Then the above comparison can be
//   simplified in the following way:
//
//   (5)  {{l1},{r1}} >> {{l2,r2}}  <==>  l1>l2 & l1>r2
//   (6)  {{l1},{r1}} << {{l2,r2}}  <==>  l1<=l2 v l1<=r2
//
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static CompareResult compare_poseqn_negeqn(OCB_p ocb,
					   Eqn_p eq1, Eqn_p eq2)
{
   CompareResult l1l2, l1r2, r1l2, r1r2;

   assert(eq1->positive);
   assert(!eq2->positive);

   l1l2 = TOCompare(ocb, eq1->lterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   if ((EqnIsOriented(eq1)) &&
       ((l1l2 == to_lesser) || (l1l2 ==to_equal))) {
      return to_lesser;                                 /* Case (6) */
   }

   l1r2 = TOCompare(ocb, eq1->lterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);
   if (EqnIsOriented(eq1)) {
      if ((l1r2 == to_lesser) || (l1r2 == to_equal)) {
	 return to_lesser;                              /* Case (6) */
      }
      if ((l1l2 == to_greater) && (l1r2 == to_greater)) {
	 return to_greater;                             /* Case (5) */
      }
      return to_uncomparable;
   }

   assert(!EqnIsOriented(eq1));

   r1l2 = TOCompare(ocb, eq1->rterm, eq2->lterm, DEREF_ALWAYS, DEREF_ALWAYS);
   r1r2 = TOCompare(ocb, eq1->rterm, eq2->rterm, DEREF_ALWAYS, DEREF_ALWAYS);
   if (((l1l2 == to_greater) && (l1r2 == to_greater)) ||
       ((r1l2 == to_greater) && (r1r2 == to_greater))) {
      return to_greater;                                /* Case (2) */
   }
   if ((((l1l2 == to_lesser) || (l1l2 == to_equal)) ||
	((l1r2 == to_lesser) || (l1r2 == to_equal))) &&
       (((r1l2 == to_lesser) || (r1l2 == to_equal)) ||
	((r1r2 == to_lesser) || (r1r2 == to_equal)))) {
      return to_greater;                                /* Case (3) */
   }
   return to_uncomparable;
                        /* Note that the `equal'-case is impossible */
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: EqnAlloc()
//
//   Allocate a literal with the given (shared terms). References for
//   the terms will be added to the bank.
//
// Global Variables: -
//
// Side Effects    : Adds references
//
/----------------------------------------------------------------------*/

Eqn_p EqnAlloc(Term_p lterm, Term_p rterm, TB_p bank,  bool positive)
{
   Eqn_p handle = EqnCellAlloc();

   handle->type =  EqUnoriented;
   handle->positive = positive;
   handle->maximal = false;
   handle->equ_lit = !TBTermEqual(rterm, bank->true_term);
   handle->bank = bank;

   handle->next = NULL;

   TermGetRef(&(handle->lterm), lterm);
   TermGetRef(&(handle->rterm), rterm);

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnFree()
//
//   Free the space taken by an equation, release the term refs and
//   try to delete the terms.
//
// Global Variables: -
//
// Side Effects    : May change term bank, removes references
//
/----------------------------------------------------------------------*/

void EqnFree(Eqn_p junk)
{
   TermReleaseRef(&(junk->lterm));
   TBDelete(junk->bank, junk->lterm);
   TermReleaseRef(&(junk->rterm));
   TBDelete(junk->bank, junk->rterm);
   EqnCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function:  EqnParse()
//
//   Parse an equation and return a pointer to the resulting cell. If
//   EqnUseInfix is true, = and != (and =>, !=>) are used as infix
//   operators, otherwise eq() and ~eq() are used. Non-equational
//   literals are transformed into equational form. Note that
//   EqnStrictEquationalRep is not enforced (it is only used for
//   output). 
//
// Global Variables: EqnUseInfix
//
// Side Effects    : Input, memory management.
//
/----------------------------------------------------------------------*/

Eqn_p EqnParse(Scanner_p in, TB_p bank)
{
   Term_p  lterm;
   Term_p  rterm;
   bool    positive = true;
   EqnType type = EqUnoriented;
   Eqn_p   handle;
   

   if(TestInpTok(in, TildeSign))
   {
      NextToken(in);
      positive = false;
   }
   if(EqnUseInfix && !TestInpId(in, EQUAL_PREDICATE))
   {
      lterm = TBTermParse(in, bank);
      if(TestInpTok(in, Exclamation | EqualSign))
      {
	 if(!positive)
	 {
	    Warning("Negation sign ('~') used in infix equational literal");
	 }
	 if(TestInpTok(in, Exclamation))
	 {
	    NextToken(in);
	    positive = !positive;
	 }
	 AcceptInpTok(in, EqualSign);	 
	 if(TestInpTok(in, GreaterSign))
	 {
	    AcceptInpTokNoSkip(in, GreaterSign);
	    type = EqOriented;
	 }
	 rterm = TBTermParse(in, bank);
      }
      else
      {
	 rterm = bank->true_term; /* Non-Equational literal */
      }
   }
   else
   {
      if(TestInpId(in, EQUAL_PREDICATE))
      {	 
	 NextToken(in);
	 AcceptInpTok(in, OpenBracket);
	 lterm = TBTermParse(in, bank);
	 AcceptInpTok(in, Comma);
	 rterm = TBTermParse(in, bank);
	 AcceptInpTok(in, CloseBracket);
      }
      else
      {
	 lterm = TBTermParse(in, bank);
	 rterm = bank->true_term; /* Non-Equational literal */
      }
   }
   handle = EqnAlloc(lterm, rterm, bank, positive);
   handle->type = type;
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnPrint()
//
//   Print an equation. If negated is true, print the negated
//   equation.
//
// Global Variables: EqnUseInfix, EqnFullEquationalRep
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EqnPrint(FILE* out, Eqn_p eq, bool negated,  bool fullterms)
{
   if(EqnUseInfix && 
      (EqnFullEquationalRep || 
       eq->rterm!=eq->bank->true_term ||
       eq->lterm==eq->bank->true_term))
   {
      TBPrintTerm(out, eq->bank, eq->lterm, fullterms);
      if(XOR(!eq->positive, negated))
      {
	 fputc('!', out);
      }
      if(eq->type == EqOriented)
      {
	 fprintf(out, "=>");
      }
      else
      {
	 fprintf(out, "=");
      }
      TBPrintTerm(out, eq->bank, eq->rterm, fullterms);
   }
   else
   {
      if(XOR(!eq->positive, negated))
      {
	 fputc('~', out);
      }
      if((eq->rterm!=eq->bank->true_term) ||
	 EqnFullEquationalRep ||
	 eq->lterm==eq->bank->true_term)
      {
	 fprintf(out, EQUAL_PREDICATE"(");
	 TBPrintTerm(out, eq->bank, eq->lterm, fullterms);
	 fprintf(out, ", ");
	 TBPrintTerm(out, eq->bank, eq->rterm, fullterms);
	 fputc(')', out);
      }
      else
      {
	 TBPrintTerm(out, eq->bank, eq->lterm, fullterms);
      }
   }
#ifdef NEVER_DEFINED
   fprintf(out, "/* %s, %s */", eq->maximal? "Max" : "Dom",
	   eq->equ_lit? " = " : " ~ ");
#endif
}


/*-----------------------------------------------------------------------
//
// Function: EqnSwapSidesSimple()
//
//   Exchange the two sides of the equation. This will lead to
//   inconsistent states if not used carefully (i.e. only temporary or
//   via EqnSwapSides() (which takes care of the attached strings).
//
// Global Variables: -
//
// Side Effects    : Exchanges sides, !!! Causes inconsistent state!!!
//
/----------------------------------------------------------------------*/

void EqnSwapSidesSimple(Eqn_p eq)
{
   Term_p term;

   term = eq->lterm;
   eq->lterm = eq->rterm;
   eq->rterm = term;
}


/*-----------------------------------------------------------------------
//
// Function: EqnSwapSides()
//
//   Exchange the two sides of an equation. Will update type and
//   references.  
//
// Global Variables: -
//
// Side Effects    : Exchanges the sides, changes type.
//
/----------------------------------------------------------------------*/

void EqnSwapSides(Eqn_p eq)
{
   if(eq->type == EqOriented)
   {
      eq->type = EqUnoriented;
   }
   TermReleaseRef(&(eq->lterm));
   TermReleaseRef(&(eq->rterm));

   EqnSwapSidesSimple(eq);

   TermGetRef(&(eq->lterm), eq->lterm);
   TermGetRef(&(eq->rterm), eq->rterm);
}


/*-----------------------------------------------------------------------
//
// Function: EqnRecordTermDates()
//
//   Set eq->[lr]date to eq->[lr]term->tb_date.
//
// Global Variables: -
//
// Side Effects    : As above
//
/----------------------------------------------------------------------*/

void EqnRecordTermDates(Eqn_p eq)
{
   assert(eq);
   assert(eq->lterm);
   assert(eq->rterm);

   eq->ldate = eq->lterm->tb_date;
   eq->rdate = eq->rterm->tb_date;
}


/*-----------------------------------------------------------------------
//
// Function:  EqnHasChanged()
//
//   Return MaxSide if a potential maximal side of eq has been
//   rewritten, MinSide if a non-maximal side has been rewritten,
//   and NoSide if equation is unchanged.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

EqnSide EqnHasChanged(Eqn_p eq)
{
   if(SysDateCompare(eq->ldate, eq->lterm->tb_date)!=DateEqual)
   {
      return MaxSide;
   }
   if(SysDateCompare(eq->rdate, eq->rterm->tb_date)!=DateEqual)
   {
      if(EqnIsOriented(eq))
      {
	 return MinSide;
      }
      return MaxSide;
   }
   return NoSide;
}


/*-----------------------------------------------------------------------
//
// Function: EqnSideHasChanged()
//
//   Query wether a given side has changed.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

bool EqnSideHasChanged(Eqn_p eq, EqnSide side)
{
   bool res = false;
   
   switch(side)
   {
   case LeftSide:
	 if(SysDateCompare(eq->ldate, eq->lterm->tb_date)!=DateEqual)
	 {
	    res = true;
	 }
	 break;
   case RightSide:
	 if(SysDateCompare(eq->rdate, eq->rterm->tb_date)!=DateEqual)
	 {
	    res = true;
	 }
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnUpdateOrientation()
//
//   If maximal side of an equation has been rewritten, change type to
//   EqUnoriented.
//
// Global Variables: -
//
// Side Effects    : Changes eq->type
//
/----------------------------------------------------------------------*/

bool EqnUpdateOrientation(Eqn_p eq)
{
   if(EqnHasChanged(eq)==LeftSide)
   {
      eq->type = EqUnoriented;
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EqnCopy()
//
//   Create a copy of eq with terms from bank. Does not copy the
//   next pointer. If keeporient is true, the copy will have the same
//   type as the old equation, otherwise, if the original equation was
//   oriented, the copy is so as well, otherwise it is
//   unoriented. This is only correct if the original equation is at
//   most modified by instantiation! 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eqn_p EqnCopy(Eqn_p eq, TB_p bank, bool keeporient)
{
   Eqn_p  handle;
   Term_p lterm, rterm;

   lterm = TBInsert(bank, eq->lterm, DEREF_ALWAYS);
   rterm = TBInsert(bank, eq->rterm, DEREF_ALWAYS);

   handle = EqnAlloc(lterm, rterm, bank, eq->positive);

   handle->maximal = eq->maximal;
   if(keeporient)
   {
      handle->type = eq->type;
   }
   else
   {
      handle->type = eq->type == EqOriented?EqOriented:EqUnoriented;
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnEqualDirected()
//
//   Test wether two equations are identical (modulo EqualTest). Does
//   not check for positive/negative, evaluates only terms
//   (i.e. treats equations as unsigned term pairs).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnEqualDirected(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest)
{
   bool res = EqualTest(eq1->lterm, eq2->lterm);
   
   if(!res)
   {
      return res;
   }
   res = EqualTest(eq1->rterm, eq2->rterm);
   
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnEqual()
//
//   Test wether two equations are equivalen (modulo EqualTest and
//   commutativity). Treats equations as unsigned term sets.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnEqual(Eqn_p eq1,  Eqn_p eq2, TermEqualTestFun EqualTest)
{
   bool res;

   res = EqnEqualDirected(eq1, eq2, EqualTest);
   
   if(res || ((eq1->type == EqOriented) && (eq2->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(eq2);
   res = EqnEqualDirected(eq1, eq2, EqualTest);
   EqnSwapSidesSimple(eq2);
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnSubsumeDirected()
//
//   Test wether an equation subsumes another one. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodified. Don't deal with commutativity of
//   equality. 
//
// Global Variables: -
//
// Side Effects    : -
//
//----------------------------------------------------------------------*/

bool EqnSubsumeDirected(Eqn_p subsumer, Eqn_p subsumed, Subst_p
			 subst, TermEqualTestFun EqualTest)
{
   PStackPointer backtrack = PStackGetSP(subst);
   bool res;

   res = SubstComputeMatch(subsumer->lterm, subsumed->lterm, subst,
			   EqualTest);
   if(res)
   {
      res = SubstComputeMatch(subsumer->rterm,
			      subsumed->rterm, subst,
			      EqualTest);
   }
   if(!res)
   {
      SubstBacktrackToPos(subst, backtrack);
   }
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: EqnSubsume()
//
//   Test wether an equation subsumes another one. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodifies. Equations are treated as 2-sets of
//   terms unless both are oriented.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnSubsume(Eqn_p subsumer, Eqn_p subsumed, Subst_p subst,
		TermEqualTestFun EqualTest)
{
   bool   res = EqnSubsumeDirected(subsumer, subsumed, subst, EqualTest);
   
   if(res || 
      ((subsumer->type == EqOriented) && 
       (subsumed->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(subsumer);
   res = EqnSubsumeDirected(subsumer, subsumed, subst, EqualTest);
   EqnSwapSidesSimple(subsumer);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnSubsumeP()
//
//   Test wether subsumer subsumes subsumed, undo all side effects.
//
// Global Variables: -
//
// Side Effects    : - (I hope)
//
/----------------------------------------------------------------------*/

bool EqnSubsumeP(Eqn_p subsumer, Eqn_p subsumed, TermEqualTestFun
		 EqualTest) 
{
   Subst_p subst = SubstAlloc();
   bool    res = EqnSubsume(subsumer, subsumed, subst, EqualTest);

   SubstDelete(subst);
   
   return res;   
}


/*-----------------------------------------------------------------------
//
// Function: LiteralSubsumeP()
//
//   Return true if subsumer subsumes subsumed, false
///  otherwise. Checks signs!
//
// Global Variables: -
//
// Side Effects    : Only temporary
//
/----------------------------------------------------------------------*/

bool LiteralSubsumeP(Eqn_p subsumer, Eqn_p subsumed, TermEqualTestFun
		     EqualTest) 
{
   if(subsumer->positive == subsumed->positive)
   {
      return EqnSubsumeP(subsumer, subsumed, EqualTest);
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EqnUnifyDirected()
//
//   Test wether two equations can be unified. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodified. Don't deal with commutativity of
//   equality. 
//
// Global Variables: -
//
// Side Effects    : -
//
//----------------------------------------------------------------------*/

bool EqnUnifyDirected(Eqn_p eq1, Eqn_p eq2, Subst_p subst)
{
   PStackPointer backtrack = PStackGetSP(subst);
   bool res;

   res = SubstComputeMgu(eq1->lterm, eq2->lterm, subst);
   if(res)
   {
      res = SubstComputeMgu(eq1->rterm,
			      eq2->rterm, subst);
   }
   if(!res)
   {
      SubstBacktrackToPos(subst, backtrack);
   }
   return res;
}




/*-----------------------------------------------------------------------
//
// Function: EqnUnify()
//
//   Test wether two equations are unifyable. If yes, return true
//   and extend subst to give the substitution, otherwise just return
//   true and let subst unmodifies. Equations are treated as 2-sets of
//   terms unless both are oriented.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnUnify(Eqn_p eq1, Eqn_p eq2, Subst_p subst)
{
   bool   res = EqnUnifyDirected(eq1, eq2, subst);
   
   if(res || 
      ((eq1->type == EqOriented) && 
       (eq2->type == EqOriented)))
   {
      return res;
   }
   EqnSwapSidesSimple(eq1);
   res = EqnUnifyDirected(eq1, eq2, subst);
   EqnSwapSidesSimple(eq1);
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: EqnUnifyP()
//
//   Test wether two equations are unifiable, undo all side effects.
//
// Global Variables: -
//
// Side Effects    : - (I hope)
//
/----------------------------------------------------------------------*/

bool  EqnUnifyP(Eqn_p eq1, Eqn_p eq2)
{
   Subst_p subst = SubstAlloc();
   bool    res = EqnUnify(eq1, eq2, subst);

   SubstDelete(subst);
   
   return res;   
}


/*-----------------------------------------------------------------------
//
// Function: EqnOrient()
//
//   Orient an equation. Return true, if sides are exchanged, false
//   otherwise. 
//
// Global Variables: -
//
// Side Effects    : May change term references
//
/----------------------------------------------------------------------*/

bool EqnOrient(OCB_p ocb, Eqn_p eq)
{
   CompareResult relation = to_uncomparable;
   bool res = false;

   assert(eq->type != EqOriented);

   if(TBTermEqual(eq->lterm, eq->rterm))
   {
      relation = to_equal;
   }
   else if(TBTermEqual(eq->lterm, eq->bank->true_term))
   {
      relation = to_lesser;
   }
   else if(TBTermEqual(eq->rterm, eq->bank->true_term))
   {
      relation = to_greater;
   }
   else
   {
      /* printf("EqnOrient: ");
      TermPrint(stdout, eq->lterm, eq->bank->sig, DEREF_ALWAYS);
      printf(" # ");
      TermPrint(stdout, eq->rterm, eq->bank->sig, DEREF_ALWAYS);      
      printf("\n"); */
      relation = TOCompare(ocb, eq->lterm, eq->rterm, DEREF_ALWAYS,
			   DEREF_ALWAYS);
   }
   
   switch(relation)
   {
   case to_uncomparable:
   case to_equal:
	 eq->type = EqUnorientable;
	 break;
   case to_greater:
	 eq->type = EqOriented;
	 break;
   case to_lesser:
	 EqnSwapSides(eq);
	 res = true;
	 eq->type = EqOriented;
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnCompare()
//
//   Compare two equations (as multisets of terms) and return the
//   result.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult EqnCompare(OCB_p ocb, Eqn_p eq1, Eqn_p eq2)
{
   CompareResult res;

   res = compare_pos_eqns(ocb, eq1, eq2);
   
  return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnGreater()
//
//   Return true if eq1 is greater than eq2, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnGreater(OCB_p ocb, Eqn_p eq1, Eqn_p eq2)
{
   bool res;

   res = EqnCompare(ocb, eq1, eq2) == to_greater;

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: LiteralCompare()
//
//   Compare two signed literals L1 and L2:
//
//     L1 > L2  <==>  rep(L1) >> rep(L2)
//
//     where >> stands for the extension of an ordering on terms to
//     multisets of (multisets of) terms
//     and rep is a function mapping (negative or positive) equations
//     to multisets (of multisets) of terms in the following way:
//
//     o rep(s=t) = rep(s=/=t) = {s,t}
//       if L1 and L2 both are positive or both are negative equations
//
//     o rep(s=t) = {{s},{t}}  and  rep(s=/=t) = {{s,t}}
//       otherwise
//
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult LiteralCompare(OCB_p ocb, Eqn_p eq1, Eqn_p eq2)
{
   if ((eq1->positive && eq2->positive) ||
       (!eq1->positive && !eq2->positive)) {
      return compare_pos_eqns(ocb, eq1, eq2);
   }
   
   /* Exactly one of the equations is negative */
   if (eq1->positive) {
      return compare_poseqn_negeqn(ocb, eq1, eq2);
   }
   
   /* eq1 is negative and eq2 is positive */
   return OCBPrecedenceInverse(compare_poseqn_negeqn(ocb, eq2, eq1));
}



/*-----------------------------------------------------------------------
//
// Function: LiteralGreater()
//
//   Return true if eq1 is greater than eq2, takes as (signed)
//   literals, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool LiteralGreater(OCB_p ocb, Eqn_p eq1, Eqn_p eq2)
{
   bool res;
   
   res = LiteralCompare(ocb, eq1, eq2) == to_greater;
   
   return res;   
}

/*-----------------------------------------------------------------------
//
// Function: SubstNormEqn()
//
//   Instantiate all variables in eq with normed variables.
//
// Global Variables: -
//
// Side Effects    : By SubstNormTerm()
//
/----------------------------------------------------------------------*/

FunCode SubstNormEqn(Eqn_p eq, Subst_p subst, VarBank_p vars)
{
  FunCode res;

   SubstNormTerm(eq->lterm, subst, vars);
   res = SubstNormTerm(eq->rterm, subst, vars);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnWeight()
//
//   Compute the weight of an equation. Weights of potentially maximal
//   sides are multiplied by max_multiplier.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

double EqnWeight(Eqn_p eq, double max_multiplier, long vweight, long
		  fweight)
{
   double res;

   if(EqnIsOriented(eq))
   {
      res = (double)TermWeight(eq->rterm, vweight, fweight);
   }
   else
   {
      res = (double)TermWeight(eq->rterm, vweight, fweight) *
	 max_multiplier;
   }
   res += ((double)TermWeight(eq->lterm, vweight, fweight) * max_multiplier);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: LiteralWeight()
//
//   Return weight of a literal. max_multipler is applied to maximal
//   sides _and_ maximal literals, pos_multiplier is applied to
//   positive literals
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double LiteralWeight(Eqn_p eq, double max_multiplier, double
		      pos_multiplier, long vweight, long fweight)
{
   double res;

   res = EqnWeight(eq, max_multiplier, vweight, fweight);
      
   if(eq->maximal)
   {
      res = res*max_multiplier;
   }
   if(eq->positive)
   {
      res = res*pos_multiplier;
   }

   return res;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


