/*-----------------------------------------------------------------------

File  : ccl_eqnlist.c

Author: Stephan Schulz

Contents
 
   Functions for dealing with (singly linked) lists of equations as
   used in clauses

Changes

<1> Fri Apr 10 21:11:18 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_eqnlist.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: eqn_list_find_last()
//
//   Find the last EqnRef in *list (may be list itself).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

EqnRef eqn_list_find_last(EqnRef list)
{
   while(*list)
   {
      list = &((*list)->next);
   }
   return list;   
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: EqnListFree()
//
//   Deallocate the list.
//
// Global Variables: -
//
// Side Effects    : Memory Management
//
/----------------------------------------------------------------------*/

void EqnListFree(Eqn_p list)
{
   Eqn_p handle;

   while(list)
   {
      handle = list;
      list = list->next;
      EqnFree(handle);
   }
}


/*-----------------------------------------------------------------------
//
// Function: EqnListLength()
//
//   Return number of equations in the list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long EqnListLength(Eqn_p list)
{
   long i=0;

   while(list)
   {
      i++;
      list = list->next;
   }
   return i;
}



/*-----------------------------------------------------------------------
//
// Function: EqnListExtractElement()
//
//   Take the given element out of the list and return a pointer to
//   it.
//
// Global Variables: -
//
// Side Effects    : Changes the list
//
/----------------------------------------------------------------------*/

Eqn_p EqnListExtractElement(EqnRef element)
{
   Eqn_p handle = *element;
   assert(handle);

   *element = handle->next;

   handle->next = NULL;
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListDeleteElement()
//
//   Delete the given element from the list.
//
// Global Variables: -
//
// Side Effects    : Via EqnListExtractElement() and EqnFree()
//
/----------------------------------------------------------------------*/

void EqnListDeleteElement(EqnRef element)
{
   Eqn_p handle;

   handle = EqnListExtractElement(element);

   EqnFree(handle);
}


/*-----------------------------------------------------------------------
//
// Function: EqnListInsertElement()
//
//   Insert the element at the position defined by pos.
//
// Global Variables: -
//
// Side Effects    : Changes the list
//
/----------------------------------------------------------------------*/

void EqnListInsertElement(EqnRef pos, Eqn_p element)
{
   element->next = *pos;
   *pos = element;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListAppend()
//
//   Append newpart at the end of *list.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Eqn_p EqnListAppend(EqnRef list, Eqn_p newpart)
{
   EqnRef result = list;
   
   list = eqn_list_find_last(list);
   assert(!(*list));
   *list = newpart;

   return *result;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListCopy()
//
//   Return a copy of the given list, with new terms from the term
//   bank. Instantiated terms are copied as instantiations.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eqn_p EqnListCopy(Eqn_p list, TB_p bank, bool keeporient)
{
   Eqn_p  newlist = NULL;
   EqnRef insert = &newlist;

   while(list)
   {
      *insert = EqnCopy(list, bank, keeporient);
      insert = &((*insert)->next);
      list = list->next;
   }
   *insert = NULL;
   
   return newlist;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListCopyExcept()
//
//   Return a copy of the given list, except for the equation given in
//   except, with new terms from the term bank. Instantiated terms are
//   copied as instantiations. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eqn_p EqnListCopyExcept(Eqn_p list, Eqn_p except, TB_p bank, bool
			keeporient)  
{
   Eqn_p  newlist = NULL;
   EqnRef insert = &newlist;

   while(list)
   {
      if(list != except)
      {
	 *insert = EqnCopy(list, bank, keeporient);
	 insert = &((*insert)->next);
      }
      list = list->next;
   }
   *insert = NULL;
   
   return newlist;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListNegateEqns()
//
//   Negate all signs in the list.
//
// Global Variables: -
//
// Side Effects    : Changes signs in the list
//
/----------------------------------------------------------------------*/

Eqn_p EqnListNegateEqns(Eqn_p list)
{
   Eqn_p handle = list;

   while(handle)
   {
      handle->positive = !(handle->positive);
      handle = handle->next;
   }
   return list;
}



/*-----------------------------------------------------------------------
//
// Function: EqnListRemoveDuplicates()
//
//   Remove all but one copy of identical (modulo commutativity)
//   elements from the list. Return new lenght of the list.
//
// Global Variables: -
//
// Side Effects    : Changes the list
//
/----------------------------------------------------------------------*/

int EqnListRemoveDuplicates(Eqn_p list, TermEqualTestFun EqualTest)
{
   EqnRef handle;
   int    len = 0;

   while(list)
   {
      len++;
      handle = &(list->next);
      while(*handle)  
      {
	 if(LiteralEqual(*handle, list, EqualTest))
	 {
	    EqnListDeleteElement(handle);
	 }
	 else
	 {
	    handle = &((*handle)->next);
	 }
      }
      list = list->next;
   }
   return len;
}


/*-----------------------------------------------------------------------
//
// Function:  EqnListRemoveResolved()
//
//   Remove trivial negative equations.
//
// Global Variables: -
//
// Side Effects    : Changes the list.
//
/----------------------------------------------------------------------*/

int EqnListRemoveResolved(EqnRef list, TermEqualTestFun EqualTest)
{
   int len = 0;

   while(*list)
   {
      if( (! ((*list)->positive)) && EqnIsTrivial(*list, EqualTest))
      {
	 EqnListDeleteElement(list);
      }
      else
      {
	 len++;
	 list = &((*list)->next);
      }
   }
   return len;
}

/*-----------------------------------------------------------------------
//
// Function: EqnListIsTautologic()
//
//   Return true if the list contains two equal literals with
//   opposing signs or a positive trivial equation. Assumes that all
//   terms are shared. 
//
// Global Variables: - 
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnListIsTautologic(Eqn_p list)
{
   Eqn_p handle;
   
   while(list)
   {
      if(list->positive 
	 && TBTermEqual(list->lterm, list->rterm))
      {
	 return true;
      }	    
      for(handle = list->next; handle; handle = handle->next)
      {
	 if(handle->positive != list->positive)
	 {
	    if(EqnEqual(handle, list, TBTermEqual))
	    {
	       return true;
	    }
	 }
      }
      list = list->next;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListIsGround()
//
//   Return true if all equations in list are true, false otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnListIsGround(Eqn_p list)
{
   bool res = true;
   
   while(list)
   {
      if(!EqnIsGround(list))
      {
	 res = false;
	 break;
      }
      list = list->next;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListResetOrientStatus()
//
//   Set all equations in list to status.
//
// Global Variables: -
//
// Side Effects    : Changes orientation status
//
/----------------------------------------------------------------------*/

void EqnListResetOrientStatus(Eqn_p list, EqnType status)
{
   Eqn_p handle;

   for(handle = list; handle; handle = handle->next)
   {
      handle->type = status;
   }
}



/*-----------------------------------------------------------------------
//
// Function: EqnListOrient()
//
//   Orient all the equations in list. Equations already oriented are
//   not reoriented! Return number of swapped equations. 
//
// Global Variables: -
//
// Side Effects    : Orients equations
//
/----------------------------------------------------------------------*/

long EqnListOrient(OCB_p ocb, Eqn_p list)
{
   long res = 0;
   Eqn_p handle;

   for(handle = list; handle; handle = handle->next)
   {
      if((handle->type!=EqOriented) && EqnOrient(ocb, handle))
      {
	 res++;
      }
      
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: EqnListUpdateOrientation()
//
//   For all equations update the type to EqUnoriented if maximal side
//   has been rewritten. Returns number of changed equations.
//
// Global Variables: -
//
// Side Effects    : By EqnUpdateOrientation()
//
/----------------------------------------------------------------------*/

long EqnListUpdateOrientation(Eqn_p list)
{
   long count = 0;
   
   while(list)
   {
      if(EqnUpdateOrientation(list))
      {
	 count++;
      }
      list = list->next;
   }
   return count;   
}


/*-----------------------------------------------------------------------
//
// Function: EqnListMaximalLiterals()
//
//   Determine for each literal wether it is maximal or not. Returns
//   number of maximal literals. This may need some work, depending on
//   the calculus...
//
// Global Variables: -
//
// Side Effects    : Sets the maximal flags of literals.
//
/----------------------------------------------------------------------*/

long EqnListMaximalLiterals(OCB_p ocb, Eqn_p list)
{
   Eqn_p handle, stepper;
   CompareResult cmp;
   long res = 0;

   for(handle = list; handle; handle = handle->next)
   {
      handle->was_max = handle->maximal;
      handle->maximal = true;
      res++;
   }
   for(handle = list; handle; handle = handle->next)
   {
      for(stepper = handle->next; stepper; stepper = stepper->next)
      {
	 if(stepper->maximal && handle->maximal)
	 {
	    cmp = LiteralCompare(ocb, handle, stepper);
	    if(cmp == to_greater)
	    {
	       stepper->maximal = false;
	       res--;
	    }
	    else if(cmp == to_lesser)
	    {
	       handle->maximal = false;
	       res--;
	    }	  
	 }
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListEqnIsMaximal()
//
//   Return true if eqn is maximal with respect to list (i.e. if there
//   are no equations that dominate it), false otherwise. As above,
//   details of this may need change if the calculus changes.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool EqnListEqnIsMaximal(OCB_p ocb, Eqn_p list, Eqn_p eqn)
{
   Eqn_p handle;
   bool res = true;
   
   for(handle=list; handle; handle = handle->next)
   {
      if(handle!=eqn && handle->maximal)
      {
	 if(LiteralGreater(ocb, handle, eqn))
	 {
	    res = false;
	    break;
	 }
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnListMarkRestrictedTerms()
//
//   Mark strict maximal terms of positive maximal literals with
//   TPRestricted. The list is supposed to be oriented and have
//   maximal terms marked.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void EqnListMarkRestrictedTerms(Eqn_p list)
{
   while(list)
   {
      if(list->positive && list->maximal && 
	 (list->type == EqOriented))
      {
	 DEBUGMARK(4, "Marking ");
	 DEBUG(4, TermPrint(stdout, list->lterm, list->bank->sig, DEREF_NEVER););
	 DEBUGOUT(4, "\n");
	 TBRefSetProp(list->bank, &(list->lterm), TPRestricted); 
      }
      else
      {
	 TBRefDelProp(list->bank, &(list->lterm), TPRestricted); 	 
      }
      list = list->next;
   }
}


/*-----------------------------------------------------------------------
//
// Function: EqnListPrint()
//
//   Print the list. Separate elements with the given separator
//   (usually "," oder ";"). If negated is true, negate equations
//   before printing (to allow for easy printing of clauses in
//   implicational form). 
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EqnListPrint(FILE* out, Eqn_p list, char* sep, 
		  bool negated, bool fullterms)
{
   Eqn_p handle = list;

   if(handle)
   {
      EqnPrint(out, handle, negated, fullterms);
   
      while(handle->next)
      {
	 handle = handle->next;
	 fprintf(out, sep);
	 EqnPrint(out, handle, negated, fullterms);
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: EqnListParse()
//
//   Parse a list of equations, separated by Tokens of type sep.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations, by EqnParse()
//
/----------------------------------------------------------------------*/

Eqn_p EqnListParse(Scanner_p in, TB_p bank, TokenType sep)
{
   Eqn_p handle = NULL, 
         list = NULL;

   if(TestInpTok(in, TermStartToken|TildeSign))
   {
      list = EqnParse(in, bank);
      handle = list;
      while(TestInpTok(in,sep))
      {
	 NextToken(in);
	 handle->next = EqnParse(in, bank);
	 handle = handle->next;
      }
   }
   return list;
}


/*-----------------------------------------------------------------------
//
// Function: NormSubstEqnListExcept()
//
//   Instantiate all variables in eqnlist (except for terms from
//   except)  with fresh variables from vars. 
//
// Global Variables: -
//
// Side Effects    : Builds substitution, instantiates variables
//
/----------------------------------------------------------------------*/

FunCode NormSubstEqnListExcept(Eqn_p list, Eqn_p except, Subst_p
			       subst, VarBank_p vars)
{
   Eqn_p   handle;
   FunCode res = VarBankGetVCount(vars);

   for(handle = list; handle; handle = handle->next)
   {
      if(handle!= except)
      {
	 res = SubstNormEqn(handle, subst, vars);
      }
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


