/*-----------------------------------------------------------------------

File  : ccl_evaluations.c

Author: Stephan Schulz

Contents
 
  Data type for representing evaluations of clauses.

Changes

<1> Thu Apr  9 02:00:51 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_evaluations.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

unsigned long EvaluationCounter = 0;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: rotate_left()
//
//   Rotate the unbalanced tree at *update left, adjusting weights
//   accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left(Eval_p *update)
{
   Eval_p store;
   int b1,b2;

   VERBOUT10("Rotate tree left\n");

   store = (*update)->rson;
   (*update)->rson = (*update)->rson->lson;
   store->lson = *update;
   *update = store;
   
   b1 = (*update)->balance;
   b2 = ((*update)->lson)->balance;
   
   switch(b1)
   {
   case -1:
	 assert(b2 == 1);
	 (*update)->balance = -2;
	 ((*update)->lson)->balance = 0;
	 break;
   case 0:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 0;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 1;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 1:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = -1;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = 0;
	    ((*update)->lson)->balance = 0;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 2:
	 assert(b2 == 2);
	 (*update)->balance = 0;
	 ((*update)->lson)->balance = -1;
	 break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right()
//
//   Symmetrical case to rotate_left()
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right(Eval_p *update)
{
   Eval_p store;
   int b1, b2;

   VERBOUT10("Rotate tree right\n");

   store = (*update)->lson;
   (*update)->lson = (*update)->lson->rson;
   store->rson = *update;
   *update = store;

   b1 = (*update)->balance;
   b2 = ((*update)->rson)->balance;

   switch(b1)
   {
   case 1:
         assert(b2 == -1);
         (*update)->balance = 2;
         ((*update)->rson)->balance = 0;
         break;
   case 0:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 0;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = -1;
         }
         else
         {
            assert(false);
         }
         break;
   case -1:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 1;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 0;
            ((*update)->rson)->balance = 0;
         }
         else
         {
            assert(false);
         }
         break;
   case -2:
         assert(b2 == -2);
         (*update)->balance = 0;
         ((*update)->rson)->balance = 1;
         break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right_left()
//
//   Perform the right-left rotation to balance a tree. More exactly:
//   Rotate the subtree at (*update)->rson right and the tree at
//   *update left, and adjust the balances accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right_left(Eval_p *update)
{
   VERBOUT10("Rotate tree right left\n");

   rotate_right(&((*update)->rson));
   rotate_left(update);
}


/*-----------------------------------------------------------------------
//
// Function: rotate_left_right()
//
//   Perform the left-right rotation to balance a tree. See
//   rotate_right_left() 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left_right(Eval_p *update)
{
   VERBOUT10("Rotate tree left right\n");

   rotate_left(&((*update)->lson));
   rotate_right(update);
}

/*-----------------------------------------------------------------------
//
// Function: rebalance_tree()
//
//   Rebalance the tree at the passed position. Assumes that the
//   weights in the tree are correct. Returns true if the tree depth
//   has changed.
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool rebalance_tree(Eval_p *update)
{
   int b2 = 0;
   bool shrunk;

   if((*update)->balance == 2)
   {
      b2 = (*update)->rson->balance;
      assert((*update)->rson);
      if((*update)->rson->balance == -1)
      {
	 rotate_right_left(update);
      }
      else
      {
	 rotate_left(update);
      }
   }
   else if((*update)->balance == -2)
   {
      b2 = (*update)->lson->balance;
      assert((*update)->lson);
      if((*update)->lson->balance == 1)
      {
	 rotate_left_right(update);
      }
      else
      {
	 rotate_right(update);
      }
   }

   shrunk = (b2 != 0);
   
   return shrunk;
}


/*-----------------------------------------------------------------------
//
// Function: find_min_cell()
//
//   Find the position of the pointer to the minimal cell in the tree
//   anchored at root. See find_max_cell().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static Eval_p *find_min_cell(Eval_p *root, bool remove_mode)
{
   if(!*root)
   {
      return NULL;
   }

   while((*root)->lson)
   {
      (*root)->passed = -1;
      root = &((*root)->lson);
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: find_cell_pos()
//
//   Starting at root, search the tree for an entry with 'key', setting
//   'passed' values along the search papth and keeping *update
//   pointing to a potential update position. If update_on_zero is
//   true, set update to the lowest cell with balance 0, otherwise to
///   the lowest cell with balance != 0. Return the pointer to
//   where the entry in anchored, or the position where it would be
//   inserted if no such entry exists.
//
// Global Variables: -
//
// Side Effects    : Sets 'passed' values and *update
//
/----------------------------------------------------------------------*/

static Eval_p *find_cell_pos(Eval_p *root, Eval_p key, 
				Eval_p **update,
				bool remove_mode)
{
   int cmpres;

   while(*root)
   {
      cmpres = EvalCompare(key, *root);
      if(cmpres == 0)
      {
	 return root;
      }
      
      if(!remove_mode && (*root)->balance)
      {
	 *update = root;
      }
      if(cmpres > 0)
      {
	 (*root)->passed = 1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == -1))
	 {
	    *update = root;
	 }
	 root = &((*root)->rson);
      }
      else
      {
	 (*root)->passed = -1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == 1))
	 {
	    *update = root;
	 }
	 root = &((*root)->lson);
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: isolate_and_mark_cell()
//
//   Find the cell with the key and move it to a position in the tree
//   where it can be deleted (i.e. a position where it has only one
//   successor). Set the passed-values leading to the new position,
//   and set the passed-value of the relevant cell to 0 so that it can
//   be found without (expensive) key comparisons.
//
// Global Variables: -
//
// Side Effects    : Changes tree and some cell entries.
//
/----------------------------------------------------------------------*/

static Eval_p isolate_and_mark_cell(Eval_p *root, Eval_p key)
{
   Eval_p *elpos = root;
   Eval_p *replace = NULL;
   Eval_p *dummy;
   Eval_p element, repl, t_lson, t_rson;
   long      help;
   
   elpos = find_cell_pos(root, key, &dummy, true);
   if(!(*elpos))
   {
      return NULL;
   }
   element = *elpos;
   
   if(element->rson&&element->lson)
   {
      replace = find_min_cell(&(element->rson), true);
      repl = *replace;
      
      t_lson = repl->lson;
      t_rson = repl->rson;
      repl->lson = element->lson;
      
      if(repl == element->rson)
      {
	 repl->rson =  element;
      }
      else
      {
	 repl->rson = element->rson;
	 *replace = element;
      }
      
      *elpos = repl;
      element->lson = t_lson;
      element->rson = t_rson;
      
      repl->passed = 1;
      element->passed = 0;
      help = repl->balance;
      repl->balance = element->balance;
      element->balance = help; 
   }
   else 
   {
      element->passed = 0;
   }
   
   return element;
}


/*-----------------------------------------------------------------------
//
// Function: extract_marked_cell()
//
//   Descend the tree guided by the passed-entries until the marked
//   cell (passed == 0) has been found, extract it, store it in *cell,
//   and ascend the tree, rebalancing as necessary. Return true if the
//   tree height has changes. The tree has to be preprocessed by
//   isolate_and_mark_cell().
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool extract_marked_cell(Eval_p *root, Eval_p *cell)
{
   bool shrunk = false;

   assert(*root);

   if(!(*root)->passed)
   {
      *cell = *root;
      if((*cell)->lson)
      {
	 *root = (*cell)->lson;
      }
      else
      {
	 *root = (*cell)->rson;
      }
      shrunk = true;
   }
   else
   {
      if((*root)->passed == 1)
      {
	 if(extract_marked_cell(&((*root)->rson), cell))
	 {
	    shrunk = true;
	    (*root)->balance--;
	 }
      }
      else if((*root)->passed == -1)
      {
	 if(extract_marked_cell(&((*root)->lson), cell))
	 {
	    shrunk = true;
	    (*root)->balance++;
	 }
      }

      switch((*root)->balance)
      {
      case 2:
      case -2:
	    shrunk = rebalance_tree(root);
	    break;
      case 1:
      case -1:
	    shrunk = false;
	    break;
      case 0:
	    break;
      default:
	    assert(false);
	    break;
      }
   }
   return shrunk;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: Eval_p EvalAlloc()
//
//   Allocate an evaluation cell with proper eval_count;
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

Eval_p EvalAlloc()
{
   Eval_p eval = EvalCellAlloc();

   eval->eval_count = EvaluationCounter++;

   return eval;
}


/*-----------------------------------------------------------------------
//
// Function: EvalPrint()
//
//   Print an evaluation to the given channel.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EvalPrint(FILE* out, Eval_p eval)
{
   fprintf(out, "[%3d:%2.4f:%ld]", eval->priority,
	   eval->heuristic, eval->eval_count);
}


/*-----------------------------------------------------------------------
//
// Function: EvalPrintComment()
//
//   Print an evaluation (as a comment) to the given channel.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EvalPrintComment(FILE* out, Eval_p eval)
{
   fprintf(out, "/*");
   EvalPrint(out, eval);
   fprintf(out, "*/");

}


/*-----------------------------------------------------------------------
//
// Function: EvalListPrint()
//
//   Print an evaluation list.
//
// Global Variables: -
//
// Side Effects    : Ouput
//
/----------------------------------------------------------------------*/

void EvalListPrint(FILE* out, Eval_p list)
{
   while(list)
   {
      EvalPrint(out, list);
      list = list->next;
   }
}


/*-----------------------------------------------------------------------
//
// Function: EvalListPrintComment()
//
//   Print an evaluation list as a comment.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EvalListPrintComment(FILE* out, Eval_p list)
{   
   fprintf(out, "/*");
   EvalListPrint(out, list);
   fprintf(out, "*/");
}

/*-----------------------------------------------------------------------
//
// Function: EvalGreater()
//
//   Compare two evaluations, return true if the first one is
//   greater. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/


bool EvalGreater(Eval_p ev1, Eval_p ev2)
{
   if(ev1->priority > ev2->priority)
   {
      return true;
   }
   else if(ev1->priority == ev2->priority)
   {
      if(ev1->heuristic > ev2->heuristic)
      {
	 return true;
      }
      else if(ev1->heuristic == ev2->heuristic)
      {
	 if(ev1->eval_count > ev2->eval_count)
	 {
	    return true;
	 }      
      }
   }
   return false;
}

/*-----------------------------------------------------------------------
//
// Function: EvalCompare()
//
//   Compare two evaluations, return a value <0, =0 or >0 if the first
//   one is smaller than, equal two, or bigger than the second one.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long EvalCompare(Eval_p ev1, Eval_p ev2)
{
   long   res;
   double tmp;
   
   res = ev1->priority - ev2->priority;
   if(res)
   {
      return res;
   }
   tmp = ev1->heuristic - ev2->heuristic;
   if(tmp < 0.0)
   {
      return -1;
   }
   if(tmp > 0.0)
   {
      return 1;
   }
   return ev1->eval_count - ev2->eval_count;
}


/*-----------------------------------------------------------------------
//
// Function: EvalListFree()
//
//   Free a list of evaluations. Does _not_ check for tree
//   dependencies. 
//
// Global Variables: -
//
// Side Effects    : Memory management.
//
/----------------------------------------------------------------------*/

void EvalListFree(Eval_p junk)
{
   Eval_p next;
   
   while(junk)
   {
      next = junk->next;
      EvalCellFree(junk);
      junk = next;
   }      
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeFree()
//
//   Free a evaluation tree (does not free the object).
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void EvalTreeFree(Eval_p junk)
{
   if(!junk)
   {
      return;
   }
   EvalTreeFree(junk->lson);
   EvalTreeFree(junk->rson);
   EvalCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeInsert()
//
//   If an entry with new exists in the tree return a
//   pointer to it. Otherwise insert new in the tree and return
//   NULL. 
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

Eval_p EvalTreeInsert(Eval_p *root, Eval_p new)
{
   Eval_p *insert = root;
   Eval_p *update = root;
   Eval_p help;

   insert = find_cell_pos(root, new, &update, false);
   if(*insert)
   {
      return *insert;
   }
   
   *insert = new;
   new->lson = new->rson = NULL;
   new->balance = 0;
   
   help = *update;
   while(help != new)
   {
      help->balance += help->passed;
      help = help->passed == 1 ? help->rson : help->lson;
   }
   rebalance_tree(update);
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeFind()
//
//   Find the entry with key key in the tree and return it. Return
//   NULL if no such key exists.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Eval_p EvalTreeFind(Eval_p root, Eval_p key)
{
   int cmpres;
   
   while(root)
   {
      cmpres = EvalCompare(key, root);

      if(cmpres == 0)
      {
	 break;
      }
      if(cmpres > 0)
      {
	 root = root->rson;
      }
      else
      {
	 root = root->lson;
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeExtractEntry()
//
//   Find the entry with key key, remove it from the tree, rebalance
//   the tree, and return the pointer to the removed element. Return
//   NULL if no matching element exists.
//
//   Basic Algorithm (obsolete, replaced by recursive Version):
//
//   - Search for element el to remove, keeping potential update position
//   - Search for the one-child element mv to take its position, again
//     keeping potential update postition:
//        + if balance == -1, search for largest element in el->lson
//        + Otherwise search for the smallest element in el->rson
//   - for update to mv update balance
//   - replace el by mv and update mv's parent.
//   - Rebalance tree at *update and hope that it works ;-)
//   - Oh yes, return el...
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/


Eval_p EvalTreeExtractEntry(Eval_p *root, Eval_p key)
{
   Eval_p cell;
   
   if(!isolate_and_mark_cell(root, key))
   {
      return NULL;
   }
   extract_marked_cell(root, &cell);

   cell->lson = cell->rson = NULL;
   return cell;
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeDeleteEntry()
//
//   Delete the entry with key key from the tree. 
//
// Global Variables: -
//
// Side Effects    : By EvalExtract(), memory operations
//
/----------------------------------------------------------------------*/

bool EvalTreeDeleteEntry(Eval_p *root, Eval_p key)
{
   Eval_p cell;
   
   cell = EvalTreeExtractEntry(root, key);
   if(cell)
   {
      EvalCellFree(cell);
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: EvalTreeFindSmallest()
//
//   Find the smallest evaluation.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Eval_p EvalTreeFindSmallest(Eval_p root)
{
   if(!root)
   {
      return NULL;
   }
   while(root->lson)
   {
      root = root->lson;
   }
   
   return root;
}


AVL_TRAVERSE_DEFINITION(EvalTree,Eval_p)


/*-----------------------------------------------------------------------
//
// Function: EvalTreePrintInOrder()
//
//   Print an evaluation tree in ascending order to stdout (mainly for
//   debugging and to test the traversal functions ;-)
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void EvalTreePrintInOrder(FILE* out, Eval_p tree)
{
   PStack_p state;
   Eval_p   handle;

   state = EvalTreeTraverseInit(tree);
   
   while((handle = EvalTreeTraverseNext(state)))
   {
      EvalPrintComment(out, handle);
   }
   EvalTreeTraverseExit(state);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/




