/*-----------------------------------------------------------------------

File  : ccl_pdtrees.c

Author: Stephan Schulz

Contents
 
  Perfect discrimination trees for optimized rewriting.

Changes

<1> Wed Jun 24 00:55:29 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_pdtrees.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool PDTreePreferGeneral = false;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: pdtree_select_alternative()
//
//   Given a pdtree and a FunCode, return the alternative in the
//   tree or NULL if no alternative exists up to now.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static PDTree_p pdtree_select_alternative(PDTree_p root, FunCode f_code)
{
   assert(root);
   assert(f_code);

   if(f_code < 0)
   {
      return PDArrayElementP(root->v_alternatives, -f_code);
   }
   assert(f_code < root->sig_size);
   return root->f_alternatives[f_code];   
}


/*-----------------------------------------------------------------------
//
// Function: remove_demod_sides()
//
//   Remove all clause positions indexed by term from the tree at
//   *root. If restrict != NULL only remove entries containing the
//   given clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

static long remove_demod_sides(PTree_p *root, Term_p term, Clause_p
			       restrict)
{
   PStack_p    stack, key_stack = PStackAlloc();
   PTree_p     cell;
   long        res = 0;
   ClausePos_p pos;
   bool        dont_delete = false;
   bool        test;

   stack = PTreeTraverseInit(*root);
   while((cell = PTreeTraverseNext(stack)))
   {
      pos = cell->key;
      if(TBTermEqual(ClausePosGetSide(pos), term))
      {	 
	 if(!restrict || (restrict == pos->clause))
	 {
	    PStackPushP(key_stack, pos);
	    ClausePosFree(pos);
	    res++;
	 }
	 else
	 {
	    dont_delete = true; /* There still is a clause left which
				   has this term in the index */
	 }
      }
   }
   
   PTreeTraverseExit(stack);
   
   while(!PStackEmpty(key_stack))
   {
      pos = PStackPopP(key_stack);
      test = PTreeDeleteEntry(root, pos);
      assert(test);
   }
   PStackFree(key_stack);
   
   if(!dont_delete)
   {
      term->is_indexed = false;
   }
   /* printf("res = %ld\n", res); */
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: pd_tree_delete_alternative()
//
//   Given a PDTreeCell and an alternative, delete the corresponding
//   sub-tree-cell. 
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

static void pd_tree_delete_alternative(PDTree_p root, FunCode f_code)
{
   PDTree_p to_delete;

   if(f_code<0)
   {
      to_delete = PDArrayElementP(root->v_alternatives, -f_code);
      assert(to_delete);
      PDTreeDefaultCellFree(to_delete);
      PDArrayAssignP(root->v_alternatives, -f_code, NULL);
      if(-f_code == root->max_var)
      {
	 long i;

	 for(i=-f_code; i>0; i--)
	 {
	    if(PDArrayElementP(root->v_alternatives, i))
	    {
	       root->max_var = i;
	       break;
	    }
	 }
      }
   }
   else
   {
      assert(f_code < root->sig_size);
      PDTreeDefaultCellFree(root->f_alternatives[f_code]);
      root->f_alternatives[f_code] = NULL;
   }
}


/*-----------------------------------------------------------------------
//
// Function: pdtree_node_find_next_var()
//
//   Given a pdtree-cell and a term, set alt_pdt->trav-state to the
//   next compatible variable position or to 0.
//
// Global Variables: -
//
// Side Effects    : Sets alt_pdt->trav_state
//
/----------------------------------------------------------------------*/

static void pdtree_node_find_next_var(PDTree_p alt_pdt, Term_p term)
{
   PDTree_p succ;
   FunCode  i = -alt_pdt->trav_state;

   while(i <= alt_pdt->max_var)
   {
      succ = PDArrayElementP(alt_pdt->v_alternatives, i);
      if(succ && (!succ->variable->binding ||
		  TBTermEqual(succ->variable->binding, term)))
      {
	 alt_pdt->trav_state = -i;
	 return;
      }
      i++;
   }
   alt_pdt->trav_state = 0;
   return;
}


/*-----------------------------------------------------------------------
//
// Function: pdtree_node_init()
//
//   Initialize a pdtree-node for traversal during search.
//
// Global Variables: -
//
// Side Effects    : Sets alt_pdt->trav_state
//
/----------------------------------------------------------------------*/

static void pdtree_node_init(PDTree_p alt_pdt, Term_p term)
{
   if(!term)
   {
      return;
   }

   if(!PDTreePreferGeneral)
   {
      if((term->f_code > 0) && alt_pdt->f_alternatives[term->f_code])
      {
	 alt_pdt->trav_state = term->f_code;
      }
      else
      {
	 alt_pdt->trav_state = -1;
	 pdtree_node_find_next_var(alt_pdt, term);
      }
   }
   else
   {
      alt_pdt->trav_state = -1;
      pdtree_node_find_next_var(alt_pdt, term);
      if(!alt_pdt->trav_state &&
	 (term->f_code > 0) &&
	 alt_pdt->f_alternatives[term->f_code])
      {
	 alt_pdt->trav_state = term->f_code;
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: pdtree_node_next()
//
//   Given a pdtree_node, return the next potential successor node or
//   NULL.
//
// Global Variables: -
//
// Side Effects    : Changes alt_pdt->trav_state
//
/----------------------------------------------------------------------*/

static FunCode pdtree_node_next(PDTree_p alt_pdt, Term_p term)
{
   FunCode ret;

   ret = alt_pdt->trav_state;

   if(!PDTreePreferGeneral)
   {
      if(ret > 0)
      {
	 alt_pdt->trav_state = -1;
      }
      else
      {
	 alt_pdt->trav_state--;
      }
      pdtree_node_find_next_var(alt_pdt, term);

   }
   else
   {
      if(ret > 0)
      {
	 alt_pdt->trav_state = 0;
      }
      else
      {
	 alt_pdt->trav_state--;
	 pdtree_node_find_next_var(alt_pdt, term);
	 if(!alt_pdt->trav_state && 
	    (term->f_code > 0) &&
	    alt_pdt->f_alternatives[term->f_code])
	 {
	    alt_pdt->trav_state = term->f_code;
	 }
      }
   }
   return ret;
}



/*-----------------------------------------------------------------------
//
// Function: pdtree_backtrack()
//
//   Backtrack over a PDTree node. Return false if backtracked
//   over to the inital node, true otherwise.
//
// Global Variables: -
//
// Side Effects    : Memory operations, may change subst.
//
/----------------------------------------------------------------------*/

static bool pdtree_backtrack(PDSearch_p state, Subst_p subst)
{
   Term_p   term;
   PDTree_p goal;

   goal = state->trav_pointer->parent;

   if(!goal)
   {
      assert(PStackEmpty(state->term_proc));
      return false;
   }
   
   term = PStackPopP(state->term_proc);
   if(state->trav_pointer->variable)
   {
      PStackPushP(state->term_state, term);
      SubstBacktrackToPos(subst, goal->var_state);
   }
   else
   {
      TermLRTraversePrev(state->term_state, term);
   }
   state->trav_pointer = goal;
   return true;
}


/*-----------------------------------------------------------------------
//
// Function: pdtree_forward()
//
//   Move one step forward in the PDTree.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static PDTree_p pdtree_forward(PDSearch_p state, Subst_p subst,
			       FunCode alt)
{
   PDTree_p alternative;
   Term_p   var, binding, term;
   
   assert(alt);
   
   alternative = pdtree_select_alternative(state->trav_pointer, alt);
   var = alternative->variable;
   
   if(var)
   {
      binding = alternative->variable->binding;      
      term = PStackPopP(state->term_state);
      state->trav_pointer->var_state = PStackGetSP(subst);
      /* printf("Var %ld...\n", var->f_code); */
      if(!binding)
      {
	 /* printf("!Binding...\n"); */
	 SubstAddBinding(subst, var, term);
      }
   }
   else
   {
      /* printf("No var\n"); */
      term = TermLRTraverseNext(state->term_state);
   }
   PStackPushP(state->term_proc, term);
   
   term = PStackEmpty(state->term_state) ?
      NULL:PStackTopP(state->term_state);
   
   pdtree_node_init(alternative, term);
   state->trav_pointer = alternative;

   return alternative;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: PDTreeAlloc()
//
//   Allocate an empty, initialized PDTreeCell for indexing terms with
//   up to max_symbol function symbols.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

PDTree_p PDTreeAlloc(FunCode sig_size)
{
   PDTree_p handle = PDTreeCellAlloc();
   FunCode  i;
   
   handle->sig_size       = sig_size;
   handle->ref_count      = 0;
   handle->f_alternatives =
      SizeMalloc(sizeof(PDTree_p)*handle->sig_size);
   for(i=0; i<handle->sig_size; i++)
   {
      handle->f_alternatives[i] = NULL;
   }
   handle->v_alternatives = PDArrayAlloc(8,4);
   handle->max_var        = 0;
   handle->variable       = NULL;
   handle->parent         = NULL;
   handle->demod_terms    = NULL;
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeDefaultCellFree()
//
//   Free a initilized but empty PDTreeCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PDTreeDefaultCellFree(PDTree_p junk)
{
   assert(junk);
   assert(junk->f_alternatives);
   SizeFree(junk->f_alternatives, sizeof(PDTree_p)*junk->sig_size);
   PDArrayFree(junk->v_alternatives);
   PDTreeCellFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function: PDTreeFree()
//
//   Return the memory taken by a PDTree.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void PDTreeFree(PDTree_p junk)
{
   FunCode      i;
   ClausePos_p  tmp;
   PDTree_p     subtree;

   for(i=0; i<junk->sig_size; i++)
   {
      if(junk->f_alternatives[i])
      {
	 PDTreeFree(junk->f_alternatives[i]);
      }
   }
   for(i=1; i<=junk->max_var; i++)
   {
      subtree = PDArrayElementP(junk->v_alternatives, i);
      if(subtree)
      {
	 PDTreeFree(subtree);
      }
   }
   while(junk->demod_terms)
   {
      tmp = PTreeExtractRootKey(&junk->demod_terms);
      ClausePosFree(tmp);
   }   
   PDTreeDefaultCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: PDSearchAlloc()
//
//   Allocate an initialized PDSearchState.
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

PDSearch_p PDSearchAlloc(PDTree_p index, Term_p term)
{
   PDSearch_p handle = PDSearchCellAlloc();
 
   assert(index);
   assert(term); 

   handle->trav_pointer = index;	 
   pdtree_node_init(index, term);
   handle->term_state = TermLRTraverseInit(term);
   handle->term_proc  = PStackAlloc();
   handle->demod_store = PStackAlloc();

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: PDSearchFree()
//
//   Free a PDSearchstate.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void PDSearchFree(PDSearch_p junk)
{
   TermLRTraverseExit(junk->term_state);
   PStackFree(junk->term_proc);
   PStackFree(junk->demod_store);
   
   PDSearchCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseInit()
//
//   Initialize a stack for left-right term traversal (i.e. traversal
//   in writing order) and return it.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

PStack_p TermLRTraverseInit(Term_p term)
{
   PStack_p stack = PStackAlloc();
   
   PStackPushP(stack, term);
   
   return stack;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraverseNext()
//
//   Return the next term node in LR-ordering and update the
//   stack. Return NULL if term traveral is complete.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

Term_p TermLRTraverseNext(PStack_p stack)
{
   int    i;
   Term_p handle;
   
   if(PStackEmpty(stack))
   {
      return NULL;
   }
   handle = PStackPopP(stack);
   for(i=handle->arity-1; i>=0; i--)
   {
      PStackPushP(stack, handle->args[i]);
   }
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermLRTraversePrev()
//
//   Undo a TermLRTraverseNext() operation by replacing terms args on
//   the stack with term.
//
// Global Variables: -
//
// Side Effects    : Stack changes
//
/----------------------------------------------------------------------*/

Term_p TermLRTraversePrev(PStack_p stack, Term_p term)
{
   Term_p tmp;
   int    i;

   for(i=0; i<term->arity; i++)
   {
      tmp = PStackPopP(stack);
      assert(tmp == term->args[i]);
   }
   PStackPushP(stack, term);

   return term;
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeInsert()
//
//   Insert a clause position into the index.
//
// Global Variables: -
//
// Side Effects    : Changes PDTree
//
/----------------------------------------------------------------------*/

void PDTreeInsert(PDTree_p root, ClausePos_p demod_side)
{
   Term_p   term = ClausePosGetSide(demod_side);
   PStack_p stack;
   PDTree_p new;
   
   assert(root);

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeInsert()...\n");

   term->is_indexed = true;
   stack = TermLRTraverseInit(term);
   while((term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(root, term->f_code);
      if(!new)
      {
	 new = PDTreeAlloc(root->sig_size);
	 new->parent = root;
	 
	 if(TermIsVar(term))
	 {
	    new->variable = term;
	    PDArrayAssignP(root->v_alternatives, -term->f_code, new);
	    root->max_var=MAX(-term->f_code, root->max_var);
	 }
	 else
	 {
	    assert(term->f_code < root->sig_size);
	    root->f_alternatives[term->f_code] = new;
	 }
      }
      new->ref_count++;
      root = new;
   }
   TermLRTraverseExit(stack);
   /* printf("In index: %d\n", (int)demod_side); */
   
   /* printf("Storing %d, %c\n", (int)demod_side, demod_side->side ==
      LeftSide?'l':'r'); */
   PTreeStore(&(root->demod_terms), demod_side);

   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeInsert()\n");
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeDelete()
//
//   Remove a term from the index.
//
// Global Variables: -
//
// Side Effects    : Modifies term.
//
/----------------------------------------------------------------------*/

void PDTreeDelete(PDTree_p root, Term_p term, Clause_p restrict)
{
   PStack_p stack, delete_stack;
   long     ref_diff;
   PDTree_p tmp, new;
   FunCode  f_code;
   Term_p   trav_term;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeDelete()...\n");

   assert(root);

   /* Hmmm....is this necessary? I suppose not....*/

   if(!term->is_indexed)
   {
      return;
   }
   
   delete_stack = PStackAlloc();
   PStackPushP(delete_stack, root);
   tmp = root;
   
   trav_term = term;
   stack = TermLRTraverseInit(trav_term);
   while((trav_term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(tmp, trav_term->f_code);
      assert(new);
      PStackPushInt(delete_stack, trav_term->f_code);
      PStackPushP(delete_stack, new);
      tmp = new;
   }
   TermLRTraverseExit(stack);
   
   ref_diff = remove_demod_sides(&(tmp->demod_terms), term, restrict);
   
   /* assert(ref_diff > 0); */
   
   while(PStackTopP(delete_stack)!=root)
   {
      tmp = PStackPopP(delete_stack);
      f_code = PStackPopInt(delete_stack);
      tmp->ref_count-=ref_diff;
      if(!tmp->ref_count)
      {
	 pd_tree_delete_alternative(PStackTopP(delete_stack), f_code);	 
      }
   }
   PStackFree(delete_stack);

   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeDelete()\n");
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeFindNextIndexedTerm()
//
//    Given a search state and a substitution, find the next search
//    state describing a term matching the term encoded in
//    state->term_state and state->term_proc.
//
//    Starting conditions:
//
//    state->term_state describes the part of the term still to
//    process, with PStackTopP(state->term_state) being the next
//    subterm to process.
//
//    state->term_proc contains subterms already worked off.
//
//    state->trav_state points to the PDTreeCell currently traversed.
//
// Global Variables: -
//
// Side Effects    : Memory operations, binds variables, changes state
//
/----------------------------------------------------------------------*/
   
PDTree_p PDTreeFindNextIndexedTerm(PDSearch_p state, Subst_p subst)
{
   PDTree_p alternative = NULL;
   FunCode  new_alt;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeFindNextIndexed()...\n");

   assert(state);
   assert(subst);
   
   if(PStackEmpty(state->term_state))
   {
      pdtree_backtrack(state, subst);
   }

   while(!PStackEmpty(state->term_state))
   {
      new_alt = pdtree_node_next(state->trav_pointer,
				 PStackTopP(state->term_state));
      
      if(!new_alt)
      {
	 if(!pdtree_backtrack(state, subst))
	 {
	    alternative = NULL;
	    break;
	 }
      }
      else
      {
	 alternative = pdtree_forward(state, subst, new_alt);
	 assert(alternative);
      }
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeFindNextIndexed()\n");
   return alternative;
}


/*-----------------------------------------------------------------------
//
// Function:  PDTreeFindNextDemodulator()
//
//   Return the next potential demodulator frpm the index.
//
// Global Variables: -
//
// Side Effects    : Changes state and subst
//
/----------------------------------------------------------------------*/

ClausePos_p PDTreeFindNextDemodulator(PDSearch_p state, Subst_p
				      subst)
{
   PDTree_p    alternative;
   PTree_p     node;
   ClausePos_p res;

   DEBUGMARK(PDT_INTERFACE_WATCH, "PDTreeFindNextDemodulator()...\n");

   assert(state);
   assert(subst);
   
   if(PStackEmpty(state->demod_store))
   {
      alternative = PDTreeFindNextIndexedTerm(state, subst);
      if(alternative)
      {
	 PStackPushP(state->demod_store, alternative->demod_terms);
      }
      else
      {
	 DEBUGMARK(PDT_INTERFACE_WATCH, "...PDTreeFindNextDemodulator()\n");   
	 return NULL;
      }
   }
   assert(!PStackEmpty(state->demod_store));

   node = PStackPopP(state->demod_store); 
   assert(node);

   if(node->lson)
   {
      PStackPushP(state->demod_store, node->lson);
   }
   if(node->rson)
   {
      PStackPushP(state->demod_store, node->rson);
   }   
   res = node->key;

   
   DEBUGMARK(PDT_INTERFACE_WATCH,
	     "...PDTreeFindNextDemodulator()\nDemod: ");
   DEBUG(PDT_INTERFACE_WATCH, 
	 ClausePrint(stdout, res->clause, true);
	 printf(" (%c)\n---> ", res->side == LeftSide? 'l':'r');
	 TermPrint(stdout, ClausePosGetSide(res),
		   res->clause->literals->bank->sig, DEREF_ONCE);
	 printf(" =?= ");
         TermPrint(stdout, PStackElementP(state->term_proc, 0),
		   res->clause->literals->bank->sig, DEREF_NEVER);
	 printf("\n"););
   assert(VerifyMatch(ClausePosGetSide(res),
		      PStackElementP(state->term_proc, 0)));

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: PDTreeFindTerm()
//
//   Find the node corresponding to a given term in the index.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

PDTree_p PDTreeFindTerm(PDTree_p index, Term_p term)
{
   PStack_p stack;
   PDTree_p tmp, new;
   Term_p   trav_term;
   
   assert(index);
   
   tmp = index;   
   trav_term = term;
   stack = TermLRTraverseInit(trav_term);
   while((trav_term = TermLRTraverseNext(stack)))
   {
      new = pdtree_select_alternative(tmp, trav_term->f_code);
      assert(new);
      tmp = new;
   }
   TermLRTraverseExit(stack);
   
   return tmp;
}

/*-----------------------------------------------------------------------
//
// Function: PDTreeStoreClauses()
//
//   Store all clauses at a node described by a term in the index into
//   a PTree. 
//
// Global Variables: -
//
// Side Effects    : Changes tree at *tree
//
/----------------------------------------------------------------------*/

void PDTreeStoreClauses(PTree_p *tree, PDTree_p index, Term_p term)
{
   PDTree_p    node;
   PStack_p    stack;
   ClausePos_p pos;
   PTree_p     entry;

   node = PDTreeFindTerm(index, term);
   
   stack = PTreeTraverseInit(node->demod_terms);
   while((entry = PTreeTraverseNext(stack)))
   {
      pos = entry->key;

      if(TBTermEqual(ClausePosGetSide(pos), term))
      {
	 PTreeStore(tree, pos->clause);
      }
   }
   PTreeTraverseExit(stack);
}

/*-----------------------------------------------------------------------
//
// Function: ClauseHasIndexedTerm()
//
//   Return true if the clause has an indexed term.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool ClauseHasIndexedTerm(Clause_p clause)
{   
   Eqn_p handle = clause->literals;

   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 return true;
      }
      if(handle->rterm->is_indexed)
      {
	 return true;
      }
      handle = handle->next;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveUnchangedIndexedTerms()
//
//   Remove all indexed terms which are not modified of clause from
//   index. Modified terms should have been removed during
//   rewriting... 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseRemoveUnchangedIndexedTerms(Clause_p clause, PDTree_p index)
{   
   Eqn_p handle = clause->literals;
   long  i = 0;

   DEBUGMARK(PDT_INTERFACE_WATCH, "ClauseRemoveUnchangedIndexedTerms()...\n");
   assert(index);
   
   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 if(!EqnSideHasChanged(handle, LeftSide))
	 {
	    /* printf("Removed: \n--->");
	    ClausePrint(stdout, clause, true);
	    printf("\n"); */
	    PDTreeDelete(index, handle->lterm, clause);
	    i++;
	 }
      }
      if(handle->rterm->is_indexed)
      {
	 if(!EqnSideHasChanged(handle, RightSide))
	 {	    
	    PDTreeDelete(index, handle->rterm, clause);
	    /* printf("Removed: \n--->");
	    ClausePrint(stdout, clause, true);
	    printf("\n"); */
	    i++;
	 }
      }
      handle = handle->next;
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...ClauseRemoveUnchangedIndexedTerms()\n");
   return i;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveFromIndexed()
//
//   Remove all indexed terms of clause from index.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long ClauseRemoveFromIndex(Clause_p clause, PDTree_p index)
{   
   Eqn_p handle = clause->literals;
   long  i = 0;

   DEBUGMARK(PDT_INTERFACE_WATCH, "ClauseRemoveFromIndex()...\n");
   assert(index);
   
   /* printf("Removing clause %d\n", (int) clause); */

   while(handle)
   {
      if(handle->lterm->is_indexed)
      {
	 /* printf("Removed: \n--->");
	 ClausePrint(stdout, clause, true);
	 printf("\n"); */
	 PDTreeDelete(index, handle->lterm, clause);
	 i++;
      }
      if(handle->rterm->is_indexed)
      {
	 /* printf("Removed: \n--->");
	 ClausePrint(stdout, clause, true);	 
	 printf("\n"); */
	 PDTreeDelete(index, handle->rterm, clause);		 
	 i++;
      }
      handle = handle->next;
   }
   DEBUGMARK(PDT_INTERFACE_WATCH, "...ClauseRemoveFromIndex()\n");
   return i;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





