/*-----------------------------------------------------------------------

File  : ccl_rewrite.c

Author: Stephan Schulz

Contents
 
  Rewriting. For all functions, if index != NULL, delete rewritten
  terms and superterms from the index!

Changes

<1> Tue May 26 19:47:52 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_rewrite.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

long RewriteAttempts = 0;
long RewriteSucesses = 0;

#ifndef NDEBUG
bool RewriteWarnLeftSide = false;
#endif

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: pdt_index_remove_superterms()
//
//   Remove term and all superterms of term from index.
//
// Global Variables: -
//
// Side Effects    : Changes index
//
/----------------------------------------------------------------------*/

void pdt_index_remove_superterms(Term_p term, PDTree_p index)
{
   PStack_p stack = PStackAlloc();
   PStack_p super_stack;
   PTree_p  cell;
   PTree_p  tmp = NULL;
   Clause_p clause;

   PStackPushP(stack, term);
   
   while(!PStackEmpty(stack))
   {
      term = PStackPopP(stack);
      super_stack = PTreeTraverseInit(term->superterms);
      while((cell = PTreeTraverseNext(super_stack)))
      {
	 PStackPushP(stack, cell->key);
      }
      PTreeTraverseExit(super_stack);
      if(term->is_indexed)
      {
	 PDTreeStoreClauses(&tmp, index, term);
      }
   }   
   PStackFree(stack);
   while(tmp)
   {
      clause = PTreeExtractRootKey(&tmp);
      ClauseRemoveFromIndex(clause, index);
   }
}


/*-----------------------------------------------------------------------
//
// Function: instance_is_rule()
//
//   Return true if lside->rside is a rule, i.e. lside>rside (for the
//   instantiated terms) and rside contains no unbound
//   variables. Assumes that uninstantated terms lside and rside are
//   uncomparable! 
//
// Global Variables: -
//
// Side Effects    : TOGreater() can manipulate TPOpFlag of terms.
//
/----------------------------------------------------------------------*/

static bool instance_is_rule(OCB_p ocb, Term_p lside, Term_p rside,
			     Subst_p subst)
{   
   if(TermHasUnboundVariables(rside))
   {
      return false;
   }
   if(SubstIsRenaming(subst)) /* Save comparisons */
   {
      return false;
   }
   return TOGreater(ocb, lside, rside, DEREF_ONCE, DEREF_ONCE);
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: TermFindRewriteClause()
//
//   Given a term and a (unit) clause set, find a unit clause that is
//   able to rewrite the term. Set pos accordingly, return matching
//   clause. The matching term is instantiated, the match is stored in
//   subst. Clauses are tried in the order defined by
//   demodulators->eval_indices[0], that is, the set must have at
//   least one evaluations (probably AddWeight()).
//
//   This function will probably be replaced by an tree indexing based
//   function.
//
// Global Variables: -
//
// Side Effects    : Instantiates, changes pos
//
/----------------------------------------------------------------------*/

Clause_p TermFindRewriteClause(OCB_p ocb, Term_p term, ClauseSet_p
			       demodulators, ClausePos_p pos, Subst_p
			       subst)
{
   Clause_p      handle, res = NULL;
   Eval_p        key;
   PStack_p      stack;
   PStackPointer state = PStackGetSP(subst);
   Eqn_p         eqn;

   DEBUGMARK(RW_INTERFACE_WATCH2, "TermFindRewriteClause()...\n");   
   assert(term);
   assert(demodulators);
   assert(pos);
   assert(demodulators->eval_indices[0]
	  ||(demodulators->anchor->succ==demodulators->anchor));
   
   RewriteAttempts++;   
   stack = EvalTreeTraverseInit(demodulators->eval_indices[0]);
   
   while((key = EvalTreeTraverseNext(stack)))
   {
      handle = key->object;
      if(handle->pos_lit_no != 1)
      {
	 printf("Problem here...\n");
	 ClausePrint(stdout, handle, true);
	 printf("\n");
	 continue;
      }
      assert(handle->pos_lit_no == 1);
      assert(handle->neg_lit_no == 0);
      
      if((SysDateCompare(demodulators->date,
			 SysDateCreationTime())!=DateEqual)
	 &&
	 (SysDateCompare(term->nf_date,
			 handle->date)!=DateEarlier))
      {
	 DEBUGMARK(RW_INTERFACE_WATCH2,
		   "...TermFindRewriteClause() (Used dates)\n");  
	 break;
      }
      
      pos->clause  = handle;
      eqn = handle->literals;
      
      DEBUGMARK(RW_MATCH_WATCH, "Matching ");
      DEBUG(RW_MATCH_WATCH, 
	    ClausePrint(stdout, pos->clause, true););
      DEBUGOUT(RW_MATCH_WATCH, " (l) onto ");
      DEBUG(RW_MATCH_WATCH, 
	    TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
      DEBUGOUT(RW_MATCH_WATCH, "\nInstantiated lside: ");
      DEBUG(RW_MATCH_WATCH, 
	    TermPrint(stdout, eqn->lterm, ocb->sig, DEREF_ONCE););
      DEBUGOUT(RW_MATCH_WATCH, "\n");
      
      if(SubstComputeMatch(eqn->lterm, term, subst, TBTermEqual))
      {
	 if(((eqn->type == EqOriented) 
	     || instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	    &&
	    (!TermCellQueryProp(term, TPRestricted) ||
	     !SubstIsRenaming(subst)))
	 {
	    res = handle;
	    pos->literal = eqn;
	       pos->side = LeftSide;
	       break;
	 }
	 SubstBacktrackToPos(subst, state);
      }

      if(eqn->type != EqOriented)
      {
	 DEBUGMARK(RW_MATCH_WATCH, "Matching ");
	 DEBUG(RW_MATCH_WATCH, 
	       ClausePrint(stdout, pos->clause, true););
	 DEBUGOUT(RW_MATCH_WATCH, " (r) onto ");
	 DEBUG(RW_MATCH_WATCH, 
	       TermPrint(stdout, term, ocb->sig, DEREF_NEVER););
	 DEBUGOUT(RW_MATCH_WATCH, "\n");
	 if(SubstComputeMatch(eqn->rterm, term, subst, TBTermEqual))
	 {
	    if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst)
	       &&
	       (!TermCellQueryProp(term, TPRestricted) ||
		!SubstIsRenaming(subst)))
	    {
	       res = handle;
	       pos->literal = eqn;
	       pos->side = RightSide;
	       break;
	    }	    
	    SubstBacktrackToPos(subst, state);
	 }	 
      }
   }
   EvalTreeTraverseExit(stack);

   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermFindClause()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermIndexedFindDemodulator()
//
//   Find a demodulator via demodulators->demod_index. This is the
//   replacement function for TermFindRewriteClause()
//   above. Unfortunatly, it has a different calling interface - which
//   should be much more efficent. TermRewriteWithClauseSet() will
//   hopefully take care of the resulting mess transparently.
//
// Global Variables: -
//
// Side Effects    : Instantiates
//
/----------------------------------------------------------------------*/

ClausePos_p TermIndexedFindDemodulator(OCB_p ocb, Term_p term,
				       ClauseSet_p demodulators,
				       Subst_p subst)
{
   Eqn_p       eqn;   
   PDSearch_p  state;  
   ClausePos_p pos, res = NULL;
   
   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "TermIndexedFindDemodulator()...\n");
   
   assert(term);
   assert(demodulators);
   assert(demodulators->demod_index);

   RewriteAttempts++;   
   
   state = PDSearchAlloc(demodulators->demod_index, term);  

   DEBUGMARK(RW_MATCH_WATCH, "\nSearching match for: ");
   DEBUG(RW_MATCH_WATCH, TermPrint(stdout, term, ocb->sig,
				   DEREF_NEVER);
	 printf("\n"););
   
   while((pos = PDTreeFindNextDemodulator(state, subst)))
   {
      eqn = pos->literal;

      switch(pos->side)
      {
      case LeftSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (l), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    if(((eqn->type == EqOriented) 
		|| instance_is_rule(ocb, eqn->lterm, eqn->rterm, subst))
	       &&
	       (!TermCellQueryProp(term, TPRestricted) ||
		!SubstIsRenaming(subst)))
	    {
	       res = pos;
	    }
	    break;
      case RightSide:
	    DEBUGMARK(RW_MATCH_WATCH, "Found: ");
	    DEBUG(RW_MATCH_WATCH, 
		  ClausePrint(stdout, pos->clause, true);
		  printf(" (r), instantiated: ");
		  TermPrint(stdout, pos->clause->literals->lterm,
			    ocb->sig, DEREF_ONCE);
		  printf(" <=> ");
		  TermPrint(stdout, pos->clause->literals->rterm,
			    ocb->sig, DEREF_ONCE);
		  printf("\n");
	       );
	    assert(eqn->type != EqOriented);
	    if(instance_is_rule(ocb, eqn->rterm, eqn->lterm, subst)
	       &&
	       !TermCellQueryProp(term, TPRestricted)) 
	       /* Case SubstIsRenaming(subst) already eleminated in
		  instance_is_rule! */
	    {
	       res = pos;
	    }
	    break;
      default: 
	    assert(false);
	    break;
      }
      if(res)
      {
	 break;
      }
   }
   PDSearchFree(state);

   DEBUGMARK(RW_INTERFACE_WATCH2,
	     "...TermIndexedFindDemodulator()\n");
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TermRewriteWithClauseSet()
//
//   Rewrite the given term at root position with the first matching,
//   orientable rule from demodulators. Return true if a matching rule
//   was found, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes term and termbank, by TBTermReplace()
//
/----------------------------------------------------------------------*/

bool TermRewriteWithClauseSet(OCB_p ocb, TB_p bank, Term_p term,
			      PDTree_p index, ClauseSet_p
			      demodulators)
{
   bool        res   = false;
   Subst_p     subst = SubstAlloc();
   ClausePos_p pos   = NULL;
   Clause_p    demod;
   Term_p      repl;
   void*       found = NULL;
   
   DEBUGMARK(RW_INTERFACE_WATCH2, "TermRewriteWithClauseSet()...\n");   
   if(demodulators->demod_index)
   {
      pos = TermIndexedFindDemodulator(ocb, term, demodulators,
				       subst);
      found = pos;
   }
   else
   {
      pos = ClausePosAlloc();
      demod = TermFindRewriteClause(ocb, term, demodulators, pos,
				    subst);
      found = demod;
   }
   if(found)
   {      
      DEBUGMARK(RW_REWRITE_WATCH, "Rewriting ");
      DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, term, bank->sig, DEREF_NEVER););
      DEBUGOUT(RW_REWRITE_WATCH, " with ");
      DEBUG(RW_REWRITE_WATCH,
	    ClausePrint(stdout, pos->clause, true);
	    printf("(%c)\n", pos->side == LeftSide? 'l':'r'););
      
      RewriteSucesses++;

      repl = ClausePosGetOtherSide(pos);
      
      if(index) /* This can destroy pos! */
      {
	 pdt_index_remove_superterms(term, index);
      }
      repl = ClausePosGetOtherSide(pos);
      
      repl = TBInsert(bank, repl, DEREF_ONCE);

      TBTermReplace(bank, term, repl);
      
      DEBUGMARK(RW_REWRITE_WATCH, "Result ");
      DEBUG(RW_REWRITE_WATCH,
	    TermPrint(stdout, repl, bank->sig, DEREF_NEVER););
      DEBUGOUT(RW_REWRITE_WATCH, "\n");
      
      res = true;
   }
   SubstDelete(subst);
   
   if(!demodulators->demod_index)
   {
      ClausePosFree(pos);
   }

   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermRewriteWithClauseSet()\n");   
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: TermComputeLINormalform()
//
//   Compute a leftmost-innermost normal form of term. This is more
//   tricky than expected, as supertems on the stack may change. I'll
//   start with a recursive version and see how it works. This uses
//   dates to minimize rewrite-attempts: If the normal form of the
//   term is younger than the clause set, no further rewrite-attempt
//   on this term is made.
//
// Global Variables: -
//
// Side Effects    : Changes terms, updates nf_date.
//
/----------------------------------------------------------------------*/

bool TermComputeLINormalform(OCB_p ocb, TB_p bank, Term_p term,
			     PDTree_p index, ClauseSet_p demodulators)
{
   bool modified = true, res = false;

   DEBUGMARK(RW_INTERFACE_WATCH2, "TermComputeLINormalform()...\n");   
      
   if((SysDateCompare(demodulators->date,
		      SysDateCreationTime())!=DateEqual)
      &&
      (SysDateCompare(term->nf_date,
		      demodulators->date)!=DateEarlier))
   {
      DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform() (Used dates)\n");  
      return false;
   }
   if(!TermIsVar(term))
   {
      while(modified)
      {
	 int i;
	 
	 TermGetRef(&term, term);
	 for(i=0; i<term->arity; i++)
	 {
	    modified = TermComputeLINormalform(ocb, bank,
					       term->args[i], index,
					       demodulators);
	    res = res || modified ;
	 }
	 modified = TermRewriteWithClauseSet(ocb, bank, term, index,
					     demodulators);
	 TermReleaseRef(&term);
	 res = res || modified;
      }
   }
   /* Temporary hack for some tests */
   if(modified || SysDateCompare(demodulators->date,
				 SysDateCreationTime())!=DateEqual)
   {
      term->nf_date = demodulators->date;
   }
   DEBUGMARK(RW_INTERFACE_WATCH2, "...TermComputeLINormalform()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: EqnComputeLINormalform()
//
//   Compute the normal form of maximal, minimal or both terms in an
//   equation. Return true if a term was rewritten.
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool EqnComputeLINormalform(OCB_p ocb, TB_p bank, Eqn_p eqn,  PDTree_p
			    index, EqnSide sides, ClauseSet_p
			    demodulators)
{   
   bool res1 = false, res2 = false;

   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   switch(sides)
   {
   case MaxSide:
	 res1 = TermComputeLINormalform(ocb, bank, eqn->lterm, index,
					demodulators);
	 if(!EqnIsOriented(eqn))
	 {
	    res2 = TermComputeLINormalform(ocb, bank, eqn->rterm, index,
					   demodulators);
	 }
	 break;	 
   case MinSide:
	 if(EqnIsOriented(eqn))
	 {
	    res2 = TermComputeLINormalform(ocb, bank, eqn->rterm, index,
					   demodulators);
	 }
	 break;
   case BothSides:
	 res1 = TermComputeLINormalform(ocb, bank, eqn->lterm, index,
					demodulators);
	 res2 = TermComputeLINormalform(ocb, bank, eqn->rterm, index,
					demodulators);
	 break;   
   default:
	 assert(false);
	 break;
   }   
   DEBUGMARK(RW_INTERFACE_WATCH1, "EqnComputeLINormalform()...\n");
   return res1 || res2;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseComputeLINormalform()
//
//   Compute the normal form of terms in a clause. Return true if a
//   term was rewritten. 
//
// Global Variables: -
//
// Side Effects    : Changes terms and term bank
//
/----------------------------------------------------------------------*/

bool ClauseComputeLINormalform(OCB_p ocb, TB_p bank, Clause_p clause,
			       PDTree_p index, EqnSide sides,
			       ClauseSet_p demodulators) 
{
   Eqn_p handle;
   bool tmp, res=false;

   DEBUGMARK(RW_INTERFACE_WATCH1, "ClauseComputeLINormalform()...\n");
   DEBUG(RW_INTERFACE_WATCH1, printf("# Normalizing ");
	 ClausePrint(stdout, clause, true);printf("\n"););

   for(handle = clause->literals; handle; handle=handle->next)
   {
      tmp = EqnComputeLINormalform(ocb, bank, handle, index, sides, 
				   demodulators);
      res = res || tmp;
   }
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseComputeLINormalform()\n");   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetComputeLINormalform()
//
//   Compute a normal form for terms in all clauses in set with
//   respect to clauses in demodulators. Returns true if a term was
//   rewritten. 
//
// Global Variables: -
//
// Side Effects    : Changes clauses
//
/----------------------------------------------------------------------*/

bool ClauseSetComputeLINormalform(OCB_p ocb, TB_p bank,
				  ClauseSet_p set,  PDTree_p
				  index,EqnSide sides,ClauseSet_p
				  demodulators)
{
   Clause_p handle;
   bool     res = false, tmp;

   DEBUGMARK(RW_INTERFACE_WATCH1,
	     "ClauseSetComputeLINormalform()...\n");

   for(handle=set->anchor->succ; handle!=set->anchor; handle =
	  handle->succ)
   {
      tmp = ClauseComputeLINormalform(ocb, bank, handle, index, sides,			     
				      demodulators);
      res = res || tmp;
   }
   DEBUGMARK(RW_INTERFACE_WATCH1, "...ClauseSetComputeLINormalform()\n");
   return res;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


