/*-----------------------------------------------------------------------

File  : ccl_subsumption.c

Author: Stephan Schulz

Contents
 
  Functions for subsumption of clauses.

Changes

<1> Sun Jun  7 15:12:29 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "ccl_subsumption.h"


/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool StrongUnitForwardSubsumption = false;


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: unit_clause_set_topsubsumes_termpair()
//
//   Return a unit clause with sign positive from set that subsumes
//   t1=t2 at top level.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static 
Clause_p unit_clause_set_topsubsumes_termpair(ClauseSet_p set, 
					      Term_p t1, Term_p t2,
					      bool positive)
{
   Subst_p     subst = SubstAlloc();
   Clause_p    res = NULL;
   
   if(set->demod_index)
   {
      PDSearch_p  state = PDSearchAlloc(set->demod_index, t1);  
      ClausePos_p pos;
      
      while((pos = PDTreeFindNextDemodulator(state, subst)))
      {
	 if(!EQUIV(pos->clause->literals->positive,positive))
	 {
	    break;
	 }
	 if(SubstComputeMatch(ClausePosGetOtherSide(pos), 
			      t2, subst, TBTermEqual))
	 {
	    res = pos->clause;
	    assert(res->set == set);
	    break;
	 }   
      }
      PDSearchFree(state);
   }
   else
   {
      Clause_p handle;
      
      for(handle = set->anchor->succ; handle != set->anchor; handle =
	     handle->succ)
      {
	 if(EQUIV(handle->literals->positive, positive))
	 {
	    if(SubstComputeMatch(handle->literals->lterm, t1, subst,
				 TBTermEqual))
	    {
	       if(SubstComputeMatch(handle->literals->rterm,
				    t2, subst, TBTermEqual))
	       {
		  res = handle;
		  break;
	       }
	    }
	    else if(SubstComputeMatch(handle->literals->lterm, t2,
				      subst, TBTermEqual))
	    {
	       if(SubstComputeMatch(handle->literals->rterm,
				    t1, subst, TBTermEqual))
	       {
		  res = handle;
		  break;
	       }
	    }
	    SubstBacktrack(subst);
	 }
      }
   }   
   SubstDelete(subst);
      
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: unit_clause_set_subsumes_termpair()
//
//   Return a unit clause with sign positive from set that subsumes
//   t1=t2.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static 
Clause_p unit_clause_set_subsumes_termpair(ClauseSet_p set, 
					   Term_p t1, Term_p t2,
					   bool positive)
{
   Term_p   tmp1, tmp2 = NULL;
   int      i;
   Clause_p res = NULL;
   
   while(!(res = unit_clause_set_topsubsumes_termpair(set, t1, t2,
						      positive)))
   {
      if(t1->f_code != t2->f_code || !t1->arity)
      {
	 break;
      }
      tmp1 = NULL; /* Used to determine if another position has
		      already been found */
      
      for(i=0; i<t1->arity; i++)
      {
	 if(!TBTermEqual(t1->args[i], t2->args[i]))
	 {
	    if(tmp1)
	    {
	       break;
	    }
	    tmp1 = t1->args[i];
	    tmp2 = t2->args[i];
	 }
      }
      t1 = tmp1;
      t2 = tmp2;
   }

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: unit_clause_set_strongsubsumes_termpair()
//
//   Return a unit clause with sign positive from set if there is a
//   subset with sign positive that shows t1=t2 in one step. Return
//   NULL otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static 
Clause_p unit_clause_set_strongsubsumes_termpair(ClauseSet_p set, 
						 Term_p t1, Term_p t2,
						 bool positive)
{
   PStack_p stack = PStackAlloc();
   int      i;
   Clause_p res = NULL;
   
   PStackPushP(stack, t1);
   PStackPushP(stack, t2);
   
   while(!PStackEmpty(stack))
   {
      t2 = PStackPopP(stack);
      t1 = PStackPopP(stack);
      res = unit_clause_set_topsubsumes_termpair(set, t1, t2,
						 positive);
      if(!res)
      {
	 if(t1->f_code != t2->f_code || !t1->arity)
	 {
	    break;
	 }
	  for(i=0; i<t1->arity; i++)
	  {
	     if(!TBTermEqual(t1->args[i], t2->args[i]))
	     {
		PStackPushP(stack, t1->args[i]);
		PStackPushP(stack, t2->args[i]);
	     }
	  }
      }
      
   }
   PStackFree(stack);
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: unit_clause_set_subsumes_clause()
//
//   Return a clause from set that subsumes clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static 
Clause_p unit_clause_set_subsumes_clause(ClauseSet_p set,
					 Clause_p clause)
{
   Eqn_p    handle = clause->literals;
   Clause_p res = NULL;
   
   while(handle)
   {
      res = StrongUnitForwardSubsumption?
	 unit_clause_set_strongsubsumes_termpair(set, handle->lterm,
						 handle->rterm,
						 handle->positive):
	 unit_clause_set_subsumes_termpair(set, handle->lterm,
					   handle->rterm,
					   handle->positive);
      if(res)
      {
	 break;
      }
      handle = handle->next;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: eqn_topsubsumes_termpair()
//
//   Return true if eqn subsumes t1=t2 at top level.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool eqn_topsubsumes_termpair(Eqn_p eqn, Term_p t1, Term_p t2)
{
   Subst_p subst = SubstAlloc();
   bool    res = false;

   if(SubstComputeMatch(eqn->lterm, t1, subst, TBTermEqual))
   {
      if(SubstComputeMatch(eqn->rterm,
			   t2, subst, TBTermEqual))
      {
	 res = true;
      }
   }
   else if(SubstComputeMatch(eqn->lterm, t2, subst, TBTermEqual))
   {
      if(SubstComputeMatch(eqn->rterm, t1, subst, TBTermEqual))
      {
	 res = true;
	 
      }
   }
   SubstDelete(subst);

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: eqn_subsumes_termpair()
//
//   Return true if the equation subsumes t1=t2.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool eqn_subsumes_termpair(Eqn_p eqn, Term_p t1, Term_p t2)
{
   Term_p   tmp1, tmp2 = NULL;
   int      i;
   bool     res = false;
   
   while(!(res = eqn_topsubsumes_termpair(eqn, t1, t2)))
   {
      if(t1->f_code != t2->f_code || !t1->arity)
      {
	 break;
      }
      tmp1 = NULL; /* Used to determine if another position has
		      already been found */
      
      for(i=0; i<t1->arity; i++)
      {
	 if(!TBTermEqual(t1->args[i], t2->args[i]))
	 {
	    if(tmp1)
	    {
	       break;
	    }
	    tmp1 = t1->args[i];
	    tmp2 = t2->args[i];
	 }
      }
      t1 = tmp1;
      t2 = tmp2;
   }

   return res;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: LiteralSubsumesClause()
//
//   Return true if literal subsumes one of the literals in clause
//   (otherwise return false).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool LiteralSubsumesClause(Eqn_p literal, Clause_p clause)
{
   Eqn_p handle;

   handle = clause->literals;
   while(handle)
   {
      if(EQUIV(literal->positive, handle->positive) &&
	 eqn_subsumes_termpair(literal, handle->lterm, handle->rterm))
      {
	 return true;
      }
      handle = handle->next;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: UnitClauseSubsumesClause()
//
//   Return true if unit subsumes clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool UnitClauseSubsumesClause(Clause_p unit, Clause_p clause)
{
   bool res;

   assert(unit->tot_lit_no == 1);
   
   res = LiteralSubsumesClause(unit->literals, clause);
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: UnitClauseSetSubsumesClause()
//
//   If a clause in set subsumes clause, return a pointer to
//   it. Otherwise return NULL.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Clause_p UnitClauseSetSubsumesClause(ClauseSet_p set, Clause_p
				     clause)
{
   return unit_clause_set_subsumes_clause(set, clause);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetFindSubsumedClause()
//
//   Return a pointer to the first clause that is subsumed by subsumer
//   in the list at set_position. Return NULL, if no such clause
//   exists. 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

Clause_p ClauseSetFindSubsumedClause(ClauseSet_p set, Clause_p
				     set_position, Clause_p subsumer)
{
   assert(subsumer->tot_lit_no == 1);

   while(set_position != set->anchor)
   {
      if(UnitClauseSubsumesClause(subsumer, set_position))
      {
	 return set_position;
      }
      set_position = set_position->succ;
   }
   return NULL;
}
   


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


