/*-----------------------------------------------------------------------

File  : cco_paramodulation.c

Author: Stephan Schulz

Contents
 
  Controling paramodulation inferences.

Changes

<1> Mon Jun  8 18:26:14 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_paramodulation.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ComputeClauseClauseParamodulants()
//
//   Compute all paramodulants between clause and with, with terms
//   from bank, and put them into store. Returns number of
//   paramodulants. 
//
// Global Variables: -
//
// Side Effects    : As above...
//
/----------------------------------------------------------------------*/

long ComputeClauseClauseParamodulants(TB_p bank, OCB_p ocb, Clause_p
				      clause, Clause_p parent_alias,
				      Clause_p with, ClauseSet_p
				      store, VarBank_p freshvars)
{
   Clause_p    paramod;
   Term_p      test;
   ClausePos_p pos1, pos2;
   long        paramod_count = 0;

   pos1 = ClausePosAlloc();
   pos2 = ClausePosAlloc();

   DEBUGMARK(32, "ComputeClauseClauseParamodulants()...\n");
   DEBUG(32,
	 ClausePrint(stdout, clause, true);
	 printf(" =|=|= ");
	 ClausePrint(stdout, parent_alias, true); 	 
	 printf("\n");
	 ClausePrint(stdout, with, true);
	 printf("\n");
	 );

   /* Paramod with into clause */

   test = ClausePosFirstParamodPair(clause, pos1, with, pos2, normal);
   
   while(test)
   {
      assert(TermPosIsTopPos(pos1->pos));
      paramod = ClauseOrderedParamod(bank, ocb, pos1, pos2, freshvars);
      if(paramod)
      {
	 paramod_count++;
         paramod->parent1 = parent_alias;
         ClauseRegisterChild(parent_alias, paramod);
	 if(parent_alias!=with)
	 {
	    paramod->parent2 = with;
	    ClauseRegisterChild(with, paramod);
	 }
         ClauseSetInsert(store, paramod);
      }
      test = ClausePosNextParamodPair(pos1, pos2, normal);
   }
   
   /* Paramod clause into with - no top positions this time ;-) */

   test = ClausePosFirstParamodPair(with, pos1, clause, pos2, true);
   
   while(test)
   {
      assert(TermPosIsTopPos(pos1->pos));
      paramod = ClauseOrderedParamod(bank, ocb, pos1, pos2, freshvars);
      if(paramod)
      {
	 paramod_count++;
         paramod->parent1 = parent_alias;
         ClauseRegisterChild(parent_alias, paramod);
	 if(parent_alias!=with)
	 {
	    paramod->parent2 = with;
	    ClauseRegisterChild(with, paramod);
	 }
         ClauseSetInsert(store, paramod);
      }
      test = ClausePosNextParamodPair(pos1, pos2, true);
   }
   ClausePosFree(pos1);
   ClausePosFree(pos2);
   
   return paramod_count;
}


/*-----------------------------------------------------------------------
//
// Function: ComputeAllParamodulants()
//
//   Compute all paramodulants between clause and with_set, put them
//   into store.
//
// Global Variables: -
//
// Side Effects    : By called functions, memory manangement.
//
/----------------------------------------------------------------------*/

long ComputeAllParamodulants(TB_p bank, OCB_p ocb, Clause_p clause,
			     Clause_p parent_alias, ClauseSet_p
			     with_set, ClauseSet_p store, VarBank_p
			     freshvars)
{
   Clause_p handle;
   long     paramod_count = 0;

   for(handle = with_set->anchor->succ; handle != with_set->anchor;
       handle = handle->succ)
   {
      paramod_count +=
	 ComputeClauseClauseParamodulants(bank, ocb, clause,
					  parent_alias, handle,
					  store, freshvars);
   }
   return paramod_count;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


