/*-----------------------------------------------------------------------

File  : cco_proofproc.c

Author: Stephan Schulz

Contents
 
  Functions realizing the proof procedure.

Changes

<1> Mon Jun  8 11:47:44 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_proofproc.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: remove_subsumed()
//
//   Remove all clauses subsumed by subsumed from set, kill their
//   children. 
//
// Global Variables: -
//
// Side Effects    : Changes set, memory operations.
//
/----------------------------------------------------------------------*/

static void remove_subsumed(Clause_p subsumer, ClauseSet_p set)
{
   Clause_p handle, next;

   assert(subsumer->tot_lit_no == 1);

   DEBUGMARK(PP_HIGHDETAILS , "remove_subsumed()...\n");
   next = set->anchor->succ;
   while((handle = ClauseSetFindSubsumedClause(set, next, subsumer)))
   {
      DEBUGMARK(PP_HIGHDETAILS, "*\n");
      next = handle->succ;      
      /* printf("Removing (subsumption): %d", (int)handle); */
      ClauseKillChildren(handle);
      if(set->demod_index)
      {	 
	 ClauseSetIndexedDeleteEntry(handle);      
      }
      else
      {
	 ClauseSetDeleteEntry(handle);  
      }
   }
   DEBUGMARK(PP_HIGHDETAILS, "...remove_subsumed()\n");
}


/*-----------------------------------------------------------------------
//
// Function: backward_subsumption()
//
//   Perform backward-subsumption (i.e. use the selected clause to
//   remove processed equations). Creates new copy of clause and
//   inserts it into the correct clause set.
//
// Global Variables: -
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static Clause_p backward_subsumption(ProofState_p state, Clause_p clause)
{
   Clause_p insert_copy;
   Subst_p  subst;

   DEBUGMARK(PP_LOWDETAILS,"$$$ backward_subsumption()...\n");
   DEBUG(PP_LOWDETAILS, ClausePrint(stdout, clause, true);
	 printf("\n"););
   
   if(clause->tot_lit_no == 1)
   {
      /* First backward-rewrite the unit clauses, so that we can use
	 them for intereduction later on... */
      
      if(ClauseIsDemodulator(clause))
      {
	 DEBUGMARK(PP_LOWDETAILS,"Clause is demodulator...\n");
	 
	 HCBClauseEvaluate(state->demod_criterion, clause);
	 ClauseSetInsert(state->tmp_store, clause);

	 state->tmp_store->date = SysDateCreationTime();

	 ClauseSetComputeLINormalform(state->ordering,
				      state->processed_terms,
				      state->processed_pos_units,
				      state->processed_pos_units->demod_index,
				      MaxSide,
				      state->tmp_store);
	 DEBUGMARK(PP_HIGHDETAILS, "Backward rewriting done()\n");
	 ClauseSetExtractEntry(clause);
	 

	 assert(state->tmp_store->members == 0);
	 
	 ClauseSetRemoveMaxChanged(state->ordering,
				   state->unprocessed_terms,
				   state->processed_pos_units,
				   state->tmp_store);
	 /* All remaining demodulators are safe - they can only be
	    rewritten on the left side, which won't hurt */

	 /* printf("After ClauseSetRemoveMaxChanged()...\n"); */
	 /* PDTreeVerifyIndex(state->processed_pos_units->demod_index,
			   state->processed_pos_units);*/
	 
	 
	 DEBUGMARK(PP_HIGHDETAILS,"Removed rewritten clauses...\n");
      }
      


      /* Now insert the new clause where it belongs. It will not be
	 rewritten (it cannot rewrite itself, and it is in normal
	 form with respect to all other demodulators */
            
      subst = SubstAlloc();
      VarBankResetVCount(state->freshvars);
      NormSubstClause(clause, subst, state->freshvars);
      insert_copy = ClauseCopy(clause, state->processed_terms);      
      SubstDelete(subst);

      ClauseMarkRestrictedTerms(insert_copy);
      ClauseRecordTermDates(insert_copy);
      DEBUGOUT(PP_LOWDETAILS,"# insert_copy: ");
      DEBUG(PP_LOWDETAILS, ClausePrint(stdout, insert_copy, true););
      DEBUG(PP_LOWDETAILS, printf(" # Copy of %ld\n", clause->ident););
      
      if(clause->pos_lit_no == 1)
      {
	 remove_subsumed(clause, state->processed_pos_units);	 
	 /* printf("After remove_subsumed()...\n"); */
	 /* PDTreeVerifyIndex(state->processed_pos_units->demod_index,
			   state->processed_pos_units); */
	 
	 HCBClauseEvaluate(state->demod_criterion, insert_copy);	
	 SysDateInc(&(state->processed_pos_units->date));
	 insert_copy->date = state->processed_pos_units->date;
	 
	 if(state->processed_pos_units->demod_index)
	 {
	    ClauseSetIndexedInsert(state->processed_pos_units,
				   insert_copy);
	 }
	 else
	 {
	    ClauseSetInsert(state->processed_pos_units,
			    insert_copy);
	 }
	 
	 
	 /* We can now use the new and improved set of demodulators to
	    a) normalize the left sides of demodulacd tors and b) rewrite
	    all other clauses */
	 
	 if(ClauseIsDemodulator(clause))
	 {
	    /* We know that no indexed term can be rewritten now! */

	    ClauseSetComputeLINormalform(state->ordering,
					 state->processed_terms,
					 state->processed_pos_units,
					 NULL,
					 MinSide,
					 state->processed_pos_units);	 
	    
	    ClauseSetComputeLINormalform(state->ordering,
					 state->processed_terms,
					 state->processed_neg_units,
					 NULL,
					 BothSides,
					 state->processed_pos_units);

	    ClauseSetRemoveMaxChanged(state->ordering,
				      state->unprocessed_terms,
				      state->processed_neg_units,
				      state->tmp_store);

	    ClauseSetComputeLINormalform(state->ordering,
					 state->processed_terms,
					 state->processed_non_units,
					 NULL,
					 BothSides,
					 state->processed_pos_units);	 
	    ClauseSetRemoveMaxChanged(state->ordering,
				      state->unprocessed_terms,
				      state->processed_non_units,
				      state->tmp_store);      
	 }
	 DEBUGMARK(PP_HIGHDETAILS, "After interreduction\n");
      }
      else
      {
	 assert(insert_copy->neg_lit_no == 1);
	 remove_subsumed(clause, state->processed_neg_units);
	 ClauseSetInsert(state->processed_neg_units, insert_copy);
      }
      remove_subsumed(clause, state->processed_non_units);   
   }
   else
   {
      assert(clause->tot_lit_no > 1);
      insert_copy = ClauseCopy(clause, state->processed_terms);
      DEBUG(PP_LOWDETAILS, ClausePrint(stdout, insert_copy, true););
      DEBUG(PP_LOWDETAILS, printf(" # Copy of %ld\n", clause->ident););
      ClauseMarkRestrictedTerms(insert_copy);
      ClauseSetInsert(state->processed_non_units, insert_copy);
   }
   DEBUGMARK(PP_LOWDETAILS,"...Backward subsumption");
   DEBUG(PP_LOWDETAILS, printf("-> pos: %ld, neg: %ld, non: %ld\n",
		   state->processed_pos_units->members,
		   state->processed_neg_units->members,
		   state->processed_non_units->members););
   return insert_copy;
}



/*-----------------------------------------------------------------------
//
// Function: insert_new_clauses()
//
//   Rewrite clauses in state->tmp_store, remove superfluous literals,
//   insert them into state->unprocessed. If an empty clause is
//   detected, return it, otherwise return NULL.
//
// Global Variables: -
//
// Side Effects    : As described.
//
/----------------------------------------------------------------------*/

Clause_p insert_new_clauses(ProofState_p state)
{
   Clause_p handle;

   while((handle = ClauseSetExtractFirst(state->tmp_store)))
   {
      state->generated_count++;
      DEBUGMARK(PP_HIGHDETAILS, "insert_new_clauses() - processing: ");
      DEBUG(64, ClausePrint(stdout, handle, true););
      DEBUG(64, printf("\n"););
      ClauseComputeLINormalform(state->ordering,
				state->unprocessed_terms, handle,
				NULL, BothSides,
				state->processed_pos_units);
      DEBUGMARK(PP_HIGHDETAILS, "insert_new_clauses() normal form build\n");
      if(ClauseRemoveSuperfluousLiterals(handle))
      {
	 return handle;	 
      }
      DEBUGMARK(PP_HIGHDETAILS, "insert_new_clauses() Literals removed\n");
      if(ClauseIsTautology(handle))
      {
	 assert(!handle->children);
	 ClauseDetachParents(handle);
	 ClauseFree(handle);
      }
      else
      {
	 state->non_trivial_generated_count++;
	 HCBClauseEvaluate(state->heuristic, handle);
	 ClauseSetInsert(state->unprocessed, handle);
      }
   }
   return NULL;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: InitProofState()
//
//   Given a proof state with axioms and a heuristic parameter
//   description, initialize the ProofStateCell, i.e. generate the
//   HCB, the ordering, and evaluate the axioms and put them in the
//   unprocessed list.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void InitProofState(ProofState_p state, HeuristicParms_p h_parms)
{
   Clause_p handle, new;

   OUTPRINT(1, "# Initializing proof state\n");
   
   if(!state->ordering)
   {
      state->ordering = ProofStateCreateOrdering(state);
   }
   if(!state->heuristic)
   {
      state->heuristic = HCBCreate(h_parms->heuristic_name,
				   state->ordering, state->signature,
				   state->axioms, h_parms->i_parms,
				   h_parms->d_parms,
				   h_parms->s_parms);
   }
   DEBUGMARK(PP_LOWDETAILS,"InitProofState()...starting to copy\n");
   
   
   if(!ClauseSetEmpty(state->processed_pos_units))
   {
      for(handle = state->processed_pos_units->anchor->succ;
	  handle != state->processed_pos_units->anchor; handle =
	     handle->succ)
      {
	 new = ClauseCopy(handle, state->unprocessed_terms);
	 HCBClauseEvaluate(state->heuristic, new);
	 ClauseSetInsert(state->unprocessed, new);
      }
      ClauseSetFree(state->processed_pos_units);
      state->processed_pos_units = ClauseSetAlloc();
   }

   if(state->use_indexing)
   {
      OUTPRINT(1, "# Using perfect discrimination trees\n");
      state->processed_pos_units->demod_index =
	 PDTreeAlloc(state->signature->f_count+1);
   }

   if(!ClauseSetEmpty(state->processed_non_units))
   {
      for(handle = state->processed_non_units->anchor->succ;
	  handle != state->processed_non_units->anchor; handle =
	     handle->succ)
      {
	 new = ClauseCopy(handle, state->unprocessed_terms);
	 HCBClauseEvaluate(state->heuristic, new);
	 ClauseSetInsert(state->unprocessed, new);
      }
      ClauseSetFree(state->processed_non_units);
      state->processed_pos_units = ClauseSetAlloc();
   }
   
   for(handle = state->axioms->anchor->succ;
       handle != state->axioms->anchor; handle = handle->succ)
   {
      new = ClauseCopy(handle, state->unprocessed_terms);
      HCBClauseEvaluate(state->heuristic, new);
      ClauseSetInsert(state->unprocessed, new);
   }
}


/*-----------------------------------------------------------------------
//
// Function: ProcessClause()
//
//   Select an unprocessed clause, process it. Return pointer to empty
//   clause if it can be derived, NULL otherwise.
//
// Global Variables: -
//
// Side Effects    : Everything ;-)
//
/----------------------------------------------------------------------*/

Clause_p ProcessClause(ProofState_p state)
{
   Clause_p clause, insert_copy, empty;

   VERBOUT2("Processing clause\n");
   
   clause = state->heuristic->hcb_select(state->heuristic,
					 state->unprocessed);
   
   if(OutputLevel)
   {
      putc('#', GlobalOut);
   }

   assert(clause);
   DEBUGOUT(PP_LOWDETAILS, "# *** Processing Clause: ");
   DEBUG(PP_LOWDETAILS, ClausePrint(stdout, clause, true);); 
   DEBUGOUT(PP_LOWDETAILS, "\n");
   DEBUGOUT(PP_HIGHDETAILS, "Processed clauses:\n");
   DEBUG(PP_HIGHDETAILS, ClauseSetPrint(stdout, state->processed_pos_units, true););
   
   state->processed_count++;

   ClauseSetExtractEntry(clause);
   ClauseDetachParents(clause);
   ClauseRemoveEvaluations(clause);

   /* Forward subsumption... */
   ClauseComputeLINormalform(state->ordering,
			     state->unprocessed_terms, clause, NULL,
			     BothSides, 
			     state->processed_pos_units);
      
   /* This is the success case! -> There is another one further down. */

   if(ClauseRemoveSuperfluousLiterals(clause))
   {
      return clause;
   }
   DEBUGMARK(PP_LOWDETAILS, "Clause is not empty.\n");
   
   if(ClauseIsTautology(clause))
   {
      ClauseFree(clause);
      DEBUGMARK(PP_LOWDETAILS, "Clause is trivial.\n");
      return NULL;
   }
   DEBUGMARK(PP_LOWDETAILS, "Clause is not trivial.\n");

   /* Now we mark maximal terms... */

   ClauseMarkMaximalTerms(state->ordering, clause);

   DEBUGOUT(PP_LOWDETAILS,"# Oriented: ");
   DEBUG(PP_LOWDETAILS,
	 ClausePrint(stdout, clause, true);printf("\n"););

   /* Still forward subsumption... */

   if((clause->pos_lit_no && 
       UnitClauseSetSubsumesClause(state->processed_pos_units, clause))||
      (clause->neg_lit_no &&
       UnitClauseSetSubsumesClause(state->processed_neg_units, clause)))
   {
      DEBUGMARK(PP_LOWDETAILS, "Clause subsumed!\n");
      ClauseFree(clause);
      return NULL;
   }
   
   if(OutputLevel)
   {
      putc('\n', GlobalOut);
      ClausePrint(GlobalOut, clause, true);
      putc('\n', GlobalOut);
   }

   state->non_trivial_count++;

   /* Now on to backward subsumption... insert_copy is the copy of
      clause that will go into the set of processed equations. It is
      computed here so that we can use the address of it for marking
      parents. */
   
   insert_copy = backward_subsumption(state, clause);
   
   DEBUGMARK(PP_LOWDETAILS, "...starting generating inferences...\n");

   state->factor_count+=
      ComputeAllOrderedFactors(state->unprocessed_terms, state->ordering,
			       clause, insert_copy, state->tmp_store,
			       state->freshvars);
   DEBUGMARK(PP_HIGHDETAILS, "...starting equality resoultion...\n");
   state->resolv_count+=
      ComputeAllEqnResolvents(state->unprocessed_terms, clause,
			      insert_copy, state->tmp_store,
			      state->freshvars);
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (pos_units)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->unprocessed_terms, state->ordering,
			      clause, insert_copy,
			      state->processed_pos_units,
			      state->tmp_store, state->freshvars); 
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (neg_units)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->unprocessed_terms, state->ordering,
			      clause, insert_copy,
			      state->processed_neg_units,
			      state->tmp_store, state->freshvars);
   DEBUGMARK(PP_HIGHDETAILS, "...starting paramodulation (non_units)...\n");
   state->paramod_count+=
      ComputeAllParamodulants(state->unprocessed_terms, state->ordering,
			      clause, insert_copy,
			      state->processed_non_units, 
			      state->tmp_store, state->freshvars);
   DEBUGMARK(PP_LOWDETAILS, "...completed generation...\n");
   ClauseFree(clause);
   
   if((empty = insert_new_clauses(state)))
   {
      return empty;
   }  
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function:  Saturate()
//
//   Process clauses until either the empty clause has been derived, a
//   specified number of clauses has been processed, or the clause set
//   is saturated. Return emoty clause (if found) or NULL. 
//
// Global Variables: -
//
// Side Effects    : Modifies state.
//
/----------------------------------------------------------------------*/

Clause_p Saturate(ProofState_p state, long step_limit, long
		  proc_limit, long unproc_limit, long total_limit)
{
   Clause_p unsatisfiable = NULL;
   long count = 0;

   while(state->unprocessed->members && 
	 step_limit>=count && 
	 proc_limit>=(state->processed_pos_units->members +
		      state->processed_neg_units->members +
		      state->processed_non_units->members) &&
	 unproc_limit>=state->unprocessed->members &&
	 total_limit>=(state->processed_pos_units->members +
		       state->processed_neg_units->members +
		       state->processed_non_units->members+
		       state->unprocessed->members))
   {     
      count++;
      unsatisfiable = ProcessClause(state);
      if(unsatisfiable)
      {
	 break;
      }
   }
   return unsatisfiable;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


