/*-----------------------------------------------------------------------

File  : cto_proofstate.c

Author: Stephan Schulz

Contents
 
  Basic functions for proof objects.

Changes

<1> Mon Jun  8 11:14:02 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_proofstate.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ProofStateAlloc()
//
//   Return an empty, initialized proof state.
//
// Global Variables: -
//
// Side Effects    : memory operations
//
/----------------------------------------------------------------------*/

ProofState_p ProofStateAlloc()
{
   ProofState_p handle = ProofStateCellAlloc();
   PStack_p pdummy = PStackAlloc();
   DStack_p ddummy = DStackAlloc();

   handle->signature           = SigAlloc();
   handle->ordering            = NULL;
   handle->heuristic           = NULL;
   handle->demod_criterion     = HCBLIFOCreate(NULL, NULL, NULL, NULL,
					       NULL, NULL);
   
   handle->original_terms      = TBAlloc(TPIgnoreProps, handle->signature);
   handle->processed_terms     = TBAlloc(TPRestricted, handle->signature);
   handle->unprocessed_terms   = TBAlloc(TPIgnoreProps, handle->signature);
   handle->freshvars           = VarBankAlloc();
   handle->axioms              = ClauseSetAlloc();
   handle->processed_pos_units = ClauseSetAlloc();
   handle->processed_neg_units = ClauseSetAlloc();
   handle->processed_non_units = ClauseSetAlloc();
   handle->unprocessed         = ClauseSetAlloc();
   handle->tmp_store           = ClauseSetAlloc();
   handle->processed_count   = 0;
   handle->non_trivial_count = 0;
   handle-> generated_count  = 0;
   handle->non_trivial_generated_count = 0;
   handle->paramod_count     = 0;
   handle->factor_count      = 0;
   handle->resolv_count      = 0;

   PStackFree(pdummy);
   DStackFree(ddummy);

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ProofStateFree()
//
//   Free a ProofStateCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ProofStateFree(ProofState_p junk)
{
   assert(junk);
   if(junk->ordering)
   {
      OCBFree(junk->ordering);
   }
   if(junk->heuristic)
   {
      HCBFree(junk->heuristic);
   }
   HCBFree(junk->demod_criterion);
   ClauseSetFree(junk->axioms);
   ClauseSetFree(junk->processed_pos_units);
   ClauseSetFree(junk->processed_neg_units);
   ClauseSetFree(junk->processed_non_units);
   ClauseSetFree(junk->unprocessed);
   ClauseSetFree(junk->tmp_store);
   junk->original_terms->sig = NULL;
   junk->processed_terms->sig = NULL;
   junk->unprocessed_terms->sig = NULL;
   SigFree(junk->signature);   
   TBFree(junk->original_terms);
   TBFree(junk->processed_terms);
   TBFree(junk->unprocessed_terms);
   VarBankFree(junk->freshvars);

   ProofStateCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: ProofStateCreateOrdering()
//
//   Given a proof state, generate a (hopefully suitable) ordering for
//   it. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

OCB_p ProofStateCreateOrdering(ProofState_p state)
{
   OCB_p handle;

   switch(state->ordertype)
   {      
   case LPO:
	 handle = OCBAlloc(LPO, state->signature);
	 TOGenerateDefaultPrecedence(handle, state->signature);
	 OUTPRINT(1, "# Creating term ordering: LPO selected\n");
	 break;
   case NoOrdering:
	 /* Do some magic...currently default to KBO! -> Fall through! */
   case KBO:
	 handle = OCBAlloc(KBO, state->signature);
	 TOGenerateDefaultPrecedence(handle, state->signature);
	 TOGenerateDefaultWeights(handle, state->signature);
	 OUTPRINT(1, "# Creating term ordering: KBO selected\n");
	 break;
   case KBO2:
	 handle = OCBAlloc(KBO2, state->signature);
	 TOGenerateDefaultPrecedence(handle, state->signature);
	 TOGenerateDefaultWeights(handle, state->signature);
	 OUTPRINT(1, "# Creating term ordering: KBO2 selected\n");
	 break;
   case RPO:
	 assert(false && "RPO not yet implemented!");
	 handle = NULL;
	 break;
   default:
	 assert(false);
	 handle = NULL;
	 break;
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: ProofStateStatisticsPrint()
//
//   Print the statistics of the proof state.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ProofStateStatisticsPrint(FILE* out, ProofState_p state)
{
   fprintf(out, "# Processed clauses       : %ld\n",
	   state->processed_count);   
   fprintf(out, "# ...of these nonredundant: %ld\n",
	   state->non_trivial_count);   
   fprintf(out, "# Generated clauses       : %ld\n",
	   state->generated_count);   
   fprintf(out, "# ...of these nonredundant: %ld\n",
	   state->non_trivial_generated_count);   
   fprintf(out, "# Paramodulations         : %ld\n",
	   state->paramod_count);   
   fprintf(out, "# Factoriziation          : %ld\n",
	   state->factor_count);    
   fprintf(out, "# Equation resolutions    : %ld\n",
	   state->resolv_count);    
}

/*-----------------------------------------------------------------------
//
// Function: ProofStatePrint()
//
//   Print the clause sets of the proof state.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void ProofStatePrint(FILE* out, ProofState_p state)
{
   fprintf(out, "\n# Processed positive unit clauses:\n");
   ClauseSetPrint(out, state->processed_pos_units, true);
   fprintf(out, "\n# Processed negative unit clauses:\n");
   ClauseSetPrint(out, state->processed_neg_units, true);
   fprintf(out, "\n# Processed non-unit clauses:\n");
   ClauseSetPrint(out, state->processed_non_units, true);
   fprintf(out, "\n# Unprocessed clauses:\n");
   ClauseSetPrint(out, state->unprocessed, true);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
