/*-----------------------------------------------------------------------

File  : cco_simplification.c

Author: Stephan Schulz

Contents
 
  Control of simplified clauses

Changes

<1> Mon Jun  8 14:49:49 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cco_simplification.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ClauseMaxHasChanged()
//
//   Return true if the set of maximal terms of a clause changes,
//   either because a maximal term has been rewritten or because a
//   non-maximal term becomes elevated to maximal status.
//
// Global Variables: -
//
// Side Effects    : May recompute the max terms of the clause.
//
/----------------------------------------------------------------------*/

bool ClauseMaxHasChanged(OCB_p ocb, Clause_p clause)
{
   Eqn_p   handle;
   EqnSide single_res;
   bool    recompute_max = false; 

   DEBUGMARK(CO_SIMPLIFICATION, "ClauseMaxHasChanged()...\n");

   for(handle = clause->literals; handle; handle = handle->next)
   {
      handle->was_max = handle->maximal;
      if(handle->maximal)
      {
	 single_res = EqnHasChanged(handle);
	 if(single_res == MaxSide)
	 {
	    DEBUGMARK(CO_SIMPLIFICATION, "...ClauseMaxHasChanged() -> true\n");
	    return true;
	 }
	 if(single_res == MinSide)
	 {
	    recompute_max = true;
	 }
      }
   }
   
   if(recompute_max && (clause->tot_lit_no > 1))
   {
      ClauseMarkMaximalTerms(ocb, clause);

      for(handle = clause->literals; handle; handle = handle->next)
      {
	 if(handle->was_max != handle->maximal)
	 {
	    DEBUGMARK(CO_SIMPLIFICATION, "...ClauseMaxHasChanged() -> true\n");
	    return true;
	 }
      }
   }
   DEBUGMARK(CO_SIMPLIFICATION, "...ClauseMaxHasChanged() -> false\n");
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRemoveOrientation()
//
//   Set literal types of left-rewritten literals to EqUnoriented.
//
// Global Variables: -
//
// Side Effects    : Changes orientation of equations in clause
//
/----------------------------------------------------------------------*/

void ClauseRemoveOrientation(Clause_p clause)
{
   EqnListUpdateOrientation(clause->literals);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseSetRemoveMaxChanged()
//
//   Remove simplified clauses from from and put them into into (with
//   new terms from new_bank). Kills children of moved clauses,
//   changes literals to EqUnoriented if maximal term has been
//   rewritten. 
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

void ClauseSetRemoveMaxChanged(OCB_p ocb, TB_p new_bank,
			       ClauseSet_p from, ClauseSet_p into)
{
   Clause_p handle, work, new;

   DEBUGMARK(CO_SIMPLIFICATION, "ClauseSetRemoveMaxChanged()...\n");

   handle = from->anchor->succ;
   
   while(handle!=from->anchor)
   {
      work = handle;
      handle = handle->succ;
      if(ClauseMaxHasChanged(ocb, work))
      {
	 DEBUGMARK(CO_SIMPLIFICATION, "Removed: ");
	 DEBUG(CO_SIMPLIFICATION, ClausePrint(stdout, work, true);
	       printf("\n"););
	 ClauseRemoveOrientation(work);
	 ClauseMarkRestrictedTerms(work);
	 ClauseKillChildren(work);
	 new = ClauseCopy(work, new_bank);
	 ClauseSetDeleteEntry(work);
	 ClauseSetInsert(into, new);
      }
   }	 
   DEBUGMARK(CO_SIMPLIFICATION, "...ClauseSetRemoveMaxChanged()\n");
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


