/*-----------------------------------------------------------------------

File  : che_clauseweight.c

Author: Stephan Schulz

Contents
 
  Functions realising clause evaluation with clause weights.

Changes

<1> Fri Jun  5 22:51:52 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_clauseweight.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ClauseWeightInit() 
//
//   Return an initialized WFCB for ClauseWeight evaluation.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

WFCB_p ClauseWeightInit(WFCBPARAMETERS)
{
   WeightParam_p data = WeightParamCellAlloc();

   if(i_parms)
   {
      data->vweight = PStackElementInt(i_parms, 0);
      data->fweight = PStackElementInt(i_parms, 1);      
   }
   else
   {
      data->vweight = DEFAULT_VWEIGHT;
      data->fweight = DEFAULT_FWEIGHT;
   }
   if(d_parms)
   {
      data->max_multiplier = DStackElement(d_parms, 0);
      data->pos_multiplier = DStackElement(d_parms, 1);  
   }	    
   else	    
   {	    
      data->max_multiplier = DEFAULT_MAX_MULT;
      data->pos_multiplier = DEFAULT_POS_MULT;
   }
   return WFCBAlloc(ClauseWeightCompute, ClauseWeightExit, data);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseGoalWeightInit() 
//
//   Return an initialized WFCB for ClauseGoalWeight evaluation.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

WFCB_p ClauseGoalWeightInit(WFCBPARAMETERS)
{
   WeightParam_p data = WeightParamCellAlloc();

   if(i_parms)
   {
      data->vweight = PStackElementInt(i_parms, 0);
      data->fweight = PStackElementInt(i_parms, 1);      
   }
   else
   {
      data->vweight = DEFAULT_VWEIGHT;
      data->fweight = DEFAULT_FWEIGHT;
   }
   if(d_parms)
   {
      data->max_multiplier = DStackElement(d_parms, 0);
      data->pos_multiplier = DStackElement(d_parms, 1);  
   }	    
   else	    
   {	    
      data->max_multiplier = DEFAULT_MAX_MULT;
      data->pos_multiplier = DEFAULT_POS_MULT;
   }
   return WFCBAlloc(ClauseGoalWeightCompute, ClauseWeightExit, data);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseAxiomWeightInit() 
//
//   Return an initialized WFCB for ClauseGoalWeight evaluation.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

WFCB_p ClauseAxiomWeightInit(WFCBPARAMETERS)
{
   WeightParam_p data = WeightParamCellAlloc();

   if(i_parms)
   {
      data->vweight = PStackElementInt(i_parms, 0);
      data->fweight = PStackElementInt(i_parms, 1);      
   }
   else
   {
      data->vweight = DEFAULT_VWEIGHT;
      data->fweight = DEFAULT_FWEIGHT;
   }
   if(d_parms)
   {
      data->max_multiplier = DStackElement(d_parms, 0);
      data->pos_multiplier = DStackElement(d_parms, 1);  
   }	    
   else	    
   {	    
      data->max_multiplier = DEFAULT_MAX_MULT;
      data->pos_multiplier = DEFAULT_POS_MULT;
   }
   return WFCBAlloc(ClauseAxiomWeightCompute, ClauseWeightExit, data);
}

/*-----------------------------------------------------------------------
//
// Function: ClauseWeightCompute()
//
//   Compute an evaluation for a clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eval_p ClauseWeightCompute(void* data, Clause_p clause)
{
   Eval_p        eval  = EvalAlloc();
   WeightParam_p local = data;
   
   if((!clause->pos_lit_no) && ClauseIsGround(clause))
   {
      eval->priority  = PrioGoal;
   }
   else
   {
      eval->priority  = PrioNormal;
   }
   eval->heuristic = ClauseWeight(clause, local->max_multiplier,
                                  local->pos_multiplier,
                                  local->vweight, local->fweight);
   return eval;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseGoalWeightCompute()
//
//   Compute an evaluation for a clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eval_p ClauseGoalWeightCompute(void* data, Clause_p clause)
{
   Eval_p        eval  = EvalAlloc();
   WeightParam_p local = data;
   
   if(!clause->pos_lit_no)
   {
      eval->priority  = PrioPrefer;
   }
   else
   {
      eval->priority  = PrioNormal;
   }
   eval->heuristic = ClauseWeight(clause, local->max_multiplier,
				  local->pos_multiplier,
				  local->vweight, local->fweight);
   return eval;
}

/*-----------------------------------------------------------------------
//
// Function: ClauseAxiomWeightCompute()
//
//   Compute an evaluation for a clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eval_p ClauseAxiomWeightCompute(void* data, Clause_p clause)
{
   Eval_p        eval  = EvalAlloc();
   WeightParam_p local = data;
   
   if(clause->pos_lit_no)
   {
      eval->priority  = PrioPrefer;
   }
   else
   {
      eval->priority  = PrioNormal;
   }
   eval->heuristic = ClauseWeight(clause, local->max_multiplier,
				  local->pos_multiplier,
				  local->vweight, local->fweight);
   return eval;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseWeightExit()
//
//   Free the data entry in a clauseweight WFCB.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseWeightExit(void* data)
{
   WeightParamCellFree(data);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


