/*-----------------------------------------------------------------------

File  : che_hcb.h

Author: Stephan Schulz

Contents
 
  Heuristic control blocks, describing heuristics for clause
  selection.

Changes

<1> Fri Jun  5 22:25:02 MET DST 1998
    New

-----------------------------------------------------------------------*/

#ifndef CHE_HCB

#define CHE_HCB

#include <clb_dstacks.h>
#include <che_wfcb.h>
#include <che_clauseweight.h>
#include <che_refinedweight.h>
#include <che_simweight.h>
#include <che_fifo.h>
#include <che_lifo.h>


/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/


/* An HCBCell describes a heuristic for clause selection. There are
   two main elemenats: A list of clause evauation functions (described
   by a WFCB-List and a clause selection function (whose data is kept
   in the HBCCell). */

typedef struct hcb_cell
{
   /* List and number of weight function used by the heuristic. Take
      care: The list of WFCBs is ordered in the opposite direction
      from the evaluation in a clause, i.e. the _last_ WCFB will
      create the _first_ evaluation. */
   WFCB_p          wfcb_list;
   int             wfcb_no;

   /* Evaluation currently used for selection. This referes to the
      order of evaluations in the clause. See above!       */
   int             current_eval;

   /* Switching for HCBStandardClauseSelect and possibly other
      selection functions: Whenever select_count modulo
      select_switch[wfcb_no-1] reaches select_switch[current_eval],
      current_eval is increased modulo wfcb_no]. */
   int             select_switch[CLAUSE_MAX_EVALUATIONS];
   long            select_count;
   
   /* Selection function, this function is called to select an
      unprocessed clause from the set */
   Clause_p        (*hcb_select)(struct hcb_cell* hcb, ClauseSet_p
				 set);
   
   /* Some HCB selection or evaluation functions may need data of
      their own. If yes, their creation function can allocate data,
      and needs to register a cleanup-function here. This function is
      only called if data != NULL. */
   GenericExitFun  hcb_exit;
   void*           data;
}HCBCell, *HCB_p;


typedef Clause_p (*ClauseSelectFun)(HCB_p hcb, ClauseSet_p set);

typedef HCB_p (*HCBCreateFun)(WFCBPARAMETERS);


/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

#define HCBCellAlloc() (HCBCell*)SizeMalloc(sizeof(HCBCell))
#define HCBCellFree(junk)        SizeFree(junk, sizeof(HCBCell))

void     HCBFree(HCB_p junk);
void     HCBClauseEvaluate(HCB_p hcb, Clause_p clause);
Clause_p HCBStandardClauseSelect(HCB_p hcb, ClauseSet_p set);
Clause_p HCBSingleWeightClauseSelect(HCB_p hcb, ClauseSet_p set);


#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





