/*-----------------------------------------------------------------------

File  : che_heu_weight.c

Author: Stephan Schulz

Contents
 
  Functions describing the Clauseweight heuristic, using only the
  weight of a clause (used for determining the order of trying
  demodulators).

Changes

<1> Tue Jun  9 01:51:42 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_heu_weight.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: HCBClauseweightCreate()
//
//   Create a HCB entry for the Clauseweight heuristic. The parameter
//   stacks are interpreted as follows:
// 
//   Stack        What?               Default
//                                    
//   i_parms:  0  VWeigth             DEFAULT_VWEIGHT
//             1  FWeight             DEFAULT_FWEIGHT
//   d_parms:  0  Max-Multiplier      DEFAULT_MAX_MULT
//             1  Pos-Multiplier      DEFAULT_POS_MULT
//   s_parms:  Ignored
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

HCB_p HCBClauseweightCreate(WFCBPARAMETERS)
{
   long          fweight, vweight;
   double        max_mult, pos_mult;
   PStackPointer psp;
   DStackPointer dsp;
   PStack_p      i_stack = PStackAlloc();
   DStack_p      d_stack = DStackAlloc();
   WFCB_p        wfcb;
   HCB_p         hcb     = HCBCellAlloc();
   

   psp = PStackGetSP(i_parms);
   dsp = DStackGetSP(d_parms);
   
   if(psp > 2)
   {
      Error("Heuristic \"clauseweight\" takes at most 2 integer "
	    "parameters", USAGE_ERROR);
   }      
   fweight   = psp >= 2 ? 
      PStackElementInt(i_parms, 1) : DEFAULT_FWEIGHT;
   vweight   = psp >= 1 ? 
      PStackElementInt(i_parms, 0) : DEFAULT_VWEIGHT;


   if(dsp > 2)
   {
      Error("Heuristic \"standard\" takes at most 2 float "
	    "parameters", USAGE_ERROR);
   }      
   max_mult = dsp >= 2 ? 
      DStackElement(d_parms, 1) : DEFAULT_MAX_MULT;
   pos_mult = dsp >= 1 ?
      DStackElement(d_parms, 0) : DEFAULT_POS_MULT;
   
   PStackPushInt(i_stack, vweight);
   PStackPushInt(i_stack, fweight);
   DStackPush(d_stack, max_mult);
   DStackPush(d_stack, pos_mult);
      
   wfcb = ClauseWeightInit(NULL, NULL, NULL, i_stack, d_stack, NULL);
   hcb->wfcb_list = wfcb;

   hcb->wfcb_no    = 1;
   hcb->hcb_select = HCBSingleWeightClauseSelect;
   hcb->hcb_exit   = NULL;
   hcb->data       = NULL;

   PStackFree(i_stack);
   DStackFree(d_stack);
   return hcb;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


