/*-----------------------------------------------------------------------

File  : che_refinedweight.c

Author: Stephan Schulz

Contents
 
  Functions realising clause evaluation with refined clause weights.

Changes

<1> Wed Jun 17 00:11:03 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_refinedweight.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: eqn_refined_weight()
//
//   Evaluate an equation...
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

static double literal_refined_weight(Eqn_p eqn, double max_multiplier,
				     double pos_multiplier, double
				     product_influence, int vweight,
				     int fweight)
{
   double res, lweight, rweight;

   lweight =  (double)TermWeight(eqn->lterm, vweight, fweight);
   rweight =  (double)TermWeight(eqn->rterm, vweight, fweight);

   if(EqnIsOriented(eqn))
   {
      res = rweight;
   }
   else
   {
      res = rweight * max_multiplier;
   }
   res += lweight * max_multiplier;
   res += (rweight * lweight * product_influence);
   
   if(eqn->maximal)
   {
      res = res*max_multiplier;
   }
   if(eqn->positive)
   {
      res = res*pos_multiplier;
   }
   
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: clause_refined_weight()
//
//   Compute the refined weight of a clause.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static double clause_refined_weight(Clause_p clause, double
				    max_multiplier, double
				    pos_multiplier, double
				    product_influence, int vweight,
				    int fweight)
{
   Eqn_p  handle;
   double res = 0;

   for(handle = clause->literals; handle; handle = handle->next)
   {
      res += literal_refined_weight(handle, max_multiplier,
				    pos_multiplier,
				    product_influence,
				    vweight, fweight);
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: ClauseRefinedWeightInit() 
//
//   Return an initialized WFCB for ClauseRefinedWeight evaluation.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

WFCB_p ClauseRefinedWeightInit(WFCBPARAMETERS)
{
   RefinedWeightParam_p data = RefinedWeightParamCellAlloc();

   if(i_parms)
   {
      assert(PStackGetSP(i_parms)==2);
      data->vweight = PStackElementInt(i_parms, 0);
      data->fweight = PStackElementInt(i_parms, 1);      
   }
   else
   {
      data->vweight = DEFAULT_VWEIGHT;
      data->fweight = DEFAULT_FWEIGHT;
   }
   if(d_parms)
   {
      assert(PStackGetSP(i_parms)==3);
      data->max_multiplier = DStackElement(d_parms, 0);
      data->pos_multiplier = DStackElement(d_parms, 1);  
      data->product_influence = DStackElement(d_parms, 2); 
   }	    
   else	    
   {	    
      data->max_multiplier = DEFAULT_MAX_MULT;
      data->pos_multiplier = DEFAULT_POS_MULT;
      data->product_influence = DEFAULT_PRODUCT_INFLUENCE;
   }
   printf("!!!!! %f %f %f\n", data->product_influence,
	  data->max_multiplier, data->pos_multiplier);
   return WFCBAlloc(ClauseRefinedWeightCompute,
		    ClauseRefinedWeightExit, data);
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRefinedWeightCompute()
//
//   Compute an evaluation for a clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eval_p ClauseRefinedWeightCompute(void* data, Clause_p clause)
{
   Eval_p               eval  = EvalAlloc();
   RefinedWeightParam_p local = data;
   
   if(!clause->pos_lit_no)
   {
      eval->priority  = PrioGoal;
   }
   else
   {
      eval->priority  = PrioNormal;
   }
   eval->heuristic = clause_refined_weight(clause,
					   local->max_multiplier,
					   local->pos_multiplier,
					   local->product_influence,
					   local->vweight,
					   local->fweight);
   return eval;
}


/*-----------------------------------------------------------------------
//
// Function: ClauseRefinedWeightExit()
//
//   Free the data entry in a clauseweight WFCB.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void ClauseRefinedWeightExit(void* data)
{
   RefinedWeightParamCellFree(data);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


