/*-----------------------------------------------------------------------

File  : che_simweight.c

Author: Stephan Schulz

Contents
 
  Functions realising clause evaluation with similarities weights.

Changes

<1> Sun Jun 28 18:18:00 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "che_simweight.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: sim_eqn_weight()
//
//   Evaluate am equation by computing term similarity of both sides.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double sim_eqn_weight(Eqn_p eqn, SimParam_p parms)
{
   double   eq_weight = 0.0, clash_weight = 0.0;
   PStack_p stack = PStackAlloc();
   Term_p   lside, rside;
   int      i;
   
   lside = eqn->lterm;
   rside = eqn->rterm;
   PStackPushP(stack, lside);
   PStackPushP(stack, rside);
   
   while(!PStackEmpty(stack))
   {
      rside = PStackPopP(stack);
      lside = PStackPopP(stack);
      
      if(lside->f_code == rside->f_code)
      {
	 eq_weight += parms->equal_weight;
	 for(i=0; i<lside->arity; i++)
	 {
	    PStackPushP(stack, lside->args[i]);
	    PStackPushP(stack, rside->args[i]);    
	 }
      }
      else
      {
	 if(TermIsVar(lside))
	 {
	    if(TermIsVar(rside))
	    {
	       clash_weight += parms->var_var_clash;
	    }
	    else
	    {
	       clash_weight += parms->var_term_clash;
	    }
	 }
	 else
	 {
	    if(TermIsVar(rside))
	    {
	       clash_weight += parms->var_term_clash;
	    }
	    else
	    {
	       clash_weight += parms->term_term_clash * 
		  (TermWeight(lside, 1, 1)+TermWeight(rside, 1, 1));
	    }
	 }
      }
   }
   PStackFree(stack);
   return clash_weight;	 
}


/*-----------------------------------------------------------------------
//
// Function: sim_weight()
//
//   Evaluate a clause by computing term similarity of equations.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

double sim_weight(Clause_p clause, SimParam_p parms)
{
   Eqn_p  handle = clause->literals;
   double res    = 0.0;

   while(handle)
   {
      res += sim_eqn_weight(handle, parms);
      handle = handle->next;
   }
   return res*20+ClauseWeight(clause, 1, 1, 1, 2);
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: SimWeightInit() 
//
//   Return an initialized WFCB for SimWeight evaluation.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

WFCB_p SimWeightInit(WFCBPARAMETERS)
{
   SimParam_p data = SimParamCellAlloc();

   if(d_parms)
   {
      data->equal_weight    = DStackElement(d_parms, 0);
      data->var_var_clash   = DStackElement(d_parms, 1);  
      data->var_term_clash  = DStackElement(d_parms, 2);
      data->term_term_clash = DStackElement(d_parms, 3);
   }  
   else
   {
      data->equal_weight    = DEFAULT_EQUAL_WEIGHT;
      data->var_var_clash   = DEFAULT_VAR_VAR_CLASH;
      data->var_term_clash  = DEFAULT_VAR_TERM_CLASH;
      data->term_term_clash = DEFAULT_TERM_TERM_CLASH;
   }
   return WFCBAlloc(SimWeightCompute, SimWeightExit, data);
}


/*-----------------------------------------------------------------------
//
// Function: SimWeightCompute()
//
//   Compute an evaluation for a clause.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Eval_p SimWeightCompute(void* data, Clause_p clause)
{
   Eval_p        eval  = EvalAlloc();
   SimParam_p    local = data;
   
   if(!clause->pos_lit_no)
   {
      eval->priority  = PrioGoal;
   }
   else
   {
      eval->priority  = PrioNormal;
   }
   eval->heuristic = sim_weight(clause, local);
   return eval;
}


/*-----------------------------------------------------------------------
//
// Function: SimWeightExit()
//
//   Free the data entry in a Simweight WFCB.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void SimWeightExit(void* data)
{
   SimParamCellFree(data);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


