/*-----------------------------------------------------------------------

File  : che_wfcb.h

Author: Stephan Schulz

Contents
 
  Weigth-function-Control blocks, functions computing weights for
  clauses. 

  The interface to an evaluation function requires three functions:
  
  WFCB_p <eval>Init(OCB_p ocb, Sig_p sig, ClauseSet_p clauses,
                    PStack_p i_parms, DStack_p *d_parms, DStack_p
		    s_parms)

     This function takes ordering, signature, initial clause set,
     two sets of parameters, and a possible file name,  and returns a
     WFCB block describing the evaluation function. In particular, it
     is responsible for creating the data block of the WFCB.

  Eval_p <eval>Compute(void *data, Clause_p clause)

     Given a clause and a data block, return an evaluation for the
     clause. 

  void <Eval>Exit(void* data)

     This function is responsible for freeing data, before the WFCB is
     deleted.

Changes

<1> Fri Jun  5 21:39:40 MET DST 1998
    New

-----------------------------------------------------------------------*/

#ifndef CHE_WFCB

#define CHE_WFCB

#include <cio_output.h>
#include <ccl_clausesets.h>
#include <clb_dstacks.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/


/* Weight functions and their data */

typedef Eval_p (*ClauseEvalFun)(void* data, Clause_p
				clause);


typedef struct wfcb_cell
{
   ClauseEvalFun     wfcb_eval;    /* Compute a clauses evaluation */
   GenericExitFun    wfcb_exit;    /* Clean up - in particular, free
				      data */
   void*             data;         /* WFCB-Data...each set of
				      evaluation functions is
				      responsible for cleaning up...*/
   struct wfcb_cell* next;         /* WFCBs can be chained to describe
				      heuristics */
}WFCBCell, *WFCB_p;



#define WFCBPARAMETERS OCB_p ocb, Sig_p sig, ClauseSet_p clauses,\
                       PStack_p i_parms, DStack_p d_parms, PStack_p\
		       s_parms

typedef WFCB_p (*WFCBCreateFun)(WFCBPARAMETERS);


/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

#define WFCBCellAlloc() (WFCBCell*)SizeMalloc(sizeof(WFCBCell))
#define WFCBCellFree(junk)         SizeFree(junk, sizeof(WFCBCell))

WFCB_p WFCBAlloc(ClauseEvalFun wfcb_eval, GenericExitFun wfcb_exit,
		 void* data);
void   WFCBFree(WFCB_p junk);

void   ClauseAddEvaluation(WFCB_p wfcb, Clause_p clause);

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/

