/*-----------------------------------------------------------------------

File  : cio_streams.c

Author: Stephan Schulz

Contents
 
  Implementation of the stream datatype for look-aheadable input.

Changes

<1> Sat Jul  5 02:28:25 MET DST 1997
    New

-----------------------------------------------------------------------*/

#include "cio_streams.h"


/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: read_char()
//
//   Read a character and return it. Return an infinite sequence of
//   EOFs after the end of file.
//
// Global Variables: -
//
// Side Effects    : May read a character
//
/----------------------------------------------------------------------*/

static int read_char(Stream_p stream)
{
   int ch;

   if(stream->eof_seen)
   {
      ch = EOF;
   }
   else
   {
      ch = getc(stream->file);
      if(ch == EOF)
      {
	 stream->eof_seen = true;
      }
   }
   return ch;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: CreateStream()
//
//   Create a stream associated with the file name. Both the
//   NULL-pointer and the name "-" are taken to mean stdin. 
//
// Global Variables: -
//
// Side Effects    : May terminate with an error message
//
/----------------------------------------------------------------------*/

Stream_p CreateStream(char* name)
{
   Stream_p handle;
   int      i;

   handle = StreamCellAlloc();

   handle->name = DStrAlloc();

   if(!name || !strcmp(name,"-"))
   {
      DStrSet(handle->name, "<stdin>");
      handle->file = stdin;
   }
   else
   {
      DStrSet(handle->name, name);
      if(! (handle->file = fopen(name,"r")))
      {
	 TmpErrno = errno;
	 sprintf(ErrStr, "Cannot open file %s", name);
	 SysError(ErrStr, FILE_ERROR);
      }
   }
   VERBOUTARG("Opening ", DStrView(handle->name));
   handle->next      = NULL;
   handle->eof_seen  = false;
   handle->line      = 1;
   handle->column    = 1;
   handle->current   = 0;

   for(i=0; i<MAXLOOKAHEAD; i++)
   {
      handle->buffer[i] = read_char(handle);
   }

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: DestroyStream()
//
//   Free all resources (memory, file handle) associated with the
//   stream. 
//
// Global Variables: -
//
// Side Effects    : Memory operations, file operations
//
/----------------------------------------------------------------------*/

void DestroyStream(Stream_p stream)
{
   if(stream->file != stdin)
   {
      if(fclose(stream->file) != 0)
      {
	 TmpErrno = errno;
	 sprintf(ErrStr, "Cannot close file %s", DStrView(stream->name));
	 SysError(ErrStr, FILE_ERROR);
      }
   }
   VERBOUTARG("Closing ", DStrView(stream->name));
   DStrFree(stream->name);
   StreamCellFree(stream);
}


/*-----------------------------------------------------------------------
//
// Function: StreamNextChar()
//
//   Move the current window on the input stream one character
//   forward. Return the new CurrChar().
//
// Global Variables: -
//
// Side Effects    : Reads one character from the input, update the
//                   stream information about the current position.
//
/----------------------------------------------------------------------*/

char StreamNextChar(Stream_p stream)
{
   if(StreamCurrChar(stream) == '\n')
   {
      stream->line++;
      stream->column = 1;
   }
   else
   {
      stream->column++;
   }
   stream->current=STREAMREALPOS(stream->current+1);
   stream->buffer[STREAMREALPOS(stream->current+MAXLOOKAHEAD-1)]
      = read_char(stream);

   return StreamCurrChar(stream);
}


/*-----------------------------------------------------------------------
//
// Function: OpenStackedInput()
//
//   Open a new input stream and put it on top of the stack. All
//   further input from this stack is read from the new top of the
//   stack. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void OpenStackedInput(Inpstack_p stack, char* name)
{
   Stream_p handle;

   handle = CreateStream(name);
   handle->next = *stack;
   *stack = handle;
}


/*-----------------------------------------------------------------------
//
// Function: CloseStackedInput()
//
//   Pop the top from the input stack and destroy the associated
//   stream. 
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void CloseStackedInput(Inpstack_p stack)
{
   Stream_p handle;

   assert(*stack);
   handle = *stack;
   *stack = handle->next;
   DestroyStream(handle);
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


