/*-----------------------------------------------------------------------

File  : cio_streams.h

Author: Stephan Schulz

Contents
 
  Definitions for a stream type, i.e. an object associated with a file
  pointer (and possibly a file name), allowing read operations,
  arbitrary look-aheads, and maintaining line and column numbers for
  error messages.

Changes

<1> Wed Aug 20 00:02:32 MET DST 1997
    New

-----------------------------------------------------------------------*/

#ifndef CIO_STREAMS

#define CIO_STREAMS

#include <clb_dstrings.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

#define MAXLOOKAHEAD 64

/* The following data structure describes the state of a (named) input
   stream with lookahead-capability. Streams are stackable, with new
   data being read from the top of the stack. The empty stack is a
   NULL-valued pointer of type Inpstack_p! */
   

typedef struct streamcell
{
   struct streamcell* next;
   DStr_p             name;
   FILE*              file;
   bool               eof_seen;
   long               line;
   long               column;
   char               buffer[MAXLOOKAHEAD];
   int                current;
}StreamCell, *Stream_p, **Inpstack_p;

/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

#define StreamCellAlloc() (StreamCell*)SizeMalloc(sizeof(StreamCell))
#define StreamCellFree(junk)         SizeFree(junk, sizeof(StreamCell))

Stream_p CreateStream(char* name);
void     DestroyStream(Stream_p stream);

#define  STREAMREALPOS(pos) ((pos) % MAXLOOKAHEAD)

#define  StreamLookChar(stream, look)\
         (assert((look)<MAXLOOKAHEAD),\
	  (stream)->buffer[STREAMREALPOS((stream)->current+(look))])
#define  StreamCurrChar(stream) ((stream)->buffer[(stream)->current])
#define  StreamCurrLine(stream)   ((stream)->line)
#define  StreamCurrColumn(stream) ((stream)->column)

char     StreamNextChar(Stream_p stream);

void     OpenStackedInput(Inpstack_p stack, char* name);
void     CloseStackedInput(Inpstack_p stack);


#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





