/*-----------------------------------------------------------------------

File:    cto_kbo.c

Author:  Joachim Steinbach

Contents:
         Implementation of a Knuth_Bendix ordering (KBO) on CLIB
	 terms. It is based on the following refined definition:

	 Let w:F -> N be a weight function assigning natural numbers
	 to operators. Let w be its usual extension to terms. Further,
	 let >F be a precedence on F where f is maximal if f is a
	 unary operator with weight zero.

	 s >KBO t    <=>    Var(s,x) >= Var(t,x)  forall variables x
	                    and

	                    - s = f^n(x) and t = x where n>=1 and
			      w(f)=0  or
			    - w(s) > w(t)  or
			    - w(s) = w(t) and Head(s) > Head(t)  or
			    - w(s) = w(t) and Head(s) ~ Head(t) and
			      Args(s) >KBO,lex Args(t)

	 where Var(r,x) stands for the number of occurrences of the
	 variable x in the term r.


Changes: <1> Thu May 28 12:18:06 MET DST 1998
         New

	 <2> Tue Jun 23 08:36:23 MET DST 1998
	 Changed


-----------------------------------------------------------------------*/

#include "cto_kbo.h"

/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/
static int getweight(OCB_p, FunCode);

static int getweightmultiplier(OCB_p);

static int gettermweight(OCB_p, Term_p, DerefType);

static CompareResult varsupseteq(Term_p, Term_p, DerefType, DerefType);

static CompareResult kbovarequal(OCB_p, Term_p, Term_p,
				 DerefType, DerefType);

static CompareResult kbovargreater(OCB_p, Term_p, Term_p,
				   DerefType, DerefType);

static CompareResult kbocomparevars(OCB_p, Term_p, Term_p,
				    DerefType, DerefType);

static void setvarfrequency(Term_p, setvarfreq, DerefType);

static CompareResult kbogtr(OCB_p, Term_p, Term_p,
			    DerefType, DerefType);

static bool setfreq(Term_p, setvarfreq, DerefType);

static CompareResult kbovargtrgreater(OCB_p, Term_p, Term_p,
				   DerefType, DerefType);


/*-----------------------------------------------------------------------
//
// Function: KBOCompare(ocb, s, t)
//
//   Compare two terms s,t in the Knuth-Bendix Ordering,
//   return the result
//                          to_greater         if s >KBO t
//                          to_equal           if s =KBO t
//                          to_lesser          if t >KBO s
//                          to_uncomparable    otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult KBOCompare(OCB_p ocb, Term_p s, Term_p t,
			 DerefType deref_s, DerefType deref_t)
{ 
   CompareResult res, vars;

   s = TermDeref(s, &deref_s); 
   t = TermDeref(t, &deref_t);
   
   /* The cases s=x v t=x are checked separately. */

   if (TermIsVar(s) || TermIsVar(t)) 
   {
      return kbocomparevars(ocb, s, t, deref_s, deref_t);
   }
   
   vars = KBOVarCompare(s, t, deref_s, deref_t);

   switch(vars)
   {
   case to_equal:
	 res = kbovarequal(ocb, s, t, deref_s, deref_t);
	 break;
   case to_uncomparable:
	 res = to_uncomparable;
	 break;
   case to_greater:
	 res = kbovargreater(ocb, s, t, deref_s, deref_t);
	 break;
   case to_lesser:
	 res = OCBPrecedenceInverse(kbovargreater(ocb, t, s,
						  deref_t, deref_s));
	 break;
   default:
	 res = to_uncomparable;       /* Should never come here ... */
	 assert(false);
	 break;
   }

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: getweight(ocb, op)
//
//   Provides the weight of the operator specified by <op>.
//   Currently, all operators have weight `1', except unary operators
//   which have weight `0'.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static int getweight(OCB_p ocb, FunCode symb)
{
   if (symb<0) {
      return 1;
   }
   return OCBFunWeight(ocb, symb);
}


/*-----------------------------------------------------------------------
//
// Function: getweightmultiplier(ocb)
//
//   Returns the weight multiplier, currently set to `1'.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static int getweightmultiplier(OCB_p ocb)
{
   return 1;
}


/*-----------------------------------------------------------------------
//
// Function: gettermweight(ocb, term)
//
//   Returns the weight of a term t=f(t_1,...,t_n):
//
//                             w(t) = w(f) + w(t_1) + ... + w(t_n)
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static int gettermweight(OCB_p ocb, Term_p t, DerefType deref)
{
   int i, weight;

   t = TermDeref(t, &deref);

   weight = getweight(ocb, t->f_code);
   
   if (!(TermIsVar(t)))
   {
      for (i=0; i<t->arity; i++) 
      {
	 weight += getweightmultiplier(ocb) 
	    * gettermweight(ocb, t->args[i], deref);
      }
   }
   return weight;
}


/*-----------------------------------------------------------------------
//
// Function: varsupseteq()
//
//   Checks whether the variable condition for the terms s and t is
//   satisfied, i.e. it returns
//
//       to_equal          if Var(s,x) = Var(t,x) for all variables x
//
//   otherwise it returns
//
//       to_greater        if Var(s,x) >= Var(t,x) for all variables x
//       to_lesser         if Var(t,x) >= Var(s,x) for all variables x
//       to_uncomparable   otherwise
//
//
//   Changed by Stephan Schulz, Thu Jun 18 23:31:42 MET DST 1998
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static CompareResult varsupseteq(Term_p s, Term_p t,
				 DerefType deref_s, DerefType deref_t)
{
   PStack_p  stack = PStackAlloc();
   Term_p    term;
   DerefType deref;
   bool      greaterzero = false, lesserzero = false;
   int       i;
   
   setvarfrequency(s, reset, deref_s);
   setvarfrequency(t, reset, deref_t);
   setvarfrequency(s, inc, deref_s);
   setvarfrequency(t, dec, deref_t);
   
   PStackPushInt(stack, deref_s);
   PStackPushP(stack, s);
   PStackPushInt(stack, deref_t);
   PStackPushP(stack, t);

   while(!PStackEmpty(stack)) {
      term  = PStackPopP(stack);
      deref = PStackPopInt(stack);
      term  = TermDeref(term, &deref);

      assert(!term->binding || deref == DEREF_NEVER);

      /* We only need to visit each node once. TPOpFlag has been reset
	 by setvarfrequency, it is set in each processed node. Thus,
	 we only need to process nodes without TPOpFlag.            */
      if (!TermCellQueryProp(term, TPOpFlag)) {
	 if (TermIsVar(term)) {
	    if (term->var_frequency > 0) {
	       greaterzero = true;
	    }
	    if (term->var_frequency < 0) {
	       lesserzero = true;
	    }
	    if (greaterzero && lesserzero) {
	       break;
	    }
	 }
	 else {
	    for(i=0; i<term->arity; i++) {
	       PStackPushInt(stack, deref);
	       PStackPushP(stack, term->args[i]);
	    }
	 }
	 TermCellSetProp(term, TPOpFlag);
      }      
   }

   PStackFree(stack);
   
   if (greaterzero && lesserzero) {
      return to_uncomparable;
   }
   if (greaterzero) {
      return to_greater;
   }
   if (lesserzero) {
      return to_lesser;
   }
   return to_equal;
}


/*-----------------------------------------------------------------------
//
// Function: kbovarequal(ocb, s, t)
//
// Compares the terms s and t wrt. the KBO whenever Var(s,x)=Var(t,x)
// holds for all variables x.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static CompareResult kbovarequal(OCB_p ocb, Term_p s, Term_p t,
				 DerefType deref_s, DerefType deref_t)
{
   int sweight, tweight, i;
   CompareResult res;

   sweight = gettermweight(ocb, s, deref_s);
   tweight = gettermweight(ocb, t, deref_t);
   
   if (sweight > tweight) {
      return to_greater;
   }
   if(sweight < tweight) {
      return to_lesser;
   }

   assert(sweight == tweight);
   
   switch (OCBFunCompare(ocb, s->f_code, t->f_code)) 
   {
   case to_greater:
	 return to_greater;
	 break;
   case to_equal:
	 for (i=0; i<MAX(s->arity,t->arity); i++) 
	 {
	    if (t->arity <= i) 
	    {
	       return to_greater;
	    }
	    if (s->arity <= i) 
	    {
	       return to_lesser;
	    }
	    if ((res = KBOCompare(ocb, s->args[i], t->args[i],
				  deref_s, deref_t)) != to_equal)
	    {
	       return res;
	    }
	 }
	 return to_equal;
	 break;
   case to_lesser:
	 return to_lesser;
   case to_uncomparable:
	 return to_uncomparable;
	 break;
   default:
	 assert(false);
	 break;
   }
   assert(false);                     /* Should never come here ... */
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: kbovargreater(ocb, s, t)
//
// Compares the terms s and t wrt. the KBO whenever Var(s,x)>=Var(t,x)
// holds for all variables x and Var(s,y)>Var(t,y) for at least one
// variable y.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static CompareResult kbovargreater(OCB_p ocb, Term_p s, Term_p t,
				   DerefType deref_s, DerefType deref_t)
{
   CompareResult res;
   int sweight, tweight, i;

   sweight = gettermweight(ocb, s, deref_s);
   tweight = gettermweight(ocb, t, deref_t);
   
   if(sweight > tweight) 
   {
      return to_greater;
   }
   if(sweight < tweight) 
   {
      return to_uncomparable;
   }
   assert(sweight == tweight);
   
   switch (OCBFunCompare(ocb, s->f_code, t->f_code)) 
   {
   case to_greater:
	 return to_greater;
	 break;
   case to_equal:
	 for (i=0; i<MAX(s->arity,t->arity); i++) 
	 {
	    if (t->arity <= i) 
	    {
	       return to_greater;
	    }
	    if (s->arity <= i) 
	    {
	       return to_uncomparable;
	    }
	    if ((res = KBOCompare(ocb, s->args[i], t->args[i],
				  deref_s, deref_t)) == to_greater) 
	    {
	       return to_greater;
	    }
	    if (res != to_equal) 
	    {
	       return to_uncomparable;
	    }
	 }
	 /* Note that s and t cannot be identical since there
	    // exists a variable which occurs more often in s than
	    // in t.                                            */
	 return to_uncomparable;
	 break;
   case to_uncomparable:
   case to_lesser:
	 return to_uncomparable;
	 break;
   default:
	 assert(false);
	 break;
   }
   assert(false);                     /* Should never come here ... */
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: kbocomparevars(ocb, s, t)
//
// Compares two terms s and t wrt. KBO if either s or t is a variable
// and returns
//               to_greater        if t is a subterm of s
//               to_equal          if s == t
//               to_lesser         if s is a subterm of t
//               to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static CompareResult kbocomparevars(OCB_p ocb, Term_p s, Term_p t,
				    DerefType deref_s, DerefType deref_t)
{
   assert(!t->binding || deref_t == DEREF_NEVER);
   assert(!s->binding || deref_s == DEREF_NEVER);

   if (TermIsVar(t)) {
      if (TBTermEqual(s,t)) {
	 return to_equal;
      }
      else {
	 if (TermIsSubterm(s, t, deref_s, TBTermEqual)) {
	    return to_greater;
	 }
      }
   }
   else {               /* Note that in this case, s is a variable. */
      assert(TermIsVar(s));
      if (TermIsSubterm(t, s, deref_t, TBTermEqual)) {
	 return to_lesser;
      }
   }   
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: setvarfrequency()
//
//    Manipulate the var_frequency counter of variables in the term
//    t. Also clears TPOpFlag (for optimised traversal later on in the
//    function varsupseteq).
//
// Global Variables: -
//
// Side Effects    : As described in the function varsupseteq.
//
-----------------------------------------------------------------------*/

static void setvarfrequency(Term_p t, setvarfreq varfreq,
			    DerefType deref_t)
{
   PStack_p stack = PStackAlloc();
   int i;
   
   assert(t);

   PStackPushInt(stack, deref_t);
   PStackPushP(stack, t);

   while (!PStackEmpty(stack)) {
      t = PStackPopP(stack);
      deref_t = PStackPopInt(stack);
      t = TermDeref(t, &deref_t);
      TermCellDelProp(t, TPOpFlag);
      
      if (TermIsVar(t)) {
	 assert(!t->binding || deref_t == DEREF_NEVER);
	 switch (varfreq) {
	 case reset:
	       t->var_frequency = 0;
	       break;
	 case inc:
	       t->var_frequency++;
	       break;
	 case dec:
	       t->var_frequency--;
	       break;
	 }
      }
      else {
         for(i=0; i<t->arity; i++) {
	    PStackPushInt(stack, deref_t);
            PStackPushP(stack, t->args[i]);
         }
      }
   }   
   PStackFree(stack);
}

/*-----------------------------------------------------------------------
//
// Function: KBOGreater(ocb, s, t)
//
//   Checks whether the term s is greater than the term t in the
//   Knuth-Bendix Ordering (KBO), i.e. returns
//
//                       true      if s >KBO t,
//      		 false     otherwise.
//
//   For a description of the KBO see the header of this file.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

bool KBOGreater(OCB_p ocb, Term_p s, Term_p t,
		DerefType deref_s, DerefType deref_t)
{
   if (kbogtr(ocb, s, t, deref_s, deref_t) == to_greater) {
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: kbogtr(ocb, s, t)
//
//   Returns
//              to_greater         if s >KBO t,
//              to_equal           if s =KBO t,
//              to_uncomparable    otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/
static CompareResult kbogtr(OCB_p ocb, Term_p s, Term_p t,
			    DerefType deref_s, DerefType deref_t)
{
   bool res;
   s = TermDeref(s, &deref_s); 
   t = TermDeref(t, &deref_t);
   
   if (TermIsVar(s)) 
   {
      if (TBTermEqual(s,t)) 
      {
	 return to_equal;
      }
      return to_uncomparable;
   }
   if ((TermIsVar(t)) && (TermIsSubterm(s, t, deref_s, TBTermEqual))) 
   {
      return to_greater;
   }

   if(KBOVarGreater(s, t, deref_s, deref_t))
   {
      return kbovargtrgreater(ocb, s, t, deref_s, deref_t);
   }
#ifdef NEVER_DEFINED   
   res = setfreq(s, reset, deref_s);
   res = setfreq(t, reset, deref_t);
   res = setfreq(s, inc, deref_s);
   if (setfreq(t, dec, deref_t)) {
      return kbovargtrgreater(ocb, s, t, deref_s, deref_t);
   }
#endif
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: setfreq(<t>, <how>)
//
//    Manipulate the var_frequency counter of variables in the term
//    t. Returns
//                 true   if <how>=reset or inc or
//                        if <how>=dec and no frequency counter of a
//                           variable is less than zero
//                 false  otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static bool setfreq(Term_p t, setvarfreq varfreq, DerefType deref_t)
{
   PStack_p stack = PStackAlloc();
   int i;
   
   assert(t);

   PStackPushInt(stack, deref_t);
   PStackPushP(stack, t);

   while (!PStackEmpty(stack)) {
      t = PStackPopP(stack);
      deref_t = PStackPopInt(stack);
      t = TermDeref(t, &deref_t);
      
      if (TermIsVar(t)) {
	 switch (varfreq) {
	 case reset:
	       t->var_frequency = 0;
	       break;
	 case inc:
	       t->var_frequency++;
	       break;
	 case dec:
	       t->var_frequency--;
	       if (t->var_frequency < 0) {
		  PStackFree(stack);
		  return false;
	       }
	       break;
	 }
      }
      else {
         for(i=0; i<t->arity; i++) {
	    PStackPushInt(stack, deref_t);
            PStackPushP(stack, t->args[i]);
         }
      }
   }   
   PStackFree(stack);
   return true;
}


/*-----------------------------------------------------------------------
//
// Function: kbovargtrgreater(ocb, s, t)
//
// Compares the terms s and t wrt. the KBO whenever Var(s,x)>=Var(t,x)
// holds for all variables x and Var(s,y)>Var(t,y) for at least one
// variable y. It returns
//                          to_greater         if s >KBO t,
//                          to_equal           if s =KBO t,
//                          to_uncomparable    otherwise.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

static CompareResult kbovargtrgreater(OCB_p ocb, Term_p s, Term_p t,
				   DerefType deref_s, DerefType deref_t)
{
   CompareResult res;
   int sweight, tweight, i;

   sweight = gettermweight(ocb, s, deref_s);
   tweight = gettermweight(ocb, t, deref_t);
   
   if(sweight > tweight) {
      return to_greater;
   }
   if(sweight < tweight) {
      return to_uncomparable;
   }
   assert(sweight == tweight);
   
   switch (OCBFunCompare(ocb, s->f_code, t->f_code)) {
   case to_greater:
	 return to_greater;
	 break;
   case to_equal:
	 for (i=0; i<MAX(s->arity,t->arity); i++) {
	    if (t->arity <= i) {
	       return to_greater;
	    }
	    if (s->arity <= i) {
	       return to_uncomparable;
	    }
	    if ((res = kbogtr(ocb, s->args[i], t->args[i],
			      deref_s, deref_t)) != to_equal) {
	       return res;
	    }
	 }
	 return to_equal;
	 break;
   case to_uncomparable:
   case to_lesser:
	 return to_uncomparable;
	 break;
   default:
	 assert(false);
	 break;
   }
   assert(false);                     /* Should never come here ... */
   return to_uncomparable;
}


/*---------------------------------------------------------------------*/
/*                      Exported Functions                             */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: KBOVarCompare()
//
//   Compare the variable occurences in two terms, return the strongest
//   KBO result comparable with the variable condition.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

CompareResult KBOVarCompare(Term_p s, Term_p t, DerefType deref_s,
			    DerefType deref_t)
{
   bool           s_gt = false, t_gt = false;
   VarHash_p      hash = VarHashAlloc();
   int            i; 
   VarHashEntry_p handle;

   VarHashAddVarDistrib(hash, s, deref_s, 1);
   VarHashAddVarDistrib(hash, t, deref_t, -1);
   
   for(i=0; i<VAR_HASH_SIZE; i++)
   {
      for(handle=hash->hash[i]; handle; handle=handle->next)
      {
	 if(handle->val > 0)
	 {
	    s_gt = true;
	 }
	 else if(handle->val < 0)
	 {
	    t_gt = true;
	 }
      }
      if(s_gt && t_gt)
      {
	 break;
      }
   }

   VarHashFree(hash);

   if(s_gt && t_gt)
   {
      return to_uncomparable;
   }
   if(s_gt)
   {
      return to_greater;
   } 
   if(t_gt)
   {
      return to_lesser;
   }
   return to_equal;
}


/*-----------------------------------------------------------------------
//
// Function: KBOVarGreater()
//
//   Return true if vars(s) multisetsubseteq vars(t), false otherwise.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

bool KBOVarGreater(Term_p s, Term_p t, DerefType deref_s, DerefType
		   deref_t)
{
   bool           t_gt = false;
   VarHash_p      hash = VarHashAlloc();
   int            i; 
   VarHashEntry_p handle;

   VarHashAddVarDistrib(hash, s, deref_s, 1);
   VarHashAddVarDistrib(hash, t, deref_t, -1);
   
   for(i=0; i<VAR_HASH_SIZE; i++)
   {
      for(handle=hash->hash[i]; handle; handle=handle->next)
      {
	 if(handle->val < 0)
	 {
	    t_gt = true;
	    break;
	 }
      }
      if(t_gt)
      {
	 break;
      }
   }

   VarHashFree(hash);

   if(t_gt)
   {
      return false;
   }
   return true;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
