/*-----------------------------------------------------------------------

File:    cto_lpo.c

Author:  Joachim Steinbach

Contents:
         Implementation of a lexicographic path ordering on CLIB terms.

Changes: <1> Tue May  5 13:45:47 MET DST 1998
         New

-----------------------------------------------------------------------*/

#include "cto_lpo.h"

CompareResult LPOFunGreater(OCB_p, Term_p, Term_p);
CompareResult LPOFunEqual(OCB_p, Term_p, Term_p);
CompareResult LPOGtr(OCB_p, Term_p, Term_p);

/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------

   Function: LPOCompare(ocb, s, t)

     Compare two terms s,t in the Lexicographic Path Ordering,
     return the result
                            to_greater         if s >LPO t
	                    to_equal           if s =LPO t
	                    to_lesser          if t >LPO s
	                    to_uncomparable    otherwise

   Global Variables: -

   Side Effects    : -

-----------------------------------------------------------------------*/


CompareResult LPOCompare(OCB_p ocb, Term_p s, Term_p t)
{
  CompareResult res, res_funs;
  int i;

  s = TermDeref(s);
  t = TermDeref(t);

  if (TermIsVar(s))
    {
      if (s == t)
	return to_equal;
      else
	{
	  if (TermIsSubterm(t,s,true,TBTermEqual))
	    return to_lesser;
	  else return to_uncomparable;
	}
    }

  if (TermIsVar(t))
    {
      if (TermIsSubterm(s,t,true,TBTermEqual))
	return to_greater;
      else return to_uncomparable;
    }

  switch ((res_funs = OCBFunCompare(ocb,s->f_code,t->f_code)))
    {
    case to_greater:
      if ((res = LPOFunGreater(ocb,s,t)) != to_uncomparable)
	return res;
    case to_lesser:
      if ((res = LPOFunGreater(ocb,t,s)) != to_uncomparable)
	return (res == to_greater) ? to_lesser : to_greater;
    case to_equal:
      if ((res = LPOFunEqual(ocb,s,t)) != to_uncomparable)
	return res;
    default:
      if (s->arity == 1)
	return to_uncomparable;
      if ((res_funs != to_greater) && ((res_funs != to_equal) || (s->arity >= 2)))
	for (i=0; i<s->arity; i++)
	  {
	    res = LPOCompare(ocb,s->args[i],t);
	    if ((res == to_greater) || (res == to_equal))
	      return to_greater;
	  }
      if (t->arity == 1)
	return to_uncomparable;
      if ((res_funs != to_lesser) && ((res_funs != to_equal) || (t->arity >= 2)))
	for (i=0; i<t->arity; i++)
	  {
	    res = LPOCompare(ocb,t->args[i],s);
	    if ((res == to_lesser) || (res == to_equal))
	      return to_lesser;
	  }
      return to_uncomparable;
      break;
    }
}

CompareResult LPOFunGreater(OCB_p ocb, Term_p s, Term_p t)
{
  bool greater = true;
  int i;

  for (i=0; i<t->arity; i++)
    switch (LPOCompare(ocb,s,t->args[i]))
      {
      case to_greater:
	break;
      case to_lesser: case to_equal:
	return to_lesser;
	break;
      case to_uncomparable:
	greater = false;
	break;
      default:
	assert(false);
	break;
      }
  if (greater)
    return to_greater;
  else return to_uncomparable;
}

CompareResult LPOFunEqual(OCB_p ocb, Term_p s, Term_p t)
{
  CompareResult res = to_equal;
  int i, j;

  for (i=0; i<MAX(s->arity,t->arity); i++)
    {
      if ((res = LPOCompare(ocb,s->args[i],t->args[i])) == to_uncomparable)
	return to_uncomparable;
      if ((res == to_greater) || (res == to_lesser))
	{
	  CompareResult res_help = res;
	  CompareResult comp_res = (res == to_greater) ? to_lesser : to_greater;

	  j = i+1;
	  while ((j<MAX(s->arity,t->arity)) && (res_help == res))
	    {
	      if (res == to_greater)
		res_help = LPOCompare(ocb,s,t->args[j]);
	      else res_help = LPOCompare(ocb,s->args[j],t);
	      if (res_help == comp_res)
		return comp_res;
	      j++;
	    }
	  if (res_help == res)
	    return res;
	}
    }
  if (res == to_equal)
    return to_equal;
  else return to_uncomparable;
}


/*-----------------------------------------------------------------------

   Function: LPOGreater(ocb, s, t)

     Checks whether the term s is greater than the term t in the
     Lexicographic Path Ordering, i.e. return
                                               true    if s >LPO t
					       false   otherwise

   Global Variables: -

   Side Effects    : -

-----------------------------------------------------------------------*/

bool LPOGreater(OCB_p ocb, Term_p s, Term_p t)
{
  if (LPOGtr(ocb,s,t) == to_greater)
    {
      return true;
    }
  else return false;
}

CompareResult LPOGtr(OCB_p ocb, Term_p s, Term_p t)
{
  CompareResult res, res_funs;
  int i;

  s = TermDeref(s);
  t = TermDeref(t);

  if (TermIsVar(s))
    {
      if (s == t)
	return to_equal;
      else return to_uncomparable;
    }

  if (TermIsVar(t))
    {
      if (TermIsSubterm(s,t,true,TBTermEqual))
	return to_greater;
      else return to_uncomparable;
    }

  switch ((res_funs = OCBFunCompare(ocb,s->f_code,t->f_code)))
    {
    case to_greater:
      res = to_greater;
      for (i=0; i<t->arity; i++)
	if ((res = LPOGtr(ocb,s,t->args[i])) != to_greater)
	  break;
      if (res == to_greater)
	return to_greater;
    case to_lesser:
      for (i=0; i<s->arity; i++)
	{
	  res = LPOGtr(ocb,s->args[i],t);
	  if ((res == to_greater) || (res == to_equal))
	    return to_greater;
	}
      return to_uncomparable;
    case to_equal:
      res = to_equal;
      for (i=0; i<MAX(s->arity,t->arity); i++)
	{
	  res = LPOGtr(ocb,s->args[i],t->args[i]);
	  if ((res != to_greater) && (res != to_equal))
	    break;
	  if (res == to_greater)
	    {
	      int j = i+1;
	      while ((j<MAX(s->arity,t->arity)) && (res == to_greater))
		{
		  res = LPOGtr(ocb,s,t->args[j]);
		  j++;
		}
	      if (res == to_greater)
		return to_greater;
	      else
		{
		  res = to_uncomparable;
		  break;
		}
	    }
	}
      if (res == to_equal)
	return to_equal;
    default:
      if (s->arity == 1)
	return to_uncomparable;
      if ((res_funs != to_greater) && ((res_funs != to_equal) || (s->arity >= 2)))
	for (i=0; i<s->arity; i++)
	  {
	    res = LPOGtr(ocb,s->args[i],t);
	    if ((res == to_greater) || (res == to_equal))
	      return to_greater;
	  }
      return to_uncomparable;
      break;
    }
}
