/*-----------------------------------------------------------------------

File:    cto_lpo.c

Author:  Joachim Steinbach

Contents:
         Implementation of a lexicographic path ordering (LPO) on CLIB
	 terms. It is based on the following refined definition:

	 s = f(s_1,...,s_m) >LPO g(t_1,...,t_n) = t  iff

	   f > g  &  forall j in [1,n]: s >LPO t_j   or
	   f ~ g  &  (s_1,...,s_m) >LPOlex (t_1,...,t_n)  &
	             forall j in [1,n]: s >LPO t_j   or
	   not(f > g)  &  (not(f ~ g) v ar(f) >= 2)  &
	             there exists i in [1,m]: s_i >=LPO t

Changes: <1> Tue May  5 13:45:47 MET DST 1998
         New
	 <2> Mon May 18 10:43:15 MET DST 1998
	 Changed

-----------------------------------------------------------------------*/

#include "cto_lpo.h"

/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/
CompareResult lpocomparevars(OCB_p ocb, Term_p s, bool varps, Term_p t);
CompareResult lpofungreater(OCB_p, Term_p, Term_p);
CompareResult lpofunequal(OCB_p, Term_p, Term_p);
CompareResult lpocheckargument(OCB_p ocb, Term_p s, Term_p t);
CompareResult lpogtr(OCB_p, Term_p, Term_p);
CompareResult lpogtrcomparevars(OCB_p ocb, Term_p s, bool varps, Term_p t);
CompareResult lpogtrfunequal(OCB_p ocb, Term_p s, Term_p t);
CompareResult lpogtrcheckargument(OCB_p ocb, Term_p s, Term_p t);


/*-----------------------------------------------------------------------
//
// Function: LPOCompare(ocb, s, t)
//
//   Compare two terms s,t in the Lexicographic Path Ordering,
//   return the result
//                          to_greater         if s >LPO t
//                          to_equal           if s =LPO t
//                          to_lesser          if t >LPO s
//                          to_uncomparable    otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult LPOCompare(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res, res_funs;
   bool varps, varpt;

   s = TermDeref(s);
   t = TermDeref(t);

printf("LPOCompare: s=");
TermPrint(stdout, s, ocb->sig, false);
printf("   t=");
TermPrint(stdout, t, ocb->sig, false);
printf("\n");

   /* The cases s=x v t=x are checked separately. */
   varpt = TermIsVar(t);
   if ((varps = TermIsVar(s)) || varpt)
   {
printf("LPOCompare returns the value of lpocomparevars:\n");
      return lpocomparevars(ocb,s,varps,t);
   }

   /* Depending on the comparison of the leading operators wrt. the
   // precedence, different conditions must be tested.           */
if (OCBFunCompare(ocb,s->f_code,t->f_code) == to_greater)
printf("topsymbolcomparison in LPOCompare: to_greater\n");
if (OCBFunCompare(ocb,s->f_code,t->f_code) == to_equal)
printf("topsymbolcomparison in LPOCompare: to_equal\n");
if (OCBFunCompare(ocb,s->f_code,t->f_code) == to_lesser)
printf("topsymbolcomparison in LPOCompare: to_lesser\n");
if (OCBFunCompare(ocb,s->f_code,t->f_code) == to_uncomparable)
printf("topsymbolcomparison in LPOCompare: to_uncomparable\n");

   switch ((res_funs = OCBFunCompare(ocb,s->f_code,t->f_code)))
   {
   case to_greater:
printf("######## jetzt in case to_greater von LPOCompare\n");
	 if ((res = lpofungreater(ocb, s, t)) != to_uncomparable)
	 {
printf("LPOCompare-to_greater returns the value of lpofungreater\n");
	    return res;
	 }
	 break;
   case to_equal:
printf("######## jetzt in case to_equal von LPOCompare\n");
	 if ((res = lpofunequal(ocb, s, t)) != to_uncomparable)
	 {
printf("LPOCompare-to_equal returns the value of lpofunequal\n");
	    return res;
	 }
	 break;
   case to_lesser:
printf("######## jetzt in case to_lesser von LPOCompare\n");
	 if ((res = lpofungreater(ocb, t, s)) != to_uncomparable)
	 {
printf("LPOCompare-to_lesser returns the complement-value of lpofungreater\n");
	    return (res == to_greater) ? to_lesser : to_greater;
	 }
	 break;
   default:
	 break;
   }
printf("######## jetzt im default case von LPOCompare\n");
   if ((res_funs != to_greater) && \
       ((res_funs != to_equal) || (s->arity >= 2)) && \
       (lpocheckargument(ocb, s, t) == to_greater))
   {
printf("LPOCompare-default returns to_greater\n");
      return to_greater;
   }
   if ((res_funs != to_lesser) && \
       ((res_funs != to_equal) || (s->arity >= 2)) && \
       (lpocheckargument(ocb, s, t) == to_lesser))
   {
printf("LPOCompare-default returns to_lesser\n");
      return to_lesser;
   }
printf("LPOCompare-default returns to_uncomparable\n");
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: lpocomparevars(ocb, s, varps, t)
//
// Compares two terms s and t wrt. LPO if either s or t is a variable
// and returns
//               to_greater        if t is a subterm of s
//               to_equal          if s == t
//               to_lesser         if s is a subterm of t
//               to_uncomparable   otherwise
//
// If s is a variable, then varps is true.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpocomparevars(OCB_p ocb, Term_p s, bool varps, Term_p t)
{
printf("--lpocomparevars: s=");
TermPrint(stdout, s, ocb->sig, false);
printf("   t=");
TermPrint(stdout, t, ocb->sig, false);
printf("\n");

   if (varps)
   {
      if (TBTermEqual(s, t))
      {
printf("----lpocomparevars returns to_equal\n");
	 return to_equal;
      }
      else
      {
	 if (TermIsSubterm(t, s, true, TBTermEqual))
	 {
printf("----lpocomparevars returns to_lesser\n");
	    return to_lesser;
	 }
	 else
	 {
printf("----lpocomparevars returns to_uncomparable\n");
	    return to_uncomparable;
	 }
      }
   }
   else
   {
      if (TermIsSubterm(s, t, true, TBTermEqual))
      {
printf("----lpocomparevars returns to_greater\n");
	 return to_greater;
      }
      else
      {
printf("----lpocomparevars returns to_uncomparable\n");
	 return to_uncomparable;
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: lpofungreater(ocb, s, t)
//
// Compares the term s with the arguments t_1,...,t_n of t and returns
//
//       to_greater        if s >LPO t_1, ..., s >LPO t_n
//       to_lesser         if there exists t_i with t_i >=LPO s
//       to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpofungreater(OCB_p ocb, Term_p s, Term_p t)
{
   bool greater = true;
   int i;
   
printf("lpofungreater: s=");
TermPrint(stdout, s, ocb->sig, false);
printf("   t=");
TermPrint(stdout, t, ocb->sig, false);
printf("\n");

   for (i=0; i<t->arity; i++)
      switch (LPOCompare(ocb, s, t->args[i]))
      {
      case to_greater:
printf("lpofungreater returns to_greater\n");
	    break;
      case to_lesser: case to_equal:
printf("lpofungreater returns to_lesser\n");
	    return to_lesser;
	    break;
      case to_uncomparable:
	    greater = false;
	    break;
      default:
	    assert(false);
	    break;
      }

   if (greater)
   {
printf("lpofungreater returns to_greater\n");
      return to_greater;
   }
   else
   {
printf("lpofungreater returns to_uncomparable\n");
      return to_uncomparable;
   }
}


/*-----------------------------------------------------------------------
//
// Function: lpofunequal(ocb, s, t)
//
// Compares the arguments s_1,...,s_m of the term s with the arguments
// t_1,...,t_n of the term t and returns
//
//        to_equal          if m=n & s_1 =LPO t_1, ..., s_m =LPO t_m
//        to_greater        if (s_1,...,s_m) >LPOlex (t_1,...,t_n)  &
//                             forall t_i: s >LPO t_i
//        to_lesser         if (t_1,...,t_n) >LPOlex (s_1,...,s_n)  &
//                             forall s_i: t >LPO s_i
//        to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpofunequal(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res = to_equal;
   int i, j;
   
printf("lpofunequal: s=");
TermPrint(stdout, s, ocb->sig, false);
printf("   t=");
TermPrint(stdout, t, ocb->sig, false);
printf("\n");

   for (i=0; i<MAX(s->arity,t->arity); i++)
   {
      if (t->arity < i)
      {
printf("lpofunequal-arity< returns to_greater\n");
	 return to_greater;
      }
      if (s->arity < i)
      {
printf("lpofunequal-arity< returns to_lesser\n");
	 return to_lesser;
      }
      if ((res = LPOCompare(ocb, s->args[i], t->args[i])) == to_uncomparable)
      {
printf("lpofunequal returns to_uncomparable\n");
	 return to_uncomparable;
      }
      if ((res == to_greater) || (res == to_lesser))
      {
	 CompareResult res_help = res;
	 CompareResult comp_res = (res == to_greater) ? to_lesser : to_greater;
	 
	 j = i+1;
	 while ((j<MAX(s->arity, t->arity)) && (res_help == res))
	 {
	    if (res == to_greater)
	    {
	       res_help = LPOCompare(ocb, s, t->args[j]);
	    }
	    else
	    {
	       res_help = LPOCompare(ocb, s->args[j], t);
	    }
	    if (res_help == comp_res)
	    {
printf("lpofunequal returns `comp_res'\n");
	       return comp_res;
	    }
	    j++;
	 }
	 if (res_help == res)
	 {
printf("lpofunequal returns `res'\n");
	    return res;
	 }
      }
   }
   if (res == to_equal)
   {
printf("lpofunequal returns to_equal\n");
      return to_equal;
   }
   else
   {
printf("lpofunequal returns to_uncomparable\n");
      return to_uncomparable;
   }
}


/*-----------------------------------------------------------------------
//
// Function: lpocheckargument(ocb, s, t)
//
// Checks the third condition of the LPO, i.e. returns
//
//        to_greater        if there is an argument s_i of s
//                             with s_i >=LPO t
//        to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpocheckargument(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res;
   int i;
   
printf("lpocheckargument: s=");
TermPrint(stdout, s, ocb->sig, false);
printf("   t=");
TermPrint(stdout, t, ocb->sig, false);
printf("\n");

   for (i=0; i<s->arity; i++)
   {
      res = LPOCompare(ocb, s->args[i], t);
      if ((res == to_greater) || (res == to_equal))
      {
printf("lpocheckargument returns to_greater\n");
	 return to_greater;
      }
   }
printf("lpocheckargument returns to_uncomparable\n");
   return to_uncomparable;
}

/*-----------------------------------------------------------------------
//
// Function: LPOGreater(ocb, s, t)
//
//   Checks whether the term s is greater than the term t in the
//   Lexicographic Path Ordering (LPO), i.e. return
//
//                       true    if s >LPO t
//      		 false   otherwise
//
//   For a description of the LPO see the header of this file.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

bool LPOGreater(OCB_p ocb, Term_p s, Term_p t)
{
   if (lpogtr(ocb, s, t) == to_greater)
   {
      return true;
   }
   else
   {
      return false;
   }
}

/*-----------------------------------------------------------------------
//
// Function: lpogtr(ocb, s, t)
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/
CompareResult lpogtr(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res, res_funs;
   bool varps, varpt;
   int i;
   
   s = TermDeref(s);
   t = TermDeref(t);
   
   /* The cases s=x v t=x are checked separately. */
   varpt = TermIsVar(t);
   if ((varps = TermIsVar(s)) || varpt)
   {
      return lpogtrcomparevars(ocb,s,varps,t);
   }

   /* Depending on the comparison of the leading operators wrt. the
   // precedence, different conditions must be tested.           */
   switch ((res_funs = OCBFunCompare(ocb, s->f_code, t->f_code)))
   {
   case to_greater:
	 res = to_greater;
	 for (i=0; i<t->arity; i++)
	    if ((res = lpogtr(ocb, s, t->args[i])) != to_greater)
	    {
	       break;
	    }
	 if (res == to_greater)
	 {
	    return to_greater;
	 }
	 break;
   case to_lesser:
	 for (i=0; i<s->arity; i++)
	 {
	    res = lpogtr(ocb, s->args[i], t);
	    if ((res == to_greater) || (res == to_equal))
	    {
	       return to_greater;
	    }
	 }
	 return to_uncomparable;
	 break;
   case to_equal:
	 if ((res = lpogtrfunequal(ocb, s, t)) != to_uncomparable)
	 {
	    return res;
	 }
	 break;
   default:
	 break;
   }

   if (s->arity == 1)
   {
      return to_uncomparable;
   }
   if ((res_funs != to_greater) && \
       ((res_funs != to_equal) || (s->arity >= 2)))
   {
      return lpogtrcheckargument(ocb, s, t);
   }
   return to_uncomparable;
}


/*-----------------------------------------------------------------------
//
// Function: lpogtrcomparevars(ocb, s, varps, t)
//
// Compares two terms s and t wrt. LPO if either s or t is a variable
// and returns
//               to_greater        if t is a subterm of s
//               to_equal          if s == t
//               to_uncomparable   otherwise
//
// If s is a variable, then varps is true.
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpogtrcomparevars(OCB_p ocb, Term_p s, bool varps, Term_p t)
{
   if (varps)
   {
      if (TBTermEqual(s, t))
      {
	 return to_equal;
      }
      else
      {
	 return to_uncomparable;
      }
   }
   else
   {
      if (TermIsSubterm(s, t, true, TBTermEqual))
      {
	 return to_greater;
      }
      else
      {
	 return to_uncomparable;
      }
   }
}


/*-----------------------------------------------------------------------
//
// Function: lpogtrfunequal(ocb, s, t)
//
// Compares the arguments s_1,...,s_m of the term s with the arguments
// t_1,...,t_n of the term t and returns
//
//        to_equal          if m=n & s_1 =LPO t_1, ..., s_m =LPO t_m
//        to_greater        if (s_1,...,s_m) >LPOlex (t_1,...,t_n)  &
//                             forall t_i: s >LPO t_i
//        to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpogtrfunequal(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res = to_equal;
   int i;

   for (i=0; i<MAX(s->arity, t->arity); i++)
   {
      if (t->arity < i)
      {
	 return to_greater;
      }
      if (s->arity < i)
      {
	 return to_uncomparable;
      }
      
      res = lpogtr(ocb, s->args[i], t->args[i]);
      if (res == to_uncomparable)
      {
	 break;
      }
      if (res == to_greater)
      {
	 int j = i+1;
	 while ((j<MAX(s->arity, t->arity)) && (res == to_greater))
	 {
	    res = lpogtr(ocb, s, t->args[j]);
	    j++;
	 }
	 if (res == to_greater)
	 {
	    return to_greater;
	 }
	 else
	 {
	    return to_uncomparable;
	 }
      }
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: lpogtrcheckargument(ocb, s, t)
//
// Checks the third condition of the LPO, i.e. returns
//
//        to_greater        if there is an argument s_i of s
//                             with s_i >=LPO t
//        to_uncomparable   otherwise
//
// Global Variables: -
//
// Side Effects    : -
//
-----------------------------------------------------------------------*/

CompareResult lpogtrcheckargument(OCB_p ocb, Term_p s, Term_p t)
{
   CompareResult res;
   int i;

   for (i=0; i<s->arity; i++)
   {
      res = lpogtr(ocb, s->args[i], t);
      if ((res == to_greater) || (res == to_equal))
      {
	 return to_greater;
      }
   }
   return to_uncomparable;
}
