/*-----------------------------------------------------------------------

File  : cto_ocb.c

Author: Stephan Schulz

Contents
 
  Functions for describing orderings, precedences and so on.

Changes

<1> Thu Apr 30 03:11:31 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include "cto_ocb.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

char* TOCompareSymbol[]=
{
   "=/=",
   " = ",
   " > ",
   " < "
};

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: ocb_trans_compute()
//
//   Given the relations between f1 and f2, and f2 and f3, compute the
//   relation between f1 and f3. Return true, if it can be set, false
//   otherwise. 
//
// Global Variables: -
//
// Side Effects    : Changes ocb->precedences
//
/----------------------------------------------------------------------*/

bool ocb_trans_compute(OCB_p ocb, FunCode f1, FunCode f2, FunCode f3)
{
   CompareResult rel12, rel23;
   bool          res = true;

   rel12 = OCBFunCompare(ocb, f1, f2);
   rel23 = OCBFunCompare(ocb, f2, f3);

   switch(rel12)
   {
   case to_uncomparable:
	 break;
   case to_equal:
	 if(rel23 != to_uncomparable)
	 {
	    res = OCBPrecedenceAddTuple(ocb, f1, f3, rel23);
	 }
	 break;
   case to_greater:
	 if(rel23 == to_equal || rel23 == to_greater)
	 {
	    res = OCBPrecedenceAddTuple(ocb, f1, f3, to_greater);
	 }
	 break;
   case to_lesser:
	 if(rel23 == to_equal || rel23 == to_lesser)
	 {
	    res = OCBPrecedenceAddTuple(ocb, f1, f3, to_lesser);
	 }
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: OCBAlloc()
//
//   Allocate an initialized order control block.
//
// Global Variables: -
//
// Side Effects    : Memory management
//
/----------------------------------------------------------------------*/

OCB_p OCBAlloc(TermOrdering type, Sig_p sig)
{
   OCB_p handle;
   int   i,j;

   handle = OCBCellAlloc();

   handle->type  = type;
   handle->sig   = sig;
   handle->sig_size = sig->f_count;
   handle->statestack = PStackAlloc();
   switch(type)
   {
   case KBO:
   case KBO2:
	 handle->weights    = SizeMalloc(sizeof(long)*handle->sig_size);
	 handle->precedence = SizeMalloc(sizeof(CompareResult)
					 *handle->sig_size
					 *handle->sig_size);
	 handle->status     = SizeMalloc(sizeof(CompareStatus)
					  *handle->sig_size);
	 break;
   case LPO:
	 handle->weights    = NULL;
	 handle->precedence = SizeMalloc(sizeof(CompareResult)
					 *handle->sig_size
					 *handle->sig_size);
	 handle->status     = NULL;
	 break;
   case RPO:
	 handle->weights    = NULL;
	 handle->precedence = SizeMalloc(sizeof(CompareResult)
					 *handle->sig_size
					 *handle->sig_size);
	 handle->status     =  SizeMalloc(sizeof(CompareStatus)
					  *handle->sig_size);
	 break; 
   default:
	 assert(false);
	 break;
   }

   if(handle->weights)
   {
      for(i=1; i<=handle->sig_size; i++)
      {
	 *OCBFunWeightPos(handle,i) = 1;
      }
   }
   if(handle->precedence)
   {
      for(i=1; i<=handle->sig_size; i++)
      {
	 for(j=1; j<=handle->sig_size; j++)
	 {
	    *OCBFunComparePos(handle, i,j) = 
	       ((i==j) ? to_equal : to_uncomparable);
	 }
      }
   }
   if(handle->status)
   {
      for(i=0; i<handle->sig_size; i++)
      {
	 *OCBFunStatusPos(handle,i) = Multiset;
      }
   }

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: OCBFree()
//
//   Free the memory taken by an order control block. Note: The
//   signature is not considered part of the ocb and is not free'd.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void OCBFree(OCB_p junk)
{
   assert(junk);

   if(junk->weights)
   {
      assert(junk->type == KBO || junk->type == KBO2);
      SizeFree(junk->weights, sizeof(long)*junk->sig_size);
   }
   if(junk->precedence)
   {
      assert(junk->type == KBO || junk->type == KBO2 || junk->type ==
	     LPO || junk->type == RPO);
      SizeFree(junk->precedence, sizeof(CompareResult) 
	       * junk->sig_size * junk->sig_size);
   }
   if(junk->status)
   {
      assert(junk->type == RPO || junk->type == KBO || junk->type ==
	     KBO2);
      SizeFree(junk->status, sizeof(CompareStatus)*junk->sig_size);
   }
   PStackFree(junk->statestack);
   OCBCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: OCBDebugPrint()
//
//   Print an OCB in debug-friendly form (not suitable for
//   re-parsing, revealing a lot of internal information).
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void OCBDebugPrint(FILE* out, OCB_p ocb)
{
   long i,j;

   fprintf(out, "# ==============OCB-Debug-Information============\n");
   fprintf(out, "# ===============================================\n");
   if(ocb->weights)
   {
      fprintf(out, "# Weights: ");
      for(i=1; i<=ocb->sig_size;i++)
      {
	 fprintf(out, " (%ld = %ld) ", i, OCBFunWeight(ocb, i));
      }
      fprintf(out, "\n\n");
   }
   else
   {
      fprintf(out, "# No weights!\n");
   } 
   fprintf(out, "# -----------------------------------------------\n");
   if(ocb->precedence)
   {
      fprintf(out, "# Precedence Matrix:\n#       ");
      for(j=1; j<=ocb->sig_size; j++)
      {
	 fprintf(out, " %2ld ", j);
      }
      fprintf(out, "\n");
      for(i=1; i<=ocb->sig_size; i++)
      {
	 fprintf(out, "# %2ld  | ", i);
	 for(j=1; j<=ocb->sig_size; j++)
	 {
	    char* symb;

	    symb = TOCompareSymbol[OCBFunCompare(ocb, i, j)];
	    fprintf(out, " %s", symb);
	 }
	 fprintf(out, "\n");
      }
   }
   else
   {
      fprintf(out, "# No precedence!\n");
   } 
   fprintf(out, "# -----------------------------------------------\n");
   if(ocb->status)
   {
      fprintf(out, "# Status: ");
      for(i=1; i<=ocb->sig_size;i++)
      {
	 fprintf(out, " (%ld = %d) ", i, OCBFunStatus(ocb, i));
      }
      fprintf(out, "\n\n");
   }
   else
   {
      fprintf(out, "# No status!\n");
   }    
   fprintf(out, "# ===============================================\n");
}



/*-----------------------------------------------------------------------
//
// Function: OCBPrecedenceInverse()
//
//   Given a precedence relation, return the inverse relation.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

CompareResult OCBPrecedenceInverse(CompareResult relation)
{
   CompareResult res = relation;

   switch(relation)
   {
   case to_equal:
   case to_uncomparable:
	 break;
   case to_greater:
	 res = to_lesser;
	 break;
   case to_lesser:
	 res = to_greater;
	 break;
   default:
	 assert(false);
	 break;
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: OCBPrecedenceAddTuple()
//
//   Add a new binary relation to the precedence stored in the ocb and
//   compute the new transitive closure of the to_greater, to_smaller
//   and to_equal. Store updated cell in ocb->statestackcell. Return
//   the new stackpointer if everything went fine, undo all changes
//   and return 0 otherwise.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

PStackPointer OCBPrecedenceAddTuple(OCB_p ocb, FunCode f1, FunCode f2,
			   CompareResult relation)
{
   FunCode       i;
   PStackPointer res = 0, old;

   assert(ocb);
   assert(ocb->precedence);
   assert(f1<=ocb->sig_size);
   assert(f2<=ocb->sig_size);
   assert(relation!=to_uncomparable);

   old = PStackGetSP(ocb->statestack);

   if(OCBFunCompare(ocb, f1, f2)==relation)
   {
      res = old;
   }
   else if(OCBFunCompare(ocb, f1, f2)!=to_uncomparable)
   {
      res = 0;
   }
   else
   {
      PStackPushInt(ocb->statestack, f1);
      PStackPushInt(ocb->statestack, f2);
      *OCBFunComparePos(ocb, f1, f2) = relation;
      *OCBFunComparePos(ocb, f2, f1) = OCBPrecedenceInverse(relation);

      for(i=1; i<=ocb->sig_size; i++)
      {
	 res = ocb_trans_compute(ocb, f1, f2, i);
	 if(!res)
	 {
	    break;
	 }
	 res = ocb_trans_compute(ocb, i, f1, f2);
	 if(!res)
	 {
	    break;
	 }
      }
      if(!res)
      { /* Error case, undo changes */
	 f2 = PStackPopInt(ocb->statestack);
	 f1 = PStackPopInt(ocb->statestack);
	 *OCBFunComparePos(ocb, f1, f2) = to_uncomparable;
	 *OCBFunComparePos(ocb, f2, f1) = to_uncomparable;
      }
   }
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: OCBPrecedenceBacktrack()
//
//   Backtrack the precedence matrix to a given state. Return true if
//   the stack is non-empty afterwards, false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes the matrix
//
/----------------------------------------------------------------------*/

bool OCBPrecedenceBacktrack(OCB_p ocb, PStackPointer state)
{
   FunCode f1,f2;
   
   while(state!=PStackGetSP(ocb->statestack))
   {
      assert(!PStackEmpty(ocb->statestack));
      f2 = PStackPopInt(ocb->statestack);
      assert(!PStackEmpty(ocb->statestack));
      f1 = PStackPopInt(ocb->statestack);

      assert(OCBFunCompare(ocb, f1, f2) != to_uncomparable);
      *OCBFunComparePos(ocb, f1, f2) = to_uncomparable;
      assert(OCBFunCompare(ocb, f2, f1) != to_uncomparable);
      *OCBFunComparePos(ocb, f2, f1) = to_uncomparable;
   }
   return !PStackEmpty(ocb->statestack);
}

/*-----------------------------------------------------------------------
//
// Function: OCBTermMaxFunCode()
//
//   Return the (or rather a) maximal function symbol (according to
//   ocb->precedence) from term. Follows bindings exactly once
//   (i.e. assumes that substitutions are matches).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

FunCode OCBTermMaxFunCode(OCB_p ocb, Term_p term)
{
   int       i;
   FunCode   res = 0, tmp ;
   DerefType deref = DEREF_ONCE;

   assert(ocb->precedence);

   term = TermDeref(term, &deref);
   
   if(TermIsVar(term))
   {
      return res;
   }
   res = term->f_code;

   for(i=1; i<term->arity; i++)
   {
      tmp = OCBTermMaxFunCode(ocb, term->args[i]);
      if(OCBFunCompare(ocb, tmp, res) == to_greater)
      {
	 res = tmp;
      }
   }
   return res;
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


