/*-----------------------------------------------------------------------

File  : eprover.c

Author: Stephan Schulz

Contents
 
  Main program for the E equational theorem prover.

Changes

<1> Tue Jun  9 01:32:15 MET DST 1998
    New

-----------------------------------------------------------------------*/

#include <sys/time.h>
#include <sys/resource.h>
#include <unistd.h>

#include <cio_commandline.h>
#include <cio_output.h>
#include <cco_proofproc.h>

/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_SATURATED,
   OPT_SILENT,
   OPT_OUTPUTLEVEL,
   OPT_MEM_LIMIT,
   OPT_STEP_LIMIT,
   OPT_PROC_LIMIT,
   OPT_UNPROC_LIMIT,
   OPT_TOTAL_LIMIT,
   OPT_DISCOUNT_VARS,
   OPT_NO_INFIX,
   OPT_FULL_EQ_REP,
   OPT_NO_NEG_PARAMOD,
   OPT_NO_NONEQ_PARAMOD,
   OPT_ORDERING,
   OPT_INDEXING,
   OPT_RULES_GENERAL,
   OPT_STRONGSUBSUMPTION,
   OPT_HEURISTIC,
   OPT_HEU_INT,
   OPT_HEU_FLOAT,
   OPT_HEU_STRING
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},

   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program. This differs "
    "from the output level (below) in that technical information is "
    "printed to stderr, while the output level determines which "
    "logical manipulations of the clauses are printed."},

   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},

   {OPT_SILENT,
    's', "silent",
    NoArg, NULL,
    "Equivalent to --output-level=0."},

   {OPT_OUTPUTLEVEL,
    'l', "output-level",
    ReqArg, NULL,
    "Select an output level, greater values imply more verbose "
    "output. Level 0 produces "
    "nearly no output, level 1 will output each clause as it is "
    "processed, level 2 will output generating inferences (not yet "
    "implemented)  and level three will give a full protocol "
    "including rewrite steps."},
   
   {OPT_PRINT_SATURATED,
    'S', "print-saturated",
    NoArg, NULL,
    "Print the (semi-) saturated clause sets after terminating the "
    "saturation process."},

   {OPT_MEM_LIMIT,
    'm', "memory-limit",
    OptArg, "64",
    "Limit the memory the prover may use. The optional argument is "
    "the allowed amount of memory in MB. If no argument is given, a "
    "limit of 64 MB is assumed. This option apparently works only "
    "under Solaris 2.6 (and up, presumely) and Linux."},

   {OPT_STEP_LIMIT,
    'C', "processed-clauses-limit",
    ReqArg, NULL,
    "Set the maximal number of clauses to process."},
   
   {OPT_PROC_LIMIT,
    'P', "processed-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of processed clauses."},

   {OPT_UNPROC_LIMIT,
    'U', "unprocessed-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of unprocessed clauses."},
   
   {OPT_TOTAL_LIMIT,
    'T', "total-clause-set-limit",
    ReqArg, NULL,
    "Set the maximal size of the set of all clauses."},

   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    OptArg, "true",
    "Read and write PCL terms (otherwise use PROLOG terms)."},

   {OPT_NO_INFIX,
    'n', "eqn-no-infix",
    OptArg, "true",
    "Print equations in prefix notation equal(x,y)."},

   {OPT_FULL_EQ_REP,
    'e', "full-equational-rep",
    OptArg, "true",
    "Print all literals as equations, even non-equational ones."},

   {OPT_NO_NEG_PARAMOD,
    'N', "no-paramod-into-negatives",
    OptArg, "true",
    "Refrain from paramodulating into negative literals."},
   
   {OPT_NO_NONEQ_PARAMOD,
    'E', "paramod-equational-literals-only",
    OptArg, "true",
    "Don't paramodulate from or into non-equational literals."},

   {OPT_HEURISTIC,
    'x', "expert-heuristic",
    ReqArg, NULL,
    "Select one of the heuristics. Currently available: Standard, "
    "Clauseweight, FIFOWeight."}, 

   {OPT_ORDERING,
    't', "term-ordering",
    ReqArg, NULL,
    "Select an ordering type (currently LPO, KBO or KBO2)."},      

   {OPT_INDEXING,
    'p', "no-pdt-indexing",
    NoArg, NULL,
    "Use conventional matching for forward rewriting and subsumption "
    "(default is to use perfect discrimination trees)."}, 

   {OPT_RULES_GENERAL,
    'g', "prefer-general-demodulators",
    NoArg, NULL,
    "If using perfect discrimination tree indexing, prefer general "
    "demodulators. By default, use specialized demodulators."},

   {OPT_STRONGSUBSUMPTION,
    'u', "strong-forward-subsumption",
    NoArg, NULL,
    "Try multiple positions and unit-equations to try to subsume a "
    "single new clause. Default is to search for a single position."},

   {OPT_HEU_INT,
    'i', "heuristic-int-arg",
    ReqArg, NULL,
    "Give an integer argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_HEU_FLOAT,
    'f', "heuristic-float-arg",
    ReqArg, NULL,
   "Give a float argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_HEU_STRING,
    's', "heuristic-string-arg",
    ReqArg, NULL,
   "Give a string argument to the selected heuristic. This option "
    "can be given multiple times, order is significant."},

   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

TermOrdering     ordertype = NoOrdering;
char             *outname = NULL;
HeuristicParms_p h_parms;
bool             indexing = true,
                 print_sat = false;
long             step_limit = LONG_MAX, 
                 proc_limit = LONG_MAX,
                 unproc_limit = LONG_MAX, 
                 total_limit = LONG_MAX;
struct rlimit    memlimit = {RLIM_INFINITY, RLIM_INFINITY};


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p        state;
   Scanner_p        in;    
   ProofState_p     proofstate;
   int              i;
   Clause_p         success;

   assert(argv[0]);
   InitError(argv[0]);

   h_parms = HeuristicParmsAlloc();
   h_parms->heuristic_name = "Clauseweight";

   state = process_options(argc, argv);

   if(memlimit.rlim_cur != RLIM_INFINITY)
   {
      if(setrlimit(RLIMIT_DATA, &memlimit))
      {
	 SysError("Unable to set memory limit", SYS_ERROR);
      }
   }

   GlobalOut = OutOpen(outname);

   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }
   
   for(i=0; state->argv[i]; i++)
   {
      in = CreateScanner(state->argv[i] , true, NULL);
      
      proofstate = ProofStateAlloc();
      proofstate->use_indexing = indexing;
      proofstate->ordertype = ordertype;
      ClauseSetParseList(in, proofstate->axioms,
			 proofstate->original_terms);
      VERBOUT2("Specification read\n");
      InitProofState(proofstate, h_parms);
      VERBOUT2("Prover state initialized\n");
      success = Saturate(proofstate, step_limit, proc_limit,
			 unproc_limit, total_limit);
      if(success)      
      {
	 fprintf(GlobalOut, "\n# Proof found!\n");
	 ClauseFree(success);
      }
      else
      {
	 fprintf(GlobalOut, "\n# No proof found!\n");
      }
      if(print_sat)
      {
	 ProofStatePrint(GlobalOut, proofstate);
	 fprintf(GlobalOut, "\n");
      }
      if(OutputLevel)
      {
	 ProofStateStatisticsPrint(GlobalOut, proofstate);
      }

      ProofStateFree(proofstate);
      DestroyScanner(in);
   }

   OutClose(GlobalOut);
   CLStateFree(state);

   HeuristicParmsFree(h_parms);

#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif

   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_SILENT:
	    OutputLevel = 0;
	    break;
      case OPT_OUTPUTLEVEL:
	    OutputLevel = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PRINT_SATURATED:
	    print_sat = true;
	    break;
      case OPT_MEM_LIMIT:
	    memlimit.rlim_cur = MEGA*CLStateGetIntArg(handle, arg);
	    break;
      case OPT_STEP_LIMIT:
	    step_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_PROC_LIMIT:
	    proc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_UNPROC_LIMIT:
	    unproc_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_TOTAL_LIMIT:
	    total_limit = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_INFIX:
	    EqnUseInfix = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FULL_EQ_REP:
	    EqnFullEquationalRep = CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_NEG_PARAMOD:
	    ParamodOverlapIntoNegativeLiterals = !CLStateGetBoolArg(handle, arg);
	    break;	    
      case OPT_NO_NONEQ_PARAMOD:
	    ParamodOverlapNonEqLiterals = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_HEURISTIC:
	    h_parms->heuristic_name = arg;
	    break;
      case OPT_ORDERING:
	    if(strcmp(arg, "LPO")==0)
	    {
	       ordertype = LPO;
	    }
	    else if(strcmp(arg, "KBO")==0)
	    {
	       ordertype = KBO;
	    }
	    else if(strcmp(arg, "KBO2")==0)
	    {
	       ordertype = KBO2;
	    }
	    else
	    {
	       Error("Option -t (--term-ordering) requires LPO or KBO"
		     " as an argument", USAGE_ERROR);
	    }
	    break;
      case OPT_INDEXING:
	    indexing = false;
	    break;
      case OPT_RULES_GENERAL:
	    PDTreePreferGeneral = true;
	    break;
      case OPT_STRONGSUBSUMPTION:
	    StrongUnitForwardSubsumption = true;
	    break;
      case OPT_HEU_INT:
	    PStackPushInt(h_parms->i_parms, CLStateGetIntArg(handle,
							     arg));
	    break;
      case OPT_HEU_FLOAT:
	    DStackPush(h_parms->d_parms, CLStateGetFloatArg(handle,
							    arg));
	    break;
      case OPT_HEU_STRING:
	    PStackPushP(h_parms->s_parms, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}

void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
E 0.2\n\
\n\
Usage: eprover [options] [files]\n\
\n\
Read sets of clauses and try to refute them.\n\
\n");
   PrintOptions(stdout, opts);
   fprintf(out, "\n\
Copyright 1998 by Stephan Schulz, schulz@informatik.tu-muenchen.de\n\
\n\
This program is free software; you can redistribute it and/or modify\n\
it under the terms of the GNU General Public License as published by\n\
the Free Software Foundation; either version 2 of the License, or\n\
(at your option) any later version.\n\
\n\
This program is distributed in the hope that it will be useful,\n\
but WITHOUT ANY WARRANTY; without even the implied warranty of\n\
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n\
GNU General Public License for more details.\n\
\n\
You should have received a copy of the GNU General Public License\n\
along with this program (it should be contained in the top level\n\
directory of the distribution in the file COPYING); if not, write to\n\
the Free Software Foundation, Inc., 59 Temple Place, Suite 330,\n\
Boston, MA  02111-1307 USA\n\
\n\
The original copyright holder can be contacted as\n\
\n\
Stephan Schulz\n\
Technische Universitaet Muenchen\n\
Fakultaet fuer Informatik\n\
Arcisstrasse 20\n\
D-80290 Muenchen\n\
Germany\n\
");

}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


