/*-----------------------------------------------------------------------

File  : cte_terms.c

Author: Stephan Schulz

Contents
 
  Basic functionality of (unshared) terms. Includes some support for
  future extensions. 


Changes

<1> Mon Nov 17 00:58:58 MET 1997
    New

-----------------------------------------------------------------------*/

#include "cte_terms.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool      TermPrintLists = true;
bool      TermPrologVars = false;
TokenType TermIdentToken = Identifier | PosInt;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: print_cons_list()
//
//   Print a list of $cons'ed terms, terminated with $nil. Abort on
//   not well-formed lists (no cons pairs!).
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void print_cons_list(FILE* out, Term_p list, Sig_p sig)
{
   assert(SigSupportLists);
   putc('[', out);
   if(list->f_code == SIG_CONS_CODE)
   {
      assert(list->args);
      TermPrint(out, list->args[0], sig);
      list = list->args[1];
      while(list->f_code == SIG_CONS_CODE)
      {
	 putc(',', out);
	 /* putc(' ', out); */
	 assert(list->args);
	 TermPrint(out, list->args[0], sig);
	 list = list->args[1];
      }
      assert(list->f_code == SIG_NIL_CODE);
   }
   putc(']', out);
}


/*-----------------------------------------------------------------------
//
// Function: parse_cons_list()
//
//   Parse a LOP list into an internal $cons list.
//
// Global Variables: -
//
// Side Effects    : Input, Memory operations
//
/----------------------------------------------------------------------*/

static Term_p parse_cons_list(Scanner_p in, Sig_p sig, VarTrans_p vars)
{
   Term_p handle;
   Term_p current;
   
   assert(SigSupportLists);
   
   AcceptInpTok(in, OpenSquare);

   handle = DefaultTermCellAlloc();
   TermSetProp(handle, TPTopPos);
   current = handle;
   
   if(!TestInpTok(in, CloseSquare))
   {

      current->f_code = SIG_CONS_CODE;
      current->arity = 2;
      current->args = TermArgArrayAlloc(2);
      current->args[0] = TermParse(in, sig, vars); 
      current->args[1] = DefaultTermCellAlloc();
      current = current->args[1];
      
      while(TestInpTok(in, Comma))
      {
	 NextToken(in);
	 current->f_code = SIG_CONS_CODE;             
	 current->arity = 2;                          
         current->args = TermArgArrayAlloc(2);
	 current->args[0] = TermParse(in, sig, vars); 
	 TermDelProp(current->args[0], TPTopPos);
	 current->args[1] = DefaultTermCellAlloc();
	 current = current->args[1];      
      }
   }
   AcceptInpTok(in, CloseSquare);
   current->f_code = SIG_NIL_CODE;             
   
   return handle;
}
   

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: DefaultTermCellAlloc()
//
//   Allocate a term cell with default values.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p DefaultTermCellAlloc()
{
   Term_p handle;
   
   handle = TermCellAlloc();

   handle->flags = 0;
   handle->arity = 0;
   handle->binding = NULL;
   handle->args = NULL;
   handle->entry_no = 0;
   handle->ref_count = 0;
   handle->superterms = NULL;
   handle->ext_refs = NULL;
   handle->balance = 0;
   handle->passed = 0;
   handle->lson = NULL;
   handle->rson = NULL;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermTopFree()
//
//   Return term cell and arg array (if it exists).
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void TermTopFree(Term_p junk)
{
   if(junk->arity)
   {
      assert(junk->args);
      TermArgArrayFree(junk->args, junk->arity);
   }
   else
   {
      assert(!junk->args);
   }
   PTreeFree(junk->superterms);
   PTreeFree(junk->ext_refs);
   TermCellFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function:  TermFree()
//
//   Return the memory taken by an (unshared) term.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void TermFree(Term_p junk)
{
   assert(junk);
   assert(junk->ref_count == 0);
   if(junk->arity)
   {
      int i;
      
      assert(junk->args);
      for(i=0; i<junk->arity; i++)
      {
	 TermFree(junk->args[i]);
      }
   }
   else
   {
      assert(!junk->args);
   }
   TermTopFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function: TermPrint()
//
//   Print a term to the given stream.
//
// Global Variables: TermPrintLists, TermPrologVars
//
// Side Effects    : Output 
//
/----------------------------------------------------------------------*/

void TermPrint(FILE* out, Term_p term, Sig_p sig)
{
   assert(term);
   assert(sig);
   
   term = TermDeref(term);
   if(SigSupportLists && TermPrintLists &&
      ((term->f_code == SIG_NIL_CODE)||
       (term->f_code == SIG_CONS_CODE)))
   {
      print_cons_list(out, term, sig);
   }
   else
   {
      if(TermIsVar(term))
      {
	 fprintf(out, "%c%ld", TermPrologVars ? 'X' : 'x',
		 -term->f_code);
      }
      else
      {
	 fprintf(out, SigFindName(sig, term->f_code));
	 if(TermIsConst(term))
	 {
	    if(!TermPrologVars)
	    {
	       fprintf(out, "()");
	    }
	 }
	 else
	 {
	    assert(term->args);
	    TermPrintArgList(out, term->args, term->arity, sig);
	 }
      }
   }
}

/*-----------------------------------------------------------------------
//
// Function: TermPrintArgList()
//
//   Print an argument list (i.e. an array with at least one term
//   element) to the given stream.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void TermPrintArgList(FILE* out, Term_p *args, int arity, Sig_p sig)
{
   int i;

   assert(arity>=1);
   putc('(', out);
 
   TermPrint(out, args[0], sig);
   
   for(i=1; i<arity; i++)
   {
      putc(',', out);
      /* putc(' ', out); */
      TermPrint(out, args[i], sig);
   }
   putc(')', out);
}



/*-----------------------------------------------------------------------
//
// Function: TermParseOperator()
//
//   Parse an operator (i.e. an optional $, followed by an
//   identifier), store the representation into id and determine
//   wether it represents a variable or a function symol by the
//   following rules:
//   - If it starts with a $, it's a function symbol (LOP global
//     variables are treated as function symbols by me).
//   - If TermPrologVars is true and it starts with a lower case
//     character, it's a function symbol.
//   - If it is followed by an opening bracket, its a function
//     symbol.
//   - Otherwise its a variable.
//   Return value is true if the symbol represents a variable.
//
// Global Variables: TermPrologVars, TermIdentToken
//
// Side Effects    : Read input
//
/----------------------------------------------------------------------*/

bool TermParseOperator(Scanner_p in, DStr_p id)
{
   bool res = true;

   if(TestInpTok(in, Dollar))
   {
      DStrAppendChar(id, '$');
      res = false;
      NextToken(in);
      if(!TestInpNoSkip(in))
      {
	 AktTokenError(in, "Build-In symbol name incomplete", false);
      
      }
   }
   if(TermPrologVars && islower(DStrView(AktToken(in)->literal)[0]))
   {
      res = false;
   }
   DStrAppendStr(id, DStrView(AktToken(in)->literal));
   AcceptInpTok(in, TermIdentToken);
   if(TestInpTok(in, OpenBracket))
   {
      res = false;
   }
   return res;
}      



/*-----------------------------------------------------------------------
//
// Function: TermParse()
//
//   Parse a term from the given scanner object into the internal term
//   representation.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations, may extend signature
//
/----------------------------------------------------------------------*/

Term_p TermParse(Scanner_p in, Sig_p sig, VarTrans_p vars)
{
   Term_p handle;
   DStr_p id;
   DStr_p source_name, errpos;
   long   line, column;

   if(SigSupportLists && TestInpTok(in, OpenSquare))
   {
      handle =  parse_cons_list(in, sig, vars);
   }
   else
   {
      id = DStrAlloc();
      line = AktToken(in)->line;
      column = AktToken(in)->column;
      source_name = DStrGetRef(AktToken(in)->source_name);

      handle = DefaultTermCellAlloc();
      TermSetProp(handle, TPTopPos);
      
      if(TermParseOperator(in, id))
      {
	 handle->f_code = VarTransInsertId(vars, DStrView(id));
      }
      else
      {
	 if(TestInpTok(in, OpenBracket))
	 {
	    handle->arity = TermParseArgList(in, &(handle->args), sig,
					     vars);
	 }
	 else
	 {
	    handle->arity = 0;
	 }
	 handle->f_code = SigInsertId(sig, DStrView(id), handle->arity);
	 if(!handle->f_code)
	 {
	    errpos = DStrAlloc();

	    DStrAppendStr(errpos, PosRep(source_name, line, column));
	    DStrAppendChar(errpos, ' ');
	    DStrAppendStr(errpos, DStrView(id));
	    DStrAppendStr(errpos, " used with arity ");
	    DStrAppendInt(errpos, (long)handle->arity);
	    DStrAppendStr(errpos, " but registered with arity ");
	    DStrAppendInt(errpos,
			  (long)SigFindArity(sig, SigFindFCode(sig, DStrView(id))));
	    Error(DStrView(errpos), SYNTAX_ERROR);
	    DStrFree(errpos);
	 }
      }
      DStrReleaseRef(source_name);
      DStrFree(id);
   }
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermParseArgList()
//
//   Parse a list of terms (comma-separated and enclosed in brackets)
//   into an array of term pointers. Return the number of terms
//   parsed, and make arg_anchor point to the array. Note: The array
//   has to have exactly the right size, as it will be handled by
//   Size[Malloc|Free] for efficiency reasons and may otherwise lead
//   to a memory leak. This leads to some complexity...
//   If the arglist is empty, return 0 and use the NULL pointer as
//   anchor. 
//
// Global Variables: -
//
// Side Effects    : Input, memory operations
//
/----------------------------------------------------------------------*/

int TermParseArgList(Scanner_p in, Term_p** arg_anchor, Sig_p sig,
		     VarTrans_p vars) 
{
   Term_p *handle;
   int    arity;
   int    size;
   int    i;

   AcceptInpTok(in, OpenBracket);
   if(TestInpTok(in, CloseBracket))
   {
      NextToken(in);
      *arg_anchor = NULL;
      return 0;
   }
   size = TERMS_INITIAL_ARGS;
   handle = (Term_p*)SizeMalloc(size*sizeof(Term_p));
   arity = 0;
   handle[arity] = TermParse(in, sig, vars);
   TermDelProp(handle[arity], TPTopPos);
   arity++;
   while(TestInpTok(in, Comma))
   {
      NextToken(in);
      if(arity==size)
      {
	 size+=TERMS_INITIAL_ARGS;
	 handle = (Term_p*)SecureRealloc(handle, size*sizeof(Term_p));
      }
      handle[arity] = TermParse(in, sig, vars);
      TermDelProp(handle[arity], TPTopPos);
      arity++;
   }
   AcceptInpTok(in, CloseBracket);
   *arg_anchor = TermArgArrayAlloc(arity);
   for(i=0;i<arity;i++)
   {
      (*arg_anchor)[i] = handle[i];
   }
   SizeFree(handle, size*sizeof(Term_p));
   
   return arity;
}

/*-----------------------------------------------------------------------
//
// Function: TermDeref()
//
//   If term->binding is NULL, return term, otherwise return
//   TermDeref(term->binding). I.e. follow instantiated variables and
//   replaced terms and return the final term.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p TermDeref(Term_p term)
{
   while(term->binding)
   {
      term = term->binding;
   }
   return term;
}


/*-----------------------------------------------------------------------
//
// Function: TermArgListCopy()
//
//   Return a copy of the argument array of source.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p* TermArgListCopy(Term_p source)
{
   Term_p *handle;
   int i;
   
   if(source->arity)
   {
      handle = TermArgArrayAlloc(source->arity);
      for(i=0; i<source->arity; i++)
      {
	 handle[i] = source->args[i];
      }
   }
   else
   {
      handle = NULL;
   }
   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: TermTopCopy()
//
//   Return a copy of the term node (and potential argument
//   pointers). Only the top node and the pointers are duplicated, the
//   arguments are shared between source and copy. As this function
//   operates on nodes, it does not follow bindings! Administrative
//   stuff (refs etc. will, of course, not be copied but initialized
//   to rational values for an unshared 
//   term).
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p TermTopCopy(Term_p source)
{
   Term_p handle;
   
   handle = DefaultTermCellAlloc();
   handle->flags  = source->flags;
   handle->f_code = source->f_code;
   handle->arity  = source->arity;
   handle->binding = NULL;
   handle->args = TermArgListCopy(source);
   handle->entry_no = 0;
   handle->ref_count = 0;
   handle->superterms = NULL;
   handle->ext_refs = NULL;
   handle->balance = 0;
   handle->passed = 0;
   handle->lson = NULL;
   handle->rson = NULL;
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermCopy()
//
//   Return a copy of a given term. The new term will be unshared
//   even if the original term was shared.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p TermCopy(Term_p source, bool deref)
{
   Term_p handle;
   int i;

   assert(source);

   if(deref)
   {
      source = TermDeref(source);
   }

   handle = TermTopCopy(source);
   
   for(i=0; i<handle->arity; i++) /* Hack: Loop will not be entered if
				     arity = 0 */
   {
      handle->args[i] = TermCopy(handle->args[i], deref);
   }
   
   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermStructEqual()
//
//   Return true if the two terms have the two structures. Follows
//   bindings. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TermStructEqual(Term_p t1, Term_p t2)
{
   PStack_p stack1 = PStackAlloc();
   PStack_p stack2 = PStackAlloc();
   int i;
   bool termeq = true;

   PStackPushP(stack1, t1);
   PStackPushP(stack2, t2);
   
   while(!PStackEmpty(stack1) && !PStackEmpty(stack2))
   {
      t1 = PStackPopP(stack1);
      t2 = PStackPopP(stack2);
      if(!t1 || !t2)
      {
	 termeq = false;
	 break;
      }
      else
      {
	 t1 = TermDeref(t1);
	 t2 = TermDeref(t2);
	 if(t1->f_code != t2->f_code)
	 {
	    termeq = false;
	    break;
	 }
	 else
	 {
	    assert(t1->arity == t2->arity);
	    for(i=0; i<t1->arity; i++)
	    {
	       PStackPushP(stack1, t1->args[i]);
	       PStackPushP(stack2, t2->args[i]);
	    }
	 }
      }
   }
   PStackFree(stack1);
   PStackFree(stack2);
   return termeq;
}


/*-----------------------------------------------------------------------
//
// Function: TermIsSubterm()
//
//   Return true if test is a subterm to super. Parameterized by the
//   Equal-Funktion for terms.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TermIsSubterm(Term_p super, Term_p test, bool deref,
		   TermEqualTestFun EqualTest)
{
   PStack_p stack = PStackAlloc();
   int i;
   bool subterm = false;

   printf("TermIsSubterm()...\n");

   PStackPushP(stack, super);
   
   while(!PStackEmpty(stack))
   {
      super = PStackPopP(stack);

      if(deref)
      {
	 super = TermDeref(super);
      }
      if(EqualTest(super, test))
      {
	 subterm = true;
	 break;
      }
      else
      {
	 for(i=0; i<super->arity; i++)
	 {
	    PStackPushP(stack, super->args[i]);
	 }
      }
   }
   PStackFree(stack);

   printf("...TermIsSubterm()\n");

   return subterm;
}

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/




