/*-----------------------------------------------------------------------

File  : cte_terms.h

Author: Stephan Schulz

Contents
 
  Basic definitions for terms. Includes some support for future
  extensions. 

Changes

<1> Sun Nov 16 20:09:26 MET 1997
    New

-----------------------------------------------------------------------*/

#ifndef CTE_TERMS

#define CTE_TERMS

#include <cte_signature.h>
#include <cte_vartrans.h>
#include <cio_scanner.h>
#include <clb_numtrees.h>
#include <clb_ptrees.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/

typedef enum 
{
   TPIgnoreProps = 0, /* For masking properties out */
   TPMaximal     = 1, /* Maximal term == Left hand side */
   TPTopPos      = 2, /* This cell is a entry point */
   TPBasicPos    = 4, /* Basic position on this term */
   TPPredPos     = 8, /* This is an previous predicate position
			 morphed into a term */
   TPOpFlag      = 256,/* For internal use */
   TPOutputFlag  = 512 /* Has this term already been printed? */
}TermProperties;


typedef struct termcell
{
   TermProperties   flags;      /* Like basic, lhs, top */
   FunCode          f_code;     /* Top symbol of term */
   int              arity;      /* Redundant, but saves handing around
				   the signature all the time */
   struct termcell* *args;      /* Pointer to array of arguments */
   struct termcell* binding;    /* For variable bindings, potentially
				   for temporary rewrites */
   unsigned long    entry_no;   /* Counter for terms in a given
				   termbank - needed for
				   administration and external
				   representation */
   long             ref_count;  /* How often is this term referenced? */
   PTree_p          superterms; /* Superterms, taking care of
				   "internal" references,
				   i.e. references resulting from
				   subterm relations in the term bank */
   PTree_p          ext_refs;   /* External references, i.e. positions
				   of term pointers that have to be
				   updated if the term is replaced */		   
   int              balance;    /* AVL stuff */
   int              passed;
   struct termcell* lson;
   struct termcell* rson;
}TermCell, *Term_p, **TermRef;

/* Functions which take two terms and return a boolean, i.e. test for
   equality */

typedef bool (*TermEqualTestFun)(Term_p t1, Term_p t2);

#define TERMS_INITIAL_ARGS 10

/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

#define TermIsVar(t) ((t)->f_code < 0)
#define TermIsConst(t)(!TermIsVar(t) && ((t)->arity==0))

#define TermSetProp(term, prop) ((term)->flags = (term)->flags | (prop))
#define TermDelProp(term, prop) ((term)->flags = (term)->flags & !(prop))
#define TermQueryProp(term, prop) (((term)->flags & (prop)) == (prop))

extern bool      TermPrintLists; /* Using [...] notation */
extern bool      TermPrologVars; /* Capital words are variables */ 
extern TokenType TermIdentToken; /* What can be used as an operator in
				    terms */

#define TermCellAlloc() (TermCell*)SizeMalloc(sizeof(TermCell))
#define TermCellFree(junk)         SizeFree(junk, sizeof(TermCell))
#define TermArgArrayAlloc(arity) ((Term_p*)SizeMalloc((arity)*sizeof(Term_p)))
#define TermArgArrayFree(junk, arity) SizeFree((junk),(arity)*sizeof(Term_p))

Term_p DefaultTermCellAlloc();

void   TermTopFree(Term_p junk); 
void   TermFree(Term_p junk); /* For unshared term only */

void   TermPrint(FILE* out, Term_p term, Sig_p sig);
void   TermPrintArgList(FILE* out, Term_p *args, int arity, Sig_p sig);
bool   TermParseOperator(Scanner_p in, DStr_p id);
Term_p TermParse(Scanner_p in, Sig_p sig, VarTrans_p vars);
int    TermParseArgList(Scanner_p in, Term_p** arg_anchor, Sig_p sig,
			 VarTrans_p vars);
Term_p  TermDeref(Term_p term);
Term_p* TermArgListCopy(Term_p source);
Term_p  TermTopCopy(Term_p source);
Term_p  TermCopy(Term_p source, bool deref);
bool    TermStructEqual(Term_p t1, Term_p t2);
bool    TermIsSubterm(Term_p super, Term_p test, bool deref,
		      TermEqualTestFun EqualTest);

#define TermIsStructSubterm(super, term) \
        TermIsSubterm((super),(term),true,TermStructEqual)

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





