/*-----------------------------------------------------------------------

File  : cte_match_mgu_1-1.c

Author: Stephan Schulz

Contents
 
  Implementation of simple, non-indexed 1-1 match and unification
  routines on shared terms (and unshared terms with shared
  variables).

Changes

<1> Wed Mar 11 16:17:33 MET 1998
    New

-----------------------------------------------------------------------*/

#include "cte_match_mgu_1-1.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: check_unify_job()
//
//   Return true is the pair (t1,t2) is already on the stacks.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static bool check_unify_job(PStack_p s1, PStack_p s2, Term_p t1,
			    Term_p t2)
{
   PStackPointer i, limit;

   assert(t1);
   assert(t2);
   assert(s1);
   assert(s2);

   limit = PStackGetSP(s1);
   assert(limit == PStackGetSP(s2));

   for(i=limit-1; i>=0; i--)
   {
      if((t1==PStackElementP(s1,i))&&(t2==PStackElementP(s2,i)))
      {
	 return true;
      }
   }
   return false;
}


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: SubstComputeMatch()
//
//   Try to compute a match from matcher onto to_match and record it in
//   subst. Return true if match exits (in this case subst is
//   changed and needs to be backtracked by the caller), false
//   otherwise (subst is unchanged). Both terms are assumed to contain
//   no bindings except those stored in subst.
// 
//   For shared terms, calling the function with TBTermEqual() as the
//   last argument (the function used to test terms for equality)
//   results in more efficient execution (otherwise call the function
//   with TermBasicStructEqual()). 
//
//   The routine will work and compute a valid match if the two terms
//   share variables. However, this will lead to temporary incorrect
//   term structures (a variable may be bound to itself or a superterm
//   containing it).
//
// Global Variables: -
//
// Side Effects    : Instantiates terms
//
/----------------------------------------------------------------------*/

bool SubstComputeMatch(Term_p matcher, Term_p to_match, Subst_p subst,
		       TermEqualTestFun EqualTest)
{
   PStack_p m_jobs = PStackAlloc();
   PStack_p t_jobs = PStackAlloc();
   PStackPointer backtrack = PStackGetSP(subst); /* For backtracking */
   int i;

   PStackPushP(m_jobs, matcher);
   PStackPushP(t_jobs, to_match);

   while(!PStackEmpty(m_jobs))
   {
      matcher  =  PStackPopP(m_jobs);
      to_match =  PStackPopP(t_jobs);
            
      if(TermIsVar(matcher))
      {
	 if(matcher->binding)
	 {
	    if(!EqualTest(matcher->binding,to_match))
	    {
	       SubstBacktrackToPos(subst,backtrack);
	       PStackFree(m_jobs);
	       PStackFree(t_jobs);
	       return false;
	    }
	 }
	 else
	 {
	    SubstAddBinding(subst, matcher, to_match);
	 }
      }
      else
      {
	 if(matcher->f_code != to_match->f_code)
	 {
	    SubstBacktrackToPos(subst,backtrack);
	    PStackFree(m_jobs);
	    PStackFree(t_jobs);
	    return false;
	 }
	 else
	 {
	    for(i=matcher->arity-1; i>=0; i--)
	    {
	       PStackPushP(m_jobs, matcher->args[i]);
	       PStackPushP(t_jobs, to_match->args[i]);
	    }
	 }
      }
   }

   PStackFree(m_jobs);
   PStackFree(t_jobs);
   return true;
}


/*-----------------------------------------------------------------------
//
// Function: SubstComputeMgu()
//
//   Compute an mgu between two terms. Currently without any special
//   optimization, except for checking for double entries in the
//   to-solve stacks. Returns true and modifies subst if sucessful,
//   false otherwise (as for match, see above). Terms have to be
//   variable disjoint, otherwise behaviour is unpredictable!
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

bool SubstComputeMgu(Term_p t1, Term_p t2, Subst_p subst)
{
   PStack_p jobs1 = PStackAlloc();
   PStack_p jobs2 = PStackAlloc();
   PStackPointer backtrack = PStackGetSP(subst); /* For backtracking */
   int i;
   DerefType deref = DEREF_ALWAYS;

   PStackPushP(jobs1, t1);
   PStackPushP(jobs2, t2);

   while(!PStackEmpty(jobs1))
   {
      t1 =  TermDeref(PStackPopP(jobs1), &deref);
      t2 =  TermDeref(PStackPopP(jobs2), &deref);
            
      if(TermIsVar(t1))
      {
	 if(!TBTermEqual(t1,t2))
	 {
	    /* Occur-Check - remember, variables are elementary and
	       shared! */
	    if(TermIsSubterm(t2, t1, DEREF_ALWAYS, TBTermEqual))
	    {
	       SubstBacktrackToPos(subst,backtrack);
	       PStackFree(jobs1);
	       PStackFree(jobs2);
	       return false;
	    }
	    else
	    {
	       SubstAddBinding(subst, t1, t2);
	    }
	 }
      }
      else if((TermIsVar(t2)))
      {
	 if(!TBTermEqual(t2,t1))
	 {
	    /* Occur-Check - remember, variables are elementary and
	       shared! */
	    if(TermIsSubterm(t1, t2, DEREF_ALWAYS, TBTermEqual))
	    {
	       SubstBacktrackToPos(subst,backtrack);
	       PStackFree(jobs1);
	       PStackFree(jobs2);
	       return false;
	    }
	    else
	    {
	       SubstAddBinding(subst, t2, t1);
	    }
	 }
      }      
      else
      {
	 if(t1->f_code != t2->f_code)
	 {
	    SubstBacktrackToPos(subst,backtrack);
	    PStackFree(jobs1);
	    PStackFree(jobs2);
	    return false;
	 }
	 else
	 {
	    for(i=t1->arity-1; i>=0; i--)
	    {
	       if(!check_unify_job(jobs1, jobs2, t1->args[i],
				   t2->args[i]))
	       {
		  PStackPushP(jobs1, t1->args[i]);
		  PStackPushP(jobs2, t2->args[i]);
	       }
	    }
	 }
      }
   }
   PStackFree(jobs1);
   PStackFree(jobs2);
   return true;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


