/*-----------------------------------------------------------------------

File  : cte_replace.c

Author: Stephan Schulz

Contents
 
  Functions for replacing and rewriting of terms.

Changes

<1> Mon Jan 26 19:16:16 MET 1998
    new

-----------------------------------------------------------------------*/

#include "cte_replace.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: new_to_replace()
//
//   Add a new ReplaceCell recording the necessity of replacing term
//   with repl to tree.
//
// Global Variables: -
//
// Side Effects    : Changes tree, gets references.
//
/----------------------------------------------------------------------*/

static void new_to_replace(PQueue_p list, Term_p term, Term_p repl)
{
   Replace_p handle;

   handle = ReplaceCellAlloc();
   handle->tb_date = term->tb_date;
   TermGetRef(&(handle->term), term);
   TermGetRef(&(handle->repl), repl);
   PQueueStoreP(list, handle);
}


/*-----------------------------------------------------------------------
//
// Function: get_to_replace()
//
//   Remove the root place cell from tree, release references, return
//   term and repl (via argument pointers)
//
// Global Variables: -
//
// Side Effects    : Changes tree, gets references.
//
/----------------------------------------------------------------------*/

static bool get_to_replace(PQueue_p list, Term_p *term, Term_p *repl)
{
   bool      res;
   Replace_p handle;

   handle = PQueueGetNextP(list);
   TermReleaseRef(&(handle->term));
   TermReleaseRef(&(handle->repl));
   *term = handle->term;
   *repl = handle->repl;
   res = SysDateCompare(handle->tb_date, handle->term->tb_date) ==
      DateEqual;

   ReplaceCellFree(handle);

   return res;
}




/*-----------------------------------------------------------------------
//
// Function: term_replace_ext_refs()
//
//   Make all external references of term point to replace instead.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void term_replace_ext_refs(Term_p term, Term_p replace)
{
   TermRef reference;

   while(term->ext_refs)
   {
      reference = term->ext_refs->key;
      TermReleaseRef(reference);
      TermGetRef(reference, replace);
   }
}



/*-----------------------------------------------------------------------
//
// Function: replace_single_cell()
//
//   Replace term with repl througout the term bank, collecting
//   conflicts in to_replace. 
//
// Global Variables: -
//
// Side Effects    : Changes the term bank and to_replace
//
/----------------------------------------------------------------------*/

static void replace_single_cell(TB_p bank, Term_p term, Term_p repl,
				PQueue_p to_replace)
{
   Term_p keyterm, super, newsuper;
   Term_p *tmp_args;
   int i;
   PTree_p tmp_tree = NULL;

   DEBUGMARK(RPL_TERM_WATCH, "replace_single_cell(");
   DEBUG(RPL_TERM_WATCH, TBPrintTermFull(stdout, bank, term);
	 printf(" <--- "); TBPrintTermFull(stdout, bank, repl);
	 printf(")...\n"););

   /*  assert(!TBVerifyBankReferences(stdout, bank)); */
   if(term == repl)
   {
      DEBUGMARK(RPL_TERM_WATCH, "...replace_single_cell() (equal-case)\n");
      return;
   }

   assert(!TBTermIsSubterm(repl,term));
   
   term_replace_ext_refs(term, repl);   
   
   while(term->superterms)
   {
      super = term->superterms->key;
      TermRemoveSuperterm(term, super);
      
      keyterm = TermTopCopy(super);
      for(i=0; i<keyterm->arity; i++)
      {
	 if(keyterm->args[i] == term)
	 {
	    keyterm->args[i] = repl;
	 }
      }
      if((newsuper = TermTreeFind(bank->term_root, keyterm,
				  bank->prop_mask)))
      {
	 PTreeStore(&tmp_tree, super);
	 new_to_replace(to_replace, super, newsuper);
      }
      else
      {	 
	 if(PTreeFind(repl->superterms, super))
	 {
	    TermRemoveSuperterm(repl, super);
	 }
	 super = TermTreeExtract(&(bank->term_root), super,
				 bank->prop_mask);
	 tmp_args = super->args;
	 super->args = keyterm->args;
	 keyterm->args = tmp_args; /* To have a complete term top
				      for freeing */
	 TermTreeInsert(&(bank->term_root), super, bank->prop_mask);
	 TermIntroduceSuperterm(repl, super);
      }
      TermTopFree(keyterm);
   }
   while(tmp_tree)
   {
      super = PTreeExtractRootKey(&tmp_tree);
      TermIntroduceSuperterm(term, super);
   }      
   DEBUGMARK(RPL_TERM_WATCH, "...replace_single_cell() (normal case)\n");
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TBPropagateReplaceDataToSupers()
//
//   Propagate information about replacing to all superterms: Set
//   nf_date to SysCreationTime() and update tb_date to reflect the
//   changes. 
//
// Global Variables: -
//
// Side Effects    : Modifies date, memory management
//
/----------------------------------------------------------------------*/

void TBPropagateReplaceDataToSupers(TB_p bank, Term_p term, SysDate
				    ref_date)
{
   PStack_p ptree_stack, term_stack;
   PTree_p node;

   assert(term);

   term_stack = PStackAlloc();
   PStackPushP(term_stack, term);

   while(!PStackEmpty(term_stack))
   {
      term = PStackPopP(term_stack);
      assert(term);
      DEBUG(16, 
	    printf("# ");SysDatePrint(stdout, ref_date); 
	    printf(" <> "); SysDatePrint(stdout, term->tb_date);
	    printf("   ");
	    TermPrint(stdout,term,bank->sig,DEREF_NEVER);
	    printf("\n"););
      if(SysDateCompare(term->tb_date, ref_date)!=DateLater)
      {
	 DEBUGMARK(16, "updated\n");
	 term->nf_date = SysDateCreationTime();
	 SysDateInc(&(bank->tb_date));
	 term->tb_date = bank->tb_date;
	 ptree_stack = PTreeTraverseInit(term->superterms);
	 while((node = PTreeTraverseNext(ptree_stack)))
	 {
	    PStackPushP(term_stack, node->key);
	 }	 
	 PTreeTraverseExit(ptree_stack);
      }
   }
   PStackFree(term_stack);   
}



/*-----------------------------------------------------------------------
//
// Function: TBTermReplace()
//
//   Replace term with replace, making sure that all references are
//   changed and that affected superterms are treated properly
//   (i.e. when replacing a() with b(), f(a()) and f(b()) will be
//   recognized as identical and treated accordingly.
//
// Global Variables: -
//
// Side Effects    : May reorganize large parts of the term bank!
//
/----------------------------------------------------------------------*/

void TBTermReplace(TB_p bank, Term_p term, Term_p repl)
{
   PQueue_p to_do_list = PQueueAlloc();
   bool     res;

   /* assert(!TBVerifyBankReferences(stdout, bank)); */

   DEBUGMARK(RPL_TERM_WATCH, "TBTermReplace(");
   DEBUG(RPL_TERM_WATCH, TBPrintTermFull(stdout, bank, term);
	 printf(" <--- "); TBPrintTermFull(stdout, bank, repl);
	 printf(")...\n"););
   
   assert(!TBTermIsSubterm(repl,term));

   TBPropagateReplaceDataToSupers(bank, term, bank->tb_date);

   new_to_replace(to_do_list, term, repl);
   
   while(!PQueueEmpty(to_do_list))
   {
      /* assert(!TBVerifyBankReferences(stdout, bank)); */
      res = get_to_replace(to_do_list, &term, &repl);
      if(!res)
      {
	 continue;
      }
      replace_single_cell(bank, term, repl, to_do_list);
      TBDelete(bank, term);
   }
   PQueueFree(to_do_list);

   DEBUGMARK(RPL_TERM_WATCH, "...TBTermReplace()\n");
}


/*-----------------------------------------------------------------------
//
// Function: TBNonCyclicTermReplace()
//
//   Replace term with replace _nearly_ everywhere in the term
//   bank. More exactly: If term is not a subterm of replace, this
//   function performs exactly as TBTermReplace(). 
//
//   If term is a subterm of replace, consequent replacing would lead
//   to an infinite term. This is correct, but now what most users
//   would expect. TBNonCyclicTermReplace() behaves more intuitively
//   and handles this case as most people expect, i.e. it replaces
//   term with replace only once and not recursively. This function
//   is, however, less efficient.
//
// Global Variables: -
//
// Side Effects    : As (and by) TBTermReplace()
//
/----------------------------------------------------------------------*/

void TBNonCyclicTermReplace(TB_p bank, Term_p term, Term_p replace)
{
   DEBUGMARK(RPL_TERM_WATCH, "TBNonCyclicTermReplace()...\n");

   if(!TBTermIsSubterm(replace,term))
   {
      TBTermReplace(bank,  term, replace);
   }
   else
   {
      Term_p tmp_term, bottom_term, bottom_shared;
   
      tmp_term = TermCopy(replace,bank->vars,DEREF_NEVER);
      bottom_term = DefaultTermCellAlloc();
      bottom_term->f_code = SIG_BOTTOM_CODE;
      bottom_shared = TBInsert(bank, bottom_term, DEREF_NEVER);
      TermFree(bottom_term);
      TBTermReplace(bank, term, bottom_shared);
      replace = TBInsert(bank, tmp_term, DEREF_NEVER);
      TermFree(tmp_term);
      TBTermReplace(bank, bottom_shared, replace);
   }
   DEBUGMARK(RPL_TERM_WATCH, "...TBNonCyclicTermReplace()\n");
}


/*-----------------------------------------------------------------------
//
// Function: TBTermPosReplace()
//
//   Create a new term by replacing the subterm designated by pos with
//   repl and insert it into the term bank. Return pointer to the new
//   term. The superterm is implicit in the position (or, in the case
//   of the empty position, is unnecessary). Does not free any terms -
//   if necessary, this is the responsibility of the calling
//   functions.
//
//   Note that this function may dereference terms more than once,
//   which will lead to strange behaviour if instantiated variables
//   appear in the term...---> This has been fixed. Each term in repl
//   and its superterms should now be dereferenced only once.
//
// Global Variables: -
//
// Side Effects    : Memory management, changes term bank
//
/----------------------------------------------------------------------*/

Term_p TBTermPosReplace(TB_p bank, Term_p repl, TermPos_p pos,
			DerefType deref)
{
   Term_p        handle, old;
   int           subscript;
   PStackPointer i;
   PStack_p      store = PStackAlloc();

   DEBUGMARK(RPL_TERM_WATCH, "TBTermPosReplace()...\n");

   assert(bank);
   assert(repl);
   assert(pos);
   
   i = PStackGetSP(pos);
   
   while(i)
   {
      i--;
      subscript = PStackElementInt(pos, i);
      assert(i);
      i--;

      old = PStackElementP(pos, i);
      handle = TermTopCopy(old);
      assert(handle->arity > subscript);
      handle->args[subscript] = repl;
      PStackPushP(store, handle);
      repl = handle;
   }
   repl = TBInsert(bank, repl, deref);
   
   while(!PStackEmpty(store))
   {
      handle = PStackPopP(store);
      TermTopFree(handle);
   }
   PStackFree(store);

   DEBUGMARK(RPL_TERM_WATCH, "...TBTermPosReplace()\n");
   return repl;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


