/*-----------------------------------------------------------------------

File  : cte_signature.h

Author: Stephan Schulz

Contents
 
  Definitions for dealing with signatures, i.e. data structures
  storing information about function symbols and their properties. 

Changes

<1> Thu Sep 18 16:54:31 MET DST 1997
    New

-----------------------------------------------------------------------*/

#ifndef CTE_SIGNATURE

#define CTE_SIGNATURE

#include <clb_stringtrees.h>
#include <cio_scanner.h>
#include <cte_types.h>

/*---------------------------------------------------------------------*/
/*                    Data type declarations                           */
/*---------------------------------------------------------------------*/



/* Keep information about function symbols: Access external name and
   arity (and possibly additional information at a later time) by
   internal numerical code for function symbol. */

typedef struct funccell
{
   /* f_code is implicit by position in the array */
   char* name;
   int   arity;
}FuncCell, *Func_p;


/* A signature contains information about function symbols with
   direct access by internal code (f_info is organized as a array,
   with f_info[f_code] being the information associated with f_code)
   and efficient access by external name (via the f_index array). 

   Function codes are integers starting at 1, while variables are
   encoded by negative integers. 0 is unused and can thus express
   error conditions when accessing something f_code. f_info[0] is
   unused. */

#define DEFAULT_SIGNATURE_SIZE 20

typedef struct sigcell
{
   long      size;     /* Size of the array */
   FunCode   f_count;  /* Largest used f_code */
   Func_p    f_info;   /* The array */
   StrTree_p f_index;  /* Back-assoc: Given a symbol, get the index */
}SigCell, *Sig_p;




/*---------------------------------------------------------------------*/
/*                Exported Functions and Variables                     */
/*---------------------------------------------------------------------*/

/* Special constant for internal operations */

#define SIG_BOTTOM_CODE 1
#define SIG_TRUE_CODE 2

extern bool      SigSupportLists; /* Auto-Insert special symbols
				     $nil=3, $cons=4 for list
				     representations */
extern TokenType SigIdentToken; /* What is a potential function symbol
				   name? */

#define SIG_NIL_CODE  3
#define SIG_CONS_CODE 4

#define SigCellAlloc() (SigCell*)SizeMalloc(sizeof(SigCell))
#define SigCellFree(junk)         SizeFree(junk, sizeof(SigCell))

Sig_p   SigAlloc();
void    SigFree(Sig_p junk);
FunCode SigFindFCode(Sig_p sig, char* name);
int     SigFindArity(Sig_p sig, FunCode f_code);
char*   SigFindName(Sig_p sig, FunCode f_code);
FunCode SigInsertId(Sig_p sig, char* name, int arity);
void    SigPrint(FILE* out, Sig_p sig);
void    SigParseOperator(Scanner_p in, DStr_p id);
FunCode SigParseKnownOperator(Scanner_p in, Sig_p sig);
FunCode SigParseSymbol(Scanner_p in, Sig_p sig);
FunCode SigParse(Scanner_p in, Sig_p sig);
int     SigFindMaxUsedArity(Sig_p sig);

#endif

/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/





