/*-----------------------------------------------------------------------

File  : cte_termbanks.c

Author: Stephan Schulz

Contents
 
  Functions for term banks (efficient dag-representations for term
  sets). 

  Bindings (i.e. values in term->binding) make only sense in dynamical
  contexts, not in (mostly) static term banks. Thus, bindings are
  followed whenever a term is treated as an individual term, but not
  when they form part of a term bank and are manipulated as such. 

Changes

<1> Sat Nov 15 16:26:30 MET 1997
    New

-----------------------------------------------------------------------*/

#include "cte_termbanks.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

bool TBPrintTermsFlat = false;    /* Always introduce abbreviations
				     for all subterms (if not set,
				     only terms referenced at least
				     twice are printed with
				     abbreviation) */
bool TBPrintInternalInfo = false; /* Print internal information about
				     term nodes, e.g. the flag field
				     and the reference counter as a
				     comment */
bool TBSupportReplace = true;     /* Maintain reference lists for
				     superterms, for replacing of
				     terms */


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*-----------------------------------------------------------------------
//
// Function: tb_print_dag()
//
//   Print the terms as a dag in the order of insertion.
//
// Global Variables: TBPrintInternalInfo
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

static void tb_print_dag(FILE *out, NumTree_p in_index, Sig_p sig)
{
   Term_p term;

   if(!in_index)
   {
      return;
   }
   tb_print_dag(out, in_index->lson, sig);
   term = in_index->val1.p_val;
   fprintf(out, "*%ld : ", term->entry_no);

   if(TermIsVar(term))
   {
      fprintf(out, "%c%ld", TermPrologVars ? 'X' : 'x',
	      -term->f_code);
   }
   else
   {
      fprintf(out, SigFindName(sig, term->f_code));
      if(TermIsConst(term))
      {
	 if(!TermPrologVars)
	 {
	    fprintf(out, "()");
	 }
      }
      else
      {
	 int i;
	 
	 assert(term->arity>=1);
	 assert(term->args);
	 putc('(', out);

	 fprintf(out, "*%ld", term->args[0]->entry_no);
	 for(i=1; i<term->arity; i++)
	 {
	    putc(',', out);
	    fprintf(out, "*%ld", term->args[i]->entry_no);
	 }
	 putc(')', out);
      }
   }
   if(TBPrintInternalInfo)
   {
      fprintf(out, "\t/*  Flags: %10d, %ld references */",
	      term->flags, term->ref_count);
      if(TBSupportReplace)
      {
	 fprintf(out, "/* SExt: %ld SSup: %ld*/",
		 PTreeNodes(term->ext_refs),
		 PTreeNodes(term->superterms));
      }
   }
   fprintf(out, "\n");
   tb_print_dag(out, in_index->rson, sig);
}


/*-----------------------------------------------------------------------
//
// Function: tb_termtop_insert()
//
//   Insert a term into the term bank for which the subterms are
//   already in the term bank. Will reuse or destroy the top cell!
//
// Global Variables: TBSupportReplace
//
// Side Effects    : Changes term bank
//
/----------------------------------------------------------------------*/

Term_p tb_termtop_insert(TB_p bank, Term_p t)
{
   int    i;
   Term_p new;

   assert(t);
   
   SysDateInc(&(bank->tb_date));
   t->tb_date = bank->tb_date;
   new = TermTreeInsert(&(bank->term_root), t, bank->prop_mask);
   
   if(new) /* Term node already existed, just add flags */
   {
      if(!TermIsVar(t))
      {
	 new->flags = (new->flags | t->flags)& bank->prop_mask;
	 TermTopFree(t);
      }
      return new;
   }
   else
   {
      IntOrP term_adr;

      t->entry_no     = ++(bank->in_count);
      t->ref_count    = 0; /* Now it becomes a shared cell! */
      t->nf_date      = SysDateCreationTime();
      term_adr.p_val  = t;
      NumTreeStore(&(bank->entry_index),
		   t->entry_no, 
		   term_adr, term_adr); /* Need for args and cannot
					   pass NULL as last arg, due
					   to ANSI nonsense */
      for(i=0; i<t->arity; i++)
      {
	 (t->args[i])->ref_count++;
	 if(TBSupportReplace)
	 {
	    PTreeStore(&((t->args[i])->superterms),  t);
	 }
      }
   }
   return t;
}


/*-----------------------------------------------------------------------
//
// Function: tb_parse_cons_list()
//
//   Parse a LOP list into an (shared) internal $cons list.
//
// Global Variables: -
//
// Side Effects    : Input, Memory operations
//
/----------------------------------------------------------------------*/

static Term_p tb_parse_cons_list(Scanner_p in, TB_p bank)
{
   Term_p   handle;
   Term_p   current;
   PStack_p stack;
   
   assert(SigSupportLists);
   
   stack = PStackAlloc();
   AcceptInpTok(in, OpenSquare);

   handle = DefaultSharedTermCellAlloc();
   current = handle;
   
   if(!TestInpTok(in, CloseSquare))
   {

      current->f_code = SIG_CONS_CODE;
      current->arity = 2;
      current->args = TermArgArrayAlloc(2);
      current->args[0] = TBTermParse(in, bank); 
      current->args[1] = DefaultSharedTermCellAlloc();
      current = current->args[1];
      PStackPushP(stack, current);
      
      while(TestInpTok(in, Comma))
      {
         NextToken(in);
         current->f_code = SIG_CONS_CODE;             
         current->arity = 2;                          
         current->args = TermArgArrayAlloc(2);
         current->args[0] = TBTermParse(in, bank); 
         current->args[1] = DefaultTermCellAlloc();
         current = current->args[1];      
	 PStackPushP(stack, current);
      }
      current = PStackPopP(stack);
   }
   AcceptInpTok(in, CloseSquare);
   current->f_code = SIG_NIL_CODE;             
   
   /* Now insert the list into the bank */
   handle = tb_termtop_insert(bank, current);

   while(!PStackEmpty(stack))
   {
      current = PStackPopP(stack);
      current->args[1] = handle;
      handle = tb_termtop_insert(bank, current);
   }
   
   PStackFree(stack);
   return handle;
}
   


/*-----------------------------------------------------------------------
//
// Function: term_remove_superterm_silently()
//
//   Remove a superterm reference silently, i.e. without printing a
//   warning if it does not exist.
//
// Global Variables: -
//
// Side Effects    : Changes term->superterms, term->ref_count
//
/----------------------------------------------------------------------*/

bool term_remove_superterm_silently(Term_p term, Term_p superterm)
{
   bool exists;
   int i, ref_count;

   exists = PTreeDeleteEntry(&(term->superterms), superterm);
   if(exists)
   {
      ref_count = 0;
      for(i=0; i< superterm->arity; i++)
      {
	 if(superterm->args[i] == term)
	 {
	    ref_count++;
	 }
      }
      assert(ref_count);
      term->ref_count-=ref_count;

      return true;
   }
   return false;
}



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function: TBAlloc()
//
//   Allocate an empty, initialized termbank.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

TB_p TBAlloc(TermProperties prop_mask, Sig_p sig)
{
   TB_p handle;
   Term_p t, term;

   handle = TBCellAlloc();

   handle->prop_mask = prop_mask;
   handle->in_count = 0;
   handle->entry_index = NULL;
   handle->ext_index = NULL;
   handle->sig = sig;
   handle->vars = VarBankAlloc();
   handle->tb_date = SysDateCreationTime();
   handle->term_root = NULL; 

   term = DefaultTermCellAlloc();
   term->f_code = SIG_TRUE_CODE;
   t = TBInsert(handle, term, DEREF_NEVER);
   TermGetRef(&(handle->true_term), t);
   TermFree(term);

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TBFree()
//
//   Free a term bank (if the signature alread has been
//   extracted). Voids all pointers to terms in the bank!
//
// Global Variables: -
//
// Side Effects    : Memory operations.
//
/----------------------------------------------------------------------*/

void TBFree(TB_p junk)
{
   assert(!junk->sig);
   
   TermTreeFree(junk->term_root);
   NumTreeFree(junk->entry_index);
   NumTreeFree(junk->ext_index);
   VarBankFree(junk->vars);
   
   TBCellFree(junk);
}


/*-----------------------------------------------------------------------
//
// Function: DefaultSharedTermCellAlloc()
//
//   Return a reasonably initialized term cell for shared terms.
//
// Global Variables: -
//
// Side Effects    : Memory Operations
//
/----------------------------------------------------------------------*/

Term_p DefaultSharedTermCellAlloc()
{
   Term_p handle;

   handle = DefaultTermCellAlloc();
   handle->ref_count = 0;
   
   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: TermGetRef()
//
//   Make ref a registered pointer to term, i.e. a pointer the term
//   node is "aware of" and that will be updated if, for some reasons,
//   the term nodes address changes.
//
// Global Variables: -
//
// Side Effects    : Changes *ref.
//
/----------------------------------------------------------------------*/

void TermGetRef(Term_p *ref, Term_p term)
{
   bool result;

   assert(ref);
   assert(term);

   *ref = term;
   term->ref_count++;
   result = PTreeStore(&(term->ext_refs), ref);
   assert(result);
}


/*-----------------------------------------------------------------------
//
// Function: TermReleaseRef()
//
//   Remove the external reference. Does not change the reference - it
//   can still be used, but will no longer be updated. This allows
//   TBDelete() to be called on this term easily.
//
// Global Variables: -
//
// Side Effects    : Changes reference tree of term.
//
/----------------------------------------------------------------------*/

void TermReleaseRef(Term_p *ref)
{
   bool result;
   
   assert(ref);
   assert(*ref);
   
   result = PTreeDeleteEntry(&((*ref)->ext_refs), ref);
   assert((*ref)->ref_count);
   (*ref)->ref_count--;
   
   assert(result);
}


/*-----------------------------------------------------------------------
//
// Function: TermIntroduceSuperterm()
//
//   Make superterm a superterm to term, change ref_count
//   accordingly. If term-supertem relation is known, do nothing.
//
// Global Variables: -
//
// Side Effects    : Changes term->superterms, term->ref_count
//
/----------------------------------------------------------------------*/

void TermIntroduceSuperterm(Term_p term, Term_p superterm)
{
   int i, ref_count;
   
   if(PTreeStore(&(term->superterms), superterm))
   {
      ref_count = 0;
      for(i=0; i< superterm->arity; i++)
      {
	 if(superterm->args[i] == term)
	 {
	    ref_count++;
	 }
      }
      assert(ref_count);
      term->ref_count+=ref_count;
   }
}


/*-----------------------------------------------------------------------
//
// Function: TermRemoveSuperterm()
//
//   If term is a (direct) subterm to superter, remove superterm from
//   the superterm list of term, change  ref_count accordingly and
//   return true. Do nothing and return false otherwise.
//
// Global Variables: -
//
// Side Effects    : Changes term->superterms, term->ref_count
//
/----------------------------------------------------------------------*/

bool TermRemoveSuperterm(Term_p term, Term_p superterm)
{
   bool exists;

   exists = term_remove_superterm_silently(term, superterm);
   if(!exists)
   {
      Warning("Superterm does not exist!");
   }
   return exists;
}


/*-----------------------------------------------------------------------
//
// Function:  TBTermEqual()
//
//   Test wether two shared terms in the same termbank are the
//   same. This is a simple pointer comparison, this function only
//   exists so that it can be passed as a function pointer.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TBTermEqual(Term_p t1, Term_p t2)
{
   return (t1==t2);
}


/*-----------------------------------------------------------------------
//
// Function: TBTermStructEqual()
//
//   Tests, wether two terms in the same termbank are structurally
//   identical (ignoring bindings).
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

bool TBTermStructEqual(Term_p t1, Term_p t2)
{
   PStack_p stack = PStackAlloc();
   int i;
   bool termeq = true;

   PStackPushP(stack, t1);
   PStackPushP(stack, t2);
   
   while(!PStackEmpty(stack))
   {
      t2 = PStackPopP(stack);
      t1 = PStackPopP(stack);

      if(t1 == t2)
      {
	 continue;
      }
      else if(t1->f_code != t2->f_code)
      {
	 termeq = false;
	 break;
      }
      assert(t1->arity == t2->arity);
      for(i=0; i<t1->arity; i++)
      {
	 PStackPushP(stack, t1->args[i]);
	 PStackPushP(stack, t2->args[i]);
      }
   }
   PStackFree(stack);
   return termeq;
}

/*-----------------------------------------------------------------------
//
// Function: TBInsert()
//
//  Insert the term into the termbank. The original term will remain
//  untouched. The routine returns a pointer to a new, shared term of
//  the same structure.
//
//  If TBSupportReplace is true, the term node "knows" about it's
//  superterms, and subsequent operations like replacing the term node
//  will change this parent pointers. Terms also administrate external
//  references (in ext_refs), which need to be explicitly administered
//  by the user.
//
//  TermProperties are masked with bank->prop_mask.
//
// Global Variables: TBSupportReplace
//
// Side Effects    : Changes term bank
//
/----------------------------------------------------------------------*/

Term_p TBInsert(TB_p bank, Term_p term, DerefType deref)
{
   int    i;
   Term_p t;

   assert(term);

   term = TermDeref(term, &deref);

   t = TermEquivCellAlloc(term, bank->vars); /* This is an unshared
						term cell at the
						moment, i.e. ref_count
						= -1! */
   t->flags = t->flags & bank->prop_mask;
   for(i=0; i<t->arity; i++)
   {
      t->args[i] = TBInsert(bank, t->args[i], deref);
   }

   return tb_termtop_insert(bank, t);
}


/*-----------------------------------------------------------------------
//
// Function: TBDelete()
//
//   Delete a term from the term bank (i.e. check the reference
//   counter and remove the cell if it is 0). Returns true if a
//   term cell has been removed, false otherwise. This function does
//   _not_ follow bindings (it would make little sense...). Do not
//   delete variable cells (but remove them from the bank).
//
// Global Variables: -
//
// Side Effects    : Changes term bank
//
/----------------------------------------------------------------------*/

bool TBDelete(TB_p bank, Term_p term)
{
   int i;
   Term_p old;

   assert(term);
   if(term->ref_count==0)
   {
      if(term->arity)
      {
	 assert(term->args);
	 for(i=0; i<term->arity; i++)
	 {
	    if(term_remove_superterm_silently(term->args[i], term))
	    {
	       TBDelete(bank, term->args[i]);
	    }
	 }
      }
      NumTreeDeleteEntry(&(bank->entry_index), term->entry_no);

      old = TermTreeExtract(&(bank->term_root), term, bank->prop_mask);
      if(!TermIsVar(old))
      {
	 TermTopFree(old);
      }
      return true;
   }
   return false;
}

/*-----------------------------------------------------------------------
//
// Function: TBPrintBankInOrder()
//
//   Print the DAG in the order of ascending entry_no. If
//   TBPrintTermsFlat is flat, print one term per node, otherwise
//   print terms refernced ones at once.
//
// Global Variables: TBPrintTermsFlat
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

void TBPrintBankInOrder(FILE* out, TB_p bank)
{
   tb_print_dag(out, bank->entry_index, bank->sig);
}


/*-----------------------------------------------------------------------
//
// Function: TBPrintTermCompact()
//
//   Print a term bank term. If TBPrintTermsFlat is set, introduce
//   abbreviations for all subterms encountered, otherwise only for
//   those with ref_count > 1. Subterms with TPOutputFlag are not
//   printed, but are assumed to be known. Does _not_ follow bindings
//   (they are temporary and as such make little sense in the term
//   bank context) 
//
// Global Variables: -
//
// Side Effects    : Output, set TPOutputFlag in subterms printed with
//                   abbreviation.
//
/----------------------------------------------------------------------*/

void TBPrintTermCompact(FILE* out, TB_p bank, Term_p term)
{
   PStack_p stack = PStackAlloc();
   int i;
   bool term_complete;
   
   PStackPushP(stack, term);

   while(!PStackEmpty(stack))
   {
      term_complete = false;
      term = PStackPopP(stack);
      if(!term)
      {
	 fprintf(out, ")");
	 term_complete = true;
      }
      else if(TermCellQueryProp(term, TPOutputFlag))
      {
	 fprintf(out, "*%ld", term->entry_no);
	 term_complete = true;
      }
      else
      {
	 if(TBPrintTermsFlat || term->ref_count >1)
	 {
	    fprintf(out, "*%ld:", term->entry_no);
	    TermCellSetProp(term, TPOutputFlag);
	 }
	 if(TermIsVar(term))
	 {
	    fprintf(out, "%c%ld", TermPrologVars ? 'X' : 'x',
		    -term->f_code);
	    term_complete = true;
	 }
	 else
	 {
	    fprintf(out, SigFindName(bank->sig, term->f_code));
	    if(TermIsConst(term))
	    {
	       if(!TermPrologVars)
	       {
		  fprintf(out, "()");
	       }
	       term_complete = true;
	    }
	    else
	    {
	       fprintf(out, "(");
	       assert(term->args);
	       PStackPushP(stack, NULL);
	       for(i=term->arity-1; i>=0; i--)
	       {
		  PStackPushP(stack, term->args[i]);
	       }
	    }
	 }
      }
      if(term_complete && !PStackEmpty(stack) && PStackTopP(stack))
      {
	 fprintf(out, ",");
      }
   }
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: TBPrintTerm()
//
//   Print a term from a term bank either in compact form (with
//   abbreviations) or as a conventional term.
//
// Global Variables: -
//
// Side Effects    : By the called functions
//
/----------------------------------------------------------------------*/

void TBPrintTerm(FILE* out, TB_p bank, Term_p term, bool fullterms)
{
   if(fullterms)
   {
      TBPrintTermFull(out, bank, term);
   }
   else
   {
      TBPrintTermCompact(out, bank, term);
   }
}



/*-----------------------------------------------------------------------
//
// Function:  TBPrintBankTerms()
//
//   Print the terms inserted into the term bank with abbreviations.
//
// Global Variables: -
//
// Side Effects    : Output, changes by TBPrintTermCompact()
//
/----------------------------------------------------------------------*/

void TBPrintBankTerms(FILE* out, TB_p bank)
{
   PStack_p stack = PStackAlloc();
   Term_p term;

   PStackPushP(stack, bank->term_root);
   
   while(!PStackEmpty(stack))
   {
      term = PStackPopP(stack);
      if(term)
      {
	 PStackPushP(stack, term->lson);
	 PStackPushP(stack, term->rson);
	 if(TermCellQueryProp(term, TPTopPos))
	 {
	    TBPrintTermCompact(out, bank, term);
	    fprintf(out, "\n");
	 }
      }
   }
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: TBTermParse()
//
//   Parse a term from the given scanner object directly into the
//   termbank. Supports abbreviations. This function will _not_ set
//   the TPTopPos property on top terms while parsing
//
// Global Variables: -
//
// Side Effects    : Input, memory operations, changes termbank.
//
/----------------------------------------------------------------------*/

Term_p TBTermParse(Scanner_p in, TB_p bank)
{
   Term_p handle;
   DStr_p id;
   DStr_p source_name, errpos;
   long   line, column;

   source_name = DStrGetRef(AktToken(in)->source_name);
   line = AktToken(in)->line;
   column = AktToken(in)->column;

   /* Test for abbreviation */
   if(TestInpTok(in, Mult))
   {
      long      abbrev;
      NumTree_p search;

      NextToken(in);
      abbrev = ParseInt(in);

      if(TestInpTok(in, Colon))
      { /* This _defines_ the abbrev! */
	 IntOrP term;
	 
	 NextToken(in);
	 term.p_val = TBTermParse(in, bank); /* Elegant, aint it? */

	 if(!NumTreeStore(&(bank->ext_index), abbrev, term, term))
	 {
	    /* Error: Abbreviation defined inconsistently */
	    errpos = DStrAlloc();
	    DStrAppendStr(errpos, PosRep(source_name, line, column));
	    DStrAppendStr(errpos, "Abbreviation *");
	    DStrAppendInt(errpos, abbrev);
	    DStrAppendStr(errpos, " already defined");
	    Error(DStrView(errpos), SYNTAX_ERROR);
	    DStrFree(errpos);
	 }
	 handle = term.p_val;
      }
      else
      { /* This references the abbrev */
	 
	 search = NumTreeFind(bank->ext_index, abbrev);
	 if(!search)
	 {
	    /* Error: Undefined abbrev */ 
	    errpos = DStrAlloc();
	    DStrAppendStr(errpos, PosRep(source_name, line, column));
	    DStrAppendStr(errpos, "Abbreviation *");
	    DStrAppendInt(errpos, abbrev);
	    DStrAppendStr(errpos, " undefined");
	    Error(DStrView(errpos), SYNTAX_ERROR);
	    DStrFree(errpos);  
	 }
	 handle = search->val1.p_val;
      }
   }
   else
   {
      /* Normal term stuff, bloated because of the nonsensical SETHEO
	 syntax */
      
      if(SigSupportLists && TestInpTok(in, OpenSquare))
      {
	 handle =  tb_parse_cons_list(in, bank);
      }
      else
      {
	 id = DStrAlloc();
	 
	 if(TermParseOperator(in, id))
	 {
	    handle = VarBankExtNameAssertAlloc(bank->vars, DStrView(id));
	 }
	 else
	 {
	    handle = DefaultSharedTermCellAlloc();
	 
	    if(TestInpTok(in, OpenBracket))
	    {
	       handle->arity = TBTermParseArgList(in, &(handle->args),
						  bank);
	    }
	    else
	    {
	       handle->arity = 0;
	    }
	    handle->f_code = SigInsertId(bank->sig, DStrView(id), handle->arity);
	    if(!handle->f_code)
	    {
	       errpos = DStrAlloc();
	       DStrAppendStr(errpos, PosRep(source_name, line, column));
	       DStrAppendStr(errpos, DStrView(id));
	       DStrAppendStr(errpos, " used with arity ");
	       DStrAppendInt(errpos, (long)handle->arity);
	       DStrAppendStr(errpos, ", but registered with arity ");
	       DStrAppendInt(errpos, 
			     (long)(bank->sig)->
			     f_info[SigFindFCode(bank->sig, DStrView(id))].arity);
	       Error(DStrView(errpos), SYNTAX_ERROR);
	       DStrFree(errpos);
	    }
	 }
	 handle = tb_termtop_insert(bank, handle);
	 DStrFree(id);
      }
   }
   DStrReleaseRef(source_name);

   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: TBTermParseArgList()
//
//   Parse a list of terms (comma-separated and enclosed in brackets)
//   into an array of (shared) term pointers. See TermParseArgList()
//   in cte_terms.c for more.
//
// Global Variables: -
//
// Side Effects    : Input, memory operations, changes term bank
//
/----------------------------------------------------------------------*/

int TBTermParseArgList(Scanner_p in, Term_p** arg_anchor, TB_p bank) 
{
   Term_p *handle;
   int    arity;
   int    size;
   int    i;

   AcceptInpTok(in, OpenBracket);
   if(TestInpTok(in, CloseBracket))
   {
      NextToken(in);
      *arg_anchor = NULL;
      return 0;
   }
   size = TERMS_INITIAL_ARGS;
   handle = (Term_p*)SizeMalloc(size*sizeof(Term_p));
   arity = 0;
   handle[arity] = TBTermParse(in, bank);
   arity++;
   while(TestInpTok(in, Comma))
   {
      NextToken(in);
      if(arity==size)
      {
         size+=TERMS_INITIAL_ARGS;
         handle = (Term_p*)SecureRealloc(handle, size*sizeof(Term_p));
 	 printf("TBTermParseArgList()-Realloc()\n");
     }
      handle[arity] = TBTermParse(in, bank);
      arity++;
   }
   AcceptInpTok(in, CloseBracket);
   *arg_anchor = TermArgArrayAlloc(arity);
   for(i=0;i<arity;i++)
   {
      (*arg_anchor)[i] = handle[i];
   }
   SizeFree(handle, size*sizeof(Term_p));
   
   return arity;
}


/*-----------------------------------------------------------------------
//
// Function: TBCountSupertermReferences()
//
//   Count the references to term in terms superterms (For debugging
//   only). 
//
// Global Variables: -
//
// Side Effects    : - (internal memory usage)
//
/----------------------------------------------------------------------*/

long TBCountSupertermReferences(Term_p term)
{
   PTree_p  cell;
   Term_p   super;
   PStack_p stack;
   long     res = 0;
   bool     local_found;

   stack = PTreeTraverseInit(term->superterms);
   while((cell = PTreeTraverseNext(stack)))
   {
      int i;

      super = cell->key;
      local_found = false;
      for(i=0; i<super->arity; i++)
      {
	 if(super->args[i]==term)
	 {
	    res++;
	    local_found = true;
	 }
      }
      assert(true);
   }
   PTreeTraverseExit(stack);
   
   return res;
}

/*-----------------------------------------------------------------------
//
// Function: TBCountTotalReferences()
//
//   Count total references to a term.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long TBCountTotalReferences(Term_p term)
{
   return PTreeNodes(term->ext_refs)+TBCountSupertermReferences(term);
}


/*-----------------------------------------------------------------------
//
// Function: TBVerifyReferences()
//
//   Check references, return true, if consistent, false
//   otherwise. Print references counts etc if inconsistent.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

bool TBVerifyReferences(FILE* out, Term_p term)
{
   bool res = true;
   long references = TBCountTotalReferences(term);

   if(references!=term->ref_count)
   {
      res = false;
      fprintf(out, "# references=%ld != ref_count=%ld\n", references,
	      term->ref_count);
   }
   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TBVerifyBankReferences()
//
//   For all term nodes in the term bank, verify the references.
//
// Global Variables: -
//
// Side Effects    : Output
//
/----------------------------------------------------------------------*/

Term_p  TBVerifyBankReferences(FILE* out, TB_p bank)
{
   PStack_p stack;
   Term_p   test, res = NULL;

   stack = TermTreeTraverseInit(bank->term_root);
   while((test = TermTreeTraverseNext(stack)))
   {
      if(!TBVerifyReferences(out, test))
      {
	 TBPrintTermFull(out, bank, test);
	 fprintf(out, "\n");
	 res = test;
      }
   }
   TermTreeTraverseExit(stack);

   return res;
}


/*-----------------------------------------------------------------------
//
// Function: TBRefSetProp()
//
//   Make ref point to a term of the same structure as *ref, but with
//   properties prop set.
//
// Global Variables: -
//
// Side Effects    : Changes bank, *ref
//
/----------------------------------------------------------------------*/

void TBRefSetProp(TB_p bank, TermRef ref, TermProperties prop)
{
   Term_p term, new;
   
   term = *ref;
   if(TermCellQueryProp(term, prop))
   {
      return;
   }
   
   new = TermTopCopy(term);
   TermReleaseRef(ref);
   TermCellSetProp(new, prop);
   new = tb_termtop_insert(bank, new);
   TermGetRef(ref, new);
   TBDelete(bank, term); /* Cannot be done earlier ->
			    tb_termtop_insert() has to register the
			    new superterm for the subterms */
}



/*-----------------------------------------------------------------------
//
// Function: TBRefDelProp()
//
//   Make ref point to a term of the same structure as *ref, but with
//   properties prop deleted.
//
// Global Variables: -
//
// Side Effects    : Changes bank, *ref
//
/----------------------------------------------------------------------*/

void TBRefDelProp(TB_p bank, TermRef ref, TermProperties prop)
{
   Term_p term, new;
   
   term = *ref;
   if(!TermCellIsAnyPropSet(term, prop))
   {
      return;
   }
   new = TermTopCopy(term);
   TermReleaseRef(ref);
   TermCellDelProp(new, prop);
   new = tb_termtop_insert(bank, new);
   TermGetRef(ref, new);
   TBDelete(bank, term);
}



/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/






