/*-----------------------------------------------------------------------

File  : cte_termtrees.c

Author: Stephan Schulz

Contents
 
  Implementation of term-top indexed AVL trees (I found that I can
  cleanly separate this from the termbank stuff). 

Changes

<1> Thu Nov 27 19:12:51 MET 1997
    New

-----------------------------------------------------------------------*/

#include "cte_termtrees.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: rotate_left()
//
//   Rotate the unbalanced tree at *update left, adjusting weights
//   accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left(Term_p *update)
{
   Term_p store;
   int b1,b2;

   VERBOUT10("Rotate tree left\n");

   store = (*update)->rson;
   (*update)->rson = (*update)->rson->lson;
   store->lson = *update;
   *update = store;
   
   b1 = (*update)->balance;
   b2 = ((*update)->lson)->balance;
   
   switch(b1)
   {
   case -1:
	 assert(b2 == 1);
	 (*update)->balance = -2;
	 ((*update)->lson)->balance = 0;
	 break;
   case 0:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 0;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = 1;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 1:
	 if(b2 == 1)
	 {
	    (*update)->balance = -1;
	    ((*update)->lson)->balance = -1;
	 }
	 else if(b2 == 2)
	 {
	    (*update)->balance = 0;
	    ((*update)->lson)->balance = 0;
	 }
	 else
	 {
	    assert(false);
	 }
	 break;
   case 2:
	 assert(b2 == 2);
	 (*update)->balance = 0;
	 ((*update)->lson)->balance = -1;
	 break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right()
//
//   Symmetrical case to rotate_left()
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right(Term_p *update)
{
   Term_p store;
   int b1, b2;

   VERBOUT10("Rotate tree right\n");

   store = (*update)->lson;
   (*update)->lson = (*update)->lson->rson;
   store->rson = *update;
   *update = store;

   b1 = (*update)->balance;
   b2 = ((*update)->rson)->balance;

   switch(b1)
   {
   case 1:
         assert(b2 == -1);
         (*update)->balance = 2;
         ((*update)->rson)->balance = 0;
         break;
   case 0:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 0;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = -1;
         }
         else
         {
            assert(false);
         }
         break;
   case -1:
         if(b2 == -1)
         {
            (*update)->balance = 1;
            ((*update)->rson)->balance = 1;
         }
         else if(b2 == -2)
         {
            (*update)->balance = 0;
            ((*update)->rson)->balance = 0;
         }
         else
         {
            assert(false);
         }
         break;
   case -2:
         assert(b2 == -2);
         (*update)->balance = 0;
         ((*update)->rson)->balance = 1;
         break;
   }
}

/*-----------------------------------------------------------------------
//
// Function: rotate_right_left()
//
//   Perform the right-left rotation to balance a tree. More exactly:
//   Rotate the subtree at (*update)->rson right and the tree at
//   *update left, and adjust the balances accordingly.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_right_left(Term_p *update)
{
   VERBOUT10("Rotate tree right left\n");

   rotate_right(&((*update)->rson));
   rotate_left(update);
}


/*-----------------------------------------------------------------------
//
// Function: rotate_left_right()
//
//   Perform the left-right rotation to balance a tree. See
//   rotate_right_left() 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static void rotate_left_right(Term_p *update)
{
   VERBOUT10("Rotate tree left right\n");

   rotate_left(&((*update)->lson));
   rotate_right(update);
}

/*-----------------------------------------------------------------------
//
// Function: rebalance_tree()
//
//   Rebalance the tree at the passed position. Assumes that the
//   weights in the tree are correct. Returns true if the tree depth
//   has changed.
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool rebalance_tree(Term_p *update)
{
   int b2 = 0;
   bool shrunk;

   if((*update)->balance == 2)
   {
      b2 = (*update)->rson->balance;
      assert((*update)->rson);
      if((*update)->rson->balance == -1)
      {
	 rotate_right_left(update);
      }
      else
      {
	 rotate_left(update);
      }
   }
   else if((*update)->balance == -2)
   {
      b2 = (*update)->lson->balance;
      assert((*update)->lson);
      if((*update)->lson->balance == 1)
      {
	 rotate_left_right(update);
      }
      else
      {
	 rotate_right(update);
      }
   }

   shrunk = (b2 != 0);
   
   return shrunk;
}

/*-----------------------------------------------------------------------
//
// Function: find_min_cell()
//
//   Find the position of the pointer to the minimal cell in the tree
//   anchored at root. See find_max_cell().
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

static Term_p *find_min_cell(Term_p *root, bool remove_mode)
{
   if(!*root)
   {
      return NULL;
   }

   while((*root)->lson)
   {
      (*root)->passed = -1;
      root = &((*root)->lson);
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: find_cell_pos()
//
//   Starting at root, search the tree for an entry with 'key', setting
//   'passed' values along the search papth and keeping *update
//   pointing to a potential update position. If update_on_zero is
//   true, set update to the lowest cell with balance 0, otherwise to
///   the lowest cell with balance != 0. Return the pointer to
//   where the entry in anchored, or the position where it would be
//   inserted if no such entry exists.
//
// Global Variables: -
//
// Side Effects    : Sets 'passed' values and *update
//
/----------------------------------------------------------------------*/

static Term_p *find_cell_pos(Term_p *root, Term_p key, 
			     Term_p **update,
			     bool remove_mode, 
			     TermProperties prop_mask)
{
   int cmpres;

   while(*root)
   {
      cmpres = TermTopCompare(*root, key, prop_mask);
      if(cmpres == 0)
      {
	 return root;
      }
      
      if(!remove_mode && (*root)->balance)
      {
	 *update = root;
      }
      if(cmpres > 0)
      {
	 (*root)->passed = 1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == -1))
	 {
	    *update = root;
	 }
	 root = &((*root)->rson);
      }
      else
      {
	 (*root)->passed = -1;
	 if(remove_mode &&
	    ((*root)->balance == 0 ||(*root)->balance == 1))
	 {
	    *update = root;
	 }
	 root = &((*root)->lson);
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: isolate_and_mark_cell()
//
//   Find the cell with the key and move it to a position in the tree
//   where it can be deleted (i.e. a position where it has only one
//   successor). Set the passed-values leading to the new position,
//   and set the passed-value of the relevant cell to 0 so that it can
//   be found without (expensive) key comparisons.
//
// Global Variables: -
//
// Side Effects    : Changes tree and some cell entries.
//
/----------------------------------------------------------------------*/

static Term_p isolate_and_mark_cell(Term_p *root, Term_p key, 
			     TermProperties prop_mask)
{
   Term_p *elpos = root;
   Term_p *replace = NULL;
   Term_p *dummy;
   Term_p element, repl, t_lson, t_rson;
   long   help;
   
   elpos = find_cell_pos(root, key, &dummy, true, prop_mask);
   if(!(*elpos))
   {
      return NULL;
   }
   element = *elpos;
   
   if(element->rson&&element->lson)
   {
      replace = find_min_cell(&(element->rson), true);
      repl = *replace;
      
      t_lson = repl->lson;
      t_rson = repl->rson;
      repl->lson = element->lson;
      
      if(repl == element->rson)
      {
	 repl->rson =  element;
      }
      else
      {
	 repl->rson = element->rson;
	 *replace = element;
      }
      
      *elpos = repl;
      element->lson = t_lson;
      element->rson = t_rson;
      
      repl->passed = 1;
      element->passed = 0;
      help = repl->balance;
      repl->balance = element->balance;
      element->balance = help; 
   }
   else 
   {
      element->passed = 0;
   }
   
   return element;
}


/*-----------------------------------------------------------------------
//
// Function: extract_marked_cell()
//
//   Descend the tree guided by the passed-entries until the marked
//   cell (passed == 0) has been found, extract it, store it in *cell,
//   and ascend the tree, rebalancing as necessary. Return true if the
//   tree height has changes. The tree has to be preprocessed by
//   isolate_and_mark_cell().
//
// Global Variables: -
//
// Side Effects    : Changes the tree
//
/----------------------------------------------------------------------*/

static bool extract_marked_cell(Term_p *root, Term_p *cell)
{
   bool shrunk = false;

   assert(*root);

   if(!(*root)->passed)
   {
      *cell = *root;
      if((*cell)->lson)
      {
	 *root = (*cell)->lson;
      }
      else
      {
	 *root = (*cell)->rson;
      }
      shrunk = true;
   }
   else
   {
      if((*root)->passed == 1)
      {
	 if(extract_marked_cell(&((*root)->rson), cell))
	 {
	    shrunk = true;
	    (*root)->balance--;
	 }
      }
      else if((*root)->passed == -1)
      {
	 if(extract_marked_cell(&((*root)->lson), cell))
	 {
	    shrunk = true;
	    (*root)->balance++;
	 }
      }

      switch((*root)->balance)
      {
      case 2:
      case -2:
	    shrunk = rebalance_tree(root);
	    break;
      case 1:
      case -1:
	    shrunk = false;
	    break;
      case 0:
	    break;
      default:
	    assert(false);
	    break;
      }
   }
   return shrunk;
}

/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: TermTreeFree(Term_p junk)
//
//   Release the memory taken by a term top AVL tree. Do not free
//   variables, as they belong to a variable bank as well. Yes, this
//   is an ugly hack! *sigh*
//
// Global Variables: -
//
// Side Effects    : Memory operations, destroys tree
//
/----------------------------------------------------------------------*/

void TermTreeFree(Term_p junk)
{
   if(junk)
   {
      TermTreeFree(junk->lson);
      TermTreeFree(junk->rson);
      if(!TermIsVar(junk))
      {
	 TermTopFree(junk);
      }
   }
}   

/*-----------------------------------------------------------------------
//
// Function: TermTopCompare()
//
//   Original version:
//
//     Compare two top level term cells as
//     f_code.masked_properties.args_as_pointers, return a value >0 if
//     t1 is greater, 0 if the terms are identical, <0 if t2 is
//     greater. 
//
//   New version: Avoid pointer arithmetic to increase portability,
//   compare entry_nos of subterm instead!
//
//   Even newer version (for increased efficiency): Compare
//   f_code.entry_nos.masked_proerties. The original pointer version
//   might be a way to get significantly more efficient (we don't need
//   to dereference the argument pointers!). I'll have to think about
//   this. 
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

int TermTopCompare(Term_p t1, Term_p t2, TermProperties prop_mask)
{
   int i, res;
   
   res = t1->f_code - t2->f_code;   
   if(res)
   {
      return res;
   }

   assert(t1->arity == t2->arity);
   for(i=0; i<t1->arity; i++)
   {
      /* res = (t1->args[i]->entry_no) - (t2->args[i]->entry_no); */
      res = PCmp(t1->args[i], t2->args[i]);
      if(res)
      {
	 return res;
      }
   }

   res = (t1->flags & prop_mask) - (t2->flags & prop_mask);
   
   return res;
}



/*-----------------------------------------------------------------------
//
// Function: TermTreeFind()
//
//   Find a entry in the term tree, given a cell with correct
//   (i.e. term-bank) argument pointers.
//   pointers
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p TermTreeFind(Term_p root, Term_p term, TermProperties prop_mask)
{
   int cmpres;
   
   while(root)
   {
      cmpres = TermTopCompare(root, term, prop_mask);

      if(cmpres == 0)
      {
         break;
      }
      if(cmpres > 0)
      {
         root = root->rson;
      }
      else
      {
         root = root->lson;
      }
   }
   return root;
}


/*-----------------------------------------------------------------------
//
// Function: TermTreeInsert() 
//
//   Insert a term with valid subterm pointers into the termtree. If
//   the entry already exists, return pointer to existing entry as
//   usual, otherwise return NULL.
//
// Global Variables: -
//
// Side Effects    : Changes tree
//
/----------------------------------------------------------------------*/

Term_p TermTreeInsert(Term_p *root, Term_p term, 
		      TermProperties prop_mask)
{
   Term_p *insert = root;
   Term_p *update = root;
   Term_p help;

   insert = find_cell_pos(root, term, &update, false, prop_mask);
   if(*insert)
   {
      return *insert;
   }
   
   *insert = term;
   term->lson = term->rson = NULL;
   term->balance = 0;
   
   help = *update;
   while(help != term)
   {
      help->balance += help->passed;
      help = help->passed == 1 ? help->rson : help->lson;
   }
   rebalance_tree(update);
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: TermTreeExtract()
//
//   Remove a top term cell from the term tree and return a pointer to
//   it.
//
// Global Variables: -
//
// Side Effects    : Changes tree
//
/----------------------------------------------------------------------*/

Term_p TermTreeExtract(Term_p *root, Term_p term, TermProperties
		       prop_mask) 
{
   Term_p cell;
   
   if(!isolate_and_mark_cell(root, term, prop_mask))
   {
      return NULL;
   }
   extract_marked_cell(root, &cell);

   cell->lson = cell->rson = NULL;
   return cell;
}


/*-----------------------------------------------------------------------
//
// Function: TermTreeDelete()
//
//   Delete a top term from the term tree.
//
// Global Variables: -
//
// Side Effects    : Changes tree, memory operations
//
/----------------------------------------------------------------------*/

bool TermTreeDelete(Term_p *root, Term_p term, TermProperties prop_mask)
{
   Term_p cell;
   
   cell = TermTreeExtract(root, term, prop_mask);
   if(cell)
   {
      TermTopFree(cell);
      return true;
   }
   return false;
}


/*-----------------------------------------------------------------------
//
// Function: TermTreeSetProp()
//
//   Set the given properties for all term cells in the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void TermTreeSetProp(Term_p root, TermProperties props)
{
   PStack_p stack = PStackAlloc();
   
   PStackPushP(stack, root);
   
   while(!PStackEmpty(stack))
   {
      root = PStackPopP(stack);
      if(root)
      {
	 TermCellSetProp(root, props);
	 PStackPushP(stack, root->lson);
	 PStackPushP(stack, root->rson);
      }
   }
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: TermTreeDelProp()
//
//   Delete the given properties for all term cells in the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

void TermTreeDelProp(Term_p root, TermProperties props)
{
   PStack_p stack = PStackAlloc();
   
   PStackPushP(stack, root);
   
   while(!PStackEmpty(stack))
   {
      root = PStackPopP(stack);
      if(root)
      {
	 TermCellDelProp(root, props);
	 PStackPushP(stack, root->lson);
	 PStackPushP(stack, root->rson);
      }
   }
   PStackFree(stack);
}

/*-----------------------------------------------------------------------
//
// Function: TermTreeNodes()
//
//   Return the number of nodes in the tree.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

long TermTreeNodes(Term_p root)
{
   PStack_p stack = PStackAlloc();
   long     res   = 0;

   PStackPushP(stack, root);
   
   while(!PStackEmpty(stack))
   {
      root = PStackPopP(stack);
      if(root)
      {
         PStackPushP(stack, root->lson);
         PStackPushP(stack, root->rson);
         res++;
      }
   }
   PStackFree(stack);

   return res;   
}


AVL_TRAVERSE_DEFINITION(TermTree, Term_p)


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/






