/*-----------------------------------------------------------------------

File  : cte_termtypes.c

Author: Stephan Schulz

Contents
 
  Declarations for the basic term type and primitive functions, mainly
  on single term cells. This module mostly provides only
  infrastructure for higher level modules.


Changes

<1> Tue Feb 24 02:17:11 MET 1998

-----------------------------------------------------------------------*/

#include "cte_termtypes.h"


/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/


/*-----------------------------------------------------------------------
//
// Function: DefaultTermCellAlloc()
//
//   Allocate a term cell with default values.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p DefaultTermCellAlloc()
{
   Term_p handle;
   
   handle = TermCellAlloc();

   handle->flags = 0;
   handle->arity = 0;
   handle->binding = NULL;
   handle->args = NULL;
   handle->entry_no = 0;
   handle->ref_count = -1;
   handle->superterms = NULL;
   handle->ext_refs = NULL;
   handle->is_indexed = false;
   handle->balance = 0;
   handle->passed = 0;
   handle->lson = NULL;
   handle->rson = NULL;

   return handle;
}


/*-----------------------------------------------------------------------
//
// Function: TermTopFree()
//
//   Return term cell and arg array (if it exists).
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void TermTopFree(Term_p junk)
{
   if(junk->arity)
   {
      assert(junk->args);
      TermArgArrayFree(junk->args, junk->arity);
   }
   else
   {
      assert(!junk->args);
   }
   PTreeFree(junk->superterms);
   PTreeFree(junk->ext_refs);
   TermCellFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function:  TermFree()
//
//   Return the memory taken by an (unshared) term. Does not free the
//   variable cells, which belong to a VarBank.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void TermFree(Term_p junk)
{
   assert(junk);
   if(!TermIsVar(junk))
   {
      assert(junk->ref_count == -1);
      if(junk->arity)
      {
	 int i;
	 
	 assert(junk->args);
	 for(i=0; i<junk->arity; i++)
	 {
	    TermFree(junk->args[i]);
	 }
      }
      else
      {
	 assert(!junk->args);
      }
      TermTopFree(junk);
   }
}


/*-----------------------------------------------------------------------
//
// Function: TermSetProp()
//
//   Set the properties in all term cells belonging to term.
//
// Global Variables: -
//
// Side Effects    : Changes properties (even in shared terms! Beware!)
//
/----------------------------------------------------------------------*/

void TermSetProp(Term_p term, DerefType deref, TermProperties prop)
{
   PStack_p stack = PStackAlloc();
   int i;

   PStackPushP(stack, term);
   PStackPushInt(stack, deref);
   
   while(!PStackEmpty(stack))
   {
      deref = PStackPopInt(stack);
      term  = PStackPopP(stack);
      term  = TermDeref(term, &deref);
      TermCellSetProp(term, prop);
      for(i=0; i<term->arity; i++)
      {
	 PStackPushP(stack, term->args[i]);
	 PStackPushInt(stack, deref);
      }
   }
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: TermDelProp()
//
//   Delete the properties in all term cells belonging to term.
//
// Global Variables: -
//
// Side Effects    : Changes properties (even in shared terms! Beware!)
//
/----------------------------------------------------------------------*/

void TermDelProp(Term_p term, DerefType deref, TermProperties prop)
{
   PStack_p stack = PStackAlloc();
   int i;

   PStackPushP(stack, term);
   PStackPushInt(stack, deref);
   
   while(!PStackEmpty(stack))
   {
      deref = PStackPopInt(stack);
      term  = PStackPopP(stack);
      term  = TermDeref(term, &deref);
      TermCellDelProp(term, prop);
      for(i=0; i<term->arity; i++)
      {
	 PStackPushP(stack, term->args[i]);
	 PStackPushInt(stack, deref);
      }
   }
   PStackFree(stack);
}


/*-----------------------------------------------------------------------
//
// Function: TermDeref()
//
//   Dereference a term. deref* tells us how many derefences to do
//   at most, it will be decremented for each dereferenciation.
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p TermDeref(Term_p term, DerefType_p deref)
{
   assert(TermIsVar(term)||!(term->binding));

   if(*deref == DEREF_ALWAYS)
   {
      while(term->binding)
      {
	 term = term->binding;
	 DEBUG(1024, printf("# TermDeref ALWAYS: %d, %ld\n",
			    (int)term, *deref););
      }
   }
   else
   {
      while(*deref)
      {
	 if(!term->binding)
	 {
	    break;
	 }
	 term = term->binding;
	 DEBUG(1024, printf("# TermDeref ELSE: %d, %ld\n",
			    (int)term, *deref););
	 (*deref)--;	 
      }
   }
   return term;
}


/*-----------------------------------------------------------------------
//
// Function: TermArgListCopy()
//
//   Return a copy of the argument array of source.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p* TermArgListCopy(Term_p source)
{
   Term_p *handle;
   int i;
   
   if(source->arity)
   {
      handle = TermArgArrayAlloc(source->arity);
      for(i=0; i<source->arity; i++)
      {
	 handle[i] = source->args[i];
      }
   }
   else
   {
      handle = NULL;
   }
   return handle;
}



/*-----------------------------------------------------------------------
//
// Function: TermTopCopy()
//
//   Return a copy of the term node (and potential argument
//   pointers). Only the top node and the pointers are duplicated, the
//   arguments are shared between source and copy. As this function
//   operates on nodes, it does not follow bindings! Administrative
//   stuff (refs etc. will, of course, not be copied but initialized
//   to rational values for an unshared 
//   term).
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

Term_p TermTopCopy(Term_p source)
{
   Term_p handle;
   
   handle = DefaultTermCellAlloc();
   handle->flags  = source->flags;
   handle->f_code = source->f_code;
   handle->arity  = source->arity;
   handle->binding = NULL;
   handle->args = TermArgListCopy(source);
   handle->entry_no = 0;
   handle->superterms = NULL;
   handle->ext_refs = NULL;
   handle->balance = 0;
   handle->passed = 0;
   handle->lson = NULL;
   handle->rson = NULL;
   
   return handle;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/




