/*-----------------------------------------------------------------------

File  : cte_termvars.c

Author: Stephan Schulz

Contents
 
  Functions for the management of shared variables.

Changes

<1> Wed Feb 25 00:48:17 MET 1998
    new

-----------------------------------------------------------------------*/

#include "cte_termvars.h"



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/


/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/



/*---------------------------------------------------------------------*/
/*                         Exported Functions                          */
/*---------------------------------------------------------------------*/

/*-----------------------------------------------------------------------
//
// Function:  VarBankAlloc()
//
//   Allocate an empty, initialized VarBank-Structure, return pointer
//   to it.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

VarBank_p VarBankAlloc()
{
   VarBank_p handle;

   handle = VarBankCellAlloc();
   handle->v_count = 0;
   handle->ext_index = NULL;
   handle->f_code_index = NULL;

   return handle;
}

/*-----------------------------------------------------------------------
//
// Function: VarBankFree()
//
//   Deallocate a VarBankCell.
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void VarBankFree(VarBank_p junk)
{
   PStack_p  stack = PStackAlloc();
   NumTree_p handle;

   assert(junk);
   StrTreeFree(junk->ext_index);
   
   PStackPushP(stack, junk->f_code_index);
   while(!PStackEmpty(stack))
   {
      handle = PStackPopP(stack);
      if(handle)
      {
	 PStackPushP(stack, handle->lson);
	 PStackPushP(stack, handle->rson);
	 TermTopFree(handle->val1.p_val);
      }
   }

   PStackFree(stack);
   NumTreeFree(junk->f_code_index);
   VarBankCellFree(junk);
}

/*-----------------------------------------------------------------------
//
// Function: VarBankClearExtNames()
//
//   Reset the External name -> FunCode association state
//
// Global Variables: -
//
// Side Effects    : Memory operations
//
/----------------------------------------------------------------------*/

void VarBankClearExtNames(VarBank_p vars)
{
   StrTreeFree(vars->ext_index);
   vars->ext_index = NULL;
   vars->v_count = 0;
}

/*-----------------------------------------------------------------------
//
// Function:  VarBankFCodeFind()
//
//   Return the pointer to the variable associated with given f_code
//   if it exists in the VarBank, NULL otherwise.  
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p VarBankFCodeFind(VarBank_p bank, FunCode f_code)
{   
   NumTree_p entry;
   
   entry = NumTreeFind(bank->f_code_index, f_code);
   
   if(entry)
   {
      return entry->val1.p_val;
   }
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function:  VarBankExtNameFind()
//
//   Return the pointer to the variable associated with given external
//   name if it exists in the VarBank, NULL otherwise.  
//
// Global Variables: -
//
// Side Effects    : -
//
/----------------------------------------------------------------------*/

Term_p VarBankExtNameFind(VarBank_p bank, char* name)
{
   StrTree_p entry;
   
   entry = StrTreeFind(bank->ext_index, name);
   
   if(entry)
   {
      return entry->val1.p_val;
   }
   return NULL;
}


/*-----------------------------------------------------------------------
//
// Function: VarBankFCodeAssertAlloc()
//
//   Return a pointer to the variable with the given f_code in the
//   variable bank. Create the variable if it does not exist.
//
// Global Variables: -
//
// Side Effects    : May change variable bank
//
/----------------------------------------------------------------------*/

Term_p VarBankFCodeAssertAlloc(VarBank_p bank, FunCode f_code)
{
   IntOrP    var_p;
   Term_p    var;
   
   var = VarBankFCodeFind(bank, f_code);
   if(!var)
   {
      var = DefaultTermCellAlloc();
      var->ref_count = 0;
      var->f_code = f_code;
      var_p.p_val = var;
      NumTreeStore(&(bank->f_code_index), f_code, var_p, var_p);
   }
   return var;
}

/*-----------------------------------------------------------------------
//
// Function: VarBankGetFreshVar()
//
//   Return a pointer to the next "fresh" variable. Freshness is
//   controlled by the v_count entry in the variable bank, which is
//   increased by this function. The variable is only guaranteed to be
//   fresh if VarBankFCodeAssertAlloc() calls are not mixed with
//   VarBankGetFreshVar() calls.
//
// Global Variables: 
//
// Side Effects    : 
//
/----------------------------------------------------------------------*/

Term_p  VarBankGetFreshVar(VarBank_p bank)
{
   Term_p var;
   
   var = VarBankFCodeAssertAlloc(bank, -(++(bank->v_count)));
   
   return var;
}

/*-----------------------------------------------------------------------
//
// Function: VarBankExtNameAssertAlloc()
//
//   Return a pointer to the variable with the given external name in
//   the variable bank. Create a new variable if none with the given
//   name exists and assign it the next unused FunCode. 
//
// Global Variables: -
//
// Side Effects    : May change variable bank
//
/----------------------------------------------------------------------*/

Term_p VarBankExtNameAssertAlloc(VarBank_p bank, char* name)
{
   Term_p    var;
   StrTree_p handle, test;

   var = VarBankExtNameFind(bank, name);
   if(!var)
   {
      var = VarBankGetFreshVar(bank);
      handle = StrTreeCellAlloc();
      handle->key = SecureStrdup(name);
      handle->val1.p_val = var;
      handle->val2.i_val = var->f_code;
      test = StrTreeInsert(&(bank->ext_index), handle);
      assert(test == NULL);
   }
   return var;
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


