/*-----------------------------------------------------------------------

File  : cl_test.c

Author: Stephan Schulz

Contents
 
  Main program for the CLIB test programm

Changes

<1> Fri Nov 28 00:27:40 MET 1997

-----------------------------------------------------------------------*/

#include <stdio.h>
#include <cio_commandline.h>
#include <cio_output.h>
#include <ccl_paramod.h>
#include <ccl_clausesets.h>


/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_REFS,
   OPT_DISCOUNT_VARS,
   OPT_NO_INFIX,
   OPT_FULL_EQ_REP
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},
   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program."},
   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},
   {OPT_PRINT_REFS,
    'r', "print-reference-number",
    OptArg, "true",
    "Print number of references for each DAG node as a comment."},
   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    OptArg, "true",
    "Read and write PCL terms (otherwise use PROLOG terms)."},
   {OPT_NO_INFIX,
    'p', "eqn-no-infix",
    OptArg, "true",
    "Print equations in prefix notation equal(x,y)."},
   {OPT_FULL_EQ_REP,
    'e', "full-equational-rep",
    OptArg, "true",
    "Print all literals as equations, even non-equational ones."},
   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

char *outname = NULL;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p   state;
   Scanner_p   in; 
   TB_p        from_bank, into_bank;
   Sig_p       sig;
   OCB_p       ocb;
   Clause_p    into, from, cp;
   ClausePos_p from_pos, into_pos;
   VarBank_p   freshvars;
   Term_p      term;
   
   assert(argv[0]);
   InitError(argv[0]);

   state = process_options(argc, argv);
   
   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }

   sig = SigAlloc();
   from_bank = TBAlloc(TPIgnoreProps, sig);
   into_bank = TBAlloc(TPIgnoreProps, sig);
   freshvars = VarBankAlloc();

   from_pos = ClausePosAlloc();
   into_pos = ClausePosAlloc();

   in = CreateScanner(state->argv[0] , true, NULL);

   while(!TestInpTok(in, NoToken))
   {
      printf("Parsing into\n");
      into = ClauseParse(in, into_bank);
      printf("Parsing from\n");
      from = ClauseParse(in, from_bank);
      printf("Parsing done...\n");
      ocb = OCBAlloc(LPO, sig);
      TOGenerateDefaultPrecedence(ocb, sig);
      printf("Marking Maximal terms...\n");
      ClauseMarkMaximalTerms(ocb, into);
      ClauseMarkMaximalTerms(ocb, from);   
      
      printf("Printing....\n");
      ClausePrint(stdout, into, true); 
      fprintf(stdout, "\n");
      ClausePrint(stdout, from, true); 
      fprintf(stdout, "\n");

      term = ClausePosFirstParamodPair(from, from_pos, into, into_pos,
				       normal);
      while(term)
      {
	 fprintf(stdout, "Overlap into: ");
	 TBPrintTermFull(stdout, into_bank, term);
	 fprintf(stdout, "\n");
	 VarBankResetVCount(freshvars);
	 assert(TermPosIsTopPos(from_pos->pos));
   	 cp = ClauseOrderedParamod(into_bank, ocb, from_pos, into_pos,
				   freshvars);
	    
	 if(cp)
	 {
	    fprintf(stdout, "CP: ");
	    ClausePrint(stdout, cp, true); 
	    fprintf(stdout, "\n");
	    ClauseFree(cp);
	 }
	 else
	 {
	    fprintf(stdout, "No cp!\n");
	 }
	 printf("vor ClausePosNextParamodPair()\n");
	 term = ClausePosNextParamodPair(from_pos, into_pos, normal);
	 printf("nach ClausePosNextParamodPair()\n");
      }
      OCBFree(ocb);
      ClauseFree(from);
      ClauseFree(into);
   }

   DestroyScanner(in);
   ClausePosFree(from_pos);
   ClausePosFree(into_pos);

   VarBankFree(freshvars);
   SigFree(sig);
   into_bank->sig = NULL; 
   TBFree(into_bank);
   from_bank->sig = NULL; 
   TBFree(from_bank);

   CLStateFree(state);

#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif

   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_INFIX:
	    EqnUseInfix = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FULL_EQ_REP:
	    EqnFullEquationalRep = CLStateGetBoolArg(handle, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}
 
void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
cl_test\n\
\n\
Usage: term2dag [options] [files]\n\
\n\
Read a set of terms and print a DAG representing it.\n\
\n");
   PrintOptions(stdout, opts);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


