/*-----------------------------------------------------------------------

File  : cl_test.c

Author: Stephan Schulz

Contents
 
  Main program for the CLIB test programm

Changes

<1> Fri Nov 28 00:27:40 MET 1997

-----------------------------------------------------------------------*/

#include <stdio.h>
#include <clb_pdarrays.h>
#include <cio_commandline.h>
#include <cio_output.h>
#include <ccl_paramod.h>
#include <ccl_clausesets.h>
#include <ccl_rewrite.h>
#include <che_hcb.h>

/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_REFS,
   OPT_DISCOUNT_VARS,
   OPT_NO_INFIX,
   OPT_FULL_EQ_REP
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},
   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program."},
   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},
   {OPT_PRINT_REFS,
    'r', "print-reference-number",
    OptArg, "true",
    "Print number of references for each DAG node as a comment."},
   {OPT_DISCOUNT_VARS,
    'd', "discount-vars",
    OptArg, "true",
    "Read and write PCL terms (otherwise use PROLOG terms)."},
   {OPT_NO_INFIX,
    'p', "eqn-no-infix",
    OptArg, "true",
    "Print equations in prefix notation equal(x,y)."},
   {OPT_FULL_EQ_REP,
    'e', "full-equational-rep",
    OptArg, "true",
    "Print all literals as equations, even non-equational ones."},
   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

char *outname = NULL;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p   state;
   Scanner_p   in; 
   TB_p        demod_bank, repl_bank;
   Sig_p       sig;
   OCB_p       ocb;
   Clause_p    clause;
   ClauseSet_p demod, repl;
   long i = 0;
   WFCB_p      wfcb;

   assert(argv[0]);
   InitError(argv[0]);

   state = process_options(argc, argv);
   
   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }

   sig = SigAlloc();
   demod_bank = TBAlloc(TPIgnoreProps, sig);
   repl_bank = TBAlloc(TPIgnoreProps, sig);
   
   demod = ClauseSetAlloc();
   demod->demod_index = PDTreeAlloc(50); /* Hack! */
   SysDateInc(&(demod->date));
   repl = ClauseSetAlloc();
   
   in = CreateScanner(state->argv[0] , true, NULL);
   
   wfcb = LIFOEvalInit(NULL, NULL, NULL, NULL, NULL, NULL);
   while(ClauseStartsMaybe(in))
   {
      clause = ClauseParse(in, demod_bank);
      ClauseAddEvaluation(wfcb, clause);
      clause->date = demod->date;
      SysDateInc(&(demod->date));
      ClausePrint(stdout, clause, true);
      printf("\n");
      ClauseSetIndexedInsert(demod, clause);
      /* ClauseSetInsert(demod, clause); */
   }
   WFCBFree(wfcb);
   
   AcceptInpTok(in, Fullstop);
   ClauseSetParseList(in, repl, repl_bank);

   printf("Verifying demod_bank\n");
   TBVerifyBankReferences(stdout, demod_bank);
   printf("Verifying repl_bank\n");
   TBVerifyBankReferences(stdout, repl_bank);

   ocb = TOGenerateDefaultOrdering(sig);
   ClauseSetMarkMaximalTerms(ocb, demod);
   
   printf("Demodulators:\n");
   ClauseSetPrint(stdout, demod, true);
   printf("\nSet to be demodulated:\n");
   ClauseSetPrint(stdout, repl, true); 
   printf("\nReady for takeoff: %ld nodes\n", TermTreeNodes(repl_bank->term_root));
   
   for(clause = repl->anchor->succ; clause != repl->anchor; clause =
	  clause->succ)
   {
      printf("# %ld.\n", i++);
      ClauseComputeLINormalform(ocb, repl_bank, clause, NULL,BothSides, demod);
   }

   printf("\nCompleted: %ld nodes\n", TermTreeNodes(repl_bank->term_root));
   printf("\nRewrite-Attempts: %ld\nRewrite-Sucesses: %ld\nResult:\n",
	  RewriteAttempts, RewriteSucesses);
   ClauseSetPrint(stdout, repl, true);
   printf("---\n");
   TBPrintBankInOrder(stdout, repl_bank);
   
   OCBFree(ocb);
   DestroyScanner(in);
   
   printf("Freigabe der Clauselmengen\n");
   ClauseSetFree(repl);
   ClauseSetFree(demod);
   
   printf("Freigabe der TermBaenke\n");
   demod_bank->sig = NULL; 
   TBFree(demod_bank);
   repl_bank->sig = NULL; 
   TBFree(repl_bank);
   printf("Freigabe der Signatur\n");
   SigFree(sig);
   
   CLStateFree(state);
   printf("Vor Memory-Debug\n");
#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif

   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	    Verbose = CLStateGetIntArg(handle, arg);
	    break;
      case OPT_HELP: 
	    print_help(stdout);
	    exit(NO_ERROR);
      case OPT_OUTPUT:
	    outname = arg;
	    break;
      case OPT_DISCOUNT_VARS:
	    TermPrologVars = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_NO_INFIX:
	    EqnUseInfix = !CLStateGetBoolArg(handle, arg);
	    break;
      case OPT_FULL_EQ_REP:
	    EqnFullEquationalRep = CLStateGetBoolArg(handle, arg);
	    break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}
 
void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
cl_test\n\
\n\
Usage: term2dag [options] [files]\n\
\n\
Read a set of terms and print a DAG representing it.\n\
\n");
   PrintOptions(stdout, opts);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/


