/*-----------------------------------------------------------------------

File  : cl_test.c

Author: Stephan Schulz

Contents
 
  Main program for the CLIB test programm

Changes

<1> Fri Nov 28 00:27:40 MET 1997

-----------------------------------------------------------------------*/

#include <stdio.h>
#include <cio_commandline.h>
#include <cio_basicparser.h>
#include <cio_output.h>
#include <cto_orderings.h>
#include <ccl_clauses.h>


/*---------------------------------------------------------------------*/
/*                  Data types                                         */
/*---------------------------------------------------------------------*/

typedef enum
{
   OPT_NOOPT=0,
   OPT_HELP,
   OPT_VERBOSE,
   OPT_OUTPUT,
   OPT_PRINT_REFS,
   OPT_PROLOG_VARS
}OptionCodes;



/*---------------------------------------------------------------------*/
/*                        Global Variables                             */
/*---------------------------------------------------------------------*/

OptCell opts[] =
{
   {OPT_HELP, 
    'h', "help", 
    NoArg, NULL,
    "Print a short description of program usage and options."},
   {OPT_VERBOSE, 
    'v', "verbose", 
    OptArg, "1",
    "Verbose comments on the progress of the program."},
   {OPT_OUTPUT,
    'o', "output-file",
    ReqArg, NULL,
   "Redirect output into the named file."},
   {OPT_PRINT_REFS,
    'r', "print-reference-number",
    OptArg, "true",
    "Print number of references for each DAG node as a comment."},
   {OPT_PROLOG_VARS,
    'p', "prolog-vars",
    OptArg, "true",
    "Read and write Prolog-like terms (otherwise use rational terms)."},
   {OPT_NOOPT,
    '\0', NULL,
    NoArg, NULL,
    NULL}
};

char *outname = NULL;

/*---------------------------------------------------------------------*/
/*                      Forward Declarations                           */
/*---------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[]);
void print_help(FILE* out);

/*---------------------------------------------------------------------*/
/*                         Internal Functions                          */
/*---------------------------------------------------------------------*/

int main(int argc, char* argv[])
{
   CLState_p state;
   Scanner_p in; 
   Sig_p sig;
   VarBank_p vars;
   OCB_p ocb;
   CompareResult rescomp, complrescomp;
   bool resgtr, resles;
   Term_p t1, t2;
   int n_of_tests = 0;

   assert(argv[0]);
   InitError(argv[0]);

   state = process_options(argc, argv);
   
   if(state->argc ==  0)
   {
      CLStateInsertArg(state, "-");
   }

   in = CreateScanner(state->argv[0] , true, NULL);

   while(!TestInpTok(in, NoToken))
   {
      sig = SigAlloc();
      SigParse(in, sig);
      SigPrint(stdout, sig);
      ocb = OCBAlloc(KBO, sig);
      TOPrecedenceParse(in, ocb, sig);
      OCBDebugPrint(stdout, ocb);

      vars = VarBankAlloc();
      t1 = TermParse(in, sig, vars);
      t2 = TermParse(in, sig, vars);
      fprintf(stdout, "Testing ");
      TermPrint(stdout, t1, sig, false);
      fprintf(stdout, " and ");
      TermPrint(stdout, t2, sig, false);
      fprintf(stdout, "\n");
      CheckInpId(in, "greater|lesser|equal|uncomparable");
      rescomp = complrescomp = to_uncomparable;
      if (TestInpId(in, "greater"))
      {
	 rescomp = to_greater;
	 complrescomp = to_lesser;
      }
      else
      {
	 if (TestInpId(in, "lesser"))
	 {
	    rescomp = to_lesser;
	    complrescomp = to_greater;
	 }
	 else
	 {
	    if (TestInpId(in, "equal"))
	    {
	       rescomp = to_equal;
	       complrescomp = to_equal;
	    }
	 }
      }
      NextToken(in);
      CheckInpId(in, "true|false");
      resgtr = false;
      if(TestInpId(in, "true"))
      {
	 resgtr = true;
      }
      NextToken(in);
      CheckInpId(in, "true|false");
      resles = false;
      if(TestInpId(in, "true"))
      {
	 resles = true;
      }
      NextToken(in);

      if (KBOCompare(ocb, t1, t2, DEREF_NEVER, DEREF_NEVER) !=
	  rescomp)
      {
	 fprintf(stdout, "   KBOCompare(");
	 TermPrint(stdout, t1, sig, false);
	 fprintf(stdout, ",");
	 TermPrint(stdout, t2, sig, false);
	 fprintf(stdout, ") =%sis not correct\n",
		 TOCompareSymbol[KBOCompare(ocb, t1, t2, DEREF_NEVER,
				DEREF_NEVER)]);
	 return 0;
      }

      if (KBOCompare(ocb, t2, t1, DEREF_NEVER, DEREF_NEVER) !=
	  complrescomp)
      {
	 fprintf(stdout, "   KBOCompare(");
	 TermPrint(stdout, t2, sig, DEREF_NEVER);
	 fprintf(stdout, ",");
	 TermPrint(stdout, t1, sig, DEREF_NEVER);
	 fprintf(stdout, ") =%sis not correct\n",
		 TOCompareSymbol[KBOCompare(ocb, t1, t2, DEREF_NEVER,
					    DEREF_NEVER)]); 
	 return 0;
      }

      if (KBOGreater(ocb, t1, t2, DEREF_NEVER, DEREF_NEVER) != resgtr)
      {
	 fprintf(stdout, "   KBOGreater(");
	 TermPrint(stdout, t1, sig, DEREF_NEVER);
	 fprintf(stdout, ",");
	 TermPrint(stdout, t2, sig, DEREF_NEVER);
	 fprintf(stdout, ") =%dis not correct\n", KBOGreater(ocb, t1,
							     t2,
							     DEREF_NEVER, DEREF_NEVER));
	 return 0;
      }

      if (KBOGreater(ocb, t2, t1, DEREF_NEVER, DEREF_NEVER) != resles)
      {
	 fprintf(stdout, "   KBOGreater(");
	 TermPrint(stdout, t2, sig, DEREF_NEVER);
	 fprintf(stdout, ",");
	 TermPrint(stdout, t1, sig, DEREF_NEVER);
	 fprintf(stdout, ") =%dis not correct\n", KBOGreater(ocb, t2,
							     t1,
							     DEREF_NEVER, DEREF_NEVER)); 
	 return 0;
      }

      OCBFree(ocb);
      TermFree(t1);
      TermFree(t2);
      SigFree(sig);
      VarBankFree(vars);

      n_of_tests++;
   }

   fprintf(stdout, "\n###############################\n");
   fprintf(stdout, "# number of sound tests:  %d  #\n", n_of_tests);
   fprintf(stdout, "###############################\n");

   DestroyScanner(in);
   CLStateFree(state);

#ifdef CLB_MEMORY_DEBUG
   MemFlushFreeList();
   MemDebugPrintStats(stdout);
#endif

   return 0;
}


/*-----------------------------------------------------------------------
//
// Function: process_options()
//
//   Read and process the command line option, return (the pointer to)
//   a CLState object containing the remaining arguments.
//
// Global Variables: opts, Verbose, TermPrologArgs,
//                   TBPrintInternalInfo 
//
// Side Effects    : Sets variables, may terminate with program
//                   description if option -h or --help was present
//
/----------------------------------------------------------------------*/

CLState_p process_options(int argc, char* argv[])
{
   Opt_p handle;
   CLState_p state;
   char*  arg;
   
   state = CLStateAlloc(argc,argv);
   
   while((handle = CLStateGetOpt(state, &arg, opts)))
   {
      switch(handle->option_code)
      {
      case OPT_VERBOSE:
	 Verbose = CLStateGetIntArg(handle, arg);
	 break;
      case OPT_HELP: 
	 print_help(stdout);
	 exit(NO_ERROR);
      case OPT_OUTPUT:
	 outname = arg;
	 break;
      case OPT_PRINT_REFS:
	 TBPrintInternalInfo = CLStateGetBoolArg(handle, arg);
	 break;
      case OPT_PROLOG_VARS:
	 TermPrologVars = CLStateGetBoolArg(handle, arg);
	 break;
      default:
	 assert(false);
	 break;
      }
   }
   return state;
}

void print_help(FILE* out)
{
   fprintf(out, "\n\
\n\
cl_test\n\
\n\
Usage: term2dag [options] [files]\n\
\n\
Read a set of terms and print a DAG representing it.\n\
\n");
   PrintOptions(stdout, opts);
}


/*---------------------------------------------------------------------*/
/*                        End of File                                  */
/*---------------------------------------------------------------------*/
